Package: affiner
Type: Package
Title: A Finer Way to Render 3D Illustrated Objects in 'grid' Using
        Affine Transformations
Version: 0.3.1
Authors@R: c(person("Trevor L.", "Davis", role = c("aut", "cre"),
                    email = "trevor.l.davis@gmail.com",
                    comment = c(ORCID = "0000-0001-6341-4639")))
Description: Dilate, permute, project, reflect, rotate, shear, and translate 2D and 3D points.  Supports parallel projections including oblique projections such as the cabinet projection as well as axonometric projections such as the isometric projection.  Use 'grid's "affine transformation" feature to render illustrated flat surfaces.
URL: https://trevorldavis.com/R/affiner/
BugReports: https://github.com/trevorld/affiner/issues
License: MIT + file LICENSE
Depends: R (>= 3.6.0)
Imports: graphics, grDevices, grid, R6, utils
Suggests: aRtsy, ggplot2, gridpattern, gtable, knitr, ragg (>= 1.3.3),
        rgl, rlang, rmarkdown, stats, testthat (>= 3.0.0), vdiffr,
        withr
VignetteBuilder: knitr, ragg, rmarkdown
Encoding: UTF-8
Config/testthat/edition: 3
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-05-10 04:57:18 UTC; trevorld
Author: Trevor L. Davis [aut, cre] (ORCID:
    <https://orcid.org/0000-0001-6341-4639>)
Maintainer: Trevor L. Davis <trevor.l.davis@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-10 06:20:02 UTC
Built: R 4.6.0; ; 2026-05-10 08:20:04 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method("!=",Coord1D)
S3method("!=",Coord2D)
S3method("!=",Coord3D)
S3method("*",Coord1D)
S3method("*",Coord2D)
S3method("*",Coord3D)
S3method("+",Coord1D)
S3method("+",Coord2D)
S3method("+",Coord3D)
S3method("-",Coord1D)
S3method("-",Coord2D)
S3method("-",Coord3D)
S3method("/",Coord1D)
S3method("/",Coord2D)
S3method("/",Coord3D)
S3method("==",Coord1D)
S3method("==",Coord2D)
S3method("==",Coord3D)
S3method("[",Coord1D)
S3method("[",Coord2D)
S3method("[",Coord3D)
S3method("[",Ellipse2D)
S3method("[",Line2D)
S3method("[",Plane3D)
S3method("[",Point1D)
S3method("[",Segment2D)
S3method("[",angle)
S3method(Arg,Coord2D)
S3method(Arg,angle)
S3method(Complex,at_matrix)
S3method(Conj,Coord2D)
S3method(Im,Coord2D)
S3method(Math,at_matrix)
S3method(Mod,Coord2D)
S3method(Ops,angle)
S3method(Ops,at_matrix)
S3method(Re,Coord2D)
S3method(abs,Coord1D)
S3method(abs,Coord2D)
S3method(abs,Coord3D)
S3method(abs,angle)
S3method(as.complex,Coord2D)
S3method(as.complex,angle)
S3method(as.data.frame,Coord1D)
S3method(as.data.frame,Coord2D)
S3method(as.data.frame,Coord3D)
S3method(as.data.frame,Ellipse2D)
S3method(as.data.frame,Line2D)
S3method(as.data.frame,Plane3D)
S3method(as.data.frame,Point1D)
S3method(as.double,angle)
S3method(as.list,Coord1D)
S3method(as.list,Coord2D)
S3method(as.list,Coord3D)
S3method(as.list,Ellipse2D)
S3method(as.list,Line2D)
S3method(as.list,Plane3D)
S3method(as.list,Point1D)
S3method(as.matrix,Coord1D)
S3method(as.matrix,Coord2D)
S3method(as.matrix,Coord3D)
S3method(as.matrix,at_matrix)
S3method(as_angle,Coord2D)
S3method(as_angle,Coord3D)
S3method(as_angle,Line2D)
S3method(as_angle,Plane3D)
S3method(as_angle,angle)
S3method(as_angle,character)
S3method(as_angle,complex)
S3method(as_angle,numeric)
S3method(as_coord1d,Coord1D)
S3method(as_coord1d,Coord2D)
S3method(as_coord1d,Point1D)
S3method(as_coord1d,character)
S3method(as_coord1d,data.frame)
S3method(as_coord1d,list)
S3method(as_coord1d,matrix)
S3method(as_coord1d,numeric)
S3method(as_coord2d,Coord2D)
S3method(as_coord2d,Coord3D)
S3method(as_coord2d,angle)
S3method(as_coord2d,character)
S3method(as_coord2d,complex)
S3method(as_coord2d,data.frame)
S3method(as_coord2d,list)
S3method(as_coord2d,matrix)
S3method(as_coord2d,numeric)
S3method(as_coord3d,Coord2D)
S3method(as_coord3d,Coord3D)
S3method(as_coord3d,angle)
S3method(as_coord3d,character)
S3method(as_coord3d,data.frame)
S3method(as_coord3d,list)
S3method(as_coord3d,matrix)
S3method(as_coord3d,numeric)
S3method(as_ellipse2d,Coord2D)
S3method(as_ellipse2d,numeric)
S3method(as_line2d,Coord2D)
S3method(as_line2d,Line2D)
S3method(as_line2d,Point1D)
S3method(as_line2d,angle)
S3method(as_line2d,character)
S3method(as_line2d,numeric)
S3method(as_plane3d,Coord3D)
S3method(as_plane3d,Line2D)
S3method(as_plane3d,Plane3D)
S3method(as_plane3d,Point1D)
S3method(as_plane3d,character)
S3method(as_plane3d,numeric)
S3method(as_point1d,Coord1D)
S3method(as_point1d,Point1D)
S3method(as_point1d,character)
S3method(as_point1d,numeric)
S3method(as_polygon2d,Coord2D)
S3method(as_polygon2d,Ellipse2D)
S3method(as_polygon2d,Polygon2D)
S3method(as_polygon2d,numeric)
S3method(as_segment2d,Coord2D)
S3method(as_segment2d,Polygon2D)
S3method(as_transform1d,default)
S3method(as_transform1d,transform1d)
S3method(as_transform2d,default)
S3method(as_transform2d,transform2d)
S3method(as_transform3d,default)
S3method(as_transform3d,transform3d)
S3method(c,Coord1D)
S3method(c,Coord2D)
S3method(c,Coord3D)
S3method(c,Line2D)
S3method(c,Plane3D)
S3method(c,Point1D)
S3method(c,angle)
S3method(ceiling,Coord1D)
S3method(ceiling,Coord2D)
S3method(ceiling,Coord3D)
S3method(convex_hull2d,Coord2D)
S3method(convex_hull2d,Polygon2D)
S3method(cos,angle)
S3method(floor,Coord1D)
S3method(floor,Coord2D)
S3method(floor,Coord3D)
S3method(format,angle)
S3method(ggplot2::autolayer,Coord1D)
S3method(ggplot2::autolayer,Coord2D)
S3method(ggplot2::autolayer,Ellipse2D)
S3method(ggplot2::autolayer,Line2D)
S3method(ggplot2::autolayer,Point1D)
S3method(ggplot2::autolayer,Polygon2D)
S3method(ggplot2::autolayer,Segment2D)
S3method(has_intersection,Line2D)
S3method(has_intersection,Plane3D)
S3method(has_intersection,Point1D)
S3method(has_intersection,default)
S3method(has_overlap2d,Ellipse2D)
S3method(has_overlap2d,Polygon2D)
S3method(intersection,Coord1D)
S3method(intersection,Coord2D)
S3method(intersection,Coord3D)
S3method(intersection,Ellipse2D)
S3method(intersection,Line2D)
S3method(intersection,Plane3D)
S3method(intersection,Point1D)
S3method(is.finite,Coord1D)
S3method(is.finite,Coord2D)
S3method(is.finite,Coord3D)
S3method(is.finite,Line2D)
S3method(is.finite,Plane3D)
S3method(is.finite,Point1D)
S3method(is.infinite,Coord1D)
S3method(is.infinite,Coord2D)
S3method(is.infinite,Coord3D)
S3method(is.infinite,Line2D)
S3method(is.infinite,Plane3D)
S3method(is.infinite,Point1D)
S3method(is.na,Coord1D)
S3method(is.na,Coord2D)
S3method(is.na,Coord3D)
S3method(is.na,Line2D)
S3method(is.na,Plane3D)
S3method(is.na,Point1D)
S3method(is.nan,Coord1D)
S3method(is.nan,Coord2D)
S3method(is.nan,Coord3D)
S3method(is.nan,Line2D)
S3method(is.nan,Plane3D)
S3method(is.nan,Point1D)
S3method(is_congruent,angle)
S3method(is_congruent,numeric)
S3method(is_degenerate,Coord1D)
S3method(is_degenerate,Coord2D)
S3method(is_degenerate,Coord3D)
S3method(is_degenerate,Line2D)
S3method(is_degenerate,Plane3D)
S3method(is_degenerate,Point1D)
S3method(is_equivalent,Coord1D)
S3method(is_equivalent,Coord2D)
S3method(is_equivalent,Coord3D)
S3method(is_equivalent,Line2D)
S3method(is_equivalent,Plane3D)
S3method(is_equivalent,Point1D)
S3method(is_equivalent,angle)
S3method(is_equivalent,numeric)
S3method(is_parallel,Line2D)
S3method(is_parallel,Plane3D)
S3method(length,Coord1D)
S3method(length,Coord2D)
S3method(length,Coord3D)
S3method(length,Line2D)
S3method(length,Plane3D)
S3method(length,Point1D)
S3method(lines,Coord2D)
S3method(lines,Ellipse2D)
S3method(lines,Line2D)
S3method(lines,Point1D)
S3method(lines,Polygon2D)
S3method(lines,Segment2D)
S3method(makeContent,affine)
S3method(makeContent,isocube)
S3method(mean,Coord1D)
S3method(mean,Coord2D)
S3method(mean,Coord3D)
S3method(normal2d,Coord2D)
S3method(normal2d,Line2D)
S3method(normal3d,Coord3D)
S3method(normal3d,Plane3D)
S3method(normal3d,character)
S3method(painter_depth,Coord2D)
S3method(painter_depth,Coord3D)
S3method(painter_depth,Polygon2D)
S3method(painter_depth,Segment2D)
S3method(painter_order,Coord2D)
S3method(painter_order,Coord3D)
S3method(painter_order,Polygon2D)
S3method(plot,Coord1D)
S3method(plot,Coord2D)
S3method(plot,Ellipse2D)
S3method(plot,Polygon2D)
S3method(plot,Segment2D)
S3method(points,Coord1D)
S3method(points,Coord2D)
S3method(print,angle)
S3method(range,Coord1D)
S3method(range,Coord2D)
S3method(range,Coord3D)
S3method(rep,Coord1D)
S3method(rep,Coord2D)
S3method(rep,Coord3D)
S3method(rep,Line2D)
S3method(rep,Plane3D)
S3method(rep,Point1D)
S3method(rep,Segment2D)
S3method(rep,angle)
S3method(rgl::plot3d,Coord3D)
S3method(rgl::plot3d,Plane3D)
S3method(round,Coord1D)
S3method(round,Coord2D)
S3method(round,Coord3D)
S3method(signif,Coord1D)
S3method(signif,Coord2D)
S3method(signif,Coord3D)
S3method(sin,angle)
S3method(solve,at_matrix)
S3method(sort,Coord2D)
S3method(sort,Coord3D)
S3method(standardize,Line2D)
S3method(standardize,Plane3D)
S3method(standardize,Point1D)
S3method(sum,Coord1D)
S3method(sum,Coord2D)
S3method(sum,Coord3D)
S3method(t,at_matrix)
S3method(tan,angle)
S3method(trunc,Coord1D)
S3method(trunc,Coord2D)
S3method(trunc,Coord3D)
export("angular_unit<-")
export(Coord1D)
export(Coord2D)
export(Coord3D)
export(Ellipse2D)
export(Line2D)
export(Plane3D)
export(Point1D)
export(Polygon2D)
export(Segment2D)
export(affineGrob)
export(affine_settings)
export(affiner_options)
export(angle)
export(angular_unit)
export(arccosecant)
export(arccosine)
export(arccotangent)
export(arcsecant)
export(arcsine)
export(arctangent)
export(as_angle)
export(as_coord1d)
export(as_coord2d)
export(as_coord3d)
export(as_ellipse2d)
export(as_line2d)
export(as_plane3d)
export(as_point1d)
export(as_polygon2d)
export(as_segment2d)
export(as_transform1d)
export(as_transform2d)
export(as_transform3d)
export(convex_hull2d)
export(cosecant)
export(cosine)
export(cotangent)
export(cross_product3d)
export(degrees)
export(distance1d)
export(distance2d)
export(distance3d)
export(dot_product1d)
export(dot_product2d)
export(dot_product3d)
export(gradians)
export(grid.affine)
export(grid.isocube)
export(has_intersection)
export(has_overlap2d)
export(intersection)
export(is_angle)
export(is_congruent)
export(is_coord1d)
export(is_coord2d)
export(is_coord3d)
export(is_ellipse2d)
export(is_equivalent)
export(is_line2d)
export(is_parallel)
export(is_plane3d)
export(is_point1d)
export(is_polygon2d)
export(is_segment2d)
export(is_transform1d)
export(is_transform2d)
export(is_transform3d)
export(isocubeGrob)
export(isotoxal_2ngon_inner_radius)
export(isotoxal_2ngon_polygon2d)
export(normal2d)
export(normal3d)
export(painter_depth)
export(painter_order)
export(permute2d)
export(permute3d)
export(pi_radians)
export(project1d)
export(project2d)
export(project3d)
export(radians)
export(rectangle_polygon2d)
export(reflect1d)
export(reflect2d)
export(reflect3d)
export(regular_ngon_polygon2d)
export(rotate2d)
export(rotate3d)
export(rotate3d_to_AA)
export(scale1d)
export(scale2d)
export(scale3d)
export(secant)
export(shear2d)
export(shear3d)
export(sine)
export(star_inner_radius)
export(star_polygon2d)
export(tangent)
export(transform1d)
export(transform2d)
export(transform3d)
export(translate1d)
export(translate2d)
export(translate3d)
export(turns)
if (getRversion() >= "4.3.0") {
  S3method("%*%",Coord1D)
}
if (getRversion() >= "4.3.0") {
  S3method("%*%",Coord2D)
}
if (getRversion() >= "4.3.0") {
  S3method("%*%",Coord3D)
}
if (getRversion() >= "4.3.0") {
  S3method("%*%",at_matrix)
}
if (getRversion() >= "4.4.0") {
  S3method("crossprod",Coord3D)
}
importFrom(R6,R6Class)
importFrom(graphics,lines)
importFrom(graphics,points)
importFrom(graphics,polygon)
importFrom(graphics,segments)
importFrom(grid,makeContent)
importFrom(utils,hasName)

## ----angles-------------------------------------------------------------------
library("affiner")
as_angle(90, "degrees") + turns(1)
is_congruent(degrees(180), radians(pi))
as.numeric(turns(1/3), "radians")

## ----trig---------------------------------------------------------------------
library("affiner")
sin(2 * pi)
sine(degrees(360))
arctangent(x = 0, y = 1)

## ----2d-----------------------------------------------------------------------
# Cartesian coordinates
library("affiner")
p <- as_coord2d(x = 1:10, y = 1:10)
print(p)

p2 <- p$
    clone()$
    scale(x = 0.5)$
    rotate(degrees(90))$
    reflect(as_line2d("y-axis"))$
    translate(as_coord2d(x = 0.5, y = 0.5))$
    print()

# Polar coordinates
theta <- degrees(seq(0, 300, by = 60))
radius <- 1
p <- as_coord2d(theta, radius = radius)
is_congruent(as_angle(p), theta) |> all()
is_congruent(abs(p), radius) |> all()

## ----3d-----------------------------------------------------------------------
# Cartesian coordinates
library("affiner")
p <- as_coord3d(x = 1:10, y = 1:10, z = 1:10)
print(p)

p2 <- p$
    clone()$
    scale(z = 0.5)$
    rotate(axis = as_coord3d("z-axis"), theta = degrees(90))$
    reflect(as_plane3d("yz-plane"))$
    shear(xy_shear = 0.5)$
    translate(as_coord3d(x = 0.5, y = 0.5, z = 0.5))$
    print()

# Spherical coordinates
inclination <- as_angle(p, type = "inclination")
azimuth <- as_angle(p, type = "azimuth")
radius <- abs(p)
ps <- as_coord3d(azimuth, radius = radius, inclination = inclination)
all.equal(p, ps)

# Cylindrical coordinates
radius <- as_coord2d(p, plane = "xy-plane") |> abs()
pc <- as_coord3d(azimuth, radius = radius, z = p$z)
all.equal(p, pc)

## ----hidden, echo = FALSE-----------------------------------------------------
knitr::opts_chunk$set(fig.cap = "", dev = "ragg_png")

## ----hex-logo2, fig.width = 4.5, fig.height = 4.5, fig.alt = "Isometric-cube hex logo", eval = getRversion() >= "4.2.0" && requireNamespace("ggplot2", quietly=TRUE) && requireNamespace("aRtsy", quietly=TRUE) && requireNamespace("gtable", quietly=TRUE), message = FALSE----
library("aRtsy")
library("ggplot2")

gg <- canvas_planet(colorPalette("lava"), threshold = 3) +
  scale_x_continuous(expand=c(0, 0)) +
  scale_y_continuous(expand=c(0, 0))
grob <- ggplotGrob(gg)
grob <- gtable::gtable_filter(grob, "panel") # grab just the panel
affiner::grid.isocube(top = grob, left = grob, right = grob,
  gp_border = grid::gpar(col = "darkorange", lwd = 12))

## ----hex-logo, fig.width = 4.5, fig.height = 4.5, fig.alt = "Isometric-cube hex logo", eval = getRversion() >= "4.2.0"----
library("affiner")
library("grid")

xy <- as_coord2d(angle(seq(90, 360 + 90, by = 60), "degrees"),
                 radius = c(rep(0.488, 6), 0))
xy$translate(x = 0.5, y = 0.5)
l_xy <- list()
l_xy$top <- xy[c(1, 2, 7, 6)]
l_xy$right <- xy[c(7, 4, 5, 6)]
l_xy$left <- xy[c(2, 3, 4, 7)]

gp_border <- gpar(fill = NA, col = "black", lwd = 12)
vp_define <- viewport(width = unit(3, "inches"), height = unit(3, "inches"))

colors <- c("#D55E00", "#56B4E9", "#009E73")
spacings <- c(0.25, 0.25, 0.2)
texts <- c("pkgname", "right\nface", "left\nface")
rots <- c(45, 0, 0)
fontsizes <- c(52, 80, 80)
sides <- c("top", "right", "left")
types <- gridpattern::names_polygon_tiling[c(5, 9, 7)]
l_grobs <- list()
grid.newpage()
for (i in 1:3) {
    side <- sides[i]
    xy_side <- l_xy[[side]]
    if (requireNamespace("gridpattern", quietly = TRUE)) {
        bg <- gridpattern::grid.pattern_polygon_tiling(
                   colour = "grey80",
                   fill = c(colors[i], "white"),
                   type = types[i],
                   spacing = spacings[i],
                   draw = FALSE)
    } else {
        bg <- rectGrob(gp = gpar(col = NA, fill = colors[i]))
    }
    text <- textGrob(texts[i], rot = rots[i],
                     gp = gpar(fontsize = fontsizes[i]))
    settings <- affine_settings(xy_side, unit = "snpc")
    grob <- l_grobs[[side]] <- grobTree(bg, text)
    grid.affine(grob,
                vp_define = vp_define,
                transform = settings$transform,
                vp_use = settings$vp)
    grid.polygon(xy_side$x, xy_side$y, gp = gp_border)
}

## ----die-faces, fig.width = 4.0, fig.height = 3.0, fig.alt = "The six die faces", eval = getRversion() >= "4.2.0"----
library("affiner")
library("grid")
xyz_face <- as_coord3d(x = c(0, 0, 1, 1) - 0.5, y = c(1, 0, 0, 1) - 0.5, z = 0.5)
l_faces <- list() # order faces for our target projections
l_faces$bottom <- xyz_face$clone()$
                    rotate("z-axis", angle(180, "degrees"))$
                    rotate("y-axis", angle(180, "degrees"))
l_faces$north <- xyz_face$clone()$
                    rotate("z-axis", angle(90, "degrees"))$
                    rotate("x-axis", angle(-90, "degrees"))
l_faces$east <- xyz_face$clone()$
                    rotate("z-axis", angle(90, "degrees"))$
                    rotate("y-axis", angle(90, "degrees"))
l_faces$west <- xyz_face$clone()$
                    rotate("y-axis", angle(-90, "degrees"))
l_faces$south <- xyz_face$clone()$
                    rotate("z-axis", angle(180, "degrees"))$
                    rotate("x-axis", angle(90, "degrees"))
l_faces$top <- xyz_face$clone()$
                    rotate("z-axis", angle(-90, "degrees"))

colors <- c("#D55E00", "#009E73", "#56B4E9", "#E69F00", "#CC79A7", "#0072B2")
spacings <- c(0.25, 0.2, 0.25, 0.25, 0.25, 0.25)
die_face_grob <- function(digit) {
    if (requireNamespace("gridpattern", quietly = TRUE)) {
        bg <- gridpattern::grid.pattern_polygon_tiling(
                   colour = "grey80",
                   fill = c(colors[digit], "white"),
                   type = gridpattern::names_polygon_tiling[digit],
                   spacing = spacings[digit],
                   draw = FALSE)
    } else {
        bg <- rectGrob(gp = gpar(col = NA, fill = colors[digit]))
    }
    digit <- textGrob(digit, gp = gpar(fontsize = 72))
    grobTree(bg, digit)
}
l_face_grobs <- lapply(1:6, function(i) die_face_grob(i))
grid.newpage()
for (i in 1:6) {
    vp <- viewport(x = unit((i - 1) %% 3 + 1, "inches"),
                   y = unit(3 - ((i - 1) %/% 3 + 1), "inches"),
                   width = unit(1, "inches"), height = unit(1, "inches"))
    pushViewport(vp)
    grid.draw(l_face_grobs[[i]])
    popViewport()
    grid.text("The six die faces", y = 0.9, 
              gp = gpar(fontsize = 18, face = "bold"))
}

## ----projected-die, fig.width = 3.0, fig.height = 3.0, fig.alt = "Parallel projection of a die", eval = getRversion() >= "4.2.0"----
# re-order face grobs for our target projections
# bottom = 6, north = 4, east = 5, west = 2, south = 3, top = 1
l_face_grobs <- l_face_grobs[c(6, 4, 5, 2, 3, 1)]
draw_die <- function(l_xy, l_face_grobs) {
    min_x <- min(vapply(l_xy, function(x) min(x$x), numeric(1)))
    min_y <- min(vapply(l_xy, function(x) min(x$y), numeric(1)))
    l_xy <- lapply(l_xy, function(xy) {
        xy$translate(x = -min_x + 0.5, y = -min_y + 0.5)
    })
    grid.newpage()
    vp_define <- viewport(width = unit(1, "inches"), height = unit(1, "inches"))
    gp_border <- gpar(col = "black", lwd = 4, fill = NA)
    for (i in 1:6) {
        xy <- l_xy[[i]]
        settings <- affine_settings(xy, unit = "inches")
        grid.affine(l_face_grobs[[i]],
                    vp_define = vp_define,
                    transform = settings$transform,
                    vp_use = settings$vp)
        grid.polygon(xy$x, xy$y, default.units = "inches", gp = gp_border)
    }
}
# oblique projection of dice onto xy-plane
l_xy_oblique1 <- lapply(l_faces, function(xyz) {
    xyz$clone() |>
        as_coord2d(scale = 0.5)
})
draw_die(l_xy_oblique1, l_face_grobs)
grid.text("Oblique projection\n(onto xy-plane)", y = 0.9,
          gp = gpar(fontsize = 18, face = "bold"))

# oblique projection of dice on xz-plane
l_xy_oblique2 <- lapply(l_faces, function(xyz) {
    xyz$clone()$
        permute("xzy") |>
        as_coord2d(scale = 0.5, alpha = angle(135, "degrees"))
})
draw_die(l_xy_oblique2, l_face_grobs)
grid.text("Oblique projection\n(onto xz-plane)", y = 0.9,
          gp = gpar(fontsize = 18, face = "bold"))

# isometric projection
l_xy_isometric <- lapply(l_faces, function(xyz) {
    xyz$clone()$
        rotate("z-axis", angle(45, "degrees"))$
        rotate("x-axis", angle(-(90 - 35.264), "degrees")) |>
        as_coord2d()
})

draw_die(l_xy_isometric, l_face_grobs)
grid.text("Isometric projection", y = 0.9,
          gp = gpar(fontsize = 18, face = "bold"))

#' Whether two 2D shapes overlap
#'
#' `has_overlap2d()` is an S3 generic that returns whether two 2D shapes
#' have a non-zero-area overlap (i.e. their interiors intersect).
#'
#' For two convex shapes the function uses the
#' [Separating Axis Theorem (SAT)](https://en.wikipedia.org/wiki/Hyperplane_separation_theorem).
#' For concave [Polygon2D] objects the function first checks whether the
#' axis-aligned bounding box and convex hull overlap; if neither rules out
#' overlap it returns `NA` with a warning because exact detection is not yet
#' supported.
#' For non-circular [Ellipse2D] objects the function approximates the ellipse
#' with inner and outer polygons: if the outer polygon does not overlap the
#' result is `FALSE`; if the inner polygon overlaps the result is `TRUE`;
#' otherwise `NA` is returned with a warning.
#'
#' @param x,y The two shapes to check.  Supported classes are [Polygon2D]
#'   and [Ellipse2D].
#' @param ... Ignored; only included for S3 method consistency.
#' @param n Number of vertices used to approximate non-circular [Ellipse2D]
#'   objects (default `64L`).  Larger values give tighter bounds.
#' @param tol Numeric tolerance for overlap comparisons (default `0`).
#'   A positive value requires projections to overlap by more than `tol`
#'   before overlap is declared, which avoids false positives caused by
#'   floating-point error accumulating in rotated coordinates.
#'   `sqrt(.Machine$double.eps)` is a reasonable choice when shapes are
#'   constructed via trigonometric transformations.
#' @return A logical vector (or `NA`) of length equal to the longer of `x`
#'   and `y` (for [Ellipse2D] objects; [Polygon2D] objects are always scalar).
#' @examples
#' # Two overlapping squares
#' sq1 <- as_polygon2d(as_coord2d(
#'   x = c(0, 1, 1, 0),
#'   y = c(0, 0, 1, 1)
#' ))
#' sq2 <- as_polygon2d(as_coord2d(
#'   x = c(0.5, 1.5, 1.5, 0.5),
#'   y = c(0.5, 0.5, 1.5, 1.5)
#' ))
#' sq3 <- as_polygon2d(as_coord2d(
#'   x = c(2, 3, 3, 2),
#'   y = c(2, 2, 3, 3)
#' ))
#' has_overlap2d(sq1, sq2)
#' has_overlap2d(sq1, sq3)
#'
#' # Circle vs polygon
#' circ <- as_ellipse2d(as_coord2d(0.5, 0.5), rx = 0.4)
#' has_overlap2d(sq1, circ)
#' has_overlap2d(sq3, circ)
#'
#' # Two circles
#' c1 <- as_ellipse2d(as_coord2d(0, 0), rx = 1)
#' c2 <- as_ellipse2d(as_coord2d(1.5, 0), rx = 1)
#' c3 <- as_ellipse2d(as_coord2d(3, 0), rx = 1)
#' has_overlap2d(c1, c2)
#' has_overlap2d(c1, c3)
#' @export
has_overlap2d <- function(x, y, ...) {
	UseMethod("has_overlap2d")
}

#' @rdname has_overlap2d
#' @export
has_overlap2d.Polygon2D <- function(x, y, ..., n = 64L, tol = sqrt(.Machine$double.eps)) {
	chkDots(...)
	if (is_polygon2d(y)) {
		polygon_polygon_overlap(x, y, tol = tol)
	} else if (is_ellipse2d(y)) {
		ne <- length(y$rx)
		vapply(
			seq_len(ne),
			function(i) ellipse_overlap_with_polygon(y[i], x, n = n, tol = tol),
			logical(1L)
		)
	} else {
		stop(paste(
			"Don't know how to check overlap with",
			dQuote(class(y)[1L])
		))
	}
}

#' @rdname has_overlap2d
#' @export
has_overlap2d.Ellipse2D <- function(x, y, ..., n = 64L, tol = sqrt(.Machine$double.eps)) {
	chkDots(...)
	nx <- length(x$rx)
	if (is_ellipse2d(y)) {
		ny <- length(y$rx)
		nout <- max(nx, ny)
		xr <- rep_len(seq_len(nx), nout)
		yr <- rep_len(seq_len(ny), nout)
		vapply(
			seq_len(nout),
			function(i) ellipse_ellipse_overlap(x[xr[i]], y[yr[i]], n = n, tol = tol),
			logical(1L)
		)
	} else if (is_polygon2d(y)) {
		vapply(
			seq_len(nx),
			function(i) ellipse_overlap_with_polygon(x[i], y, n = n, tol = tol),
			logical(1L)
		)
	} else {
		stop(paste(
			"Don't know how to check overlap with",
			dQuote(class(y)[1L])
		))
	}
}

# Internal: do two [min, max] ranges overlap by more than tol?
ranges_overlap_strictly <- function(r1, r2, tol = sqrt(.Machine$double.eps)) {
	r1[2L] > r2[1L] + tol && r2[2L] > r1[1L] + tol
}

# Internal: SAT for two convex polygons
sat_convex_polygon_polygon <- function(p1, p2, tol = sqrt(.Machine$double.eps)) {
	axes <- c(p1$normals, p2$normals)
	n <- length(axes)
	for (i in seq_len(n)) {
		ax <- axes[i]
		if (
			!ranges_overlap_strictly(
				range(p1 * ax),
				range(p2 * ax),
				tol = tol
			)
		) {
			return(FALSE)
		}
	}
	TRUE
}

# Internal: SAT for a convex polygon and a single circle
# center: Coord2D of length 1; radius: numeric scalar
sat_convex_polygon_circle <- function(p, center, radius, tol = sqrt(.Machine$double.eps)) {
	# Test polygon edge normals
	poly_axes <- p$normals
	n_edges <- length(poly_axes)
	for (i in seq_len(n_edges)) {
		ax <- poly_axes[i]
		cp <- center * ax
		if (
			!ranges_overlap_strictly(
				range(p * ax),
				c(cp - radius, cp + radius),
				tol = tol
			)
		) {
			return(FALSE)
		}
	}
	# Test vertex-to-center axes (direction from each vertex toward the center)
	n_verts <- length(p)
	for (i in seq_len(n_verts)) {
		# vector from vertex[i] to center
		vc <- center - p[i]
		len <- abs(vc)
		if (len < .Machine$double.eps) {
			return(TRUE) # center coincides with vertex → overlap
		}
		vc$scale(1 / len) # normalize in place (vc is a temporary clone)
		cp <- center * vc
		if (
			!ranges_overlap_strictly(
				range(p * vc),
				c(cp - radius, cp + radius),
				tol = tol
			)
		) {
			return(FALSE)
		}
	}
	TRUE
}

# Internal: do axis-aligned bounding boxes of two polygons overlap?
bbox_overlap_pp <- function(p1, p2, tol = sqrt(.Machine$double.eps)) {
	ranges_overlap_strictly(range(p1$x), range(p2$x), tol = tol) &&
		ranges_overlap_strictly(range(p1$y), range(p2$y), tol = tol)
}

# Internal: does a polygon's bbox overlap with a circle's bbox?
bbox_overlap_pc <- function(p, center, radius, tol = sqrt(.Machine$double.eps)) {
	cx <- center$x
	cy <- center$y
	ranges_overlap_strictly(
		range(p$x),
		c(cx - radius, cx + radius),
		tol = tol
	) &&
		ranges_overlap_strictly(
			range(p$y),
			c(cy - radius, cy + radius),
			tol = tol
		)
}

# Internal: polygon vs polygon (handles concave via hull fallback)
polygon_polygon_overlap <- function(p1, p2, tol = sqrt(.Machine$double.eps)) {
	if (isTRUE(p1$is_convex) && isTRUE(p2$is_convex)) {
		return(sat_convex_polygon_polygon(p1, p2, tol = tol))
	}
	# At least one is concave or unknown: conservative bbox + hull checks
	if (!bbox_overlap_pp(p1, p2, tol = tol)) {
		return(FALSE)
	}
	h1 <- if (isTRUE(p1$is_convex)) p1 else p1$convex_hull
	h2 <- if (isTRUE(p2$is_convex)) p2 else p2$convex_hull
	if (!sat_convex_polygon_polygon(h1, h2, tol = tol)) {
		return(FALSE)
	}
	warning(
		"Exact overlap detection is not yet supported for concave polygons; ",
		"their convex hulls overlap."
	)
	NA
}

# Internal: polygon vs single circle (handles concave via hull fallback)
polygon_circle_overlap <- function(p, center, radius, tol = sqrt(.Machine$double.eps)) {
	if (isTRUE(p$is_convex)) {
		return(sat_convex_polygon_circle(p, center, radius, tol = tol))
	}
	# Concave or unknown: conservative bbox + hull checks
	if (!bbox_overlap_pc(p, center, radius, tol = tol)) {
		return(FALSE)
	}
	h <- p$convex_hull
	if (!sat_convex_polygon_circle(h, center, radius, tol = tol)) {
		return(FALSE)
	}
	warning(
		"Exact overlap detection is not yet supported for concave polygons; ",
		"the convex hull overlaps with the circle."
	)
	NA
}

# Internal: single Ellipse2D (length 1) vs Polygon2D
ellipse_overlap_with_polygon <- function(e, p, n, tol = sqrt(.Machine$double.eps)) {
	if (e$is_circle) {
		center <- as_coord2d(e$x, e$y)
		return(polygon_circle_overlap(p, center, e$rx, tol = tol))
	}
	ellipse_bracket_overlap(e, p, n, tol = tol)
}

# Internal: single Ellipse2D vs single Ellipse2D
ellipse_ellipse_overlap <- function(e1, e2, n, tol = sqrt(.Machine$double.eps)) {
	if (e1$is_circle && e2$is_circle) {
		center1 <- as_coord2d(e1$x, e1$y)
		center2 <- as_coord2d(e2$x, e2$y)
		return(distance2d(center1, center2) < e1$rx + e2$rx - tol)
	}
	# Use the non-circle ellipse as the one we approximate with polygons
	if (!e1$is_circle) {
		ellipse_bracket_overlap(e1, e2, n, tol = tol)
	} else {
		ellipse_bracket_overlap(e2, e1, n, tol = tol)
	}
}

# Internal: bracket overlap of a single non-circle Ellipse2D vs another shape
# (Polygon2D or Ellipse2D) using inner/outer polygon approximations.
ellipse_bracket_overlap <- function(e, other, n, tol = sqrt(.Machine$double.eps)) {
	outer_poly <- as_polygon2d(e, n = n, type = "outer")
	outer_res <- has_overlap2d(outer_poly, other, n = n, tol = tol)
	if (!isTRUE(outer_res)) {
		return(FALSE)
	}
	inner_poly <- as_polygon2d(e, n = n, type = "inner")
	inner_res <- has_overlap2d(inner_poly, other, n = n, tol = tol)
	if (isTRUE(inner_res)) {
		return(TRUE)
	}
	warning(
		"Exact overlap detection is not yet supported for non-circular ellipses; ",
		"the result is in the boundary region between the inner and outer polygon approximations. ",
		"Increase `n` for a tighter approximation."
	)
	NA
}

#' 1D coordinate vector R6 Class
#'
#' `Coord1D` is an [R6::R6Class()] object representing one-dimensional points
#' represented by Cartesian Coordinates.
#'
#' @examples
#' p <- as_coord1d(x = rnorm(100, 2))
#' print(p, n = 10L)
#' pc <- mean(p) # Centroid
#' # method chained affine transformation matrices are auto-pre-multiplied
#' p$
#'   translate(-pc)$
#'   reflect("origin")$
#'   print(n = 10L)
#' @export
Coord1D <- R6Class(
	"Coord1D",
	public = list(
		#' @param xw A matrix with three columns representing (homogeneous) coordinates.
		#'           The first column represents x coordinates and
		#'           the last column is all ones.
		#'           Column names should be "x" and "w".
		initialize = function(xw) {
			private$mat_xw <- xw
			private$mat_trans <- NULL
		},
		#' @param n Number of coordinates to print. If `NULL` print all of them.
		#' @param ... Passed to [format.default()].
		print = function(n = NULL, ...) {
			if (is.null(n) || n > nrow(private$mat_xw)) {
				n <- nrow(private$mat_xw)
			}
			cat("<Coord1D[", nrow(private$mat_xw), "]>\n", sep = "")
			if (n > 0) {
				print_mat(self$xw[1:n, , drop = FALSE], ...)
			}
			invisible(self)
		},
		#' @param point `r r2i_transform1d_point`
		#' @param ... Passed to [project1d()].
		project = function(point = as_point1d("origin"), ...) {
			self$transform(project1d(point, ...))
		},
		#' @param point `r r2i_transform1d_point`
		#' @param ... Passed to [reflect1d()].
		reflect = function(point = as_point1d("origin"), ...) {
			self$transform(reflect1d(point, ...))
		},
		#' @param x_scale `r r2i_transform_x_scale`
		scale = function(x_scale = 1) {
			if (length(x_scale) == 1L) {
				self$transform(scale1d(x_scale))
			} else {
				private$apply_any_delayed_transformations()
				private$mat_xw[, 1L] <- x_scale * private$mat_xw[, 1L]
				invisible(self)
			}
		},
		#' @param x `r r2i_transform1d_x`
		#' @param ... Passed to `as_coord1d(x, ...)` if `x` is not a [Coord1D] object
		translate = function(x = as_coord1d(0), ...) {
			if (!is_coord1d(x)) {
				x <- as_coord1d(x, ...)
			}
			if (length(x) == 1L) {
				self$transform(translate1d(x))
			} else {
				private$apply_any_delayed_transformations()
				private$mat_xw[, 1L] <- private$mat_xw[, 1L] + x$x
				invisible(self)
			}
		},
		#' @param mat `r r2i_transform1d_mat`
		transform = function(mat = transform1d()) {
			if (!is_transform1d(mat)) {
				mat <- as_transform1d(mat)
			}
			if (is.null(private$mat_trans)) {
				private$mat_trans <- mat
			} else {
				private$mat_trans <- private$mat_trans %*% mat
			}
			invisible(self)
		}
	),
	active = list(
		#' @field xw A two-column matrix representing the homogeneous coordinates.
		#'           The first column is the "x" coordinates
		#'           and the second column is all ones.
		xw = function() {
			private$apply_any_delayed_transformations()
			private$mat_xw
		},
		#' @field x A numeric vector of x-coordinates.
		x = function() {
			as.vector(self$xw[, 1L])
		}
	),
	private = list(
		mat_xw = NULL,
		mat_trans = NULL,
		apply_any_delayed_transformations = function() {
			if (!is.null(private$mat_trans)) {
				private$mat_xw <- private$mat_xw %*% private$mat_trans
				private$mat_trans <- NULL
				colnames(private$mat_xw) <- c("x", "w")
			}
		}
	)
)

#' 2D coordinate vector R6 Class
#'
#' `Coord2D` is an [R6::R6Class()] object representing two-dimensional points
#' represented by Cartesian Coordinates.
#'
#' @examples
#' p <- as_coord2d(x = rnorm(100, 2), y = rnorm(100, 2))
#' print(p, n = 10)
#' pc <- mean(p) # Centroid
#' # method chained affine transformation matrices are auto-pre-multiplied
#' p$
#'   translate(-pc)$
#'   shear(x = 1, y = 0)$
#'   reflect("x-axis")$
#'   rotate(90, "degrees")$
#'   print(n = 10)
#' @export
Coord2D <- R6Class(
	"Coord2D",
	public = list(
		#' @param xyw A matrix with three columns representing (homogeneous) coordinates.
		#'             The first two columns represent x and y coordinates and
		#'             the last column is all ones.
		#'             Column names should be "x", "y", and "w".
		initialize = function(xyw) {
			private$mat_xyw <- xyw
			private$mat_trans <- NULL
		},
		#' @param permutation `r r2i_transform2d_permutation`
		permute = function(permutation = c("xy", "yx")) {
			self$transform(permute2d(permutation))
		},
		#' @param n Number of coordinates to print.  If `NULL` print all of them.
		#' @param ... Passed to [format.default()].
		print = function(n = NULL, ...) {
			if (is.null(n) || n > nrow(private$mat_xyw)) {
				n <- nrow(private$mat_xyw)
			}
			cat("<Coord2D[", nrow(private$mat_xyw), "]>\n", sep = "")
			if (n > 0) {
				print_mat(self$xyw[1:n, , drop = FALSE], ...)
			}
			invisible(self)
		},
		#' @param line `r r2i_transform2d_line`
		#' @param ... Passed to [project2d()]
		#' @param scale `r r2i_transform2d_scale`
		project = function(line = as_line2d("x-axis"), ..., scale = 0) {
			self$transform(project2d(line, ..., scale = scale))
		},
		#' @param line `r r2i_transform2d_line`
		#' @param ... Passed to [reflect2d()].
		reflect = function(line = as_line2d("x-axis"), ...) {
			self$transform(reflect2d(line, ...))
		},
		#' @param theta `r r2i_transform_theta`
		#' @param ... Passed to [as_angle()].
		rotate = function(theta = angle(0), ...) {
			if (!is_angle(theta)) {
				theta <- as_angle(theta, ...)
			}
			if (length(theta) == 1L) {
				self$transform(rotate2d(theta))
			} else {
				private$apply_any_delayed_transformations()
				x <- private$mat_xyw[, 1L]
				y <- private$mat_xyw[, 2L]
				private$mat_xyw[, 1L] <- x * cos(theta) - y * sin(theta)
				private$mat_xyw[, 2L] <- x * sin(theta) + y * cos(theta)
				invisible(self)
			}
		},
		#' @param x_scale `r r2i_transform_x_scale`
		#' @param y_scale `r r2i_transform_y_scale`
		scale = function(x_scale = 1, y_scale = x_scale) {
			if (length(x_scale) == 1L && length(y_scale) == 1L) {
				self$transform(scale2d(x_scale, y_scale))
			} else {
				private$apply_any_delayed_transformations()
				private$mat_xyw[, 1L] <- x_scale * private$mat_xyw[, 1L]
				private$mat_xyw[, 2L] <- y_scale * private$mat_xyw[, 2L]
				invisible(self)
			}
		},
		#' @param xy_shear `r r2i_transform2d_xy_shear`
		#' @param yx_shear `r r2i_transform2d_yx_shear`
		shear = function(xy_shear = 0, yx_shear = 0) {
			self$transform(shear2d(xy_shear, yx_shear))
		},
		#' @param x `r r2i_transform2d_x`
		#' @param ... Passed to `as_coord2d(x, ...)` if `x` is not a [Coord2D] object
		translate = function(x = as_coord2d(0, 0), ...) {
			if (!is_coord2d(x)) {
				x <- as_coord2d(x, ...)
			}
			if (length(x) == 1L) {
				self$transform(translate2d(x))
			} else {
				private$apply_any_delayed_transformations()
				private$mat_xyw[, 1L] <- private$mat_xyw[, 1L] + x$x
				private$mat_xyw[, 2L] <- private$mat_xyw[, 2L] + x$y
				invisible(self)
			}
		},
		#' @param mat `r r2i_transform2d_mat`
		transform = function(mat = transform2d()) {
			if (!is_transform2d(mat)) {
				mat <- as_transform2d(mat)
			}
			if (is.null(private$mat_trans)) {
				private$mat_trans <- mat
			} else {
				private$mat_trans <- private$mat_trans %*% mat
			}
			invisible(self)
		}
	),
	active = list(
		#' @field xyw A three-column matrix representing the homogeneous coordinates.
		#'             The first two columns are "x" and "y" coordinates
		#'             and the third column is all ones.
		xyw = function() {
			private$apply_any_delayed_transformations()
			private$mat_xyw
		},
		#' @field x A numeric vector of x-coordinates.
		x = function() {
			as.vector(self$xyw[, 1L])
		},
		#' @field y A numeric vector of y-coordinates.
		y = function() {
			as.vector(self$xyw[, 2L])
		}
	),
	private = list(
		mat_xyw = NULL,
		mat_trans = NULL,
		apply_any_delayed_transformations = function() {
			if (!is.null(private$mat_trans)) {
				private$mat_xyw <- private$mat_xyw %*% private$mat_trans
				private$mat_trans <- NULL
				colnames(private$mat_xyw) <- c("x", "y", "w")
			}
		}
	)
)

#' 3D coordinate vector R6 Class
#'
#' `Coord3D` is an [R6::R6Class()] object representing three-dimensional points
#' represented by Cartesian Coordinates.
#'
#' @examples
#' p <- as_coord3d(x = rnorm(100, 2), y = rnorm(100, 2), z = rnorm(100, 2))
#' print(p, n = 10)
#' pc <- mean(p) # Centroid
#' # method chained affine transformation matrices are auto-pre-multiplied
#' p$
#'   translate(-pc)$
#'   reflect("xy-plane")$
#'   rotate("z-axis", degrees(90))$
#'   print(n = 10)
#' @export
Coord3D <- R6Class(
	"Coord3D",
	public = list(
		#' @param xyzw A matrix with four columns representing (homogeneous) coordinates.
		#'             The first three columns represent x, y, and z coordinates and
		#'             the last column is all ones.
		#'             Column names should be "x", "y", "z", and "w".
		initialize = function(xyzw) {
			private$mat_xyzw <- xyzw
			private$mat_trans <- NULL
		},
		#' @param permutation `r r2i_transform3d_permutation`
		permute = function(permutation = c("xyz", "xzy", "yxz", "yzx", "zyx", "zxy")) {
			self$transform(permute3d(permutation))
		},
		#' @param n Number of coordinates to print.  If `NULL` print all of them.
		#' @param ... Passed to [format.default()].
		print = function(n = NULL, ...) {
			if (is.null(n) || n > nrow(private$mat_xyzw)) {
				n <- nrow(private$mat_xyzw)
			}
			cat("<Coord3D[", nrow(private$mat_xyzw), "]>\n", sep = "")
			if (n > 0) {
				print_mat(self$xyzw[1:n, , drop = FALSE], ...)
			}
			invisible(self)
		},
		#' @param plane `r r2i_transform3d_plane`
		#' @param ... Passed to [project3d()].
		#' @param scale `r r2i_transform3d_scale`
		#' @param alpha `r r2i_transform3d_alpha`
		project = function(
			plane = as_plane3d("xy-plane"),
			...,
			scale = 0,
			alpha = angle(45, "degrees")
		) {
			self$transform(project3d(plane, ..., scale = scale, alpha = alpha))
		},
		#' @param plane `r r2i_transform3d_plane`
		#' @param ... Passed to [reflect3d()].
		reflect = function(plane = as_plane3d("xy-plane"), ...) {
			self$transform(reflect3d(plane, ...))
		},
		#' @param axis `r r2i_transform3d_axis`
		#' @param theta `r r2i_transform_theta`
		#' @param ... Passed to [rotate3d()].
		rotate = function(axis = as_coord3d("z-axis"), theta = angle(0), ...) {
			self$transform(rotate3d(axis, theta, ...))
		},
		#' @param x_scale `r r2i_transform_x_scale`
		#' @param y_scale `r r2i_transform_y_scale`
		#' @param z_scale `r r2i_transform_z_scale`
		scale = function(x_scale = 1, y_scale = x_scale, z_scale = x_scale) {
			if (length(x_scale) == 1L && length(y_scale) == 1L && length(z_scale) == 1L) {
				self$transform(scale3d(x_scale, y_scale, z_scale))
			} else {
				private$apply_any_delayed_transformations()
				private$mat_xyzw[, 1L] <- x_scale * private$mat_xyzw[, 1L]
				private$mat_xyzw[, 2L] <- y_scale * private$mat_xyzw[, 2L]
				private$mat_xyzw[, 3L] <- z_scale * private$mat_xyzw[, 3L]
				invisible(self)
			}
		},
		#' @param xy_shear `r r2i_transform3d_xy_shear`
		#' @param xz_shear `r r2i_transform3d_xz_shear`
		#' @param yx_shear `r r2i_transform3d_yx_shear`
		#' @param yz_shear `r r2i_transform3d_yz_shear`
		#' @param zx_shear `r r2i_transform3d_zx_shear`
		#' @param zy_shear `r r2i_transform3d_zy_shear`
		shear = function(
			xy_shear = 0,
			xz_shear = 0,
			yx_shear = 0,
			yz_shear = 0,
			zx_shear = 0,
			zy_shear = 0
		) {
			self$transform(shear3d(xy_shear, xz_shear, yx_shear, yz_shear, zx_shear, zy_shear))
		},
		#' @param x `r r2i_transform3d_x`
		#' @param ... Passed to `as_coord3d(x, ...)` if `x` is not a [Coord3D] object
		translate = function(x = as_coord3d(0, 0, 0), ...) {
			if (!is_coord3d(x)) {
				x <- as_coord3d(x, ...)
			}
			if (length(x) == 1L) {
				self$transform(translate3d(x))
			} else {
				private$apply_any_delayed_transformations()
				private$mat_xyzw[, 1L] <- private$mat_xyzw[, 1L] + x$x
				private$mat_xyzw[, 2L] <- private$mat_xyzw[, 2L] + x$y
				private$mat_xyzw[, 3L] <- private$mat_xyzw[, 3L] + x$z
				invisible(self)
			}
		},
		#' @param mat `r r2i_transform3d_mat`
		transform = function(mat = transform3d()) {
			if (!is_transform3d(mat)) {
				mat <- as_transform3d(mat)
			}
			if (is.null(private$mat_trans)) {
				private$mat_trans <- mat
			} else {
				private$mat_trans <- private$mat_trans %*% mat
			}
			invisible(self)
		}
	),
	active = list(
		#' @field xyzw A four-column matrix representing the homogeneous coordinates.
		#'             The first three columns are "x", "y", and "z" coordinates
		#'             and the fourth column is all ones.
		xyzw = function() {
			private$apply_any_delayed_transformations()
			private$mat_xyzw
		},
		#' @field x A numeric vector of x-coordinates.
		x = function() {
			as.vector(self$xyzw[, 1L])
		},
		#' @field y A numeric vector of y-coordinates.
		y = function() {
			as.vector(self$xyzw[, 2L])
		},
		#' @field z A numeric vector of z-coordinates.
		z = function() {
			as.vector(self$xyzw[, 3L])
		}
	),
	private = list(
		mat_xyzw = NULL,
		mat_trans = NULL,
		apply_any_delayed_transformations = function() {
			if (!is.null(private$mat_trans)) {
				private$mat_xyzw <- private$mat_xyzw %*% private$mat_trans
				private$mat_trans <- NULL
				colnames(private$mat_xyzw) <- c("x", "y", "z", "w")
			}
		}
	)
)

p1 <- as_coord2d(x = c(0, 2), y = c(0, 0))
p2 <- as_coord2d(x = c(2, 2), y = c(0, 2))
s <- as_segment2d(p1, p2 = p2)

test_that("as_segment2d() constructs a Segment2D", {
	expect_r6_class(s, "Segment2D")
	expect_r6_class(s, "Coord2D")
	expect_snapshot(print(s))
})

test_that("is_segment2d() works", {
	expect_true(is_segment2d(s))
	expect_false(is_segment2d(p1))
})

test_that("Segment2D$p1 returns first endpoints", {
	expect_r6_class(s$p1, "Coord2D")
	expect_equal(s$p1$x, c(0, 2))
	expect_equal(s$p1$y, c(0, 0))
})

test_that("Segment2D$p2 returns second endpoints", {
	expect_r6_class(s$p2, "Coord2D")
	expect_equal(s$p2$x, c(2, 2))
	expect_equal(s$p2$y, c(0, 2))
})

test_that("Segment2D$mid_point returns midpoints and caches", {
	m <- s$mid_point
	expect_r6_class(m, "Coord2D")
	expect_equal(m$x, c(1, 2))
	expect_equal(m$y, c(0, 1))
	expect_identical(s$mid_point, m)
})

test_that("length() returns number of segments", {
	expect_equal(length(s), 2L)
})

test_that("[.Segment2D subsets correctly", {
	s1 <- s[1]
	expect_r6_class(s1, "Segment2D")
	expect_equal(length(s1), 1L)
	expect_equal(s1$p1$x, 0)
	expect_equal(s1$p2$x, 2)
})

test_that("rep.Segment2D works", {
	sr <- rep(s, 2L)
	expect_equal(length(sr), 4L)
	expect_equal(sr$p1$x, c(0, 2, 0, 2))
})

test_that("as_segment2d() vec argument works", {
	vec <- as_coord2d(x = c(2, 0), y = c(0, 2))
	sv <- as_segment2d(p1, vec = vec)
	expect_equal(sv$p2$x, c(2, 2))
	expect_equal(sv$p2$y, c(0, 2))
})

test_that("as_segment2d.Polygon2D() returns edges", {
	poly <- as_polygon2d(as_coord2d(x = c(0, 1, 1, 0), y = c(0, 0, 1, 1)))
	expect_r6_class(as_segment2d(poly), "Segment2D")
	expect_equal(length(as_segment2d(poly)), 4L)
})

test_that("as_segment2d.default() errors", {
	expect_error(as_segment2d(1))
})

test_that("transformations update p1, p2, and mid_point; cache is cleared", {
	s2 <- as_segment2d(p1, p2 = p2)
	s2$translate(as_coord2d(1, 1))
	expect_equal(s2$p1$x, c(1, 3))
	expect_equal(s2$p1$y, c(1, 1))
	expect_equal(s2$p2$x, c(3, 3))
	expect_equal(s2$p2$y, c(1, 3))
	expect_equal(s2$mid_point$x, c(2, 3))
	expect_equal(s2$mid_point$y, c(1, 2))
})

test_that("rotation updates p2 and mid_point correctly", {
	s2 <- as_segment2d(
		as_coord2d(0, 0),
		p2 = as_coord2d(1, 0)
	)
	s2$rotate(degrees(90))
	expect_equal(s2$p1$x, 0, tolerance = 1e-9)
	expect_equal(s2$p1$y, 0, tolerance = 1e-9)
	expect_equal(s2$p2$x, 0, tolerance = 1e-9)
	expect_equal(s2$p2$y, 1, tolerance = 1e-9)
	expect_equal(s2$mid_point$x, 0, tolerance = 1e-9)
	expect_equal(s2$mid_point$y, 0.5, tolerance = 1e-9)
})

test_that("Polygon2D$edges returns a Segment2D of n edges", {
	poly <- as_polygon2d(as_coord2d(x = c(0, 1, 1, 0), y = c(0, 0, 1, 1)))
	e <- poly$edges
	expect_r6_class(e, "Segment2D")
	expect_equal(length(e), 4L)
	# first edge: (0,0) -> (1,0)
	expect_equal(e$p1$x[1], 0)
	expect_equal(e$p1$y[1], 0)
	expect_equal(e$p2$x[1], 1)
	expect_equal(e$p2$y[1], 0)
	# edges are cached
	expect_identical(poly$edges, e)
})

test_that("Polygon2D$edges cache is invalidated by transformation", {
	poly <- as_polygon2d(as_coord2d(x = c(0, 1, 1, 0), y = c(0, 0, 1, 1)))
	e1 <- poly$edges
	poly$translate(as_coord2d(1, 0))
	e2 <- poly$edges
	expect_equal(e2$p1$x[1], 1)
	expect_false(identical(e1, e2))
})

test_that("sort() on Segment2D orders farthest first by default via Coord2D inheritance", {
	p1 <- as_coord2d(x = c(0, 2, 1), y = c(0, 0, 2))
	p2 <- as_coord2d(x = c(2, 1, 0), y = c(0, 2, 0))
	s <- as_segment2d(p1, p2 = p2)
	s_sorted <- sort(s, alpha = degrees(45))
	depths <- painter_depth(s, scale = 1, alpha = degrees(45))
	expect_r6_class(s_sorted, "Segment2D")
	expect_equal(length(s_sorted), length(s))
	expect_equal(painter_depth(s_sorted, scale = 1, alpha = degrees(45)), sort(depths))
})

test_that("sort() on Segment2D with decreasing = TRUE gives closest first", {
	p1 <- as_coord2d(x = c(0, 2, 1), y = c(0, 0, 2))
	p2 <- as_coord2d(x = c(2, 1, 0), y = c(0, 2, 0))
	s <- as_segment2d(p1, p2 = p2)
	s_asc <- sort(s, alpha = degrees(45))
	s_desc <- sort(s, decreasing = TRUE, alpha = degrees(45))
	expect_equal(
		painter_depth(s_asc, scale = 1, alpha = degrees(45)),
		rev(painter_depth(s_desc, scale = 1, alpha = degrees(45)))
	)
})

test_that("as_ellipse2d()", {
	# from Coord2D
	e <- as_ellipse2d(as_coord2d(0, 0), rx = 2, ry = 1, theta = degrees(45))
	expect_r6_class(e, "Ellipse2D")
	expect_true(is_ellipse2d(e))
	expect_true(is_coord2d(e))
	expect_equal(length(e), 1L)
	expect_equal(e$rx, 2)
	expect_equal(e$ry, 1)
	expect_true(is_angle(e$theta))
	expect_equal(as.numeric(e$theta, "degrees"), 45)
	expect_false(e$is_circle)

	# from numeric
	e2 <- as_ellipse2d(x = c(0, 1), y = c(0, 1), rx = c(1, 2), ry = c(0.5, 1))
	expect_equal(length(e2), 2L)

	# recycling: one center, two radii
	e3 <- as_ellipse2d(as_coord2d(0, 0), rx = c(1, 2), ry = c(0.5, 1))
	expect_equal(length(e3), 2L)

	# [.angle preserves class when indexing theta
	expect_true(is_angle(e3$theta[1L]))

	# print
	expect_snapshot(print(e))
})

test_that("mixed circle/non-circle ellipses()", {
	e1 <- as_ellipse2d(as_coord2d(0, 0), rx = 2, ry = 1)
	expect_true(is_ellipse2d(e1))
	# mixed vector: not all circles → not all is_circle
	mixed <- as_ellipse2d(x = c(0, 1), y = c(0, 0), rx = c(1, 1), ry = c(1, 0.5))
	expect_false(all(mixed$is_circle))
})

test_that("Ellipse2D$is_circle active binding", {
	e <- as_ellipse2d(x = c(0, 1), y = c(0, 0), rx = c(1, 2), ry = c(1, 1))
	expect_equal(e$is_circle, c(TRUE, FALSE))
})

test_that("Ellipse2D transform() updates shape parameters", {
	# Uniform scaling: rx and ry both scaled
	c1 <- as_ellipse2d(as_coord2d(0, 0), r = 1)
	c1$scale(2)
	expect_equal(c1$rx, 2)
	expect_equal(c1$ry, 2)
	expect_true(c1$is_circle)

	# Non-uniform scaling turns circle into ellipse
	c2 <- as_ellipse2d(as_coord2d(0, 0), r = 1)
	c2$scale(2, 1)
	expect_equal(c2$rx, 2)
	expect_equal(c2$ry, 1)
	expect_false(c2$is_circle)

	# Rotation of a circle stays a circle; theta is still an angle
	c3 <- as_ellipse2d(as_coord2d(1, 0), r = 1)
	c3$rotate(degrees(45))
	expect_true(c3$is_circle)
	expect_equal(c3$rx, 1, tolerance = 1e-10)
	expect_equal(c3$ry, 1, tolerance = 1e-10)
	expect_true(is_angle(c3$theta))

	# Translation only moves center, not radii
	c4 <- as_ellipse2d(as_coord2d(0, 0), r = 1)
	c4$translate(as_coord2d(3, 4))
	expect_equal(c4$x, 3)
	expect_equal(c4$y, 4)
	expect_equal(c4$rx, 1)
})

test_that("as.list.Ellipse2D() and as.data.frame.Ellipse2D()", {
	e <- as_ellipse2d(as_coord2d(1, 2), rx = 3, ry = 2, theta = degrees(30))
	lst <- as.list(e)
	expect_equal(lst$x, 1)
	expect_equal(lst$y, 2)
	expect_equal(lst$rx, 3)
	expect_equal(lst$ry, 2)
	expect_true(is_angle(lst$theta))
	expect_equal(as.numeric(lst$theta, "degrees"), 30)

	df <- as.data.frame(e)
	expect_equal(names(df), c("x", "y", "rx", "ry", "theta"))
	expect_true(is_angle(df$theta))
})

test_that("as_polygon2d.Ellipse2D()", {
	c1 <- as_ellipse2d(as_coord2d(0, 0), r = 1)

	inner <- as_polygon2d(c1, n = 8L, type = "inner")
	expect_r6_class(inner, "Polygon2D")
	expect_true(inner$is_convex)
	expect_equal(length(inner), 8L)
	# inner polygon vertices lie on the circle: abs should equal 1
	expect_equal(abs(inner - as_coord2d(0, 0)), rep(1, 8L), tolerance = 1e-10)

	outer <- as_polygon2d(c1, n = 8L, type = "outer")
	expect_r6_class(outer, "Polygon2D")
	expect_true(outer$is_convex)
	expect_equal(length(outer), 8L)
	# outer polygon vertices lie outside the circle
	expect_true(all(abs(outer - as_coord2d(0, 0)) > 1))

	# inner contains circle, outer contains inner
	# (all inner vertices closer than all outer vertices to center)
	expect_true(
		all(abs(inner - as_coord2d(0, 0)) < abs(outer - as_coord2d(0, 0)))
	)

	# only length-1 Ellipse2D supported
	e2 <- as_ellipse2d(x = c(0, 1), y = c(0, 0), rx = c(1, 2), ry = c(1, 1))
	expect_error(as_polygon2d(e2))
})

test_that("has_overlap2d() circle-circle", {
	c1 <- as_ellipse2d(as_coord2d(0, 0), r = 1)
	c2 <- as_ellipse2d(as_coord2d(1.5, 0), r = 1)
	c3 <- as_ellipse2d(as_coord2d(3, 0), r = 1)

	expect_true(has_overlap2d(c1, c2))
	expect_false(has_overlap2d(c1, c3))

	# Tangent circles: no non-zero-area overlap
	c4 <- as_ellipse2d(as_coord2d(2, 0), r = 1)
	expect_false(has_overlap2d(c1, c4))

	# Vectorized
	c_many <- as_ellipse2d(x = c(1.5, 3), y = c(0, 0), r = c(1, 1))
	res <- has_overlap2d(c1, c_many)
	expect_equal(res, c(TRUE, FALSE))
})

test_that("has_overlap2d() ellipse-ellipse (non-circle branches)", {
	e <- as_ellipse2d(as_coord2d(0, 0), rx = 1, ry = 0.5)
	c_near <- as_ellipse2d(as_coord2d(0.5, 0), r = 0.4)
	c_far <- as_ellipse2d(as_coord2d(5, 0), r = 0.3)

	# e1 non-circle, e2 circle
	expect_true(has_overlap2d(e, c_near))
	expect_false(has_overlap2d(e, c_far))

	# e1 circle, e2 non-circle
	expect_true(has_overlap2d(c_near, e))
	expect_false(has_overlap2d(c_far, e))

	# both non-circle
	e_near <- as_ellipse2d(as_coord2d(0.5, 0), rx = 1, ry = 0.5)
	e_far <- as_ellipse2d(as_coord2d(5, 0), rx = 1, ry = 0.5)
	expect_true(has_overlap2d(e, e_near))
	expect_false(has_overlap2d(e, e_far))
})

test_that("has_overlap2d() ellipse-polygon (circle case uses SAT)", {
	sq <- as_polygon2d(as_coord2d(
		x = c(0, 1, 1, 0),
		y = c(0, 0, 1, 1)
	))
	circ_in <- as_ellipse2d(as_coord2d(0.5, 0.5), r = 0.3)
	circ_out <- as_ellipse2d(as_coord2d(3, 3), r = 0.3)

	expect_true(has_overlap2d(sq, circ_in))
	expect_false(has_overlap2d(sq, circ_out))

	# Dispatching from Ellipse2D side
	expect_true(has_overlap2d(circ_in, sq))
	expect_false(has_overlap2d(circ_out, sq))

	# Vectorized
	circs <- as_ellipse2d(x = c(0.5, 3), y = c(0.5, 3), r = c(0.3, 0.3))
	expect_equal(has_overlap2d(sq, circs), c(TRUE, FALSE))
})

test_that("has_overlap2d() non-circle ellipse uses bracketing", {
	sq <- as_polygon2d(as_coord2d(
		x = c(0, 1, 1, 0),
		y = c(0, 0, 1, 1)
	))
	# Clearly inside
	e_in <- as_ellipse2d(as_coord2d(0.5, 0.5), rx = 0.3, ry = 0.1)
	expect_true(has_overlap2d(e_in, sq))

	# Clearly outside
	e_out <- as_ellipse2d(as_coord2d(5, 5), rx = 0.3, ry = 0.1)
	expect_false(has_overlap2d(e_out, sq))
})

test_that("has_overlap2d.Ellipse2D() errors on unsupported y class", {
	circ <- as_ellipse2d(as_coord2d(0, 0), r = 1)
	expect_error(has_overlap2d(circ, "foo"), "Don't know how to check overlap with")
})

utils::globalVariables(c("group", "x", "xend", "y", "yend"))

#' @exportS3Method ggplot2::autolayer
autolayer.Coord1D <- function(x, ...) {
	ggplot2::geom_point(ggplot2::aes(x = x, y = y), ..., data = data.frame(x = x$x, y = 0))
}

#' @exportS3Method ggplot2::autolayer
autolayer.Coord2D <- function(x, ...) {
	ggplot2::geom_point(ggplot2::aes(x = x, y = y), ..., data = as.data.frame(x))
}

#' @exportS3Method ggplot2::autolayer
autolayer.Point1D <- function(x, ...) {
	ggplot2::geom_vline(xintercept = -x$b / x$a, ...)
}

#' @exportS3Method ggplot2::autolayer
autolayer.Polygon2D <- function(x, ...) {
	ggplot2::geom_polygon(ggplot2::aes(x = x, y = y), ..., data = data.frame(x = x$x, y = x$y))
}

#' @exportS3Method ggplot2::autolayer
autolayer.Ellipse2D <- function(x, n = 60L, ...) {
	polys <- lapply(seq_len(length(x)), function(i) {
		p <- as_polygon2d(x[i], n = n)
		data.frame(x = p$x, y = p$y, group = i)
	})
	df <- do.call(rbind, polys)
	ggplot2::geom_polygon(ggplot2::aes(x = x, y = y, group = group), ..., data = df)
}

#' @exportS3Method ggplot2::autolayer
autolayer.Segment2D <- function(x, ...) {
	df <- data.frame(x = x$p1$x, y = x$p1$y, xend = x$p2$x, yend = x$p2$y)
	ggplot2::geom_segment(ggplot2::aes(x = x, y = y, xend = xend, yend = yend), ..., data = df)
}

#' @exportS3Method ggplot2::autolayer
autolayer.Line2D <- function(x, ...) {
	index_v <- which(x$b == 0)
	l <- list()
	if (length(index_v)) {
		l[[1L]] <- ggplot2::geom_vline(xintercept = -x$c / x$a, ...)
		x <- x[-index_v]
	}
	if (length(x)) {
		slope <- -x$a / x$b
		intercept <- -x$c / x$b
		l[[2L]] <- ggplot2::geom_abline(slope = slope, intercept = intercept, ...)
	}
	l
}

#' @importFrom R6 R6Class
#' @importFrom utils hasName
NULL

# variant of `cli::is_utf8_output()` without the `knitr::is_latex_output()` check
is_utf8_output <- function() {
	opt <- getOption("cli.unicode", NULL)
	if (!is.null(opt)) {
		isTRUE(opt)
	} else {
		l10n_info()[["UTF-8"]]
	}
}

# pretty-print matrix output
print_mat <- function(x, ...) {
	print(apply(x, 2L, format, ...), quote = FALSE, right = TRUE)
}

#' Angle vector aware trigonometric functions
#'
#' `sine()`, `cosine()`, `tangent()`, `secant()`, `cosecant()`, and `cotangent()` are
#' [angle()] aware trigonometric functions that allow for a user chosen angular unit.
#'
#' @inheritParams angle
#' @return A numeric vector
#'
#' @examples
#' sine(pi, "radians")
#' cosine(180, "degrees")
#' tangent(0.5, "turns")
#'
#' a <- angle(0.5, "turns")
#' secant(a)
#' cosecant(a)
#' cotangent(a)
#' @name trigonometric-functions
NULL

#' @rdname trigonometric-functions
#' @export
sine <- function(x, unit = getOption("affiner_angular_unit", "degrees")) {
	if (!is_angle(x)) {
		x <- as_angle(x, unit = unit)
	}
	sin.angle(x)
}

#' @rdname trigonometric-functions
#' @export
cosine <- function(x, unit = getOption("affiner_angular_unit", "degrees")) {
	if (!is_angle(x)) {
		x <- as_angle(x, unit = unit)
	}
	cos.angle(x)
}

#' @rdname trigonometric-functions
#' @export
tangent <- function(x, unit = getOption("affiner_angular_unit", "degrees")) {
	if (!is_angle(x)) {
		x <- as_angle(x, unit = unit)
	}
	tan.angle(x)
}

#' @rdname trigonometric-functions
#' @export
secant <- function(x, unit = getOption("affiner_angular_unit", "degrees")) {
	if (!is_angle(x)) {
		x <- as_angle(x, unit = unit)
	}
	1 / sin.angle(x)
}

#' @rdname trigonometric-functions
#' @export
cosecant <- function(x, unit = getOption("affiner_angular_unit", "degrees")) {
	if (!is_angle(x)) {
		x <- as_angle(x, unit = unit)
	}
	1 / cos.angle(x)
}

#' @rdname trigonometric-functions
#' @export
cotangent <- function(x, unit = getOption("affiner_angular_unit", "degrees")) {
	if (!is_angle(x)) {
		x <- as_angle(x, unit = unit)
	}
	1 / tan.angle(x)
}

#' Angle vector aware inverse trigonometric functions
#'
#' `arcsine()`, `arccosine()`, `arctangent()`,
#' `arcsecant()`, `arccosecant()`, and `arccotangent()` are
#' inverse trigonometric functions that return [angle()] vectors
#' with a user chosen angular unit.
#'
#' @inheritParams angle
#' @param x A numeric vector
#' @return An [angle()] vector
#'
#' @examples
#' arccosine(-1, "degrees")
#' arcsine(0, "turns")
#' arctangent(0, "gradians")
#' arccosecant(-1, "degrees")
#' arcsecant(1, "degrees")
#' arccotangent(1, "half-turns")
#'
#' # `base::atan2(y, x)` computes the angle of the vector from origin to (x, y)
#' as_angle(as_coord2d(x = 1, y = 1), "degrees")
#' @name inverse-trigonometric-functions
NULL

#' @rdname inverse-trigonometric-functions
#' @param tolerance If `x` greater than 1 (or less than -1) but is within a `tolerance`
#'                  of 1 (or -1) then it will be treated as 1 (or -1)
#' @export
arcsine <- function(
	x,
	unit = getOption("affiner_angular_unit", "degrees"),
	tolerance = sqrt(.Machine$double.eps)
) {
	idp1 <- which(x > 1 & x < 1 + tolerance)
	x[idp1] <- 1
	idn1 <- which(x < -1 & x > -1 - tolerance)
	x[idn1] <- -1
	unit <- standardize_angular_unit(unit)
	new_angle(from_radians(asin(x), unit), unit)
}

#' @rdname inverse-trigonometric-functions
#' @export
arccosine <- function(
	x,
	unit = getOption("affiner_angular_unit", "degrees"),
	tolerance = sqrt(.Machine$double.eps)
) {
	idp1 <- which(x > 1 & x < 1 + tolerance)
	x[idp1] <- 1
	idn1 <- which(x < -1 & x > -1 - tolerance)
	x[idn1] <- -1
	unit <- standardize_angular_unit(unit)
	new_angle(from_radians(acos(x), unit), unit)
}

#' @rdname inverse-trigonometric-functions
#' @param y A numeric vector or `NULL`.
#'          If `NULL` (default) we compute the 1-argument arctangent
#'          else we compute the 2-argument arctangent.
#'          For positive coordinates `(x, y)` then `arctangent(x = y/x) == arctangent(x = x, y = y)`.
#' @export
arctangent <- function(x, unit = getOption("affiner_angular_unit", "degrees"), y = NULL) {
	unit <- standardize_angular_unit(unit)
	if (is.null(y)) {
		a <- from_radians(atan(x), unit)
	} else {
		a <- from_radians(atan2(y, x), unit)
	}
	new_angle(a, unit)
}

#' @rdname inverse-trigonometric-functions
#' @export
arcsecant <- function(x, unit = getOption("affiner_angular_unit", "degrees")) {
	arcsine(1 / x, unit)
}

#' @rdname inverse-trigonometric-functions
#' @export
arccosecant <- function(x, unit = getOption("affiner_angular_unit", "degrees")) {
	arccosine(1 / x, unit)
}

#' @rdname inverse-trigonometric-functions
#' @export
arccotangent <- function(x, unit = getOption("affiner_angular_unit", "degrees")) {
	arctangent(1 / x, unit)
}

test_that("trigonometry functions", {
	expect_equal(cosine(angle(0.5, "turns")), -1)

	expect_equal(cosine(200, "gradians"), -1)
	expect_equal(sine(pi, "radians"), 0)
	expect_equal(tangent(1, "pi-radians"), 0)
	expect_equal(secant(90, "degrees"), 1)
	expect_equal(cosecant(180, "degrees"), -1)
	expect_equal(cotangent(0.25, "half-turns"), 1)

	expect_equal(arccosine(-1, "degrees"), angle(180, "degrees"))
	expect_equal(arcsine(0, "turns"), angle(0, "turns"))
	expect_equal(arctangent(0, "gradians"), angle(0, "gradians"))
	expect_equal(arctangent(x = 0, y = -1, unit = "degrees"), angle(-90, "degrees"))
	expect_equal(arctangent(x = -1, y = 0, unit = "degrees"), angle(180, "degrees"))
	expect_equal(arccosecant(-1, "degrees"), angle(180, "degrees"))
	expect_equal(arcsecant(1, "degrees"), angle(90, "degrees"))
	expect_equal(arccotangent(1, "half-turns"), angle(0.25, "half-turns"))
})

#' @export
length.Point1D <- function(x) {
	length(x$a)
}

#' @export
length.Line2D <- function(x) {
	length(x$a)
}

#' @export
length.Plane3D <- function(x) {
	length(x$a)
}

#' @export
`[.Point1D` <- function(x, i) {
	Point1D$new(x$a[i], x$b[i])
}

#' @export
`[.Line2D` <- function(x, i) {
	Line2D$new(x$a[i], x$b[i], x$c[i])
}

#' @export
`[.Plane3D` <- function(x, i) {
	Plane3D$new(x$a[i], x$b[i], x$c[i], x$d[i])
}

#' @export
as.data.frame.Point1D <- function(x, ...) {
	chkDots(...)
	data.frame(a = x$a, b = x$b)
}

#' @export
as.data.frame.Line2D <- function(x, ...) {
	chkDots(...)
	data.frame(a = x$a, b = x$b, c = x$c)
}

#' @export
as.data.frame.Plane3D <- function(x, ...) {
	chkDots(...)
	data.frame(a = x$a, b = x$b, c = x$c, d = x$d)
}

#' @export
as.list.Point1D <- function(x, ...) {
	chkDots(...)
	list(a = x$a, b = x$b)
}

#' @export
as.list.Line2D <- function(x, ...) {
	chkDots(...)
	list(a = x$a, b = x$b, c = x$c)
}

#' @export
as.list.Plane3D <- function(x, ...) {
	chkDots(...)
	list(a = x$a, b = x$b, c = x$c, d = x$d)
}

#' @export
rep.Point1D <- function(x, ..., length.out = NA_integer_) {
	if (isTRUE(length(x) == length.out)) {
		return(x)
	}
	id <- rep(seq.int(length(x)), ..., length.out = length.out)
	Point1D$new(x$a[id], x$b[id])
}

#' @export
rep.Line2D <- function(x, ..., length.out = NA_integer_) {
	if (isTRUE(length(x) == length.out)) {
		return(x)
	}
	id <- rep(seq.int(length(x)), ..., length.out = length.out)
	Line2D$new(x$a[id], x$b[id], x$c[id])
}

#' @export
rep.Plane3D <- function(x, ..., length.out = NA_integer_) {
	if (isTRUE(length(x) == length.out)) {
		return(x)
	}
	id <- rep(seq.int(length(x)), ..., length.out = length.out)
	Plane3D$new(x$a[id], x$b[id], x$c[id], x$d[id])
}

#' @export
c.Point1D <- function(...) {
	l <- list(...)
	stopifnot(all(vapply(l, is_point1d, logical(1))))
	a <- unlist(lapply(l, function(x) x$a))
	b <- unlist(lapply(l, function(x) x$b))
	Point1D$new(a, b)
}

#' @export
c.Line2D <- function(...) {
	l <- list(...)
	stopifnot(all(vapply(l, is_line2d, logical(1))))
	a <- unlist(lapply(l, function(x) x$a))
	b <- unlist(lapply(l, function(x) x$b))
	c <- unlist(lapply(l, function(x) x$c))
	Line2D$new(a, b, c)
}

#' @export
c.Plane3D <- function(...) {
	l <- list(...)
	stopifnot(all(vapply(l, is_plane3d, logical(1))))
	a <- unlist(lapply(l, function(x) x$a))
	b <- unlist(lapply(l, function(x) x$b))
	c <- unlist(lapply(l, function(x) x$c))
	d <- unlist(lapply(l, function(x) x$d))
	Plane3D$new(a, b, c, d)
}

#' @export
is.na.Point1D <- function(x) is.na(x$a) | is.na(x$b)

#' @export
is.na.Line2D <- function(x) is.na(x$a) | is.na(x$b) | is.na(x$c)

#' @export
is.na.Plane3D <- function(x) is.na(x$a) | is.na(x$b) | is.na(x$c) | is.na(x$d)

#' @export
is.nan.Point1D <- function(x) is.nan(x$a) | is.nan(x$b)

#' @export
is.nan.Line2D <- function(x) is.nan(x$a) | is.nan(x$b) | is.nan(x$c)

#' @export
is.nan.Plane3D <- function(x) is.nan(x$a) | is.nan(x$b) | is.nan(x$c) | is.nan(x$d)

#' @export
is.finite.Point1D <- function(x) is.finite(x$a) & is.finite(x$b)

#' @export
is.finite.Line2D <- function(x) is.finite(x$a) & is.finite(x$b) & is.finite(x$c)

#' @export
is.finite.Plane3D <- function(x) is.finite(x$a) & is.finite(x$b) & is.finite(x$c) & is.finite(x$d)

#' @export
is.infinite.Point1D <- function(x) is.infinite(x$a) | is.infinite(x$b)

#' @export
is.infinite.Line2D <- function(x) is.infinite(x$a) | is.infinite(x$b) | is.infinite(x$c)

#' @export
is.infinite.Plane3D <- function(x) {
	is.infinite(x$a) | is.infinite(x$b) | is.infinite(x$c) | is.infinite(x$d)
}

# "equality" is tricky unless we "normalize" the lines
# `==.Line2D`
# `==.Plane3D`
# `!=.Line2D`
# `!=.Plane3D`

#' Cast to coord1d object
#'
#' `as_coord1d()` casts to a [Coord1D] class object
#'
#' @param x An object that can be cast to a [Coord1D] class object
#'          such as a numeric vector of x-coordinates.
#' @param ... Further arguments passed to or from other methods
#' @return A [Coord1D] class object
#' @examples
#' as_coord1d(x = rnorm(10))
#' @export
as_coord1d <- function(x, ...) {
	UseMethod("as_coord1d")
}

#' Cast to coord2d object
#'
#' `as_coord2d()` casts to a [Coord2D] class object
#'
#' @param x An object that can be cast to a [Coord2D] class object
#'          such as a matrix or data frame of coordinates.
#' @param ... Further arguments passed to or from other methods
#' @return A [Coord2D] class object
#' @examples
#' df <- data.frame(x = sample.int(10, 3),
#'                  y = sample.int(10, 3))
#' as_coord2d(df)
#' as_coord2d(complex(real = 3, imaginary = 2))
#' as_coord2d(angle(90, "degrees"), radius = 2)
#' as_coord2d(as_coord3d(1, 2, 2), alpha = degrees(90), scale = 0.5)
#' @export
as_coord2d <- function(x, ...) {
	UseMethod("as_coord2d")
}

#' Cast to coord3d object
#'
#' `as_coord3d()` casts to a [Coord3D] class object
#'
#' @param x An object that can be cast to a [Coord3D] class object
#'          such as a matrix or data frame of coordinates.
#' @param ... Further arguments passed to or from other methods
#' @return A [Coord3D] class object
#' @examples
#' as_coord3d(x = 1, y = 2, z = 3)
#' df <- data.frame(x = sample.int(10, 3),
#'                  y = sample.int(10, 3),
#'                  z = sample.int(10, 3))
#' as_coord3d(df)
#' # Cylindrical coordinates
#' as_coord3d(degrees(90), z = 1, radius = 1)
#' # Spherical coordinates
#' as_coord3d(degrees(90), inclination = degrees(90), radius = 1)
#' @export
as_coord3d <- function(x, ...) {
	UseMethod("as_coord3d")
}

#' @rdname as_coord2d
#' @param radius A numeric vector of radial distances.
#' @export
as_coord2d.angle <- function(x, radius = 1, ...) {
	n <- max(length(x), length(radius))
	x <- rep_len(x, n)
	radius <- rep_len(radius, n)
	as_coord2d(radius * cos(x), radius * sin(x))
}

#' @rdname as_coord3d
#' @param radius A numeric vector.  If `inclination` is not `NULL` represents spherical distances
#'               of spherical coordinates and if `z` is not `NULL` represents
#'               radial distances of cylindrical coordinates.
#' @param inclination Spherical coordinates inclination angle aka polar angle.
#'                    `x` represents the azimuth aka azimuthal angle.
#' @export
as_coord3d.angle <- function(x, radius = 1, inclination = NULL, z = NULL, ...) {
	stopifnot(!(is.null(inclination) && is.null(z)), is.null(inclination) || is.null(z))
	if (!is.null(inclination) && !is_angle(inclination)) {
		inclination <- as_angle(inclination, ...)
	}
	if (!is.null(z) && !is.numeric(z)) {
		z <- as.numeric(z)
	}
	if (is.null(inclination)) {
		# cylindrical coordinates
		as_coord3d(as_coord2d(x, radius = radius), z = z)
	} else {
		# spherical coordinates
		as_coord3d(
			x = radius * sin(inclination) * cos(x),
			y = radius * sin(inclination) * sin(x),
			z = radius * cos(inclination)
		)
	}
}

#' @rdname as_coord1d
#' @export
as_coord1d.character <- function(x, ...) {
	xc <- vapply(x, as_coord1d_character_x, double(1), USE.NAMES = FALSE)
	p <- as_coord1d(xc)
	if (any(is.na(p) & !is.na(x))) {
		warning("NAs introduced by coercion")
	}
	p
}

as_coord1d_character_x <- function(x) {
	switch(x, "origin" = 0, NA_real_)
}

#' @rdname as_coord2d
#' @export
as_coord2d.character <- function(x, ...) {
	xc <- vapply(x, as_coord2d_character_x, double(1), USE.NAMES = FALSE)
	yc <- vapply(x, as_coord2d_character_y, double(1), USE.NAMES = FALSE)
	p <- as_coord2d(xc, yc)
	if (any(is.na(p) & !is.na(x))) {
		warning("NAs introduced by coercion")
	}
	p
}

as_coord2d_character_x <- function(x) {
	switch(x, "origin" = 0, "x-axis" = 1, "y-axis" = 0, NA_real_)
}

as_coord2d_character_y <- function(x) {
	switch(x, "origin" = 0, "x-axis" = 0, "y-axis" = 1, NA_real_)
}

#' @rdname as_coord3d
#' @export
as_coord3d.character <- function(x, ...) {
	xc <- vapply(x, as_coord3d_character_x, double(1), USE.NAMES = FALSE)
	yc <- vapply(x, as_coord3d_character_y, double(1), USE.NAMES = FALSE)
	zc <- vapply(x, as_coord3d_character_z, double(1), USE.NAMES = FALSE)
	p <- as_coord3d(xc, yc, zc)
	if (any(is.na(p) & !is.na(x))) {
		warning("NAs introduced by coercion")
	}
	p
}

as_coord3d_character_x <- function(x) {
	switch(x, "origin" = 0, "x-axis" = 1, "y-axis" = 0, "z-axis" = 0, NA_real_)
}

as_coord3d_character_y <- function(x) {
	switch(x, "origin" = 0, "x-axis" = 0, "y-axis" = 1, "z-axis" = 0, NA_real_)
}

as_coord3d_character_z <- function(x) {
	switch(x, "origin" = 0, "x-axis" = 0, "y-axis" = 0, "z-axis" = 1, NA_real_)
}

#' @rdname as_coord2d
#' @export
as_coord2d.complex <- function(x, ...) {
	as_coord2d(Re(x), Im(x))
}

# #' `as_coord1d.Coord2D()` computes the 1D projection of a [Coord2D] object
# #' onto a line.  By default will do an orthographic projection onto the x-axis
# #' but can do oblique projections onto arbitrary lines.

#' @rdname as_coord1d
#' @inheritParams project2d
#' @export
as_coord1d.Coord2D <- function(
	x,
	permutation = c("xy", "yx"),
	...,
	line = as_line2d("x-axis"),
	scale = 0
) {
	if (!is_line2d(line)) {
		line <- as_line2d(line, ...)
	}
	stopifnot(length(line) == 1, scale == 0 || line$a == 0)
	permutation <- match.arg(permutation)
	denom <- line$a^2 + line$b^2
	closest <- as_coord2d(-line$a * line$c / denom, -line$b * line$c / denom)
	theta <- as_angle(line)
	xs <- x$clone()$permute(permutation)$translate(-closest)$rotate(-theta)$shear(
		xy_shear = scale
	)$x
	as_coord1d.numeric(x = xs)
}

#' @rdname as_coord2d
#' @param permutation Either "xyz" (no permutation), "xzy" (permute y and z axes),
#'                    "yxz" (permute x and y axes), "yzx" (x becomes z, y becomes x, z becomes y),
#'                    "zxy" (x becomes y, y becomes z, z becomes x), "zyx" (permute x and z axes).
#'                    This permutation is applied before the (oblique) projection.
#' @param plane A [Plane3D] class object representing the plane
#'         you wish to project to or an object coercible to one using `as_plane3d(plane, ...)`
#'         such as "xy-plane", "xz-plane", or "yz-plane".
#' @param scale Oblique projection foreshortening scale factor.
#'   A (degenerate) `0` value indicates an orthographic projection.
#'   A value of `0.5` is used by a \dQuote{cabinet projection}
#'   while a value of `1.0` is used by a \dQuote{cavalier projection}.
#' @param alpha Oblique projection angle (the angle the third axis is projected going off at).
#'              An [angle()] object or one coercible to one with `as_angle(alpha, ...)`.
#'              Popular angles are 45 degrees, 60 degrees, and `arctangent(2)` degrees.
#' @param roll Rotation of the in-plane coordinate frame around the plane normal
#'             after the azimuth/inclination alignment.
#'             An [angle()] object or one coercible to one with `as_angle(roll, ...)`.
#'             Defaults to `angle(0)` (no roll), which preserves the azimuth/inclination convention.
#' @export
as_coord2d.Coord3D <- function(
	x,
	permutation = c("xyz", "xzy", "yxz", "yzx", "zyx", "zxy"),
	...,
	plane = as_plane3d("xy-plane"),
	scale = 0,
	alpha = angle(45, "degrees"),
	roll = angle(0, "degrees")
) {
	if (!is_plane3d(plane)) {
		plane <- as_plane3d(plane, ...)
	}
	if (!is_angle(alpha)) {
		alpha <- as_angle(alpha, ...)
	}
	if (!is_angle(roll)) {
		roll <- as_angle(roll, ...)
	}
	stopifnot(length(plane) == 1, length(alpha) == 1, length(roll) == 1)
	permutation <- match.arg(permutation)

	denom <- plane$a^2 + plane$b^2 + plane$c^2
	closest <- as_coord3d(
		-plane$a * plane$d / denom,
		-plane$b * plane$d / denom,
		-plane$c * plane$d / denom
	)
	azimuth <- as_angle(plane, type = "azimuth")
	inclination <- as_angle(plane, type = "inclination")
	z_axis <- Coord3D$new(matrix(
		c(0, 0, 1, 1),
		nrow = 1,
		dimnames = list(NULL, c("x", "y", "z", "w"))
	))
	y_axis <- Coord3D$new(matrix(
		c(0, 1, 0, 1),
		nrow = 1,
		dimnames = list(NULL, c("x", "y", "z", "w"))
	))
	p <- x$clone()$permute(permutation)$translate(-closest)$rotate(z_axis, -azimuth)$rotate(
		y_axis,
		-inclination
	)$rotate(z_axis, roll)$shear(xz_shear = scale * cos(alpha), yz_shear = scale * sin(alpha))
	as_coord2d(p$x, p$y)
}

#' @rdname as_coord1d
#' @export
as_coord1d.data.frame <- function(x, ...) {
	stopifnot(hasName(x, "x"))
	as_coord1d(x[, "x"])
}

#' @rdname as_coord2d
#' @export
as_coord2d.data.frame <- function(x, ...) {
	stopifnot(all(hasName(x, c("x", "y"))))
	Coord2D$new(as_xyw_matrix(x[, c("x", "y")]))
}

#' @rdname as_coord3d
#' @export
as_coord3d.data.frame <- function(x, ..., z = NULL) {
	stopifnot(all(hasName(x, c("x", "y"))), is.null(z) || !hasName(x, "z"))
	if (!is.null(z)) {
		x$z <- z
	}
	if (hasName(x, "z")) {
		nms <- c("x", "y", "z")
	} else {
		nms <- c("x", "y")
	}
	Coord3D$new(as_xyzw_matrix(x[, nms]))
}

#' @rdname as_coord1d
#' @export
as_coord1d.list <- function(x, ...) {
	as_coord1d.data.frame(as.data.frame(x, ...))
}

#' @rdname as_coord2d
#' @export
as_coord2d.list <- function(x, ...) {
	as_coord2d.data.frame(as.data.frame(x, ...))
}

#' @rdname as_coord3d
#' @export
as_coord3d.list <- function(x, ..., z = NULL) {
	if (is.null(z)) {
		as_coord3d.data.frame(as.data.frame(x, ...))
	} else {
		as_coord3d.data.frame(as.data.frame(x, ...), z = z)
	}
}

#' @rdname as_coord1d
#' @export
as_coord1d.matrix <- function(x, ...) {
	Coord1D$new(as_xw_matrix(x))
}

#' @rdname as_coord2d
#' @export
as_coord2d.matrix <- function(x, ...) {
	Coord2D$new(as_xyw_matrix(x))
}

#' @rdname as_coord3d
#' @export
as_coord3d.matrix <- function(x, ...) {
	Coord3D$new(as_xyzw_matrix(x))
}

#' @rdname as_coord1d
#' @export
as_coord1d.numeric <- function(x, ...) {
	xw <- cbind(x, rep_len(1, length(x)))
	Coord1D$new(as_xw_matrix(xw))
}

as_xw_matrix <- function(x) {
	stopifnot(ncol(x) == 1L || ncol(x) == 2L, is.numeric(x))
	if (ncol(x) < 2L) {
		x <- cbind(x, 1)
	} else {
		stopifnot(all(x[, 2L] == 1))
	}
	colnames(x) <- c("x", "w")
	x
}

#' @rdname as_coord2d
#' @param y Numeric vector of y-coordinates to be used.
#' @export
as_coord2d.numeric <- function(x, y = rep_len(0, length(x)), ...) {
	xyw <- cbind(x, y, rep_len(1, max(length(x), length(y))))
	Coord2D$new(as_xyw_matrix(xyw))
}

as_xyw_matrix <- function(x) {
	if (!is.matrix(x)) {
		x <- as.matrix(x)
	}
	stopifnot(ncol(x) == 2L || ncol(x) == 3L, is.numeric(x))
	if (ncol(x) < 3L) {
		x <- cbind(x, 1)
	} else {
		stopifnot(all(x[, 3L] == 1))
	}
	colnames(x) <- c("x", "y", "w")
	x
}

#' @rdname as_coord3d
#' @param y Numeric vector of y-coordinates to be used
#'          if `hasName(x, "z")` is `FALSE`.
#' @export
as_coord3d.numeric <- function(x, y = rep_len(0, length(x)), z = rep_len(0, length(x)), ...) {
	xyzw <- cbind(x, y, z, rep_len(1, max(length(x), length(y), length(z))))
	Coord3D$new(as_xyzw_matrix(xyzw))
}

as_xyzw_matrix <- function(x) {
	if (!is.matrix(x)) {
		x <- as.matrix(x)
	}
	stopifnot(ncol(x) >= 2L, ncol(x) <= 4L, is.numeric(x))
	if (ncol(x) == 2L) {
		x <- cbind(x, 0, 1)
	} else if (ncol(x) == 3L) {
		x <- cbind(x, 1)
	} else {
		stopifnot(all(x[, 4L] == 1))
	}
	colnames(x) <- c("x", "y", "z", "w")
	x
}

#' @rdname as_coord1d
#' @export
as_coord1d.Coord1D <- function(x, ...) {
	Coord1D$new(x$xw)
}

#' @rdname as_coord2d
#' @export
as_coord2d.Coord2D <- function(x, ...) {
	Coord2D$new(x$xyw)
}

#' @rdname as_coord3d
#' @export
as_coord3d.Coord3D <- function(x, ...) {
	Coord3D$new(x$xyzw)
}

#' @rdname as_coord3d
#' @param z Numeric vector of z-coordinates to be used
#' @export
as_coord3d.Coord2D <- function(x, z = rep_len(0, length(x)), ...) {
	as_coord3d(x = x$x, y = x$y, z = z)
}

#' @rdname as_coord1d
#' @export
as_coord1d.Point1D <- function(x, ...) {
	as_coord1d(-x$b / x$a)
}

test_that("distance1d()", {
	x <- c(2, 5, 7)
	p1 <- as_coord1d(x = x)
	p0 <- as_coord1d("origin")
	expect_equal(distance1d(p1, p0), abs(p1))

	pt1 <- as_point1d(a = 1, b = -2)
	expect_equal(distance1d(pt1, p0), 2)
	pt2 <- as_point1d(a = 1, b = 3)
	expect_equal(distance1d(p0, pt2), 3)

	expect_error(distance1d(as_coord3d("origin"), p0))
	expect_error(distance1d(p0, as_coord3d("origin")))
	expect_error(distance1d(pt1, as_coord3d("origin")))
})

test_that("distance2d()", {
	x <- c(2, 5, 7)
	y <- c(3, 4, 6)
	p2 <- as_coord2d(x = x, y = y)
	p0 <- as_coord2d("origin")
	expect_equal(distance2d(p2, p0), abs(p2))

	l1 <- as_line2d(a = 0, b = 2, c = 4)
	expect_equal(distance2d(l1, p0), 2)
	l2 <- as_line2d(a = 2, b = 0, c = 4)
	expect_equal(distance2d(p0, l2), 2)

	expect_error(distance2d(as_coord3d("origin"), p0))
	expect_error(distance2d(p0, as_coord3d("origin")))
	expect_error(distance2d(l1, as_coord3d("origin")))
})

test_that("distance3d()", {
	x <- c(2, 5, 7)
	y <- c(3, 4, 6)
	z <- c(2, 7, 3)
	p3 <- as_coord3d(x = x, y = y, z = z)
	p0 <- as_coord3d("origin")

	expect_equal(distance3d(p3, p0), abs(p3))

	pl1 <- as_plane3d(a = 0, b = 2, c = 0, d = 4)
	expect_equal(distance3d(pl1, p0), 2)

	pl2 <- as_plane3d(a = 0, b = 0, c = 2, d = 4)
	expect_equal(distance3d(p0, pl2), 2)

	expect_error(distance3d(pl2, "foobar"))

	expect_error(distance3d(as_coord2d("origin"), p0))
	expect_error(distance3d(p0, as_coord2d("origin")))
})

test_that("`plot()` and `autolayer()` methods", {
	skip_if_not_installed("ggplot2")
	skip_if_not_installed("vdiffr")
	suppressPackageStartupMessages(library("ggplot2"))
	library("vdiffr")

	p1 <- as_coord1d(x = seq.int(-4, -1))
	pt <- as_point1d(a = 1, b = 0)
	p2 <- p1$clone()$reflect(pt)
	expect_doppelganger("plot.Coord1D", function() {
		plot(p1, xlim = c(-5, 5))
		lines(pt)
		points(p2, col = "red")
	})
	expect_doppelganger("autolayer.Coord1D", {
		ggplot() +
			autolayer(p1) +
			autolayer(pt) +
			autolayer(p2, color = "red")
	})

	p1 <- as_coord2d(x = 0, y = 1:10)
	l <- as_line2d(a = 1, b = -1, c = 0)
	p2 <- p1$clone()$reflect(l)
	expect_doppelganger("plot.Coord2D reflect", function() {
		plot(p1, xlim = c(-1, 11), ylim = c(-1, 11))
		lines(l)
		points(p2, col = "red")
	})
	expect_doppelganger("autolayer.Coord2D reflect", {
		ggplot() +
			autolayer(p1) +
			autolayer(l) +
			autolayer(p2, color = "red")
	})

	p1 <- as_coord2d(x = 1:10, y = 1:10)
	l <- as_line2d(a = -1, b = 0, c = 0)
	p2 <- p1$clone()$project(l)
	expect_doppelganger("plot.Coord2D project", function() {
		plot(p1, xlim = c(-1, 11))
		lines(l)
		points(p2, col = "red")
	})
	expect_doppelganger("autolayer.Coord2D project", {
		ggplot() +
			autolayer(p1) +
			autolayer(l) +
			autolayer(p2, color = "red")
	})
})

test_that("lines.Coord2D, plot.Polygon2D, lines/plot.Ellipse2D", {
	skip_if_not_installed("vdiffr")
	library("vdiffr")

	pts <- as_coord2d(x = c(0, 0.5, 1, 0.5), y = c(0.5, 1, 0.5, 0))
	sq <- as_polygon2d(as_coord2d(x = c(0, 1, 1, 0), y = c(0, 0, 1, 1)))
	circ <- as_ellipse2d(as_coord2d(0.5, 0.5), r = 0.4)
	ell <- as_ellipse2d(as_coord2d(0.5, 0.5), rx = 0.4, ry = 0.2, theta = pi / 6)

	expect_doppelganger("lines.Coord2D", function() {
		plot(pts, xlim = c(0, 1), ylim = c(0, 1))
		lines(pts)
	})
	expect_doppelganger("lines.Polygon2D", function() {
		plot(pts, xlim = c(0, 1), ylim = c(0, 1))
		lines(sq)
	})
	expect_doppelganger("plot.Polygon2D", function() {
		plot(sq, col = "lightblue")
	})
	expect_doppelganger("lines.Ellipse2D circle", function() {
		plot(sq, col = "lightblue")
		lines(circ, col = "red")
	})
	expect_doppelganger("lines.Ellipse2D ellipse", function() {
		plot(sq, col = "lightblue")
		lines(ell, col = "darkgreen")
	})
	expect_doppelganger("plot.Ellipse2D circle", function() {
		plot(circ, col = "lightyellow")
	})
	expect_doppelganger("plot.Ellipse2D multiple", function() {
		e2 <- as_ellipse2d(x = c(0, 1), y = c(0, 1), rx = c(0.4, 0.3), ry = c(0.2, 0.3))
		plot(e2, col = "lightblue")
	})
})

test_that("autolayer.Polygon2D and autolayer.Ellipse2D", {
	skip_if_not_installed("ggplot2")
	skip_if_not_installed("vdiffr")
	suppressPackageStartupMessages(library("ggplot2"))
	library("vdiffr")

	sq <- as_polygon2d(as_coord2d(x = c(0, 1, 1, 0), y = c(0, 0, 1, 1)))
	circ <- as_ellipse2d(as_coord2d(0.5, 0.5), r = 0.4)
	ell <- as_ellipse2d(as_coord2d(0.5, 0.5), rx = 0.4, ry = 0.2, theta = pi / 6)

	expect_doppelganger("autolayer.Polygon2D", {
		ggplot() + autolayer(sq, fill = "lightblue")
	})
	expect_doppelganger("autolayer.Ellipse2D circle", {
		ggplot() + autolayer(circ, fill = "lightyellow")
	})
	expect_doppelganger("autolayer.Ellipse2D multiple", {
		e2 <- as_ellipse2d(x = c(0, 1), y = c(0, 1), rx = c(0.4, 0.3), ry = c(0.2, 0.3))
		ggplot() + autolayer(e2, fill = "lightblue")
	})
})

test_that("plot.Segment2D, lines.Segment2D, autolayer.Segment2D", {
	skip_if_not_installed("vdiffr")
	library("vdiffr")

	s <- as_segment2d(
		as_coord2d(x = c(0, 0), y = c(0, 1)),
		p2 = as_coord2d(x = c(1, 1), y = c(1, 0))
	)
	sq <- as_polygon2d(as_coord2d(x = c(0, 1, 1, 0), y = c(0, 0, 1, 1)))

	expect_doppelganger("plot.Segment2D", function() {
		plot(s)
	})
	expect_doppelganger("lines.Segment2D", function() {
		plot(sq, col = "lightblue")
		lines(s, col = "red")
	})
})

test_that("autolayer.Segment2D", {
	skip_if_not_installed("ggplot2")
	skip_if_not_installed("vdiffr")
	suppressPackageStartupMessages(library("ggplot2"))
	library("vdiffr")

	s <- as_segment2d(
		as_coord2d(x = c(0, 0), y = c(0, 1)),
		p2 = as_coord2d(x = c(1, 1), y = c(1, 0))
	)

	expect_doppelganger("autolayer.Segment2D", {
		ggplot() + autolayer(s, color = "blue")
	})
})

test_that("rgl works", {
	skip_on_cran()
	skip_if_not_installed("rgl")
	rgl::open3d(useNULL = TRUE)
	c1 <- as_coord3d(x = 1:10, y = 1:10, z = 1:10)
	pl <- as_plane3d(a = 0, b = 0, c = -1, d = 2) # z = 2
	c2 <- c1$clone()$reflect(pl)
	rgl::plot3d(c1, size = 8)
	rgl::plot3d(pl, color = "grey", alpha = 0.6)
	rgl::points3d(as.data.frame(c2), col = "red", size = 8)
	expect_true(nrow(rgl::ids3d()) >= 6L)
	rgl::close3d()
})

test_that("`has_intersection()`", {
	line1 <- as_line2d("x-axis")
	line2 <- as_line2d("y-axis")
	line3 <- as_line2d(a = 0, b = 1, c = 2) # y + 2 = 0
	expect_true(has_intersection(line1, line1))
	expect_true(has_intersection(line1, line2))
	expect_true(has_intersection(line1, "y-axis"))
	expect_false(has_intersection(line1, line3))

	plane1 <- as_plane3d("xy-plane")
	plane2 <- as_plane3d("xz-plane")
	plane3 <- as_plane3d(a = 0, b = 0, c = 1, d = 2) # z + 2 = 0
	expect_true(has_intersection(plane1, plane1))
	expect_true(has_intersection(plane1, plane2))
	expect_true(has_intersection(plane1, "xz-plane"))
	expect_false(has_intersection(plane1, plane3))

	point1d <- as_point1d("origin")
	point2d <- as_point1d(a = 1, b = 2)
	expect_true(has_intersection(point1d, "origin"))
	expect_false(has_intersection(point1d, point2d))

	p1 <- as_coord1d("origin")
	p2 <- as_coord1d(x = 3)
	expect_true(has_intersection(p1, "origin"))
	expect_false(has_intersection(p1, p2))

	p1 <- as_coord2d("origin")
	p2 <- as_coord2d(x = 3)
	expect_true(has_intersection(p1, "origin"))
	expect_false(has_intersection(p1, p2))

	p1 <- as_coord3d("origin")
	p2 <- as_coord3d(x = 3)
	expect_true(has_intersection(p1, "origin"))
	expect_false(has_intersection(p1, p2))

	line <- as_line2d("x-axis")
	coord <- as_coord2d(x = 0, y = 0:2)
	expect_equal(has_intersection(line, coord), c(TRUE, FALSE, FALSE))
	expect_equal(has_intersection(coord, line), c(TRUE, FALSE, FALSE))

	plane <- as_plane3d("xy-plane")
	coord <- as_coord3d(x = 0, y = 0, z = 0:2)
	expect_equal(has_intersection(plane, coord), c(TRUE, FALSE, FALSE))
	expect_equal(has_intersection(coord, plane), c(TRUE, FALSE, FALSE))
})

test_that("`intersection()`", {
	line1 <- as_line2d("x-axis")
	line2 <- as_line2d("y-axis")
	line3 <- as_line2d(a = 0, b = 1, c = 2) # y + 2 = 0
	line4 <- as_line2d(a = 1, b = 0, c = 2) # x + 2 = 0
	expect_equal(intersection(line1, line1), list(standardize(line1)))
	expect_equal(intersection(line1, line2), list(as_coord2d("origin")))
	expect_equal(intersection(line1, "y-axis"), list(as_coord2d("origin")))
	expect_equal(intersection(line1, line3), list(NULL))
	expect_equal(intersection(line3, line4), list(as_coord2d(x = -2, y = -2)))

	plane1 <- as_plane3d("xy-plane")
	plane2 <- as_plane3d("xz-plane")
	plane3 <- as_plane3d(a = 0, b = 0, c = 1, d = 2) # z + 2 = 0
	expect_equal(intersection(plane1, plane1), list(standardize(plane1)))
	expect_error(intersection(plane1, plane2))
	expect_error(intersection(plane1, "xz-plane"))
	expect_equal(intersection(plane1, plane3), list(NULL))

	point1 <- as_point1d("origin")
	point2 <- as_point1d(a = 1, b = 2)
	expect_equal(intersection(point1, "origin"), list(standardize(point1)))
	expect_equal(intersection(point1, point2), list(NULL))

	p1 <- as_coord1d("origin")
	p2 <- as_coord1d(x = 3)
	expect_equal(intersection(p1, "origin"), list(p1))
	expect_equal(intersection(p1, p2), list(NULL))

	p1 <- as_coord2d("origin")
	p2 <- as_coord2d(x = 3)
	expect_equal(intersection(p1, "origin"), list(p1))
	expect_equal(intersection(p1, p2), list(NULL))

	p1 <- as_coord3d("origin")
	p2 <- as_coord3d(x = 3)
	expect_equal(intersection(p1, "origin"), list(p1))
	expect_equal(intersection(p1, p2), list(NULL))
})

test_that("intersection() Line2D and Ellipse2D", {
	circ <- as_ellipse2d(as_coord2d("origin"), r = 1)
	xax <- as_line2d("x-axis") # y = 0

	# Two crossing points: (1, 0) and (-1, 0)
	res <- intersection(circ, xax)
	expect_equal(length(res), 1L)
	expect_equal(length(res[[1L]]), 2L)
	expect_equal(sort(res[[1L]]$x), c(-1, 1), tolerance = 1e-10)
	expect_equal(res[[1L]]$y, c(0, 0), tolerance = 1e-10)

	# Both dispatch directions give same result
	expect_equal(intersection(xax, circ), res)

	# Tangent: y = 1 touches top of unit circle at (0, 1)
	line_top <- as_line2d(a = 0, b = 1, c = -1)
	res_tan <- intersection(circ, line_top)
	expect_equal(length(res_tan[[1L]]), 1L)
	expect_equal(res_tan[[1L]]$x, 0, tolerance = 1e-10)
	expect_equal(res_tan[[1L]]$y, 1, tolerance = 1e-10)

	# No intersection: y = 2 is above the unit circle
	line_above <- as_line2d(a = 0, b = 1, c = -2)
	expect_equal(intersection(circ, line_above), list(NULL))

	# Rotated ellipse: center (0,0), rx=2, ry=1, theta=0 vs x-axis
	e <- as_ellipse2d(as_coord2d(0, 0), rx = 2, ry = 1)
	res_e <- intersection(e, xax)
	expect_equal(length(res_e[[1L]]), 2L)
	expect_equal(sort(res_e[[1L]]$x), c(-2, 2), tolerance = 1e-10)
	expect_equal(res_e[[1L]]$y, c(0, 0), tolerance = 1e-10)

	# has_intersection uses intersection() internally
	expect_true(has_intersection(circ, xax))
	expect_false(has_intersection(circ, line_above))
})

#' Affine transformation grob
#'
#' `affineGrob()` is a grid grob function to facilitate
#' using the group affine transformation features introduced in R 4.2.
#'
#' `r affine_transformation_support`
#' @param grob A grid grob to perform affine transformations on.  Passed to [grid::defineGrob()] as its `src` argument.
#' @param vp_define [grid::viewport()] to define grid group in.  Passed to [grid::defineGrob()] as its `vp` argument.
#'                  This will cumulative with the current viewport and the `vp` argument (if any),
#'                  if this cumulative viewport falls outside the graphics device drawing area this
#'                  grob may be clipped on certain graphics devices.
#' @param transform An affine transformation function.
#'                  If `NULL` default to [grid::viewportTransform()].
#'                  Passed to [grid::useGrob()] as its `transform` argument.
#' @param vp_use [grid::viewport()] passed to [grid::useGrob()] as its `vp` argument.
#' @param name A character identifier (for grid).
#' @param gp A [grid::gpar()] object.
#' @param vp A [grid::viewport()] object (or `NULL`).
#' @return A [grid::gTree()] (grob) object of class "affine".
#'         As a side effect `grid.affine()` draws to the active graphics device.
#' @examples
#' if (require("grid")) {
#'   grob <- grobTree(rectGrob(gp = gpar(fill = "blue", col = NA)),
#'                    circleGrob(gp=gpar(fill="yellow", col = NA)),
#'                    textGrob("RSTATS", gp=gpar(fontsize=32)))
#'   grid.newpage()
#'   pushViewport(viewport(width=unit(4, "in"), height=unit(2, "in")))
#'   grid.draw(grob)
#'   popViewport()
#' }
#'
#' if (require("grid") &&
#'     getRversion() >= "4.2.0" &&
#'     isTRUE(dev.capabilities()$transformations)) {
#'   # Only works if active graphics device supports affine transformations
#'   # such as `png(type="cairo")` on R 4.2+
#'   vp_define <- viewport(width=unit(2, "in"), height=unit(2, "in"))
#'   affine <- affineGrob(grob, vp_define=vp_define)
#'   grid.newpage()
#'   pushViewport(viewport(width=unit(4, "in"), height=unit(2, "in")))
#'   grid.draw(affine)
#'   popViewport()
#' }
#' if (require("grid") &&
#'     getRversion() >= "4.2.0" &&
#'     isTRUE(dev.capabilities()$transformations)) {
#'   # Only works if active graphics device supports affine transformations
#'   # such as `png(type="cairo")` on R 4.2+
#'   settings <- affine_settings(xy = list(x = c(3/3, 2/3, 0/3, 1/3),
#'                                         y = c(2/3, 1/3, 1/3, 2/3)),
#'                               unit = "snpc")
#'   affine <- affineGrob(grob,
#'                        vp_define = vp_define,
#'                        transform = settings$transform,
#'                        vp_use = settings$vp)
#'   grid.newpage()
#'   grid.draw(affine)
#' }
#' @seealso See [affine_settings()] for computing good `transform` and `vp_use` settings.
#'          See <https://www.stat.auckland.ac.nz/~paul/Reports/GraphicsEngine/groups/groups.html>
#'          for more information about the group affine transformation feature.
#'          See [isocubeGrob()] which wraps this function to render isometric cubes.
#' @export
affineGrob <- function(
	grob,
	vp_define = NULL,
	transform = NULL,
	vp_use = NULL,
	name = NULL,
	gp = grid::gpar(),
	vp = NULL
) {
	stopifnot(getRversion() >= "4.2.0")
	if (is.null(transform)) {
		transform <- grid::viewportTransform
	}
	grid::gTree(
		grob = grob,
		vp_define = vp_define,
		transform = transform,
		vp_use = vp_use,
		name = name,
		gp = gp,
		vp = vp,
		cl = "affine"
	)
}

dev_supports_transformations <- function() {
	isTRUE(grDevices::dev.capabilities()$transformations)
}

#' @importFrom grid makeContent
#' @export
makeContent.affine <- function(x) {
	if (!dev_supports_transformations()) {
		stop(paste(
			"This graphics device does not support the affine transformation feature.",
			"See the Details section of `help(\"affineGrob\")` for more info."
		))
	}
	define <- grid::defineGrob(x$grob, vp = x$vp_define)
	use <- grid::useGrob(define$name, transform = x$transform, vp = x$vp_use)
	gl <- grid::gList(define, use)
	grid::setChildren(x, gl)
}

#' @rdname affineGrob
#' @param ... Passed to `affineGrob()`
#' @export
grid.affine <- function(...) {
	af_grob <- affineGrob(...)
	grid::grid.draw(af_grob)
	invisible(af_grob)
}

test_that("makeContent.affine() error when device lacks support", {
	skip_if_not(getRversion() >= "4.2.0")
	local_mocked_bindings(dev_supports_transformations = function() FALSE)
	old_dev <- dev.cur()
	pdf(NULL)
	on.exit(
		{
			dev.off()
			if (old_dev > 1L) dev.set(old_dev)
		},
		add = TRUE
	)
	expect_error(
		grid::grid.draw(affineGrob(grid::nullGrob())),
		"does not support the affine transformation feature"
	)
})

test_that("affineGrob() works", {
	skip_if_not(getRversion() >= "4.3.0")
	skip_if_not(isTRUE(all(capabilities(c("cairo", "png")))))
	old_dev <- dev.cur()
	f <- tempfile(fileext = ".png")
	png(f, type = "cairo")
	on.exit(
		{
			if (old_dev > 1L) {
				dev.set(old_dev)
			}
			unlink(f)
		},
		add = TRUE
	)
	grid.affine(grid::nullGrob())
	dev.off()
	expect_true(file.size(f) > 0)
})

#' Get affiner options
#'
#' `affiner_options()` returns the `affiner` package's global options.
#'
#' @param ... `affiner` package options using `name = value`.
#'            The return list will use any of these instead of the current/default values.
#' @param default If `TRUE` return the default values instead of current values.
#' @return A list of option values.
#'         Note this function **does not** set option values itself but
#'         this list can be passed to [options()], [withr::local_options()], or [withr::with_options()].
#' @examples
#'   affiner_options()
#'
#'   affiner_options(default = TRUE)
#'
#'   affiner_options(affiner_angular_unit = "pi-radians")
#' @seealso [affiner] for a high-level description of relevant global options.
#' @export
affiner_options <- function(..., default = FALSE) {
	afo <- list(affiner_angular_unit = "degrees", affiner_grid_unit = "inches")
	l <- list(...)
	stopifnot(all(names(l) %in% names(afo)))
	if (isFALSE(default)) {
		for (n in names(afo)) {
			afo[n] <- list(getOption(n, afo[[n]]))
		}
	}
	if (length(names(l))) {
		for (n in names(l)) {
			afo[n] <- l[n]
		}
	}
	afo
}

test_that("`affiner_options()`", {
	skip_if_not_installed("withr")
	withr::local_options(affiner_options(default = TRUE))
	expect_equal(affiner_options(), affiner_options(default = TRUE))
	expect_true(is.list(affiner_options(affiner_angular_unit = "gradians")))
})

standardize <- function(x) {
	UseMethod("standardize")
}

# x + d = 0
#' @export
standardize.Point1D <- function(x) {
	as_point1d(a = 1, b = x$b / x$a)
}

# a * x + y + c = 0
# x + c = 0
#' @export
standardize.Line2D <- function(x) {
	a <- ifelse(is_congruent(x$b, 0), 1, x$a / x$b)
	b <- ifelse(is_congruent(x$b, 0), 0, 1)
	c <- ifelse(is_congruent(x$b, 0), x$c / x$a, x$c / x$b)
	as_line2d(a, b, c)
}

# a * x + b * y + z + d = 0
# a * x + y + d = 0
# x + d = 0
#' @export
standardize.Plane3D <- function(x) {
	a <- ifelse(is_congruent(x$c, 0), ifelse(is_congruent(x$b, 0), 1, x$a / x$b), x$a / x$c)
	b <- ifelse(is_congruent(x$c, 0), ifelse(is_congruent(x$b, 0), 0, 1), x$b / x$c)
	c <- ifelse(is_congruent(x$c, 0), 0, 1)
	d <- ifelse(is_congruent(x$c, 0), ifelse(is_congruent(x$b, 0), x$d / a, x$d / b), x$d / x$c)
	as_plane3d(a, b, c, d)
}

#' @section Package options:
#' The following affiner function arguments may be set globally via [base::options()]:
#'  \describe{
#'    \item{affiner_angular_unit}{The default for the `unit` argument used by [angle()] and [as_angle()].
#'                                The default for this option is "degrees".}
#'    \item{affiner_grid_unit}{The default for the `unit` argument used by [affine_settings()].
#'                                The default for this option is "inches".}
#'  }
#'  The following `cli` options may also be of interest:
#'   \describe{
#'     \item{cli.unicode}{Whether UTF-8 character support should be assumed.
#'                        Along with [l10n_info()] used to determine the default of the
#'                        `use_unicode` argument of [format.angle()] and [print.angle()].}
#'   }
"_PACKAGE"

#' 1D affine transformation matrices
#'
#' `transform1d()`, `reflect1d()`, `scale1d()`,
#'  and `translate1d()` create 1D affine transformation matrix objects.
#'
#' \describe{
#' \item{`transform1d()`}{User supplied (post-multiplied) affine transformation matrix}.
#' \item{`reflect1d()`}{Reflections across a point.}
#' \item{`scale1d()`}{Scale the x-coordinates by multiplicative scale factors.}
#' \item{`translate1d()`}{Translate the coordinates by a [Coord1D] class object parameter.}
#' }
#'
#' `transform1d()` 1D affine transformation matrix objects are meant to be
#' post-multiplied and therefore should **not** be multiplied in reverse order.
#' Note the [Coord1D] class object methods auto-pre-multiply affine transformations
#' when "method chaining" so pre-multiplying affine transformation matrices
#' to do a single cumulative transformation instead of a method chain of multiple transformations
#' will not improve performance as much as it does in other R packages.
#'
#' To convert a pre-multiplied 1D affine transformation matrix to a post-multiplied one
#' simply compute its transpose using [t()].  To get an inverse transformation matrix
#' from an existing transformation matrix that does the opposite transformations
#' simply compute its inverse using [solve()].
#'
#' @return A 2x2 post-multiplied affine transformation matrix with classes "transform1d" and "at_matrix"
#' @param mat `r r2i_transform1d_mat`
#' @examples
#' p <- as_coord1d(x = sample(1:10, 3))
#'
#' # {affiner} affine transformation matrices are post-multiplied
#' # and therefore should **not** go in reverse order
#' mat <- transform1d(diag(2)) %*%
#'          scale1d(2) %*%
#'          translate1d(x = -1)
#' p1 <- p$
#'   clone()$
#'   transform(mat)
#'
#' # The equivalent result applying affine transformations via method chaining
#' p2 <- p$
#'   clone()$
#'   transform(diag(2))$
#'   scale(2)$
#'   translate(x = -1)
#'
#' all.equal(p1, p2)
#' @export
transform1d <- function(mat = diag(2L)) {
	if (all(dim(mat) == c(1, 1))) {
		mat <- rbind(mat, 0)
		mat <- cbind(mat, c(0, 1))
	}
	validate_transform1d(mat)
	new_transform1d(mat)
}

#' 2D affine transformation matrices
#'
#' `transform2d()`, `project2d()`, `reflect2d()`, `rotate2d()`, `scale2d()`, `shear2d()`,
#'  and `translate2d()` create 2D affine transformation matrix objects.
#'
#' \describe{
#' \item{`transform2d()`}{User supplied (post-multiplied) affine transformation matrix}.
#' \item{`project2d()`}{Oblique vector projections onto a line parameterized by
#'                      an oblique projection scale factor.
#'                      A (degenerate) scale factor of zero results in an orthogonal projection.}
#' \item{`reflect2d()`}{Reflections across a line.
#'                      To "flip" across both the x-axis and the y-axis use `scale2d(-1)`.}
#' \item{`rotate2d()`}{Rotations around the origin parameterized by an [angle()].}
#' \item{`scale2d()`}{Scale the x-coordinates and/or the y-coordinates by multiplicative scale factors.}
#' \item{`shear2d()`}{Shear the x-coordinates and/or the y-coordinates using shear factors.}
#' \item{`translate2d()`}{Translate the coordinates by a [Coord2D] class object parameter.}
#' }
#'
#' `transform2d()` 2D affine transformation matrix objects are meant to be
#' post-multiplied and therefore should **not** be multiplied in reverse order.
#' Note the [Coord2D] class object methods auto-pre-multiply affine transformations
#' when "method chaining" so pre-multiplying affine transformation matrices
#' to do a single cumulative transformation instead of a method chain of multiple transformations
#' will not improve performance as much as it does in other R packages.
#'
#' To convert a pre-multiplied 2D affine transformation matrix to a post-multiplied one
#' simply compute its transpose using [t()].  To get an inverse transformation matrix
#' from an existing transformation matrix that does the opposite transformations
#' simply compute its inverse using [solve()].
#'
#' @return A 3x3 post-multiplied affine transformation matrix with classes "transform2d" and "at_matrix"
#' @param mat `r r2i_transform2d_mat`
#' @examples
#' p <- as_coord2d(x = sample(1:10, 3), y = sample(1:10, 3))
#'
#' # {affiner} affine transformation matrices are post-multiplied
#' # and therefore should **not** go in reverse order
#' mat <- transform2d(diag(3)) %*%
#'          reflect2d(as_coord2d(-1, 1)) %*%
#'          rotate2d(90, "degrees") %*%
#'          scale2d(1, 2) %*%
#'          shear2d(0.5, 0.5) %*%
#'          translate2d(x = -1, y = -1)
#' p1 <- p$
#'   clone()$
#'   transform(mat)
#'
#' # The equivalent result applying affine transformations via method chaining
#' p2 <- p$
#'   clone()$
#'   transform(diag(3L))$
#'   reflect(as_coord2d(-1, 1))$
#'   rotate(90, "degrees")$
#'   scale(1, 2)$
#'   shear(0.5, 0.5)$
#'   translate(x = -1, y = -1)
#'
#' all.equal(p1, p2)
#'
#' @export
transform2d <- function(mat = diag(3L)) {
	if (all(dim(mat) == c(2, 2))) {
		mat <- rbind(mat, c(0, 0))
		mat <- cbind(mat, c(0, 0, 1))
	}
	validate_transform2d(mat)
	new_transform2d(mat)
}

#' 3D affine transformation matrices
#'
#' `transform3d()`, `project3d()`, `reflect3d()`, `rotate3d()`, `scale3d()`, `shear3d()`,
#'  and `translate3d()` create 3D affine transformation matrix objects.
#'
#' \describe{
#' \item{`transform3d()`}{User supplied (post-multiplied) affine transformation matrix}.
#' \item{`scale3d()`}{Scale the x-coordinates and/or the y-coordinates and/or the z-coordinates
#'                    by multiplicative scale factors.}
#' \item{`shear3d()`}{Shear the x-coordinates and/or the y-coordinates
#'                    and/or the z-coordinates using shear factors.}
#' \item{`translate3d()`}{Translate the coordinates by a [Coord3D] class object parameter.}
#' }
#'
#' `transform3d()` 3D affine transformation matrix objects are meant to be
#' post-multiplied and therefore should **not** be multiplied in reverse order.
#' Note the [Coord3D] class object methods auto-pre-multiply affine transformations
#' when "method chaining" so pre-multiplying affine transformation matrices
#' to do a single cumulative transformation instead of a method chain of multiple transformations
#' will not improve performance as much as it does in other R packages.
#'
#' To convert a pre-multiplied 3D affine transformation matrix to a post-multiplied one
#' simply compute its transpose using [t()].  To get an inverse transformation matrix
#' from an existing transformation matrix that does the opposite transformations
#' simply compute its inverse using [solve()].
#'
#' @return A 4x4 post-multiplied affine transformation matrix with classes "transform3d" and "at_matrix"
#' @param mat `r r2i_transform3d_mat`
#' @examples
#' p <- as_coord3d(x = sample(1:10, 3), y = sample(1:10, 3), z = sample(1:10, 3))
#'
#' # {affiner} affine transformation matrices are post-multiplied
#' # and therefore should **not** go in reverse order
#' mat <- transform3d(diag(4L)) %*%
#'          rotate3d("z-axis", degrees(90)) %*%
#'          scale3d(1, 2, 1) %*%
#'          translate3d(x = -1, y = -1, z = -1)
#' p1 <- p$
#'   clone()$
#'   transform(mat)
#'
#' # The equivalent result applying affine transformations via method chaining
#' p2 <- p$
#'   clone()$
#'   transform(diag(4L))$
#'   rotate("z-axis", degrees(90))$
#'   scale(1, 2, 1)$
#'   translate(x = -1, y = -1, z = -1)
#'
#' all.equal(p1, p2)
#'
#' @export
transform3d <- function(mat = diag(4L)) {
	if (all(dim(mat) == c(3, 3))) {
		mat <- rbind(mat, c(0, 0, 0))
		mat <- cbind(mat, c(0, 0, 0, 1))
	}
	validate_transform3d(mat)
	new_transform3d(mat)
}

validate_transform1d <- function(x) {
	stopifnot(is.matrix(x), all(dim(x) == c(2L, 2L)))
	if (!all(x[, 2L] == c(0, 1))) {
		msg <- "The last column must be equal to `c(0, 1)`."
		if (all(x[2L, ] == c(0, 1))) {
			stop(paste(
				msg,
				"Do you need to transpose a pre-multiplied affine transformation matrix",
				"with `t()` to convert it into a post-multiplied one?"
			))
		} else {
			stop(msg)
		}
	}
}

validate_transform2d <- function(x) {
	stopifnot(is.matrix(x), all(dim(x) == c(3, 3)))
	if (!all(x[, 3L] == c(0, 0, 1))) {
		msg <- "The last column must be equal to `c(0, 0, 1)`."
		if (all(x[3L, ] == c(0, 0, 1))) {
			stop(paste(
				msg,
				"Do you need to transpose a pre-multiplied affine transformation matrix",
				"with `t()` to convert it into a post-multiplied one?"
			))
		} else {
			stop(msg)
		}
	}
}

validate_transform3d <- function(x) {
	stopifnot(is.matrix(x), all(dim(x) == c(4, 4)))
	if (!all(x[, 4L] == c(0, 0, 0, 1))) {
		msg <- "The last column must be equal to `c(0, 0, 0, 1)`."
		if (all(x[4L, ] == c(0, 0, 0, 1))) {
			stop(paste(
				msg,
				"Do you need to transpose a pre-multiplied affine transformation matrix",
				"with `t()` to convert it into a post-multiplied one?"
			))
		} else {
			stop(msg)
		}
	}
}

new_transform1d <- function(mat) {
	if (!inherits(mat, "transform1d")) {
		class(mat) <- c("transform1d", "at_matrix", class(matrix()))
	}
	mat
}

new_transform2d <- function(mat) {
	if (!inherits(mat, "transform2d")) {
		class(mat) <- c("transform2d", "at_matrix", class(matrix()))
	}
	mat
}

new_transform3d <- function(mat) {
	if (!inherits(mat, "transform3d")) {
		class(mat) <- c("transform3d", "at_matrix", class(matrix()))
	}
	mat
}

#' @rdname transform2d
#' @param permutation `r r2i_transform2d_permutation`
#' @export
permute2d <- function(permutation = c("xy", "yx")) {
	permutation <- match.arg(permutation)
	mat <- switch(
		permutation,
		xy = diag(3L),
		yx = matrix(c(0, 1, 0, 1, 0, 0, 0, 0, 1), byrow = TRUE, ncol = 3, nrow = 3)
	)
	new_transform2d(mat)
}

#' @rdname transform3d
#' @param permutation `r r2i_transform3d_permutation`
#' @export
permute3d <- function(permutation = c("xyz", "xzy", "yxz", "yzx", "zyx", "zxy")) {
	permutation <- match.arg(permutation)
	mat <- switch(
		permutation,
		xyz = diag(4L),
		xzy = matrix(
			c(1, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 1),
			byrow = TRUE,
			ncol = 4,
			nrow = 4
		),
		yxz = matrix(
			c(0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1),
			byrow = TRUE,
			ncol = 4,
			nrow = 4
		),
		yzx = matrix(
			c(0, 0, 1, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1),
			byrow = TRUE,
			ncol = 4,
			nrow = 4
		),
		zxy = matrix(
			c(0, 1, 0, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 1),
			byrow = TRUE,
			ncol = 4,
			nrow = 4
		),
		zyx = matrix(
			c(0, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1),
			byrow = TRUE,
			ncol = 4,
			nrow = 4
		)
	)
	new_transform3d(mat)
}

#' @rdname transform1d
#' @export
project1d <- function(point = as_point1d("origin"), ...) {
	if (!is_point1d(point)) {
		point <- as_point1d(point, ...)
	}
	stopifnot(length(point) == 1L)
	x <- as_coord1d(point)
	mat <- scale1d(0) %*% translate1d(x)
	new_transform1d(mat)
}

#' @rdname transform2d
#' @param scale `r r2i_transform2d_scale`
#' @export
project2d <- function(line = as_line2d("x-axis"), ..., scale = 0) {
	if (!is_line2d(line)) {
		line <- as_line2d(line, ...)
	}
	stopifnot(length(line) == 1L, scale == 0 || line$a == 0)
	denom <- line$a^2 + line$b^2
	closest <- as_coord2d(-line$a * line$c / denom, -line$b * line$c / denom)
	theta <- as_angle(line)
	mat <- translate2d(-closest) %*%
		rotate2d(-theta) %*%
		shear2d(xy_shear = scale) %*%
		scale2d(y_scale = 0) %*%
		rotate2d(theta) %*%
		translate2d(closest)
	new_transform2d(mat)
}

#' @rdname transform3d
#' @export
#' @param scale `r r2i_transform3d_scale`
#' @param alpha `r r2i_transform3d_alpha`
project3d <- function(
	plane = as_plane3d("xy-plane"),
	...,
	scale = 0,
	alpha = angle(45, "degrees")
) {
	if (!is_plane3d(plane)) {
		plane <- as_plane3d(plane, ...)
	}
	if (!is_angle(alpha)) {
		alpha <- as_angle(alpha, ...)
	}
	stopifnot(
		length(plane) == 1L,
		scale == 0 || (plane$a == 0 && plane$b == 0),
		length(alpha) == 1L
	)
	denom <- plane$a^2 + plane$b^2 + plane$c^2
	closest <- as_coord3d(
		-plane$a * plane$d / denom,
		-plane$b * plane$d / denom,
		-plane$c * plane$d / denom
	)
	azimuth <- as_angle(plane, type = "azimuth")
	inclination <- as_angle(plane, type = "inclination")
	z_axis <- Coord3D$new(matrix(
		c(0, 0, 1, 1),
		nrow = 1,
		dimnames = list(NULL, c("x", "y", "z", "w"))
	))
	y_axis <- Coord3D$new(matrix(
		c(0, 1, 0, 1),
		nrow = 1,
		dimnames = list(NULL, c("x", "y", "z", "w"))
	))
	mat <- translate3d(-closest) %*%
		rotate3d(z_axis, -azimuth) %*%
		rotate3d(y_axis, -inclination) %*%
		shear3d(xz_shear = scale * cos(alpha), yz_shear = scale * sin(alpha)) %*%
		scale3d(z_scale = 0) %*%
		rotate3d(y_axis, inclination) %*%
		rotate3d(z_axis, azimuth) %*%
		translate3d(closest)
	new_transform3d(mat)
}

#' @rdname transform1d
#' @param point `r r2i_transform1d_point`
#' @export
reflect1d <- function(point = as_point1d("origin"), ...) {
	if (!is_point1d(point)) {
		point <- as_point1d(point, ...)
	}
	stopifnot(length(point) == 1)
	x <- as_coord1d(point)
	mat <- translate1d(-x) %*%
		scale1d(-1) %*%
		translate1d(x)
	new_transform1d(mat)
}

#' @rdname transform2d
#' @param line `r r2i_transform2d_line`
#' @export
reflect2d <- function(line = as_line2d("x-axis"), ...) {
	if (!is_line2d(line)) {
		line <- as_line2d(line, ...)
	}
	stopifnot(length(line) == 1L)
	denom <- line$a^2 + line$b^2
	closest <- as_coord2d(-line$a * line$c / denom, -line$b * line$c / denom)
	theta <- as_angle(line)
	mat <- translate2d(-closest) %*%
		rotate2d(-theta) %*%
		scale2d(1, -1) %*%
		rotate2d(theta) %*%
		translate2d(closest)
	new_transform2d(mat)
}

# https://en.wikipedia.org/wiki/Transformation_matrix#Reflection_2

#' @rdname transform3d
#' @param plane `r r2i_transform3d_plane`
#' @export
reflect3d <- function(plane = as_plane3d("xy-plane"), ...) {
	if (!is_plane3d(plane)) {
		plane <- as_plane3d(plane, ...)
	}
	stopifnot(length(plane) == 1)
	normal <- normal3d(plane)
	denom <- plane$a^2 + plane$b^2 + plane$c^2
	closest <- as_coord3d(
		-plane$a * plane$d / denom,
		-plane$b * plane$d / denom,
		-plane$c * plane$d / denom
	)

	mat <- diag(4L)
	mat[1L, 1L] <- 1 - 2 * normal$x^2
	mat[1L, 2L] <- mat[2L, 1L] <- -2 * normal$x * normal$y
	mat[1L, 3L] <- mat[3L, 1L] <- -2 * normal$x * normal$z
	mat[2L, 2L] <- 1 - 2 * normal$y^2
	mat[2L, 3L] <- mat[3L, 2L] <- -2 * normal$y * normal$z
	mat[3L, 3L] <- 1 - 2 * normal$z^2

	mat <- translate3d(-closest) %*% mat %*% translate3d(closest)
	new_transform3d(mat)
}

#' @rdname transform2d
#' @param theta `r r2i_transform_theta`
#' @param ... Passed to [as_angle()] or [as_coord2d()].
#' @export
rotate2d <- function(theta = angle(0), ...) {
	if (!is_angle(theta)) {
		theta <- as_angle(theta, ...)
	}
	stopifnot(length(theta) == 1)
	mat <- matrix(
		c(cos(theta), sin(theta), 0, -sin(theta), cos(theta), 0, 0, 0, 1),
		nrow = 3L,
		ncol = 3L,
		byrow = TRUE
	)
	new_transform2d(mat)
}

# Axis-angle representation to rotation matrix
# https://en.wikipedia.org/wiki/Axis-angle_representation
# Because we do rotation matrix post-multiplication instead of pre-multiplication we usually need to multiply angles
# in following algorithms by -1

#' @param axis `r r2i_transform3d_axis`
#' @param theta `r r2i_transform_theta`
#' @rdname transform3d
#' @export
rotate3d <- function(axis = as_coord3d("z-axis"), theta = angle(0), ...) {
	if (!is_coord3d(axis)) {
		axis <- as_coord3d(axis, ...)
	}
	stopifnot(length(axis) == 1)
	axis <- axis / abs(axis)

	if (!is_angle(theta)) {
		theta <- as_angle(theta, ...)
	}
	stopifnot(length(theta) == 1)

	I <- diag(3L)
	K <- cross_matrix(axis)
	c <- cos(-theta)
	s <- sin(-theta)
	R <- I + s * K + (1 - c) * K %*% K

	mat <- diag(4L)
	mat[1:3, 1:3] <- R
	new_transform3d(mat)
}

# "cross" product matrix
# https://en.wikipedia.org/wiki/Cross_product#Conversion_to_matrix_multiplication
cross_matrix <- function(x) {
	stopifnot(is_coord3d(x) && length(x) == 1L)
	m <- matrix(0, nrow = 3L, ncol = 3L)
	m[1, 2] <- -x$z
	m[1, 3] <- x$y
	m[2, 1] <- x$z
	m[2, 3] <- -x$x
	m[3, 1] <- -x$y
	m[3, 2] <- x$x
	m
}

#' @rdname transform1d
#' @param x_scale `r r2i_transform_x_scale`
#' @export
scale1d <- function(x_scale = 1) {
	stopifnot(length(x_scale) == 1L)
	mat <- diag(2L)
	mat[1L, 1L] <- x_scale
	new_transform1d(mat)
}

#' @rdname transform2d
#' @param x_scale `r r2i_transform_x_scale`
#' @param y_scale `r r2i_transform_y_scale`
#' @export
scale2d <- function(x_scale = 1, y_scale = x_scale) {
	stopifnot(length(x_scale) == 1L && length(y_scale) == 1L)
	mat <- diag(3L)
	mat[1L, 1L] <- x_scale
	mat[2L, 2L] <- y_scale
	new_transform2d(mat)
}

#' @rdname transform3d
#' @param x_scale `r r2i_transform_x_scale`
#' @param y_scale `r r2i_transform_y_scale`
#' @param z_scale `r r2i_transform_z_scale`
#' @export
scale3d <- function(x_scale = 1, y_scale = x_scale, z_scale = x_scale) {
	stopifnot(length(x_scale) == 1L && length(y_scale) == 1L && length(z_scale) == 1L)
	mat <- diag(4L)
	mat[1L, 1L] <- x_scale
	mat[2L, 2L] <- y_scale
	mat[3L, 3L] <- z_scale
	new_transform3d(mat)
}

#' @rdname transform2d
#' @param xy_shear `r r2i_transform2d_xy_shear`
#' @param yx_shear `r r2i_transform2d_yx_shear`
#' @export
shear2d <- function(xy_shear = 0, yx_shear = 0) {
	stopifnot(length(xy_shear) == 1L && length(yx_shear) == 1L)
	mat <- diag(3L)
	mat[2L, 1L] <- xy_shear
	mat[1L, 2L] <- yx_shear
	new_transform2d(mat)
}

#' @rdname transform3d
#' @param xy_shear `r r2i_transform3d_xy_shear`
#' @param xz_shear `r r2i_transform3d_xz_shear`
#' @param yx_shear `r r2i_transform3d_yx_shear`
#' @param yz_shear `r r2i_transform3d_yz_shear`
#' @param zx_shear `r r2i_transform3d_zx_shear`
#' @param zy_shear `r r2i_transform3d_zy_shear`
#' @export
shear3d <- function(
	xy_shear = 0,
	xz_shear = 0,
	yx_shear = 0,
	yz_shear = 0,
	zx_shear = 0,
	zy_shear = 0
) {
	stopifnot(all(lengths(list(xy_shear, xz_shear, yx_shear, yz_shear, zx_shear, zy_shear)) == 1))
	mat <- diag(4L)
	mat[2L, 1L] <- xy_shear
	mat[3L, 1L] <- xz_shear
	mat[1L, 2L] <- yx_shear
	mat[3L, 2L] <- yz_shear
	mat[1L, 3L] <- zx_shear
	mat[2L, 3L] <- zy_shear
	new_transform3d(mat)
}

#' @rdname transform1d
#' @param x `r r2i_transform1d_x`
#' @param ... Passed to [as_coord1d()].
#' @export
translate1d <- function(x = as_coord1d(0), ...) {
	if (!is_coord1d(x)) {
		x <- as_coord1d(x, ...)
	}
	stopifnot(length(x) == 1L)
	mat <- diag(2L)
	mat[2L, 1L] <- x$x
	new_transform1d(mat)
}

#' @rdname transform2d
#' @param x `r r2i_transform2d_x`
#' @export
translate2d <- function(x = as_coord2d(0, 0), ...) {
	if (!is_coord2d(x)) {
		x <- as_coord2d(x, ...)
	}
	stopifnot(length(x) == 1L)
	mat <- diag(3L)
	mat[3L, 1L] <- x$x
	mat[3L, 2L] <- x$y
	new_transform2d(mat)
}

#' @rdname transform3d
#' @param x `r r2i_transform3d_x`
#' @param ... Passed to [as_angle()] or [as_coord3d()].
#' @export
translate3d <- function(x = as_coord3d(0, 0, 0), ...) {
	if (!is_coord3d(x)) {
		x <- as_coord3d(x, ...)
	}
	stopifnot(length(x) == 1L)
	mat <- diag(4L)
	mat[4L, 1L] <- x$x
	mat[4L, 2L] <- x$y
	mat[4L, 3L] <- x$z
	new_transform3d(mat)
}

test_that("transform1d()", {
	m <- transform1d(diag(2L))
	expect_true(is_transform1d(m))
	expect_equal(m, as_transform1d(m))
	expect_equal(m, as_transform1d(diag(1L)))
	expect_error(as_transform1d(t(translate1d(as_coord1d(1)))))
	expect_error(as_transform1d(m + 2))

	x <- c(2, 5, 7)
	p1 <- as_coord1d(x = x)

	expect_equal(p1, p1$transform()$transform(diag(2L)))

	skip_if_not(getRversion() >= "4.3.0")
	expect_true(is_transform1d(m %*% m))
})

test_that("transform2d()", {
	m <- transform2d(diag(3))
	expect_true(is_transform2d(m))
	expect_equal(m, as_transform2d(m))
	expect_equal(m, as_transform2d(diag(2)))

	x <- c(2, 5, 7)
	y <- c(3, 4, 6)
	p1 <- as_coord2d(x = x, y = y)

	expect_equal(p1, p1$transform()$transform(diag(3)))

	skip_if_not(getRversion() >= "4.3.0")
	expect_true(is_transform2d(m %*% m))
})

test_that("transform3d()", {
	m <- transform3d(diag(4))
	expect_true(is_transform3d(m))
	expect_true(is_transform3d(solve(m)))
	expect_error(as_transform3d(t(translate3d(as_coord3d(1, 2, 3)))))
	expect_error(as_transform3d(m + 2))
	expect_equal(m, as_transform3d(m))
	expect_equal(m, as_transform3d(diag(3)))

	x <- c(2, 5, 7)
	y <- c(3, 4, 6)
	z <- c(4, 9, 3)
	p1 <- as_coord3d(x = x, y = y, z = z)

	expect_equal(p1, p1$transform()$transform(diag(4)))

	skip_if_not(getRversion() >= "4.3.0")
	expect_true(is_transform3d(m %*% m))
})

test_that("at_matrix methods", {
	m <- transform2d(diag(3))
	expect_true(is_transform2d(solve(m)))
	expect_false(is_transform2d(as.matrix(m)))
	expect_false(is_transform2d(t(m)))
	expect_false(is_transform2d(m + 2))
	expect_error(as_transform2d(m + 2))
	expect_false(is_transform2d(m + m))
	expect_false(is_transform2d(abs(m)))
	expect_false(is_transform2d(Re(m)))

	expect_error(as_transform2d(t(translate2d(x = 2, y = 2))))
})

test_that("permute2d()", {
	x <- c(2, 5, 7)
	y <- c(3, 4, 6)
	p <- as_coord2d(x, y)
	expect_equal(p, p$clone()$permute())
	expect_equal(as_coord2d(y, x), p$clone()$permute("yx"))
})

test_that("permute3d()", {
	expect_true(is_transform3d(permute3d("xzy")))

	x <- c(2, 5, 7)
	y <- c(3, 4, 6)
	z <- c(4, 9, 3)
	p <- as_coord3d(x, y, z)
	expect_equal(p, p$clone()$permute())
	expect_equal(as_coord3d(x, z, y), p$clone()$permute("xzy"))
	expect_equal(as_coord3d(y, x, z), p$clone()$permute("yxz"))
	expect_equal(as_coord3d(y, z, x), p$clone()$permute("yzx"))
	expect_equal(as_coord3d(z, x, y), p$clone()$permute("zxy"))
	expect_equal(as_coord3d(z, y, x), p$clone()$permute("zyx"))
})

test_that("project1d()", {
	x <- c(2, 5, 7)
	p1 <- as_coord1d(x = x)$project(as_coord1d(x = 3))
	expect_equal(p1$x, rep(3, 3))
})

test_that("project2d()", {
	x <- c(2, 5, 7)
	y <- c(3, 4, 6)
	p1 <- as_coord2d(x = x, y = y)$project(as_line2d("x-axis"))
	expect_equal(p1$x, x)
	expect_equal(p1$y, rep(0, 3))

	p2 <- as_coord2d(x = x, y = y)$project(degrees(90))
	expect_equal(p2$x, rep(0, 3))
	expect_equal(p2$y, y)

	p3 <- as_coord2d(x = x, y = y)$project(scale = 0.5)
	expect_equal(p3$x, x + 0.5 * y)
	expect_equal(p3$y, rep(0, 3))

	l4 <- as_line2d(a = 0, b = 1, c = -4)
	p4 <- as_coord2d(x = x, y = y)$project(l4)
	expect_equal(p4$x, x)
	expect_equal(p4$y, rep(4, 3))

	l5 <- as_line2d(a = 1, b = 0, c = -4)
	p5 <- as_coord2d(x = x, y = y)$project(l5)
	expect_equal(p5$x, rep(4, 3))
	expect_equal(p5$y, y)

	x6 <- c(-5, 5, 5)
	y6 <- c(5, -5, 5)
	l6 <- as_line2d(a = 1, b = -1, c = 0)
	p6 <- as_coord2d(x = x6, y = y6)$project(l6)
	expect_equal(p6$x, c(0, 0, 5))
	expect_equal(p6$y, c(0, 0, 5))

	l7 <- as_line2d(a = 1, b = -1, c = 2)
	p7 <- as_coord2d(x = x6, y = y6)$project(l7)
	expect_equal(p7$x, c(-1, -1, 4))
	expect_equal(p7$y, c(1, 1, 6))
})

test_that("project3d()", {
	x <- c(2, 5, 7)
	y <- c(3, 4, 6)
	z <- c(4, 9, 3)
	p1 <- as_coord3d(x, y, z)$project()
	expect_equal(p1$x, x)
	expect_equal(p1$y, y)
	expect_equal(p1$z, rep(0, 3))

	p2 <- as_coord3d(x, y, z)$project("xz-plane")
	expect_equal(p2$x, x)
	expect_equal(p2$y, rep(0, 3))
	expect_equal(p2$z, z)

	p3 <- as_coord3d(x, y, z)$project("yz-plane")
	expect_equal(p3$x, rep(0, 3))
	expect_equal(p3$y, y)
	expect_equal(p3$z, z)

	alpha <- angle(60, "degrees")
	scale <- 0.5
	p4 <- as_coord3d(x, y, z)$project(scale = scale, alpha = alpha)
	expect_equal(p4$x, x + scale * cos(alpha) * z)
	expect_equal(p4$y, y + scale * sin(alpha) * z)
	expect_equal(p4$z, rep(0, 3))

	p42 <- as_coord3d(x, y, z)$project(scale = scale, alpha = 60, unit = "degrees")

	p5 <- as_coord3d(x, y, z)$permute("xzy")$project(scale = scale, alpha = alpha)
	expect_equal(p5$x, x + scale * cos(alpha) * y)
	expect_equal(p5$y, z + scale * sin(alpha) * y)
	expect_equal(p5$z, rep(0, 3))

	# Planes not on origin
	plane <- as_plane3d(a = 0, b = 0, c = 1, d = -2)
	p6 <- as_coord3d(x, y, z)$project(plane = plane, scale = scale, alpha = alpha)
	expect_equal(p6$x, x + scale * cos(alpha) * (z - 2))
	expect_equal(p6$y, y + scale * sin(alpha) * (z - 2))
	expect_equal(p6$z, rep(2, 3))

	# Plane through origin, not axis-aligned
	pl10 <- as_plane3d(a = -1, b = -1, c = 1, d = 0)
	p10 <- as_coord3d(x = -5, y = -5, z = 5)$project(pl10)
	expect_equal(p10$x, 0)
	expect_equal(p10$y, 0)
	expect_equal(p10$z, 0)
})

test_that("reflect1d()", {
	x <- c(2, 5, 7)
	p1 <- as_coord1d(x = x)$reflect("origin")
	expect_equal(p1$x, -x)
	p2 <- as_coord1d(x = x)$reflect(as_coord1d(x = 5))
	expect_equal(p2$x, c(8, 5, 3))
})

test_that("reflect2d()", {
	x <- c(2, 5, 7)
	y <- c(3, 4, 6)
	p1 <- as_coord2d(x = x, y = y)$reflect("x-axis")
	expect_equal(p1$x, x)
	expect_equal(p1$y, -y)

	p1 <- as_coord2d(x = x, y = y)$reflect(as_line2d(a = 0, b = 1, c = 0))
	expect_equal(p1$x, x)
	expect_equal(p1$y, -y)

	p2 <- as_coord2d(x = x, y = y)$reflect("y-axis")
	expect_equal(p2$x, -x)
	expect_equal(p2$y, y)

	p2 <- as_coord2d(x = x, y = y)$reflect(as_line2d(a = 1, b = 0, c = 0))
	expect_equal(p2$x, -x)
	expect_equal(p2$y, y)

	p3 <- as_coord2d(x = x, y = y)$reflect(degrees(135))
	expect_equal(p3$x, -y)
	expect_equal(p3$y, -x)
	p4 <- as_coord2d(x = x, y = y)$reflect(as_coord2d(1, 1))
	expect_equal(p4$x, -y)
	expect_equal(p4$y, -x)

	lh <- as_line2d(a = 0, b = 1, c = -2) # y = 2
	p5 <- as_coord2d(x = x, y = y)$reflect(lh)
	expect_equal(p5$x, x)
	expect_equal(p5$y, 2 - (y - 2))

	lv <- as_line2d(a = 1, b = 0, c = -2) # x = 2
	p6 <- as_coord2d(x = x, y = y)$reflect(lv)
	expect_equal(p6$x, 2 - (x - 2))
	expect_equal(p6$y, y)

	x6 <- c(-5, 5, 5)
	y6 <- c(5, -5, 5)
	l7 <- as_line2d(a = 1, b = -1, c = 2)
	p7 <- as_coord2d(x = x6, y = y6)$reflect(l7)
	expect_equal(p7$x, c(3, -7, 3))
	expect_equal(p7$y, c(-3, 7, 7))
})

test_that("reflect3d()", {
	x <- c(2, 5, 7)
	y <- c(3, 4, 6)
	z <- c(4, 9, 3)
	p1 <- as_coord3d(x = x, y = y, z = z)$reflect("xy-plane")
	expect_equal(p1$x, x)
	expect_equal(p1$y, y)
	expect_equal(p1$z, -z)

	p1a <- as_coord3d(x = x, y = y, z = z)$reflect(as_plane3d(a = 0, b = 0, c = 1, d = -2))
	expect_equal(p1a$x, x)
	expect_equal(p1a$y, y)
	expect_equal(p1a$z, 2 - (z - 2))

	p2 <- as_coord3d(x = x, y = y, z = z)$reflect("xz-plane")
	expect_equal(p2$x, x)
	expect_equal(p2$y, -y)
	expect_equal(p2$z, z)

	p2a <- as_coord3d(x = x, y = y, z = z)$reflect(as_plane3d(a = 0, b = 1, c = 0, d = -2))
	expect_equal(p2a$x, x)
	expect_equal(p2a$y, 2 - (y - 2))
	expect_equal(p2a$z, z)

	p3 <- as_coord3d(x = x, y = y, z = z)$reflect("yz-plane")
	expect_equal(p3$x, -x)
	expect_equal(p3$y, y)
	expect_equal(p3$z, z)

	p3a <- as_coord3d(x = x, y = y, z = z)$reflect(as_plane3d(a = 1, b = 0, c = 0, d = -2))
	expect_equal(p3a$x, 2 - (x - 2))
	expect_equal(p3a$y, y)
	expect_equal(p3a$z, z)

	p4 <- as_coord3d(x = x, y = y, z = z)$reflect(as_coord3d(1, 0, 1))
	expect_equal(p4$x, -z)
	expect_equal(p4$y, y)
	expect_equal(p4$z, -x)

	# Planes not on origin
	plh <- as_plane3d(a = 0, b = 1, c = 0, d = -2) # y = 2
	p5 <- as_coord3d(x = x, y = y, z = 0)$reflect(plh)
	expect_equal(p5$x, x)
	expect_equal(p5$y, 2 - (y - 2))
	expect_equal(p5$z, rep_len(0, 3))

	plv <- as_plane3d(a = 1, b = 0, c = 0, d = -2) # x = 2
	p6 <- as_coord3d(x = x, y = y, z = 0)$reflect(plv)
	expect_equal(p6$x, 2 - (x - 2))
	expect_equal(p6$y, y)
	expect_equal(p6$z, rep_len(0, 3))

	# Planes not on origin, not axis-aligned
	x6 <- c(-5, 5, 5)
	y6 <- c(5, -5, 5)
	pl7 <- as_plane3d(a = 1, b = -1, c = 0, d = 2) # y = x + 2
	p7 <- as_coord3d(x = x6, y = y6, z = 0)$reflect(pl7)
	expect_equal(p7$x, c(3, -7, 3))
	expect_equal(p7$y, c(-3, 7, 7))
	expect_equal(p7$z, rep_len(0, 3))

	pl8 <- as_plane3d(a = 1, b = 0, c = -1, d = 2) # z = x + 2
	p8 <- as_coord3d(x = x6, y = 0, z = y6)$reflect(pl8)
	expect_equal(p8$x, c(3, -7, 3))
	expect_equal(p8$y, rep_len(0, 3))
	expect_equal(p8$z, c(-3, 7, 7))

	pl9 <- as_plane3d(a = 0, b = -1, c = 1, d = 2) # y = z + 2
	p9 <- as_coord3d(x = 0, y = y6, z = x6)$reflect(pl9)
	expect_equal(p9$x, rep_len(0, 3))
	expect_equal(p9$y, c(-3, 7, 7))
	expect_equal(p9$z, c(3, -7, 3))

	# Plane through origin, not axis-aligned
	pl10 <- as_plane3d(a = -1, b = -1, c = 1, d = 0)
	p10 <- as_coord3d(x = -5, y = -5, z = 5)$reflect(pl10)
	expect_equal(p10$x, 5)
	expect_equal(p10$y, 5)
	expect_equal(p10$z, -5)
})

test_that("rotate2d()", {
	x <- c(2, 5, 7)
	y <- c(3, 4, 6)
	p1 <- as_coord2d(x = x, y = y)$rotate(90, "degrees")
	expect_equal(p1$x, c(-3, -4, -6))
	expect_equal(p1$y, c(2, 5, 7))

	p2 <- as_coord2d(x = x, y = y)$rotate(-0.5, "pi-radians")
	expect_equal(p2$x, c(3, 4, 6))
	expect_equal(p2$y, c(-2, -5, -7))

	p3 <- as_coord2d(x = x, y = y)$rotate(c(90, 0, -90), "degrees")
	expect_equal(p3$x, c(-3, 5, 6))
	expect_equal(p3$y, c(2, 4, -7))

	skip_if_not(getRversion() >= "4.3.0")
	expect_equal(rotate2d(90, "degrees"), permute2d("yx") %*% scale2d(-1, 1))
})

test_that("rotate3d()", {
	expect_true(is_transform3d(rotate3d("z-axis", degrees(90))))

	skip_if_not_installed("withr")
	withr::local_options(affiner_options(default = TRUE))
	x <- c(2, 5, 7)
	y <- c(3, 4, 6)
	z <- c(5, 2, 3)
	p1 <- as_coord3d(x = x, y = y, z = z)$rotate("z-axis", 90, "degrees")
	expect_equal(p1$x, -y)
	expect_equal(p1$y, x)
	expect_equal(p1$z, z)

	p2 <- as_coord3d(x = x, y = y, z = z)$rotate("x-axis", 90, "degrees")
	expect_equal(p2$x, x)
	expect_equal(p2$y, -z)
	expect_equal(p2$z, y)

	expect_equal(
		rotate3d_to_AA(rotate3d("x-axis", 90, "degrees")),
		list(axis = as_coord3d(1, 0, 0), theta = angle(90, "degrees"))
	)
	expect_equal(
		rotate3d_to_AA(rotate3d("y-axis", 90, "degrees")),
		list(axis = as_coord3d(0, 1, 0), theta = angle(90, "degrees"))
	)
	expect_equal(
		rotate3d_to_AA(rotate3d(as_coord3d(0.5, 0.5, 0.7071068), 180, "degrees")),
		list(axis = as_coord3d(0.5, 0.5, 0.7071068), theta = angle(180, "degrees")),
		tolerance = 1e-6
	)
	expect_equal(
		rotate3d_to_AA(rotate3d(as_coord3d(-0.6, 0.6, 0.8), 180, "degrees")),
		list(axis = as_coord3d(-0.5144958, 0.5144958, 0.6859943), theta = angle(180, "degrees")),
		tolerance = 1e-6
	)
	expect_equal(
		rotate3d_to_AA(rotate3d(as_coord3d(0.6, -0.6, 0.5291503), 180, "degrees")),
		list(axis = as_coord3d(0.6, -0.6, 0.5291503), theta = angle(180, "degrees")),
		tolerance = 1e-6
	)
	expect_equal(
		rotate3d_to_AA(rotate3d(as_coord3d(-0.6, -0.6, 0.5291503), 180, "degrees")),
		list(axis = as_coord3d(-0.6, -0.6, 0.5291503), theta = angle(180, "degrees")),
		tolerance = 1e-6
	)
	expect_equal(
		rotate3d_to_AA(rotate3d(as_coord3d(-0.6, 0.6, 0.5291503), 0, "degrees")),
		list(axis = as_coord3d(0, 0, 1), theta = angle(0, "degrees")),
		tolerance = 1e-6
	)
})

test_that("scale1d()", {
	x <- c(2, 5, 7)
	p1 <- as_coord1d(x = x)$scale(2)
	expect_equal(p1$x, 2 * x)

	p2 <- as_coord1d(x = x)$scale(rep(2, 3))
	expect_equal(p1, p2)
})

test_that("scale2d()", {
	x <- c(2, 5, 7)
	y <- c(3, 4, 6)
	p1 <- as_coord2d(x = x, y = y)$scale(2, 3)
	expect_equal(p1$x, 2 * x)
	expect_equal(p1$y, 3 * y)

	p2 <- as_coord2d(x = x, y = y)$scale(rep(2, 3), rep(3, 3))
	expect_equal(p1, p2)
})

test_that("scale3d()", {
	x <- c(2, 5, 7)
	y <- c(3, 4, 6)
	z <- c(5, 2, 3)
	p1 <- as_coord3d(x = x, y = y, z = z)$scale(2, 3, 4)
	expect_equal(p1$x, 2 * x)
	expect_equal(p1$y, 3 * y)
	expect_equal(p1$z, 4 * z)
	p2 <- as_coord3d(x = x, y = y, z = z)$scale(rep(2, 3), rep(3, 3), rep(4, 3))
	expect_equal(p1, p2)
})

test_that("shear2d()", {
	x <- c(2, 5, 7)
	y <- c(3, 4, 6)
	p1 <- as_coord2d(x = x, y = y)$shear(x = 1)
	expect_equal(p1$x, x + y)
	expect_equal(p1$y, y)

	p2 <- as_coord2d(x = x, y = y)$shear(y = 1)
	expect_equal(p2$x, x)
	expect_equal(p2$y, y + x)

	p3 <- as_coord2d(x = x, y = y)$shear(x = 1, y = 1)
	expect_equal(p3$x, x + y)
	expect_equal(p3$y, y + x)
})

test_that("shear3d()", {
	expect_true(is_transform3d(shear3d()))

	x <- c(2, 5, 7)
	y <- c(3, 4, 6)
	z <- c(5, 2, 3)
	p1 <- as_coord3d(x = x, y = y, z = z)$shear(xy = 1, xz = 0.5)
	expect_equal(p1$x, x + y + 0.5 * z)
	expect_equal(p1$y, y)
	expect_equal(p1$z, z)

	p2 <- as_coord3d(x = x, y = y, z = z)$shear(zx = 1, yx = 1)
	expect_equal(p2$x, x)
	expect_equal(p2$y, y + x)
	expect_equal(p2$z, z + x)
})

test_that("translate1d()", {
	x <- c(2, 5, 7)
	p0 <- as_coord1d(x = x)

	vec <- as_coord1d(x = 2)
	p1 <- p0$clone()$translate(vec)
	expect_equal(p1$x, x + 2)

	p2 <- p0$clone()$translate("origin")
	expect_equal(p2$x, x)

	p3 <- p0$clone()$translate(p0)
	expect_equal(p3$x, x + x)

	expect_equal(translate1d("origin"), translate1d(as_coord1d(0)))
})

test_that("translate2d()", {
	x <- c(2, 5, 7)
	y <- c(3, 4, 6)
	p0 <- as_coord2d(x = x, y = y)

	vec <- as_coord2d(x = 2, y = 3)
	p1 <- p0$clone()$translate(vec)
	expect_equal(p1$x, x + 2)
	expect_equal(p1$y, y + 3)

	p2 <- p0$clone()$translate(angle(90, "degrees"), radius = 2)
	expect_equal(p2$x, x)
	expect_equal(p2$y, y + 2)

	p3 <- p0$clone()$translate(p0)
	expect_equal(p3$x, x + x)
	expect_equal(p3$y, y + y)

	expect_equal(translate2d(angle(90, "degrees"), radius = 1), translate2d(as_coord2d(0, 1)))
})

test_that("translate3d()", {
	x <- c(2, 5, 7)
	y <- c(3, 4, 6)
	z <- c(5, 2, 3)
	p0 <- as_coord3d(x = x, y = y, z = z)
	expect_equal(p0, p0$clone()$translate("origin"))

	vec <- as_coord3d(x = 2, y = 3, z = 1)
	p1 <- p0$clone()$translate(vec)
	expect_equal(p1$x, x + 2)
	expect_equal(p1$y, y + 3)
	expect_equal(p1$z, z + 1)

	p3 <- p0$clone()$translate(p0)
	expect_equal(p3$x, x + x)
	expect_equal(p3$y, y + y)
	expect_equal(p3$z, z + z)

	expect_equal(translate3d(as_coord3d(3, 2, 1)), translate3d(as_coord2d(3, 2), z = 1))
})

# {roxygen2} doesn't allow R6 class documentation to
# inherit from S3 method param documentation

r2i_transform1d_mat <- paste(
	"A 2x2 matrix representing a post-multiplied affine transformation matrix.",
	"The last **column** must be equal to `c(0, 1)`.",
	"If the last **row** is `c(0, 1)` you may need to transpose it",
	"to convert it from a pre-multiplied affine transformation matrix to a post-multiplied one.",
	"If a 1x1 matrix we'll quietly add a final column/row equal to `c(0, 1)`.",
	sep = "\n"
)

r2i_transform2d_mat <- paste(
	"A 3x3 matrix representing a post-multiplied affine transformation matrix.",
	"The last **column** must be equal to `c(0, 0, 1)`.",
	"If the last **row** is `c(0, 0, 1)` you may need to transpose it",
	"to convert it from a pre-multiplied affine transformation matrix to a post-multiplied one.",
	"If a 2x2 matrix (such as a 2x2 post-multiplied 2D rotation matrix)",
	"we'll quietly add a final column/row equal to `c(0, 0, 1)`.",
	sep = "\n"
)

r2i_transform3d_mat <- paste(
	"A 4x4 matrix representing a post-multiplied affine transformation matrix.",
	"The last **column** must be equal to `c(0, 0, 0, 1)`.",
	"If the last **row** is `c(0, 0, 0, 1)` you may need to transpose it",
	"to convert it from a pre-multiplied affine transformation matrix to a post-multiplied one.",
	"If a 3x3 matrix (such as a 3x3 post-multiplied 3D rotation matrix)",
	"we'll quietly add a final column/row equal to `c(0, 0, 0, 1)`.",
	sep = "\n"
)

r2i_transform2d_permutation <- 'Either "xy" (no permutation) or "yx" (permute x and y axes)'

r2i_transform3d_permutation <- paste(
	'Either "xyz" (no permutation), "xzy" (permute y and z axes),',
	'"yxz" (permute x and y axes), "yzx" (x becomes z, y becomes x, z becomes y),',
	'"zxy" (x becomes y, y becomes z, z becomes x), "zyx" (permute x and z axes)',
	sep = "\n"
)

r2i_transform3d_axis <- paste(
	"A [Coord3D] class object or one that can coerced to one by `as_coord3d(axis, ...)`.",
	"The `axis` represents the axis to be rotated around.",
	sep = "\n"
)

r2i_transform_theta <- "An [angle()] object of length one or an object coercible to one by `as_angle(theta, ...)`."

r2i_transform2d_scale <- paste(
	"Oblique projection scale factor.",
	"A degenerate `0` value indicates an orthogonal projection.",
	sep = "\n"
)

r2i_transform3d_scale <- paste(
	"Oblique projection foreshortening scale factor.",
	"A (degenerate) `0` value indicates an orthographic projection.",
	"A value of `0.5` is used by a \\dQuote{cabinet projection}",
	"while a value of `1.0` is used by a \\dQuote{cavalier projection}.",
	sep = "\n"
)

r2i_transform3d_alpha <- paste(
	"Oblique projection angle (the angle the third axis is projected going off at).",
	"An [angle()] object or one coercible to one with `as_angle(alpha, ...)`.",
	"Popular angles are 45 degrees, 60 degrees, and `arctangent(2)` degrees.",
	sep = "\n"
)

r2i_transform_x_scale <- "Scaling factor to apply to x coordinates"
r2i_transform_y_scale <- "Scaling factor to apply to y coordinates"
r2i_transform_z_scale <- "Scaling factor to apply to z coordinates"

r2i_transform2d_xy_shear <- "Horizontal shear factor: `x = x + xy_shear * y`"
r2i_transform2d_yx_shear <- "Vertical shear factor: `y = yx_shear * x + y`"

r2i_transform3d_xy_shear <- "Shear factor: `x = x + xy_shear * y + xz_shear * z`"
r2i_transform3d_xz_shear <- "Shear factor: `x = x + xy_shear * y + xz_shear * z`"
r2i_transform3d_yx_shear <- "Shear factor: `y = yx_shear * x + y + yz_shear * z`"
r2i_transform3d_yz_shear <- "Shear factor: `y = yx_shear * x + y + yz_shear * z`"
r2i_transform3d_zx_shear <- "Shear factor: `z = zx_shear * x + zy_shear * y + z`"
r2i_transform3d_zy_shear <- "Shear factor: `z = zx_shear * x + zy_shear * y + z`"

r2i_transform1d_point <- paste(
	"A [Point1D] object of length one representing the point",
	"you with to reflect across or project to or an object coercible to one by `as_point1d(point, ...)`",
	' such as "origin".',
	sep = "\n"
)
r2i_transform2d_line <- paste(
	"A [Line2D] object of length one representing the line",
	"you with to reflect across or project to or an object coercible to one by `as_line2d(line, ...)`",
	' such as "x-axis" or "y-axis".',
	sep = "\n"
)
r2i_transform3d_plane <- paste(
	'A [Plane3D] object of length one representing the plane',
	'you wish to reflect across or project to or an object coercible to one using `as_plane3d(plane, ...)`',
	'such as "xy-plane", "xz-plane", or "yz-plane".',
	sep = "\n"
)

r2i_transform1d_x <- "A [Coord1D] object of length one or an object coercible to one by `as_coord1d(x, ...)`."
r2i_transform2d_x <- "A [Coord2D] object of length one or an object coercible to one by `as_coord2d(x, ...)`."
r2i_transform3d_x <- "A [Coord3D] object of length one or an object coercible to one by `as_coord3d(x, ...)`."

affine_transformation_support <- paste(
	"Not all graphics devices provided by `grDevices` or other R packages support the [affine transformation feature introduced in R 4.2](https://www.stat.auckland.ac.nz/~paul/Reports/GraphicsEngine/groups/groups.html).",
	"If `isTRUE(getRversion() >= '4.2.0')` then the active graphics device should support this feature if `isTRUE(grDevices::dev.capabilities()$transformations)`.",
	"In particular the following graphics devices should support the affine transformation feature:",
	"",
	"* R's [grDevices::pdf()] device",
	"* R's 'cairo' devices e.g. [grDevices::cairo_pdf()], `grDevices::png(type = 'cairo')`, [grDevices::svg()], `grDevices::x11(type = 'cairo')`, etc. If `isTRUE(capabilities('cairo'))` then R was compiled with support for the 'cairo' devices .",
	"* R's 'quartz' devices (since R 4.3.0) e.g. [grDevices::quartz()], `grDevices::png(type = 'quartz')`, etc. If `isTRUE(capabilities('aqua'))` then R was compiled with support for the 'quartz' devices (generally only `TRUE` on macOS systems).",
	"* `ragg`'s devices (since v1.3.0) e.g. [ragg::agg_png()], [ragg::agg_capture()], etc.",
	sep = "\n"
)

#' Whether two objects are parallel
#'
#' `is_parallel()` is a S3 method that tests whether two objects are parallel.
#'
#' @param x,y The two objects to compute if they are parallel.
#' @param ... Passed to other methods (or ignored).
#' @return A logical vector.
#' @examples
#' line1 <- as_line2d("x-axis")
#' line2 <- as_line2d("y-axis")
#' line3 <- as_line2d(a = 0, b = 1, c = 2) # y + 2 = 0
#' is_parallel(line1, line1)
#' is_parallel(line1, line2)
#' is_parallel(line1, line3)
#' @export
is_parallel <- function(x, y, ...) {
	UseMethod("is_parallel")
}

#' @rdname is_parallel
#' @param tolerance Numerics with differences smaller
#'                  than `tolerance` will be considered \dQuote{equivalent}.
#' @export
is_parallel.Line2D <- function(x, y, ..., tolerance = sqrt(.Machine$double.eps)) {
	if (!is_line2d(y)) {
		y <- as_line2d(y)
	}
	stopifnot(!any(is_degenerate(x)), !any(is_degenerate(y)))
	n <- max(length(x), length(y))
	x <- rep(standardize(x), length.out = n)
	y <- rep(standardize(y), length.out = n)
	is_equivalent(x$a, y$a, tolerance = tolerance) &
		is_equivalent(x$b, y$b, tolerance = tolerance)
}

#' @rdname is_parallel
#' @export
is_parallel.Plane3D <- function(x, y, ..., tolerance = sqrt(.Machine$double.eps)) {
	if (!is_plane3d(y)) {
		y <- as_plane3d(y)
	}
	stopifnot(!any(is_degenerate(x)), !any(is_degenerate(y)))
	n <- max(length(x), length(y))
	x <- rep(standardize(x), length.out = n)
	y <- rep(standardize(y), length.out = n)
	is_equivalent(x$a, y$a, tolerance = tolerance) &
		is_equivalent(x$b, y$b, tolerance = tolerance) &
		is_equivalent(x$c, y$c, tolerance = tolerance)
}

#' Test whether two objects are equivalent
#'
#' `is_equivalent()` is a S3 generic that tests whether two different objects are \dQuote{equivalent}.
#' The `is_equivalent()` method for [angle()] classes tests whether two angles are congruent.
#' The `is_equivalent()` method for [Point1D], [Line2D], [Plane3D] classes tests whether they are the same point/line/plane after standardization.
#'
#' @param x,y Two objects to test whether they are \dQuote{"equivalent"}.
#' @param ... Further arguments passed to or from other methods.
#' @return A logical vector
#'
#' @examples
#' line1 <- as_line2d(a = 1, b = 2, c = 3) # 1 * x + 2 * y + 3 = 0
#' line2 <- as_line2d(a = 2, b = 4, c = 6) # 2 * x + 4 * y + 6 = 0
#' is_equivalent(line1, line2)
#' @seealso [is_congruent()], [all.equal()]
#' @export
is_equivalent <- function(x, y, ...) {
	UseMethod("is_equivalent")
}

#' @param mod_turns If `TRUE` angles that are congruent modulo full turns will be considered \dQuote{congruent}.
#' @param tolerance Numerics with differences smaller
#'                  than `tolerance` will be considered \dQuote{equivalent}.
#' @rdname is_equivalent
#' @export
is_equivalent.angle <- function(
	x,
	y,
	...,
	mod_turns = TRUE,
	tolerance = sqrt(.Machine$double.eps)
) {
	is_congruent.angle(x, y, ..., mod_turns = mod_turns, tolerance = tolerance)
}

#' @rdname is_equivalent
#' @export
is_equivalent.numeric <- function(x, y, ..., tolerance = sqrt(.Machine$double.eps)) {
	is_congruent.numeric(x, y, ..., tolerance = tolerance)
}

#' @rdname is_equivalent
#' @export
is_equivalent.Coord1D <- function(x, y, ..., tolerance = sqrt(.Machine$double.eps)) {
	if (!is_coord1d(y)) {
		y <- as_coord1d(y)
	}
	stopifnot(!any(is_degenerate(x)), !any(is_degenerate(y)))
	n <- max(length(x), length(y))
	x <- rep(x, length.out = n)
	y <- rep(y, length.out = n)
	is_equivalent(x$x, y$x, tolerance = tolerance)
}

#' @rdname is_equivalent
#' @export
is_equivalent.Coord2D <- function(x, y, ..., tolerance = sqrt(.Machine$double.eps)) {
	if (!is_coord2d(y)) {
		y <- as_coord2d(y)
	}
	stopifnot(!any(is_degenerate(x)), !any(is_degenerate(y)))
	n <- max(length(x), length(y))
	x <- rep(x, length.out = n)
	y <- rep(y, length.out = n)
	is_equivalent(x$x, y$x, tolerance = tolerance) &
		is_equivalent(x$y, y$y, tolerance = tolerance)
}

#' @rdname is_equivalent
#' @export
is_equivalent.Coord3D <- function(x, y, ..., tolerance = sqrt(.Machine$double.eps)) {
	if (!is_coord3d(y)) {
		y <- as_coord3d(y)
	}
	stopifnot(!any(is_degenerate(x)), !any(is_degenerate(y)))
	n <- max(length(x), length(y))
	x <- rep(x, length.out = n)
	y <- rep(y, length.out = n)
	is_equivalent(x$x, y$x, tolerance = tolerance) &
		is_equivalent(x$y, y$y, tolerance = tolerance) &
		is_equivalent(x$z, y$z, tolerance = tolerance)
}

#' @rdname is_equivalent
#' @export
is_equivalent.Point1D <- function(x, y, ..., tolerance = sqrt(.Machine$double.eps)) {
	if (!is_point1d(y)) {
		y <- as_point1d(y)
	}
	stopifnot(!any(is_degenerate(x)), !any(is_degenerate(y)))
	n <- max(length(x), length(y))
	x <- rep(standardize(x), length.out = n)
	y <- rep(standardize(y), length.out = n)
	is_equivalent(x$a, y$a, tolerance = tolerance) &
		is_equivalent(x$b, y$b, tolerance = tolerance)
}

#' @rdname is_equivalent
#' @export
is_equivalent.Line2D <- function(x, y, ..., tolerance = sqrt(.Machine$double.eps)) {
	if (!is_line2d(y)) {
		y <- as_line2d(y)
	}
	stopifnot(!any(is_degenerate(x)), !any(is_degenerate(y)))
	n <- max(length(x), length(y))
	x <- rep(standardize(x), length.out = n)
	y <- rep(standardize(y), length.out = n)
	is_equivalent(x$a, y$a, tolerance = tolerance) &
		is_equivalent(x$b, y$b, tolerance = tolerance) &
		is_equivalent(x$c, y$c, tolerance = tolerance)
}

#' @rdname is_equivalent
#' @export
is_equivalent.Plane3D <- function(x, y, ..., tolerance = sqrt(.Machine$double.eps)) {
	if (!is_plane3d(y)) {
		y <- as_plane3d(y)
	}
	stopifnot(!any(is_degenerate(x)), !any(is_degenerate(y)))
	n <- max(length(x), length(y))
	x <- rep(standardize(x), length.out = n)
	y <- rep(standardize(y), length.out = n)
	is_equivalent(x$a, y$a, tolerance = tolerance) &
		is_equivalent(x$b, y$b, tolerance = tolerance) &
		is_equivalent(x$c, y$c, tolerance = tolerance) &
		is_equivalent(x$d, y$d, tolerance = tolerance)
}

test_that("as_point1d()", {
	x <- c(2, 5, 7)
	p1 <- as_point1d(as_coord1d(x = x))
	p0a <- as_point1d("origin")
	p0b <- as_point1d("zero")
	p0c <- as_point1d(as_coord1d("origin"))

	expect_equal(p0a, p0b)
	expect_equal(p0a, p0c)
	expect_equal(p1, as_point1d(p1))
	expect_equal(p1, p1[1:3])
	expect_snapshot(print(p1))

	expect_warning(as_point1d("boo"))

	expect_equal(length(rep_len(p1, 3)), 3L)
	expect_equal(length(rep(p1, 2)), 6L)
	expect_equal(length(c(p1, p1)), 6L)
	expect_equal(is.na(p1), rep(FALSE, 3L))
	expect_equal(is.nan(p1), rep(FALSE, 3L))
	expect_equal(is.infinite(p1), rep(FALSE, 3L))
	expect_equal(is.finite(p1), rep(TRUE, 3L))
	expect_true(is.data.frame(as.data.frame(p1)))
	expect_true(is.list(as.list(p1)))
})

test_that("`is_equivalent.Point1D()`", {
	x <- c(2, 5, 7)
	c1 <- as_coord1d(x = x)
	p1 <- as_point1d(c1)
	expect_equal(is_equivalent(p1, as_point1d(a = 2, b = -10)), c(FALSE, TRUE, FALSE))
	expect_equal(is_equivalent(as_point1d(a = 2, b = -10), p1), c(FALSE, TRUE, FALSE))
	expect_equal(is_equivalent(p1, c1), rep(TRUE, 3L))
})

#' 2D polygon R6 Class
#'
#' `Polygon2D` is an [R6::R6Class()] object representing a two-dimensional
#' polygon.  It inherits from [Coord2D] so all transformation methods
#' (e.g. `$rotate()`, `$translate()`) and arithmetic operators work directly
#' on the polygon.
#'
#' @examples
#' vertices <- as_coord2d(x = c(0, 0.5, 1, 0.5), y = c(0.5, 1, 0.5, 0))
#' p <- as_polygon2d(vertices)
#' print(p)
#' p$is_convex
#' p$normals
#' h <- p$convex_hull
#'
#' # Transformations are inherited from Coord2D
#' p$rotate(degrees(45))
#' @export
Polygon2D <- R6Class(
	"Polygon2D",
	inherit = Coord2D,
	public = list(
		#' @param xyw A matrix with three columns representing (homogeneous)
		#'   coordinates.  The first two columns are x/y coordinates and the
		#'   last column is all ones.  Column names should be `"x"`, `"y"`,
		#'   and `"w"`.
		#' @param convex `NA` (default) to auto-detect convexity from `xyw`,
		#'   `TRUE` to assert convex (skip detection), or `FALSE` to mark as
		#'   concave.
		initialize = function(xyw, convex = NA) {
			super$initialize(xyw)
			private$convex <- as.logical(convex)
		},
		#' @param n Number of vertices to print.  If `NULL` print all.
		#' @param ... Passed to [format.default()].
		print = function(n = NULL, ...) {
			nv <- nrow(private$mat_xyw)
			if (is.null(n) || n > nv) {
				n <- nv
			}
			convex_str <- self$is_convex
			cat(
				"<Polygon2D[",
				nv,
				" vertices, convex = ",
				convex_str,
				"]>\n",
				sep = ""
			)
			if (n > 0) {
				print_mat(self$xyw[1:n, , drop = FALSE], ...)
			}
			invisible(self)
		},
		#' @param mat `r r2i_transform2d_mat`
		transform = function(mat = transform2d()) {
			private$hull_cache <- NULL
			private$normals_cache <- NULL
			private$edges_cache <- NULL
			super$transform(mat)
		}
	),
	private = list(
		convex = NA,
		hull_cache = NULL,
		normals_cache = NULL,
		edges_cache = NULL
	),
	active = list(
		#' @field is_convex Logical indicating whether the polygon is convex
		#'   (`TRUE`), concave (`FALSE`), or unknown (`NA`).
		is_convex = function() {
			if (is.na(private$convex)) {
				private$convex <- is_convex_xyw(private$mat_xyw)
			}
			private$convex
		},
		#' @field convex_hull A [Polygon2D] object of the convex hull.
		#'  If the polygon is already convex (`$convex == TRUE`) it returns itself.
		#'  For concave polygons the hull is computed on demand and cached until the next transformation.
		convex_hull = function() {
			if (isTRUE(private$convex)) {
				return(invisible(self))
			}
			if (!is.null(private$hull_cache)) {
				return(private$hull_cache)
			}
			# self$x / self$y apply any pending transforms
			idx <- rev(grDevices::chull(list(x = self$x, y = self$y)))
			hull <- Polygon2D$new(private$mat_xyw[idx, , drop = FALSE], convex = TRUE)
			private$hull_cache <- hull
			hull
		},
		#' @field normals A [Coord2D] object of unit edge normals
		#  (one per edge, computed on demand, and cached until the next transformation).
		normals = function() {
			if (!is.null(private$normals_cache)) {
				return(private$normals_cache)
			}
			private$apply_any_delayed_transformations()
			xyw <- private$mat_xyw
			n <- nrow(xyw)
			idx <- c(seq_len(n - 1L) + 1L, 1L)
			dx <- xyw[idx, 1L] - xyw[, 1L]
			dy <- xyw[idx, 2L] - xyw[, 2L]
			# Plain Coord2D (not Polygon2D) so normal2d() returns Coord2D
			edges <- Coord2D$new(cbind(x = dx, y = dy, w = rep.int(1, n)))
			norms <- normal2d(edges)
			private$normals_cache <- norms
			norms
		},
		#' @field edges A [Segment2D] object of the `n` polygon edges
		#'   (computed on demand and cached until the next transformation).
		edges = function() {
			if (!is.null(private$edges_cache)) {
				return(private$edges_cache)
			}
			private$apply_any_delayed_transformations()
			xyw <- private$mat_xyw
			n <- nrow(xyw)
			idx <- c(seq_len(n - 1L) + 1L, 1L)
			vec <- cbind(
				x = xyw[idx, 1L] - xyw[, 1L],
				y = xyw[idx, 2L] - xyw[, 2L]
			)
			seg <- Segment2D$new(xyw, vec)
			private$edges_cache <- seg
			seg
		}
	)
)

#' Cast to Polygon2D object
#'
#' `as_polygon2d()` casts to a [Polygon2D] object.
#'
#' @param x Object to cast.  Either a [Coord2D] object of vertices or a
#'   numeric vector of x-coordinates.
#' @param ... Ignored; only included for S3 method consistency.
#' @return A [Polygon2D] object.
#' @seealso [isotoxal_2ngon_polygon2d()] and [regular_ngon_polygon2d()] to
#'   directly construct common polygon shapes.
#' @examples
#' vertices <- as_coord2d(x = c(0, 1, 1, 0), y = c(0, 0, 1, 1))
#' p <- as_polygon2d(vertices)
#' p$is_convex
#' print(p)
#' plot(p)
#' @export
as_polygon2d <- function(x, ...) {
	UseMethod("as_polygon2d")
}

#' @rdname as_polygon2d
#' @param convex `NA` to auto-detect (default), `TRUE` to assert convex,
#'   `FALSE` to mark as concave.
#' @export
as_polygon2d.Coord2D <- function(x, convex = NA, ...) {
	chkDots(...)
	Polygon2D$new(x$xyw, convex = convex)
}

#' @rdname as_polygon2d
#' @export
as_polygon2d.Polygon2D <- function(x, convex = NA, ...) {
	chkDots(...)
	if (is.na(convex)) {
		return(x)
	}
	Polygon2D$new(x$xyw, convex = convex)
}

#' @rdname as_polygon2d
#' @param y Numeric vector of y-coordinates (used when `x` is numeric).
#' @export
as_polygon2d.numeric <- function(x, y = 0, convex = NA, ...) {
	chkDots(...)
	as_polygon2d(as_coord2d(x = x, y = y), convex = convex)
}

#' @rdname as_polygon2d
#' @param n Number of vertices in the approximating polygon (default `60L`).
#' @param type `"inner"` (default) for an inscribed polygon whose vertices lie
#'   on the ellipse, or `"outer"` for a circumscribed polygon whose edges are
#'   tangent to the ellipse.  Both polygons are convex.
#' @export
as_polygon2d.Ellipse2D <- function(x, n = 60L, type = c("inner", "outer"), ...) {
	chkDots(...)
	stopifnot(length(x) == 1L)
	type <- match.arg(type)
	by <- 360 / n
	from <- switch(type, inner = 0, outer = -by / 2)
	t <- seq(from, by = by, length.out = n)
	scale <- if (type == "outer") 1 / cospi(1 / n) else 1
	rx <- x$rx * scale
	ry <- x$ry * scale
	ct <- cos(x$theta)
	st <- sin(x$theta)
	vx <- x$x + rx * cospi(t / 180) * ct - ry * sinpi(t / 180) * st
	vy <- x$y + rx * cospi(t / 180) * st + ry * sinpi(t / 180) * ct
	as_polygon2d(as_coord2d(vx, vy), convex = TRUE)
}

# Internal: check convexity of a polygon from its raw xyw matrix
is_convex_xyw <- function(xyw) {
	n <- nrow(xyw)
	if (n < 3L) {
		return(FALSE)
	}
	x <- xyw[, 1L]
	y <- xyw[, 2L]
	i2 <- c(seq_len(n - 1L) + 1L, 1L)
	i3 <- c(seq_len(n - 2L) + 2L, 1L, 2L)
	dx1 <- x[i2] - x
	dy1 <- y[i2] - y
	dx2 <- x[i3] - x[i2]
	dy2 <- y[i3] - y[i2]
	# z-component of the 3D cross product of those two edge vectors.
	# Its sign tells you which way you turned at vertex i+1.
	# - cross > 0: left turn (counterclockwise)
	# - cross < 0: right turn (clockwise)
	# - cross = 0: straight (collinear)
	cross <- dx1 * dy2 - dy1 * dx2
	# polygon is convex if and only if all turns go the same direction
	all(cross >= 0) || all(cross <= 0)
}

test_that("as_polygon2d()", {
	# from Coord2D
	v <- as_coord2d(x = c(0, 1, 1, 0), y = c(0, 0, 1, 1))
	p <- as_polygon2d(v)
	expect_r6_class(p, "Polygon2D")
	expect_true(is_polygon2d(p))
	expect_true(is_coord2d(p)) # Polygon2D is a Coord2D subclass
	expect_false(is_polygon2d(v))
	expect_true(p$is_convex)
	expect_equal(length(p), 4L)

	# from numeric
	p2 <- as_polygon2d(x = c(0, 1, 1, 0), y = c(0, 0, 1, 1))
	expect_r6_class(p2, "Polygon2D")

	# explicit convex override
	p3 <- as_polygon2d(v, convex = TRUE)
	expect_true(p3$is_convex)
	p4 <- as_polygon2d(v, convex = FALSE)
	expect_false(p4$is_convex)

	# print
	expect_snapshot(print(p))
})

test_that("as_polygon2d.Polygon2D()", {
	p <- as_polygon2d(as_coord2d(x = c(0, 1, 1, 0), y = c(0, 0, 1, 1)))
	expect_identical(as_polygon2d(p), p)
	expect_false(as_polygon2d(p, convex = FALSE)$is_convex)
})

test_that("Polygon2D normals", {
	# Unit square vertices CCW
	v <- as_coord2d(x = c(0, 1, 1, 0), y = c(0, 0, 1, 1))
	p <- as_polygon2d(v)
	norms <- p$normals
	expect_r6_class(norms, "Coord2D")
	expect_false(is_polygon2d(norms)) # normals are plain Coord2D, not Polygon2D
	expect_equal(length(norms), 4L)
	expect_equal(abs(norms), rep(1, 4L))

	# Cached: second access returns same object
	expect_identical(p$normals, norms)

	# Cache invalidated after transform
	p$rotate(degrees(45))
	expect_false(identical(p$normals, norms))
})

test_that("Polygon2D$convex_hull", {
	# Convex polygon: convex_hull returns itself
	v <- as_coord2d(x = c(0, 1, 1, 0), y = c(0, 0, 1, 1))
	p <- as_polygon2d(v, convex = TRUE)
	expect_identical(p$convex_hull, p)

	# Concave polygon (star-like): hull is a different polygon
	# Simple concave: a "C" shape or indented square
	vv <- as_coord2d(
		x = c(0, 0.5, 0.5, 1, 1, 0.5, 0.5, 0),
		y = c(0, 0, 0.4, 0.4, 0.6, 0.6, 1, 1)
	)
	pc <- as_polygon2d(vv, convex = FALSE)
	hull <- pc$convex_hull
	expect_r6_class(hull, "Polygon2D")
	expect_true(hull$is_convex)
	expect_true(length(hull) <= length(pc))

	# Cached: second call returns same object
	expect_identical(pc$convex_hull, hull)

	# Cache invalidated after transform
	pc$rotate(degrees(45))
	expect_false(identical(pc$convex_hull, hull))
})

test_that("convex_hull2d() returns Polygon2D", {
	# From Coord2D
	pts <- as_coord2d(x = rnorm(20), y = rnorm(20))
	hull <- convex_hull2d(pts)
	expect_r6_class(hull, "Polygon2D")
	expect_true(hull$is_convex)
	expect_true(is_coord2d(hull)) # backwards compatible

	# From convex Polygon2D: returns itself
	v <- as_coord2d(x = c(0, 1, 1, 0), y = c(0, 0, 1, 1))
	p <- as_polygon2d(v, convex = TRUE)
	expect_identical(convex_hull2d(p), p)
})

test_that("Polygon2D convex auto-detection", {
	# Convex (CCW square)
	expect_true(
		as_polygon2d(as_coord2d(
			x = c(0, 1, 1, 0),
			y = c(0, 0, 1, 1)
		))$is_convex
	)

	# Convex (CW square)
	expect_true(
		as_polygon2d(as_coord2d(
			x = c(0, 0, 1, 1),
			y = c(0, 1, 1, 0)
		))$is_convex
	)

	# Degenerate: fewer than 3 vertices → FALSE
	expect_false(as_polygon2d(as_coord2d(x = 1:2, y = 1:2))$is_convex)

	# Concave C-shape
	expect_false(
		as_polygon2d(as_coord2d(
			x = c(0, 0.5, 0.5, 1, 1, 0.5, 0.5, 0),
			y = c(0, 0, 0.4, 0.4, 0.6, 0.6, 1, 1)
		))$is_convex
	)
})

test_that("has_overlap2d() polygon-polygon", {
	sq1 <- as_polygon2d(as_coord2d(
		x = c(0, 1, 1, 0),
		y = c(0, 0, 1, 1)
	))
	sq2 <- as_polygon2d(as_coord2d(
		x = c(0.5, 1.5, 1.5, 0.5),
		y = c(0.5, 0.5, 1.5, 1.5)
	))
	sq3 <- as_polygon2d(as_coord2d(
		x = c(2, 3, 3, 2),
		y = c(2, 2, 3, 3)
	))

	expect_true(has_overlap2d(sq1, sq2))
	expect_false(has_overlap2d(sq1, sq3))

	# Touching corners: strictly no non-zero-area overlap
	sq4 <- as_polygon2d(as_coord2d(
		x = c(1, 2, 2, 1),
		y = c(1, 1, 2, 2)
	))
	expect_false(has_overlap2d(sq1, sq4))
})

test_that("has_overlap2d() polygon-circle", {
	sq <- as_polygon2d(as_coord2d(
		x = c(0, 1, 1, 0),
		y = c(0, 0, 1, 1)
	))
	circ_in <- as_ellipse2d(as_coord2d(0.5, 0.5), r = 0.3)
	circ_out <- as_ellipse2d(as_coord2d(3, 3), r = 0.3)
	circ_corner <- as_ellipse2d(as_coord2d(1.5, 1.5), r = 0.3)

	expect_true(has_overlap2d(sq, circ_in))
	expect_false(has_overlap2d(sq, circ_out))
	expect_false(has_overlap2d(sq, circ_corner))

	# Dispatching from Circle2D side
	expect_true(has_overlap2d(circ_in, sq))
	expect_false(has_overlap2d(circ_out, sq))

	# Vectorized circles
	circs <- as_ellipse2d(
		x = c(0.5, 3),
		y = c(0.5, 3),
		r = c(0.3, 0.3)
	)
	expect_equal(has_overlap2d(sq, circs), c(TRUE, FALSE))
})

test_that("has_overlap2d() concave polygon returns NA with warning", {
	# Concave polygon (C-shape)
	vc <- as_coord2d(
		x = c(0, 0.5, 0.5, 1, 1, 0.5, 0.5, 0),
		y = c(0, 0, 0.4, 0.4, 0.6, 0.6, 1, 1)
	)
	pc <- as_polygon2d(vc, convex = FALSE)
	sq <- as_polygon2d(as_coord2d(
		x = c(0.3, 0.7, 0.7, 0.3),
		y = c(0.3, 0.3, 0.7, 0.7)
	))

	# Overlapping bounding boxes + convex hulls → NA with warning
	expect_warning(res <- has_overlap2d(pc, sq), "concave")
	expect_true(is.na(res))

	# Clearly disjoint → FALSE without warning
	sq_far <- as_polygon2d(as_coord2d(
		x = c(5, 6, 6, 5),
		y = c(5, 5, 6, 6)
	))
	expect_false(has_overlap2d(pc, sq_far))
})

test_that("has_overlap2d() concave polygon and circle (bbox_overlap_pc)", {
	pc <- as_polygon2d(
		as_coord2d(
			x = c(0, 0.5, 0.5, 1, 1, 0.5, 0.5, 0),
			y = c(0, 0, 0.4, 0.4, 0.6, 0.6, 1, 1)
		),
		convex = FALSE
	)
	# Circle outside bbox → FALSE via bbox_overlap_pc
	expect_false(has_overlap2d(pc, as_ellipse2d(as_coord2d(5, 5), r = 0.3)))
	# Circle inside bbox, overlapping convex hull → NA with warning
	expect_warning(
		res <- has_overlap2d(pc, as_ellipse2d(as_coord2d(0.5, 0.5), r = 0.3)),
		"concave"
	)
	expect_true(is.na(res))
})

test_that("regular_ngon_polygon2d()", {
	p6 <- regular_ngon_polygon2d(6)
	expect_r6_class(p6, "Polygon2D")
	expect_true(p6$is_convex)
	expect_equal(length(p6), 6L)

	# center and radius are respected
	p <- regular_ngon_polygon2d(4, x = 1, y = 2, radius = 2)
	expect_equal(mean(p)$x, 1, tolerance = 1e-10)
	expect_equal(mean(p)$y, 2, tolerance = 1e-10)
	expect_equal(distance2d(as_coord2d(1, 2), p), rep(2, 4L), tolerance = 1e-10)

	# theta shifts the first vertex
	p0 <- regular_ngon_polygon2d(4, theta = degrees(0))
	expect_equal(p0$x[1], 0.5, tolerance = 1e-10)
	expect_equal(p0$y[1], 0, tolerance = 1e-10)
})

test_that("rectangle_polygon2d()", {
	p <- rectangle_polygon2d()
	expect_r6_class(p, "Polygon2D")
	expect_true(p$is_convex)
	expect_equal(length(p), 4L)
	expect_equal(range(p)$x, c(-0.5, 0.5), tolerance = 1e-10)
	expect_equal(range(p)$y, c(-0.5, 0.5), tolerance = 1e-10)

	# width, height, center respected
	p2 <- rectangle_polygon2d(width = 2, height = 3, x = 1, y = -1)
	expect_equal(range(p2)$x, c(0, 2), tolerance = 1e-10)
	expect_equal(range(p2)$y, c(-2.5, 0.5), tolerance = 1e-10)

	# rotation: a 45-degree square has equal x and y extents
	p3 <- rectangle_polygon2d(width = 1, height = 1, theta = degrees(45))
	expect_equal(range(p3)$x, range(p3)$y, tolerance = 1e-10)
})

test_that("isotoxal_2ngon_polygon2d()", {
	p <- isotoxal_2ngon_polygon2d(5, d = 2)
	expect_r6_class(p, "Polygon2D")
	expect_equal(length(p), 10L)
	expect_false(p$is_convex)

	# alpha and beta_ext routes give same result as d
	s <- isotoxal_2ngon_inner_radius(5, d = 2)
	p_alpha <- isotoxal_2ngon_polygon2d(5, alpha = 180 * (1 - 4 / 5))
	p_scale <- isotoxal_2ngon_polygon2d(5, radial_scale = s)
	expect_equal(p$x, p_alpha$x, tolerance = 1e-10)
	expect_equal(p$x, p_scale$x, tolerance = 1e-10)

	# star_polygon2d() is an alias
	p2 <- star_polygon2d(5, d = 2)
	expect_equal(p$x, p2$x)
	expect_equal(p$y, p2$y)

	# outer vertices sit on circumradius circle
	outer_idx <- seq(1, 10, by = 2)
	center <- as_coord2d(0, 0)
	expect_equal(
		distance2d(center, p)[outer_idx],
		rep(0.5, 5L),
		tolerance = 1e-10
	)

	# convex when radial_scale >= cos(pi/n)
	p_convex <- isotoxal_2ngon_polygon2d(5, radial_scale = cospi(1 / 5))
	expect_true(p_convex$is_convex)
	p_wide <- isotoxal_2ngon_polygon2d(5, radial_scale = 0.9)
	expect_true(p_wide$is_convex)
})

test_that("has_overlap2d.Polygon2D() errors on unsupported y class", {
	sq <- as_polygon2d(as_coord2d(x = c(0, 1, 1, 0), y = c(0, 0, 1, 1)))
	expect_error(has_overlap2d(sq, "foo"), "Don't know how to check overlap with")
})

#' Rectangle polygon
#'
#' `rectangle_polygon2d()` creates a rectangle [Polygon2D] object
#' centered at `(x, y)`.
#'
#' @param width Width of the rectangle (default `1`).
#' @param height Height of the rectangle (default `1`).
#' @param x X-coordinate of the center (default `0`).
#' @param y Y-coordinate of the center (default `0`).
#' @param theta Rotation angle (default `degrees(0)`).
#'   Will be coerced by [degrees()].
#' @return A [Polygon2D] object with `$is_convex == TRUE`.
#' @seealso [isotoxal_2ngon_polygon2d()] for isotoxal star polygons.
#'   [regular_ngon_polygon2d()] for regular convex polygons.
#' @examples
#' # A unit square
#' p <- rectangle_polygon2d()
#' p$is_convex
#' plot(p)
#'
#' # A rotated rectangle
#' p2 <- rectangle_polygon2d(width = 2, height = 1, theta = degrees(45))
#' plot(p2)
#' @export
rectangle_polygon2d <- function(
	width = 1,
	height = 1,
	x = 0,
	y = 0,
	theta = degrees(0)
) {
	hw <- width / 2
	hh <- height / 2
	dx <- c(-hw, hw, hw, -hw)
	dy <- c(-hh, -hh, hh, hh)
	ct <- cos(degrees(theta))
	st <- sin(degrees(theta))
	vx <- x + dx * ct - dy * st
	vy <- y + dx * st + dy * ct
	as_polygon2d(as_coord2d(vx, vy), convex = TRUE)
}

# The code for this function is in `standalone-affine-settings.r`

#' Compute `grid` affine transformation feature viewports and transformation functions
#'
#' `affine_settings()` computes `grid` group affine transformation feature viewport and transformation
#' function settings given the (x,y) coordinates of the corners of the
#' affine transformed "viewport" one wishes to draw in.
#'
#' @section Usage in other packages:
#'
#' To avoid taking a dependency on `affiner` you may copy the source of `affine_settings()`
#' into your own package under the permissive Unlicense.  Either use
#' `usethis::use_standalone("trevorld/affiner", "standalone-affine-settings.r")` or
#' copy the file `standalone-affine-settings.r` into your `R` directory and add `grid`
#' to the `Imports` of your `DESCRIPTION` file.
#'
#' @param xy An R object with named elements `x` and `y` representing the (x,y) coordinates
#'           of the affine transformed "viewport" one wishes to draw in.
#'           The (x,y) coordinates of the "viewport" should be in
#'           "upper left", "lower left", "lower right", and "upper right" order
#'           (this ordering should be from the perspective of **before**
#'           the "affine transformation" of the "viewport").
#' @param unit Which [grid::unit()] to assume the `xy` "x" and "y" coordinates are expressed in.
#' @examples
#' if (require("grid")) {
#'   grob <- grobTree(rectGrob(gp = gpar(fill = "blue", col = NA)),
#'                    circleGrob(gp=gpar(fill="yellow", col = NA)),
#'                    textGrob("RSTATS", gp=gpar(fontsize=32)))
#'   grid.newpage()
#'   pushViewport(viewport(width=unit(4, "in"), height=unit(2, "in")))
#'   grid.draw(grob)
#'   popViewport()
#' }
#' if (require("grid") &&
#'     getRversion() >= "4.2.0" &&
#'     isTRUE(dev.capabilities()$transformations)) {
#'   # Only works if active graphics device supports affine transformations
#'   # such as `png(type="cairo")` on R 4.2+
#'   vp_define <- viewport(width=unit(2, "in"), height=unit(2, "in"))
#'   settings <- affine_settings(xy = list(x = c(1/3, 0/3, 2/3, 3/3),
#'                                         y = c(2/3, 1/3, 1/3, 2/3)),
#'                               unit = "snpc")
#'   affine <- affineGrob(grob,
#'                        vp_define=vp_define,
#'                        transform = settings$transform,
#'                        vp_use = settings$vp)
#'   grid.newpage()
#'   grid.draw(affine)
#' }
#' if (require("grid") &&
#'     getRversion() >= "4.2.0" &&
#'     isTRUE(dev.capabilities()$transformations)) {
#'   # Only works if active graphics device supports affine transformations
#'   # such as `png(type="cairo")` on R 4.2+
#'   settings <- affine_settings(xy = list(x = c(3/3, 2/3, 0/3, 1/3),
#'                                         y = c(2/3, 1/3, 1/3, 2/3)),
#'                               unit = "snpc")
#'   affine <- affineGrob(grob,
#'                        vp_define=vp_define,
#'                        transform = settings$transform,
#'                        vp_use = settings$vp)
#'   grid.newpage()
#'   grid.draw(affine)
#' }
#' @return A named list with the following group affine transformation feature viewport
#'         and functions settings:\describe{
#'   \item{transform}{An affine transformation function to pass to [affineGrob()] or [useGrob()].
#'                    If `getRversion()` is less than `"4.2.0"` will instead be `NULL`.}
#'   \item{vp}{A [grid::viewport()] object to pass to [affineGrob()] or [useGrob()].}
#'   \item{sx}{x-axis sx factor}
#'   \item{flipX}{whether the affine transformed "viewport" is "flipped" horizontally}
#'   \item{x}{x-coordinate for viewport}
#'   \item{y}{y-coordinate for viewport}
#'   \item{width}{Width of viewport}
#'   \item{height}{Height of viewport}
#'   \item{default.units}{Default [grid::unit()] for viewport}
#'   \item{angle}{angle for viewport}
#' }
#' @seealso Intended for use with [affineGrob()] and [grid::useGrob()].
#'          See <https://www.stat.auckland.ac.nz/~paul/Reports/GraphicsEngine/groups/groups.html>
#'          for more information about the group affine transformation feature.
#' @export
affine_settings <- affine_settings

not_finite <- Negate(is.finite) # also tests if NA or NaN

is_degenerate <- function(x) {
	UseMethod("is_degenerate")
}

#' @export
is_degenerate.Point1D <- function(x) {
	not_finite(x) | is_congruent(x$a, 0)
}

#' @export
is_degenerate.Line2D <- function(x) {
	not_finite(x) | (is_congruent(x$a, 0) & is_congruent(x$b, 0))
}

#' @export
is_degenerate.Plane3D <- function(x) {
	not_finite(x) | (is_congruent(x$a, 0) & is_congruent(x$b, 0) & is_congruent(x$c, 0))
}

#' @export
is_degenerate.Coord1D <- function(x) {
	not_finite(x)
}

#' @export
is_degenerate.Coord2D <- function(x) {
	not_finite(x)
}

#' @export
is_degenerate.Coord3D <- function(x) {
	not_finite(x)
}

#' Angle vectors
#'
#' `angle()` creates angle vectors with user specified angular unit.
#` `degrees()`, `gradians()`,`pi_radians()`, radians()`, and `turns()` are convenience wrappers
#' around [as_angle()] for those angular units.
#'
#' @param x An angle vector or an object to convert to it (such as a numeric vector)
#' @param unit A string of the desired angular unit.  Supports the following strings
#'            (note we ignore any punctuation and space characters as well as any trailing `s`'s
#'             e.g. "half turns" will be treated as equivalent to "halfturn"):
#'
#'   * "deg" or "degree"
#'   * "half-revolution", "half-turn", or "pi-radian"
#'   * "gon", "grad", "grade", or "gradian"
#'   * "rad" or "radian"
#'   * "rev", "revolution", "tr", or "turn"
#'
#' @return A numeric vector of class "angle".
#'         Its "unit" attribute is a standardized string of the specified angular unit.
#'
#' @seealso [as_angle()], [angular_unit()], and [angle-methods].
#'   <https://en.wikipedia.org/wiki/Angle#Units> for more information about angular units.
#' @examples
#'   # Different representations of the "same" angle
#'   angle(180, "degrees")
#'   angle(pi, "radians")
#'   angle(0.5, "turns")
#'   angle(200, "gradians")
#'   pi_radians(1)
#'
#'   a1 <- angle(180, "degrees")
#'   angular_unit(a1)
#'   is_angle(a1)
#'   as.numeric(a1, "radians")
#'   cos(a1)
#'
#'   a2 <- as_angle(a1, "radians")
#'   angular_unit(a2)
#'   is_congruent(a1, a2)
#' @export
angle <- function(x = numeric(), unit = getOption("affiner_angular_unit", "degrees")) {
	as_angle(x, unit = unit)
}

new_angle <- function(x, unit) {
	class(x) <- c("angle", class(x))
	attr(x, "unit") <- unit
	x
}

#' @rdname angle
#' @export
degrees <- function(x) {
	as_angle(x, "degrees")
}

#' @rdname angle
#' @export
gradians <- function(x) {
	as_angle(x, "gradians")
}

#' @rdname angle
#' @export
pi_radians <- function(x) {
	as_angle(x, "pi-radians")
}

#' @rdname angle
#' @export
radians <- function(x) {
	as_angle(x, "radians")
}

#' @rdname angle
#' @export
turns <- function(x) {
	as_angle(x, "turns")
}

#' Test whether an object is an angle vector
#'
#' `is_angle()` tests whether an object is an angle vector
#'
#' @param x An object
#' @return A logical value
#' @examples
#' a <- angle(180, "degrees")
#' is_angle(a)
#' is_angle(pi)
#' @export
is_angle <- function(x) inherits(x, "angle")

standardize_angular_unit <- function(unit) {
	stopifnot(length(unit) == 1L)
	if (unit %in% c("degrees", "pi-radians", "gradians", "radians", "turns")) {
		unit
	} else {
		switch(
			gsub("[[:punct:]]|[[:space:]]|s$", "", unit),
			degree = "degrees",
			deg = "degrees",

			halfturn = "pi-radians",
			halfrevolution = "pi-radians",
			piradian = "pi-radians",

			gon = "gradians",
			grad = "gradians",
			grade = "gradians",
			gradian = "gradians",

			radian = "radians",
			rad = "radians",

			rev = "turns",
			revolution = "turns",
			tr = "turns",
			turn = "turns",
			stop(paste("Do not recognize angular unit", sQuote(unit)))
		)
	}
}

#' Get/set angular unit of angle vectors
#'
#' `angular_unit()` gets/sets the angular unit of [angle()] vectors.
#'
#' @param x An [angle()] vector
#' @param value A string of the desired angular unit.  See [angle()] for supported strings.
#' @return `angular_unit()` returns a string of `x`'s angular unit.
#' @examples
#' a <- angle(seq(0, 360, by = 90), "degrees")
#' angular_unit(a)
#' print(a)
#' angular_unit(a) <- "turns"
#' angular_unit(a)
#' print(a)
#' @export
angular_unit <- function(x) {
	stopifnot(is_angle(x))
	attr(x, "unit")
}

#' @rdname angular_unit
#' @export
`angular_unit<-` <- function(x, value) {
	stopifnot(is_angle(x))
	unit <- standardize_angular_unit(value)
	if (unit == attr(x, "unit")) {
		x
	} else {
		new_angle(as.numeric(x, unit = unit), unit)
	}
}

#' Implemented base methods for angle vectors
#'
#' We implemented methods for several base generics for the [angle()] vectors.
#'
#' * Mathematical [Ops] (in particular `+` and `-`)
#'   for two angle vectors will (if necessary)
#'   set the second vector's [angular_unit()] to match the first.
#' * [as.numeric()] takes a `unit` argument which can be used to convert angles into other angular units
#'    e.g. `angle(x, "degrees") |> as.numeric("radians")` to cast a numeric vector `x` from degrees to radians.
#' * [abs()] will calculate the angle modulo full turns.
#' * Use [is_congruent()] to test if two angles are congruent instead of `==` or `all.equal()`.
#' * Not all implemented methods are documented here and since [angle()] is a
#'   [numeric()] class many other S3 generics
#'   besides the explicitly implemented ones should also work with it.
#'
#' @return Typical values as usually returned by these base generics.
#' @examples
#'   # Two "congruent" angles
#'   a1 <- angle(180, "degrees")
#'   a2 <- angle(pi, "radians")
#'
#'   print(a1)
#'   print(a1, unit = "radians")
#'   print(a1, unit = "pi-radians")
#'
#'   cos(a1)
#'   sin(a1)
#'   tan(a1)
#'
#'   # mathematical operations will coerce second `angle()` object to
#'   # same `angular_unit()` as the first one
#'   a1 + a2
#'   a1 - a2
#'
#'   as.numeric(a1)
#'   as.numeric(a1, "radians")
#'   as.numeric(a1, "turns")
#'
#'   # Use `is_congruent()` to check if two angles are "congruent"
#'   a1 == a2
#'   isTRUE(all.equal(a1, a2))
#'   is_congruent(a1, a2)
#'   is_congruent(a1, a2, mod_turns = FALSE)
#'   a3 <- angle(-180, "degrees") # Only congruent modulus full turns
#'   a1 == a3
#'   isTRUE(all.equal(a1, a2))
#'   is_congruent(a1, a3)
#'   is_congruent(a1, a3, mod_turns = FALSE)
#'
#' @param x [angle()] vector
#' @param ... Passed to [print.default()]
#' @inheritParams angle
#' @name angle-methods
NULL

#' @export
c.angle <- function(...) {
	l_dots <- list(...)
	unit <- angular_unit(l_dots[[1]])
	l_angles <- lapply(l_dots, as.numeric, unit = unit)
	new_angle(unlist(l_angles), unit)
}

#' @export
`[.angle` <- function(x, i) {
	new_angle(unclass(x)[i], angular_unit(x))
}

#' @export
rep.angle <- function(x, ..., length.out = NA_integer_) {
	if (isTRUE(length(x) == length.out)) {
		return(x)
	}
	unit <- angular_unit(x)
	new_angle(rep(as.numeric(x), ..., length.out = length.out), unit)
}

#' @rdname angle-methods
#' @export
as.double.angle <- function(x, unit = angular_unit(x), ...) {
	a <- NextMethod()
	if (missing(unit)) {
		return(a)
	}
	unit <- standardize_angular_unit(unit)
	unit0 <- angular_unit(x)
	if (unit == unit0) {
		a
	} else if (unit == "radians") {
		to_radians(a, unit0)
	} else if (unit0 == "radians") {
		from_radians(a, unit)
	} else {
		from_piradians(to_piradians(a, unit0), unit)
	}
}

#' @rdname angle-methods
#' @param modulus Numeric vector representing the complex numbers' modulus
#' @export
as.complex.angle <- function(x, modulus = 1, ...) {
	complex(modulus = modulus, argument = as.numeric(x, "radians"))
}

#' @rdname angle-methods
#' @param use_unicode If `TRUE` use Unicode symbols as appropriate.
#' @export
format.angle <- function(x, unit = angular_unit(x), ..., use_unicode = is_utf8_output()) {
	if (!missing(unit)) {
		x <- as_angle(x, unit = standardize_angular_unit(unit))
	}
	suffix <- switch(
		angular_unit(x),
		degrees = ifelse(use_unicode, "\u00b0", " deg"),
		"pi-radians" = ifelse(use_unicode, "\u03c0 rad", "*pi rad"),
		gradians = " gon",
		radians = " rad",
		turns = " tr"
	)
	if (length(x)) {
		s <- paste0(format(as.numeric(x), ...), suffix)
		is.na(s) <- is.na(x)
		s
	} else {
		character()
	}
}

#' @rdname angle-methods
#' @export
print.angle <- function(x, unit = angular_unit(x), ..., use_unicode = is_utf8_output()) {
	cat("<angle<", attr(x, "unit"), ">[", length(x), "]>\n", sep = "")
	if (length(x)) {
		print.default(format.angle(x, unit = unit, ..., use_unicode = use_unicode), quote = FALSE)
	}
}

one_turn <- function(unit) {
	switch(unit, degrees = 360, "pi-radians" = 2, gradians = 400, radians = 2 * pi, turns = 1)
}

#' @rdname angle-methods
#' @export
abs.angle <- function(x) {
	unit <- angular_unit(x)
	new_angle(as.numeric(x) %% one_turn(unit), unit)
}

#' @export
cos.angle <- function(x) {
	cospi(as.numeric(x, "pi-radians"))
}

#' @export
sin.angle <- function(x) {
	sinpi(as.numeric(x, "pi-radians"))
}

#' @export
tan.angle <- function(x) {
	tanpi(as.numeric(x, "pi-radians"))
}

#' @export
Arg.angle <- function(z) {
	as.numeric(z, "radians")
}

#' @export
Ops.angle <- function(e1, e2) {
	if (missing(e2)) {
		NextMethod()
	} else {
		if (is_angle(e1) && is_angle(e2)) {
			if (angular_unit(e1) != angular_unit(e2)) {
				angular_unit(e2) <- angular_unit(e1)
			}
			NextMethod()
		} else {
			NextMethod()
		}
	}
}

to_piradians <- function(x, unit = "pi-radians") {
	switch(
		unit,
		degrees = x / 180,
		"pi-radians" = x,
		gradians = x / 200,
		radians = x / pi,
		turns = 2 * x
	)
}

from_piradians <- function(x, unit = "pi-radians") {
	switch(
		unit,
		degrees = 180 * x,
		"pi-radians" = x,
		gradians = 200 * x,
		radians = pi * x,
		turns = 0.5 * x
	)
}

to_radians <- function(x, unit = "radians") {
	switch(
		unit,
		degrees = pi * x / 180,
		"pi-radians" = pi * x,
		gradians = pi * x / 200,
		radians = x,
		turns = 2 * pi * x
	)
}

from_radians <- function(x, unit = "radians") {
	switch(
		unit,
		degrees = 180 * x / pi,
		"pi-radians" = x / pi,
		gradians = 200 * x / pi,
		radians = x,
		turns = x / (2 * pi)
	)
}

#' Cast to angle vector
#'
#' `as_angle()` casts to an [angle()] vector
#'
#' @param x An R object to convert to a [angle()] vector
#' @param ... Further arguments passed to or from other methods
#' @inheritParams angle
#' @return An [angle()] vector
#' @examples
#' as_angle(angle(pi, "radians"), "pi-radians")
#' as_angle(complex(real = 0, imaginary = 1), "degrees")
#' as_angle(as_coord2d(x = 0, y = 1), "turns")
#' as_angle(200, "gradians")
#'
#' @export
as_angle <- function(x, unit = getOption("affiner_angular_unit", "degrees"), ...) {
	UseMethod("as_angle")
}

#' @rdname as_angle
#' @export
as_angle.angle <- function(x, unit = getOption("affiner_angular_unit", "degrees"), ...) {
	unit <- standardize_angular_unit(unit)
	if (unit == angular_unit(x)) {
		x
	} else {
		new_angle(as.numeric(x, unit), unit)
	}
}

#' @rdname as_angle
#' @export
as_angle.character <- function(x, unit = getOption("affiner_angular_unit", "degrees"), ...) {
	unit <- standardize_angular_unit(unit)
	a <- vapply(x, as_angle_character_helper, double(1), USE.NAMES = FALSE, unit = unit)
	if (any(is.na(a) & !is.na(x))) {
		warning("NAs introduced by coercion")
	}
	new_angle(a, unit)
}

as_angle_character_helper <- function(x, unit) {
	switch(x, "x-axis" = 0, "y-axis" = from_piradians(0.5, unit), NA_real_)
}

#' @rdname as_angle
#' @export
as_angle.complex <- function(x, unit = getOption("affiner_angular_unit", "degrees"), ...) {
	unit <- standardize_angular_unit(unit)
	a <- from_radians(Arg(x), unit)
	new_angle(a, unit)
}

#' @rdname as_angle
#' @export
as_angle.Coord2D <- function(x, unit = getOption("affiner_angular_unit", "degrees"), ...) {
	unit <- standardize_angular_unit(unit)
	a <- from_radians(atan2(x$y, x$x), unit)
	new_angle(a, unit)
}

#' @rdname as_angle
#' @param type Use "azimuth" to calculate the azimuthal angle and "inclination" to calculate the inclination angle aka polar angle.
#' @export
as_angle.Coord3D <- function(
	x,
	unit = getOption("affiner_angular_unit", "degrees"),
	type = c("azimuth", "inclination"),
	...
) {
	unit <- standardize_angular_unit(unit)
	type <- match.arg(type)
	switch(
		type,
		azimuth = arctangent(x = x$x, y = x$y, unit = unit),
		inclination = arccosine(x$z / abs(x), unit = unit)
	)
}

#' @rdname as_angle
#' @export
as_angle.Line2D <- function(x, unit = getOption("affiner_angular_unit", "degrees"), ...) {
	arctangent(y = x$a, x = -x$b, unit = unit)
}

#' @rdname as_angle
#' @export
as_angle.Plane3D <- function(
	x,
	unit = getOption("affiner_angular_unit", "degrees"),
	type = c("azimuth", "inclination"),
	...
) {
	as_angle.Coord3D(normal3d(x), unit = unit, type = type)
}

#' @rdname as_angle
#' @export
as_angle.numeric <- function(x, unit = getOption("affiner_angular_unit", "degrees"), ...) {
	unit <- standardize_angular_unit(unit)
	new_angle(x, unit)
}

expect_congruent <- function(x, y, ...) expect_true(is_congruent(x, y, ...))

test_that("standardize_angular_unit()", {
	expect_equal(standardize_angular_unit("degrees"), "degrees")
	expect_equal(standardize_angular_unit("degree"), "degrees")
	expect_equal(standardize_angular_unit("deg"), "degrees")

	expect_equal(standardize_angular_unit("half-turn"), "pi-radians")
	expect_equal(standardize_angular_unit("half-revolution"), "pi-radians")
	expect_equal(standardize_angular_unit("pi * radians"), "pi-radians")

	expect_equal(standardize_angular_unit("gradians"), "gradians")
	expect_equal(standardize_angular_unit("gradian"), "gradians")
	expect_equal(standardize_angular_unit("grade"), "gradians")
	expect_equal(standardize_angular_unit("grad"), "gradians")
	expect_equal(standardize_angular_unit("gon"), "gradians")

	expect_equal(standardize_angular_unit("radians"), "radians")
	expect_equal(standardize_angular_unit("radian"), "radians")
	expect_equal(standardize_angular_unit("rad"), "radians")

	expect_equal(standardize_angular_unit("revolution"), "turns")
	expect_equal(standardize_angular_unit("rev"), "turns")
	expect_equal(standardize_angular_unit("turn"), "turns")
	expect_equal(standardize_angular_unit("tr"), "turns")

	expect_error(standardize_angular_unit("foobars"), "Do not recognize angular unit 'foobars'")
})

test_that("format.angle() and print.angle()", {
	skip_if_not(l10n_info()[["UTF-8"]])
	skip_if_not_installed("withr")
	withr::local_options(cli.unicode = TRUE, affiner_angular_unit = "degrees")
	expect_equal(format(angle(180, "degrees")), "180\u00b0")
	expect_equal(format(angle(200, "gradians")), "200 gon")
	expect_equal(format(angle(pi, "radians")), "3.141593 rad")
	expect_equal(format(angle(1, "pi-radians")), "1\u03c0 rad")
	expect_equal(format(angle(0.5, "turns")), "0.5 tr")
	expect_equal(format(angle()), character())

	expect_output(print(angle()), "<angle<degrees>\\[0\\]>")
	expect_output(print(angle(200, "gradians")), "\\[1\\] 200 gon")
})

test_that("to_radians() / from_radians()", {
	expect_equal(from_radians(to_radians(180, "degrees"), "degrees"), 180)
	expect_equal(from_radians(to_radians(1, "pi-radians"), "pi-radians"), 1)
	expect_equal(from_radians(to_radians(100, "gradians"), "gradians"), 100)
	expect_equal(from_radians(to_radians(pi, "radians"), "radians"), pi)
	expect_equal(from_radians(to_radians(0.5, "turns"), "turns"), 0.5)
})

test_that("to_piradians() / from_piradians()", {
	expect_equal(from_piradians(to_piradians(180, "degrees"), "degrees"), 180)
	expect_equal(from_piradians(to_piradians(1, "pi-radians"), "pi-radians"), 1)
	expect_equal(from_piradians(to_piradians(100, "gradians"), "gradians"), 100)
	expect_equal(from_piradians(to_piradians(pi, "radians"), "radians"), pi)
	expect_equal(from_piradians(to_piradians(0.5, "turns"), "turns"), 0.5)
})

test_that("angular_unit()", {
	a <- angle(seq(0, 360, by = 90), "degrees")
	expect_equal(angular_unit(a), "degrees")
	angular_unit(a) <- "degrees"
	angular_unit(a) <- "turns"
	expect_equal(angular_unit(a), "turns")
	expect_equal(as.numeric(a), seq(0, 1, by = 0.25))
	expect_equal(angular_unit(gradians(0)), "gradians")
	expect_equal(angular_unit(turns(0)), "turns")
})

test_that("is_congruent()", {
	a1 <- angle(180, "degrees")
	a2 <- angle(0.5, "turns")
	a3 <- angle(-180, "degrees")
	a4 <- angle(90, "degrees")
	expect_congruent(a1, a2)
	expect_congruent(a1, a3)
	expect_false(is_congruent(a1, a3, mod_turns = FALSE))
	expect_equal(a1, -a3)
	expect_congruent(a1, 180)
	expect_congruent(2 * a1, 360)
	expect_congruent(a1 - a2, 0)
	expect_congruent(a4, as_angle("y-axis", "degrees"))
	expect_congruent(degrees(0), as_angle("x-axis", "degrees"))
	expect_congruent(angle(1e-9, "half-turns"), angle(-1e-9, "half-turns"))
	expect_false(is_congruent(a1, a4))
	expect_congruent(2, 2)
})

test_that("abs()", {
	expect_equal(angle(90, "degrees"), abs(angle(-270, "degrees")))
	expect_equal(angle(100, "gradians"), abs(angle(-300, "gradians")))
	expect_equal(angle(1, "pi-radians"), abs(angle(3, "pi-radians")))
	expect_equal(angle(0.5, "turns"), abs(angle(-0.5, "turns")))
	expect_equal(angle(pi, "radians"), abs(angle(-pi, "radians")))
})

test_that("as.numeric.angle()", {
	a1 <- angle(180, "degrees")
	expect_identical(as.numeric(a1), 180)
	expect_identical(as.numeric(a1, "deg"), 180)
	expect_equal(as.numeric(a1, "turns"), 0.5)
	expect_equal(as.numeric(a1, "radians"), pi)
	a2 <- angle(pi, "radians")
	expect_equal(as.numeric(a2), pi)
	expect_equal(Arg(a2), pi)
	expect_equal(as.numeric(a2, "degrees"), 180)

	c1 <- complex(real = 0, imaginary = 1)
	expect_equal(c1, as.complex(angle(90, "degrees"), modulus = 1))
	expect_congruent(as_angle(c1), angle(90, "degrees"))
})

test_that("base R rounding functions", {
	expect_equal(round(degrees(90.7), 0), degrees(91))
	expect_equal(signif(degrees(184), 2), degrees(180))
	expect_equal(ceiling(degrees(184.2)), degrees(185))
	expect_equal(floor(degrees(184.7)), degrees(184))
	expect_equal(trunc(degrees(184.7)), degrees(184))
})

test_that("base R trigonometric functions", {
	a1 <- angle(180, "degrees")
	expect_equal(cos(a1), cos(pi))
	expect_equal(sin(a1), sin(pi))
	expect_equal(tan(a1), tan(pi))
	a2 <- angle(pi, "radians")
	expect_equal(cos(a2), cos(pi))
	expect_equal(sin(a2), sin(pi))
	expect_equal(tan(a2), tan(pi))
})

test_that("as_angle()", {
	skip_if_not_installed("withr")
	withr::local_options(affiner_options(default = TRUE))
	a1 <- angle(180, "degrees")
	expect_equal(a1, (as_angle(a1)))
	expect_equal(a1, (as_angle(a1, "degrees")))
	expect_equal(a1, degrees(180))
	expect_equal(a1, degrees(angle(1, "pi-radians")))
	expect_congruent(a1, as_angle(a1, "turns"))
	expect_congruent(a1, pi_radians(1))
	expect_congruent(a1, radians(pi))
	expect_false(isTRUE(all.equal(a1, as_angle(a1, "turns"), check.attributes = TRUE)))
	expect_warning(as_angle("foobar"))

	expect_equal(as_angle(as_coord3d("z-axis"), "degrees", type = "azimuth"), angle(0, "degrees"))
	expect_equal(
		as_angle(as_coord3d("z-axis"), "degrees", type = "inclination"),
		angle(0, "degrees")
	)
	expect_equal(as_angle(as_coord3d("x-axis"), "degrees", type = "azimuth"), angle(0, "degrees"))
	expect_equal(
		as_angle(as_coord3d("x-axis"), "degrees", type = "inclination"),
		angle(90, "degrees")
	)
	expect_equal(as_angle(as_coord3d("y-axis"), "degrees", type = "azimuth"), angle(90, "degrees"))
	expect_equal(
		as_angle(as_coord3d("y-axis"), "degrees", type = "inclination"),
		angle(90, "degrees")
	)
	expect_equal(as_angle(as_coord3d(1, 1, -1), "degrees", type = "azimuth"), angle(45, "degrees"))
	expect_equal(
		as_angle(as_coord3d(1, 1, -1), "degrees", type = "inclination"),
		arccosine(-1 / sqrt(3), "degrees")
	)
	expect_equal(
		as_angle(as_coord3d(0, 1, -1), "degrees", type = "inclination"),
		angle(135, "degrees")
	)
})

test_that("c.angle()", {
	a1 <- angle(180, "degrees")
	a2 <- angle(1, "pi-radians")
	expect_equal(c(a1, a2), angle(c(180, 180), "degrees"))

	expect_equal(rep_len(a1, 4), angle(rep(180, 4), "degrees"))
})

#' Isotoxal `2n`-gon and star polygon
#'
#' `isotoxal_2ngon_polygon2d()` creates an isotoxal `2n`-gon [Polygon2D]
#' object centered at `(x, y)`.
#' `star_polygon2d()` is an alias.
#'
#' Isotoxal `2n`-gon polygons have `2n` vertices alternating between `n`
#' outer vertices on a circle of radius `radius` and `n` inner vertices on
#' a concentric circle of radius `radial_scale * radius`.
#'
#' @param n Number of outer vertices.
#' @param x X-coordinate of the center (default `0`).
#' @param y Y-coordinate of the center (default `0`).
#' @param radius Outer circumradius (default `0.5`).
#' @param radial_scale Inner radius as a fraction of `radius`.  Defaults to
#'   `isotoxal_2ngon_inner_radius(n, alpha = alpha, beta_ext = beta_ext, d = d)`.
#'   Exactly one of `radial_scale`, `alpha`, `beta_ext`, or `d` must be
#'   supplied.
#' @param theta Angle of the first outer vertex (default `degrees(90)`).
#'   Will be coerced by [degrees()].
#' @param ... Ignored.
#' @param alpha Interior angle of an outer vertex.  Will be coerced by
#'   [degrees()].
#' @param beta_ext Exterior angle of an inner vertex.  Will be coerced by
#'   [degrees()].
#' @param d Density (winding number) of the star polygon `|n/d|`.
#' @return A [Polygon2D] object.
#' @seealso [isotoxal_2ngon_inner_radius()] to compute the radial scale.
#'   [rectangle_polygon2d()] for rectangles.
#'   [regular_ngon_polygon2d()] for regular convex polygons.
#'   <https://en.wikipedia.org/wiki/Isotoxal_figure#Isotoxal_polygons> and
#'   <https://en.wikipedia.org/wiki/Star_polygon#Isotoxal_star_simple_polygons>
#'   for more information on isotoxal polygons.
#' @examples
#' # |5/2| star (the verda stelo)
#' p <- isotoxal_2ngon_polygon2d(5, d = 2)
#' p$is_convex
#' plot(p, col = "#008000", border = NA)
#'
#' # `star_polygon2d()` is an alias
#' p2 <- star_polygon2d(5, d = 2)
#' all.equal(p, p2)
#' @export
isotoxal_2ngon_polygon2d <- function(
	n,
	x = 0,
	y = 0,
	radius = 0.5,
	radial_scale = isotoxal_2ngon_inner_radius(n, alpha = alpha, beta_ext = beta_ext, d = d),
	theta = degrees(90),
	...,
	alpha = NULL,
	beta_ext = NULL,
	d = NULL
) {
	chkDots(...)
	t_outer <- degrees(theta) + degrees(seq(0, by = 360 / n, length.out = n))
	t_inner <- degrees(theta) + degrees(seq(180 / n, by = 360 / n, length.out = n))
	r_inner <- radius * radial_scale
	vx <- c(rbind(radius * cos(t_outer), r_inner * cos(t_inner)))
	vy <- c(rbind(radius * sin(t_outer), r_inner * sin(t_inner)))
	convex <- radial_scale >= cospi(1 / n)
	as_polygon2d(as_coord2d(x + vx, y + vy), convex = convex)
}

#' @rdname isotoxal_2ngon_polygon2d
#' @export
star_polygon2d <- isotoxal_2ngon_polygon2d

#' Isotoxal `2n`-gon inner radius
#'
#' `isotoxal_2ngon_inner_radius()` computes the inner radius
#' of an isotoxal `2n`-gon polygon.
#' `star_inner_radius()` is an alias.
#'
#' Isotoxal `2n`-gon polygons are polygons with:
#'
#' * `2n` vertices alternating between `n` "outer" vertices evenly spaced on one circle and `n` "inner" vertices evenly spaced on smaller circle with the same center.
#' * Each edge of the polygon is of the same length.
#' * The outer vertices all have the same interior angle `alpha` and the
#'   inner vertices all have the same interior angle `beta`.
#' * They are a generalization of (the outlines of) concave simple "star" polygons that also includes convex polygons with an even number of vertices.
#' @param n The number of outer vertices.
#' @param outer_radius The outer radius of the isotoxal `2n`-gon.
#' @param ... Ignored.
#' @param alpha The interior angle of an outer vertex.  Will be coerced by [degrees()].
#' @param beta_ext The exterior angle of an inner vertex.  Will be coerced by [degrees()].
#' @param d The density aka winding number of the regular star polygon (outline) in which case this star is represented by `|n/d|`.
#' @examples
#' # |8/3| star has outer vertex internal angle 45 degrees
#' # and inner vertex external angle 90 degrees
#' isotoxal_2ngon_inner_radius(8, d = 3)
#' isotoxal_2ngon_inner_radius(8, alpha = 45)
#' isotoxal_2ngon_inner_radius(8, beta_ext = 90)
#' @seealso [isotoxal_2ngon_polygon2d()] to construct an isotoxal `2n`-gon
#'   [Polygon2D] object.
#'   <https://en.wikipedia.org/wiki/Isotoxal_figure#Isotoxal_polygons> and
#'   <https://en.wikipedia.org/wiki/Star_polygon#Isotoxal_star_simple_polygons>
#'   for more information on isotoxal polygons.
#' @return A numeric vector
#' @export
isotoxal_2ngon_inner_radius <- function(
	n,
	outer_radius = 1,
	...,
	alpha = NULL,
	beta_ext = NULL,
	d = NULL
) {
	stopifnot(is.null(alpha) + is.null(beta_ext) + is.null(d) == 2L)

	if (!is.null(alpha)) {
		alpha <- as.numeric(degrees(alpha), "degrees")
	}
	if (!is.null(beta_ext)) {
		beta_ext <- as.numeric(degrees(beta_ext), "degrees")
		alpha <- beta_ext_to_alpha(n, beta_ext)
	}
	if (!is.null(d)) {
		alpha <- 180 * (1 - 2 * d / n)
	}
	if (n == 2L) {
		return(outer_radius * isotoxal_2_scale(alpha))
	}
	stopifnot(alpha >= 0, alpha <= 180 * (1 - 2 / n))
	# we'll work with external degree
	beta_ext <- alpha_to_beta_ext(n, alpha)
	t <- 360 / n
	xy1 <- as_coord2d(x = 1, y = 0)
	xy2 <- as_coord2d(degrees(t), radius = 1)
	xyc <- mean(c(xy1, xy2))
	xyf <- as_coord2d(degrees(t / 2), radius = 1)
	dist_f <- distance2d(xyc, xyf)
	a2 <- distance2d(xy1, xyc)
	beta <- (180 - beta_ext) / 2
	b <- a2 * sin(degrees(beta)) / sin(degrees(beta_ext / 2))
	r <- 1 - b - dist_f
	stopifnot(r >= 0)
	outer_radius * r
}

#' @rdname isotoxal_2ngon_inner_radius
#' @export
star_inner_radius <- isotoxal_2ngon_inner_radius

isotoxal_2_scale <- function(alpha) {
	stopifnot(alpha >= 0, alpha <= 90)
	a1 <- alpha / 2
	a2 <- 180 - 90 - a1
	r <- sin(degrees(a1)) / sin(degrees(a2))
	stopifnot(r >= 0)
	r
}

beta_ext_to_alpha <- function(n, beta_ext) {
	total <- (2 * n - 2) * 180
	inverse <- 360 - beta_ext
	alpha <- (total - n * inverse) / n
	alpha %% 360
}

alpha_to_beta_ext <- function(n, alpha) {
	total <- (2 * n - 2) * 180
	inverse <- (total - n * alpha) / n
	beta_ext <- 360 - inverse
	beta_ext %% 360
}

test_that("`isotoxal_2ngon_inner_radius()`", {
	expect_equal(isotoxal_2ngon_inner_radius(2, alpha = 90), 1)
	expect_equal(isotoxal_2ngon_inner_radius(2, pi, alpha = 90), pi)
	expect_equal(isotoxal_2ngon_inner_radius(2, beta_ext = 270), 1)
	# |2_60| rhombus has internal angle 60
	scale_2_60 <- 0.57735026918962573106 # `gridpattern::star_scale(2, 60)`
	scale_alpha <- isotoxal_2ngon_inner_radius(2, alpha = 60)
	scale_beta_ext <- isotoxal_2ngon_inner_radius(2, beta_ext = 240)
	expect_equal(scale_2_60, scale_alpha)
	expect_equal(scale_2_60, scale_beta_ext)

	# |8/3| star has outer internal angle 45 degrees
	# and inner external angle 90 degrees
	scale_8_3_star <- 0.54119610014619701222 # `gridpattern::star_scale(8, 45)`
	scale_alpha <- isotoxal_2ngon_inner_radius(8, alpha = 45)
	scale_beta_ext <- isotoxal_2ngon_inner_radius(8, beta_ext = 90)
	scale_d <- isotoxal_2ngon_inner_radius(8, d = 3)
	expect_equal(scale_8_3_star, scale_alpha)
	expect_equal(scale_8_3_star, scale_beta_ext)
	expect_equal(scale_8_3_star, scale_d)
	expect_equal(isotoxal_2ngon_inner_radius(8, pi, d = 3), pi * scale_8_3_star)
})

test_that("`alpha_to_beta_ext()` and `beta_ext_to_alpha()`", {
	expect_equal(alpha_to_beta_ext(2, 90), 270)
	expect_equal(beta_ext_to_alpha(2, 270), 90)
	expect_equal(alpha_to_beta_ext(8, 45), 90)
	expect_equal(beta_ext_to_alpha(8, 90), 45)
	expect_equal(alpha_to_beta_ext(5, 36), 108)
	expect_equal(beta_ext_to_alpha(5, 108), 36)
	expect_equal(alpha_to_beta_ext(6, 60), 120)
	expect_equal(beta_ext_to_alpha(6, 120), 60)
})

#' @export
`[.Coord1D` <- function(x, i) {
	Coord1D$new(x$xw[i, , drop = FALSE])
}

#' @export
`[.Coord2D` <- function(x, i) {
	Coord2D$new(x$xyw[i, , drop = FALSE])
}

#' @export
`[.Segment2D` <- function(x, i) {
	xyw <- x$xyw
	vec <- x$.__enclos_env__$private$vec
	Segment2D$new(xyw[i, , drop = FALSE], vec[i, , drop = FALSE])
}

#' @export
`[.Ellipse2D` <- function(x, i) {
	Ellipse2D$new(x$xyw[i, , drop = FALSE], rx = x$rx[i], ry = x$ry[i], theta = x$theta[i])
}

#' @export
`[.Coord3D` <- function(x, i) {
	Coord3D$new(x$xyzw[i, , drop = FALSE])
}

#' @export
as.complex.Coord2D <- function(x, ...) {
	chkDots(...)
	complex(real = x$x, imaginary = x$y)
}

#' @export
as.data.frame.Coord1D <- function(x, ...) {
	chkDots(...)
	as.data.frame(x$xw)
}

#' @export
as.data.frame.Coord2D <- function(x, ...) {
	chkDots(...)
	as.data.frame(x$xyw)
}

#' @export
as.data.frame.Ellipse2D <- function(x, ...) {
	chkDots(...)
	data.frame(x = x$x, y = x$y, rx = x$rx, ry = x$ry, theta = x$theta)
}

#' @export
as.data.frame.Coord3D <- function(x, ...) {
	chkDots(...)
	as.data.frame(x$xyzw)
}

#' @export
as.list.Coord1D <- function(x, ...) {
	chkDots(...)
	as.list(as.data.frame(x$xw))
}

#' @export
as.list.Coord2D <- function(x, ...) {
	chkDots(...)
	as.list(as.data.frame(x$xyw))
}

#' @export
as.list.Ellipse2D <- function(x, ...) {
	chkDots(...)
	list(x = x$x, y = x$y, rx = x$rx, ry = x$ry, theta = x$theta)
}

#' @export
as.list.Coord3D <- function(x, ...) {
	chkDots(...)
	as.list(as.data.frame(x$xyzw))
}

#' @export
as.matrix.Coord1D <- function(x, ...) {
	chkDots(...)
	x$xw
}

#' @export
as.matrix.Coord2D <- function(x, ...) {
	chkDots(...)
	x$xyw
}

#' @export
as.matrix.Coord3D <- function(x, ...) {
	chkDots(...)
	x$xyzw
}

#' @export
c.Coord1D <- function(...) {
	l <- list(...)
	stopifnot(all(vapply(l, is_coord1d, logical(1))))
	m <- do.call(rbind, lapply(l, as.matrix))
	Coord1D$new(m)
}

#' @export
c.Coord2D <- function(...) {
	l <- list(...)
	stopifnot(all(vapply(l, is_coord2d, logical(1))))
	m <- do.call(rbind, lapply(l, as.matrix))
	Coord2D$new(m)
}

#' @export
c.Coord3D <- function(...) {
	l <- list(...)
	stopifnot(all(vapply(l, is_coord3d, logical(1))))
	m <- do.call(rbind, lapply(l, as.matrix))
	Coord3D$new(m)
}

#' @export
length.Coord1D <- function(x) {
	nrow(x$.__enclos_env__$private$mat_xw)
}

#' @export
length.Coord2D <- function(x) {
	nrow(x$.__enclos_env__$private$mat_xyw)
}

#' @export
length.Coord3D <- function(x) {
	nrow(x$.__enclos_env__$private$mat_xyzw)
}

#' @export
rep.Coord1D <- function(x, ..., length.out = NA_integer_) {
	if (isTRUE(length(x) == length.out)) {
		return(x)
	}
	id <- rep(seq.int(length(x)), ..., length.out = length.out)
	Coord1D$new(x$xw[id, , drop = FALSE])
}

#' @export
rep.Coord2D <- function(x, ..., length.out = NA_integer_) {
	if (isTRUE(length(x) == length.out)) {
		return(x)
	}
	id <- rep(seq.int(length(x)), ..., length.out = length.out)
	Coord2D$new(x$xyw[id, , drop = FALSE])
}

#' @export
rep.Segment2D <- function(x, ..., length.out = NA_integer_) {
	if (isTRUE(length(x) == length.out)) {
		return(x)
	}
	id <- rep(seq.int(length(x)), ..., length.out = length.out)
	x[id]
}

#' @export
rep.Coord3D <- function(x, ..., length.out = NA_integer_) {
	if (isTRUE(length(x) == length.out)) {
		return(x)
	}
	id <- rep(seq.int(length(x)), ..., length.out = length.out)
	Coord3D$new(x$xyzw[id, , drop = FALSE])
}

#' Compute centroids of coordinates
#'
#' `mean()`computes centroids for [Coord1D], [Coord2D], and [Coord3D] class objects
#'
#' @param x A [Coord1D], [Coord2D], or [Coord3D] object
#' @param ... Passed to [base::mean()]
#' @return A [Coord1D], [Coord2D], or [Coord3D] class object of length one
#' @examples
#' p <- as_coord2d(x = 1:4, y = 1:4)
#' print(mean(p))
#' print(sum(p) / length(p)) # less efficient alternative
#'
#' p <- as_coord3d(x = 1:4, y = 1:4, z = 1:4)
#' print(mean(p))
#' @name centroid
#' @export
mean.Coord1D <- function(x, ...) {
	as_coord1d(mean(x$x, ...))
}

#' @rdname centroid
#' @export
mean.Coord2D <- function(x, ...) {
	as_coord2d(mean(x$x, ...), mean(x$y, ...))
}

#' @rdname centroid
#' @export
mean.Coord3D <- function(x, ...) {
	as_coord3d(mean(x$x, ...), mean(x$y, ...), mean(x$z, ...))
}

#' Compute 2D convex hulls
#'
#' `convex_hull2d()` is a S3 generic for computing the convex hull of an
#' object.  It is implemented for [Coord2D] and [Polygon2D] objects using
#' [grDevices::chull()] and returns a [Polygon2D] whose vertices are in
#' counter-clockwise order.
#'
#' @param x An object to compute the convex hull of,
#'   such as a [Coord2D] or [Polygon2D] object.
#' @param ... Further arguments passed to or from other methods.
#' @return A [Polygon2D] object representing the convex hull.
#' @examples
#' pts <- as_coord2d(x = rnorm(20), y = rnorm(20))
#' hull <- convex_hull2d(pts)
#' is_polygon2d(hull)
#' hull$is_convex
#' plot(hull)
#' points(pts)
#' @export
convex_hull2d <- function(x, ...) {
	UseMethod("convex_hull2d")
}

#' @rdname convex_hull2d
#' @export
convex_hull2d.Coord2D <- function(x, ...) {
	chkDots(...)
	idx <- rev(grDevices::chull(as.list(x)))
	Polygon2D$new(x$xyw[idx, , drop = FALSE], convex = TRUE)
}

#' @rdname convex_hull2d
#' @export
convex_hull2d.Polygon2D <- function(x, ...) {
	chkDots(...)
	x$convex_hull
}

# Group "Summary"

#' @export
is.na.Coord1D <- function(x) is.na(x$x)

#' @export
is.na.Coord2D <- function(x) is.na(x$x) | is.na(x$y)

#' @export
is.na.Coord3D <- function(x) is.na(x$x) | is.na(x$y) | is.na(x$z)

#' @export
is.nan.Coord1D <- function(x) is.nan(x$x)

#' @export
is.nan.Coord2D <- function(x) is.nan(x$x) | is.nan(x$y)

#' @export
is.nan.Coord3D <- function(x) is.nan(x$x) | is.nan(x$y) | is.nan(x$z)

#' @export
is.finite.Coord1D <- function(x) is.finite(x$x)

#' @export
is.finite.Coord2D <- function(x) is.finite(x$x) & is.finite(x$y)

#' @export
is.finite.Coord3D <- function(x) is.finite(x$x) & is.finite(x$y) & is.finite(x$z)

#' @export
is.infinite.Coord1D <- function(x) is.infinite(x$x)

#' @export
is.infinite.Coord2D <- function(x) is.infinite(x$x) | is.infinite(x$y)

#' @export
is.infinite.Coord3D <- function(x) is.infinite(x$x) | is.infinite(x$y) | is.infinite(x$z)

#' @export
sum.Coord1D <- function(..., na.rm = FALSE) {
	l <- list(...)
	if (na.rm) {
		l <- lapply(l, function(p) p[!is.na(p)])
	}
	xs <- sum(sapply(l, function(p) sum(p$x)))
	as_coord1d(xs)
}

#' @export
sum.Coord2D <- function(..., na.rm = FALSE) {
	l <- list(...)
	if (na.rm) {
		l <- lapply(l, function(p) p[!is.na(p)])
	}
	xs <- sum(sapply(l, function(p) sum(p$x)))
	ys <- sum(sapply(l, function(p) sum(p$y)))
	as_coord2d(xs, ys)
}

#' @export
sum.Coord3D <- function(..., na.rm = FALSE) {
	l <- list(...)
	if (na.rm) {
		l <- lapply(l, function(p) p[!is.na(p)])
	}
	xs <- sum(sapply(l, function(p) sum(p$x)))
	ys <- sum(sapply(l, function(p) sum(p$y)))
	zs <- sum(sapply(l, function(p) sum(p$z)))
	as_coord3d(xs, ys, zs)
}

#' @export
`==.Coord1D` <- function(e1, e2) {
	stopifnot(is_coord1d(e1) && is_coord1d(e2))
	e1$x == e2$x
}

#' @export
`==.Coord2D` <- function(e1, e2) {
	stopifnot(is_coord2d(e1) && is_coord2d(e2))
	(e1$x == e2$x) & (e1$y == e2$y)
}

#' @export
`==.Coord3D` <- function(e1, e2) {
	stopifnot(is_coord3d(e1) && is_coord3d(e2))
	(e1$x == e2$x) & (e1$y == e2$y) & (e1$z == e2$z)
}

#' @export
`!=.Coord1D` <- function(e1, e2) {
	stopifnot(is_coord1d(e1) && is_coord1d(e2))
	e1$x != e2$x
}

#' @export
`!=.Coord2D` <- function(e1, e2) {
	stopifnot(is_coord2d(e1) && is_coord2d(e2))
	(e1$x != e2$x) | (e1$y != e2$y)
}

#' @export
`!=.Coord3D` <- function(e1, e2) {
	stopifnot(is_coord3d(e1) && is_coord3d(e2))
	(e1$x != e2$x) | (e1$y != e2$y) | (e1$z != e2$z)
}

#' @export
`+.Coord1D` <- function(e1, e2) {
	if (missing(e2)) {
		e1
	} else {
		plus_coord1d(e1, e2)
	}
}

#' @export
`+.Coord2D` <- function(e1, e2) {
	if (missing(e2)) {
		e1
	} else {
		plus_coord2d(e1, e2)
	}
}

#' @export
`+.Coord3D` <- function(e1, e2) {
	if (missing(e2)) {
		e1
	} else {
		plus_coord3d(e1, e2)
	}
}

#' @export
`-.Coord1D` <- function(e1, e2) {
	if (missing(e2)) {
		e1$clone()$scale(-1)
	} else {
		minus_coord1d(e1, e2)
	}
}

#' @export
`-.Coord2D` <- function(e1, e2) {
	if (missing(e2)) {
		e1$clone()$scale(-1)
	} else {
		minus_coord2d(e1, e2)
	}
}

#' @export
`-.Coord3D` <- function(e1, e2) {
	if (missing(e2)) {
		e1$clone()$scale(-1)
	} else {
		minus_coord3d(e1, e2)
	}
}

#' @export
`*.Coord1D` <- function(e1, e2) {
	if (is_coord1d(e1) && is_coord1d(e2)) {
		dot_product1d(e1, e2)
	} else if (is_coord1d(e1) && is.numeric(e2)) {
		e1$clone()$scale(e2)
	} else if (is.numeric(e1) && is_coord1d(e2)) {
		e2$clone()$scale(e1)
	} else {
		stop(paste(
			"Don't know how to multiply objects of class",
			sQuote(class(e1)[1]),
			"and class",
			sQuote(class(e2)[1])
		))
	}
}

#' @export
`*.Coord2D` <- function(e1, e2) {
	if (is_coord2d(e1) && is_coord2d(e2)) {
		dot_product2d(e1, e2)
	} else if (is_coord2d(e1) && is.numeric(e2)) {
		e1$clone()$scale(e2)
	} else if (is.numeric(e1) && is_coord2d(e2)) {
		e2$clone()$scale(e1)
	} else {
		stop(paste(
			"Don't know how to multiply objects of class",
			sQuote(class(e1)[1]),
			"and class",
			sQuote(class(e2)[1])
		))
	}
}

#' @export
`*.Coord3D` <- function(e1, e2) {
	if (is_coord3d(e1) && is_coord3d(e2)) {
		dot_product3d(e1, e2)
	} else if (is_coord3d(e1) && is.numeric(e2)) {
		e1$clone()$scale(e2)
	} else if (is.numeric(e1) && is_coord3d(e2)) {
		e2$clone()$scale(e1)
	} else {
		stop(paste(
			"Don't know how to multiply objects of class",
			sQuote(class(e1)[1]),
			"and class",
			sQuote(class(e2)[1])
		))
	}
}

#' @export
`/.Coord1D` <- function(e1, e2) {
	if (is_coord1d(e1) && is.numeric(e2)) {
		e1$clone()$scale(1 / e2)
	} else {
		stop(paste(
			"Don't know how to divide objects of class",
			sQuote(class(e1)[1]),
			"and class",
			sQuote(class(e2)[1])
		))
	}
}

#' @export
`/.Coord2D` <- function(e1, e2) {
	if (is_coord2d(e1) && is.numeric(e2)) {
		e1$clone()$scale(1 / e2)
	} else {
		stop(paste(
			"Don't know how to divide objects of class",
			sQuote(class(e1)[1]),
			"and class",
			sQuote(class(e2)[1])
		))
	}
}

#' @export
`/.Coord3D` <- function(e1, e2) {
	if (is_coord3d(e1) && is.numeric(e2)) {
		e1$clone()$scale(1 / e2)
	} else {
		stop(paste(
			"Don't know how to divide objects of class",
			sQuote(class(e1)[1]),
			"and class",
			sQuote(class(e2)[1])
		))
	}
}

#' Compute axis-aligned ranges
#'
#' `range()` computes axis-aligned ranges for
#' [Coord1D], [Coord2D], and [Coord3D] class objects.
#' @param na.rm logical, indicating if `NA`'s should be omitted
#' @param ... [Coord1D], [Coord2D], or [Coord3D] object(s)
#' @name bounding_ranges
#' @return Either a [Coord1D], [Coord2D], or [Coord3D] object of length two.
#'         The first element will have the minimum x/y(/z) coordinates
#'         and the second element will have the maximum x/y(/z) coordinates
#'         of the axis-aligned ranges.
#' @examples
#' range(as_coord2d(rnorm(5), rnorm(5)))
#' range(as_coord3d(rnorm(5), rnorm(5), rnorm(5)))
#' @export
range.Coord1D <- function(..., na.rm = FALSE) {
	x <- c.Coord1D(...)
	if (na.rm) {
		x <- x[!is.na(x)]
	}
	as_coord1d(range(x$x))
}

#' @rdname bounding_ranges
#' @export
range.Coord2D <- function(..., na.rm = FALSE) {
	x <- c.Coord2D(...)
	if (na.rm) {
		x <- x[!is.na(x)]
	}
	as_coord2d(range(x$x), range(x$y))
}

#' @rdname bounding_ranges
#' @export
range.Coord3D <- function(..., na.rm = FALSE) {
	x <- c.Coord3D(...)
	if (na.rm) {
		x <- x[!is.na(x)]
	}
	as_coord3d(range(x$x), range(x$y), range(x$z))
}

#' Compute dot (inner) products
#'
#' `dot_product1d()`, `dot_product2d()`, and `dot_product3d()` compute the
#' dot (inner) product of two coordinate vectors.
#' You may also use the `*` operator and if R >= 4.3.0 you may also use the `%*%` operator to compute the dot (inner) product.
#'
#' @param x A [Coord1D], [Coord2D], or [Coord3D] object.
#' @param y A [Coord1D], [Coord2D], or [Coord3D] object.
#' @return A numeric vector of dot products.
#' @examples
#' p1 <- as_coord2d(x = c(1, 2), y = c(3, 4))
#' p2 <- as_coord2d(x = c(5, 6), y = c(7, 8))
#' dot_product2d(p1, p2)
#' p1 * p2 # equivalent
#' if (getRversion() >= "4.3.0") {
#'   p1 %*% p2
#' }
#' @seealso [cross_product3d()] for the cross product of two [Coord3D] vectors.
#' @name dot_product
#' @export
dot_product1d <- function(x, y) {
	stopifnot(is_coord1d(x), is_coord1d(y))
	n <- max(length(x), length(y))
	x <- rep_len(x, n)
	y <- rep_len(y, n)
	x$x * y$x
}

#' @rdname dot_product
#' @export
dot_product2d <- function(x, y) {
	stopifnot(is_coord2d(x), is_coord2d(y))
	n <- max(length(x), length(y))
	x <- rep_len(x, n)
	y <- rep_len(y, n)
	rowSums(x$xyw[, 1:2, drop = FALSE] * y$xyw[, 1:2, drop = FALSE])
}

#' @rdname dot_product
#' @export
dot_product3d <- function(x, y) {
	stopifnot(is_coord3d(x), is_coord3d(y))
	n <- max(length(x), length(y))
	x <- rep_len(x, n)
	y <- rep_len(y, n)
	rowSums(x$xyzw[, 1:3, drop = FALSE] * y$xyzw[, 1:3, drop = FALSE])
}

#' @rawNamespace if (getRversion() >= "4.3.0") {
#'   S3method("%*%",Coord1D)
#' }
`%*%.Coord1D` <- function(x, y) dot_product1d(x, y)

#' @rawNamespace if (getRversion() >= "4.3.0") {
#'   S3method("%*%",Coord2D)
#' }
`%*%.Coord2D` <- function(x, y) dot_product2d(x, y)

#' @rawNamespace if (getRversion() >= "4.3.0") {
#'   S3method("%*%",Coord3D)
#' }
`%*%.Coord3D` <- function(x, y) dot_product3d(x, y)

plus_coord1d <- function(p1, p2) {
	stopifnot(is_coord1d(p1), is_coord1d(p2))
	if (length(p1) == 1) {
		p2$clone()$translate(p1)
	} else {
		p1$clone()$translate(p2)
	}
}

plus_coord2d <- function(p1, p2) {
	stopifnot(is_coord2d(p1), is_coord2d(p2))
	if (length(p1) == 1) {
		p2$clone()$translate(p1)
	} else {
		p1$clone()$translate(p2)
	}
}

plus_coord3d <- function(p1, p2) {
	stopifnot(is_coord3d(p1), is_coord3d(p2))
	if (length(p1) == 1) {
		p2$clone()$translate(p1)
	} else {
		p1$clone()$translate(p2)
	}
}

minus_coord1d <- function(p1, p2) {
	stopifnot(is_coord1d(p1), is_coord1d(p2))
	p1$clone()$translate(p2$clone()$scale(-1))
}

minus_coord2d <- function(p1, p2) {
	stopifnot(is_coord2d(p1), is_coord2d(p2))
	p1$clone()$translate(p2$clone()$scale(-1))
}

minus_coord3d <- function(p1, p2) {
	stopifnot(is_coord3d(p1), is_coord3d(p2))
	p1$clone()$translate(p2$clone()$scale(-1))
}

# Base rounding function

#' @export
ceiling.Coord1D <- function(x) {
	as_coord1d(ceiling(x$x))
}

#' @export
ceiling.Coord2D <- function(x) {
	as_coord2d(ceiling(x$x), ceiling(x$y))
}

#' @export
ceiling.Coord3D <- function(x) {
	as_coord3d(ceiling(x$x), ceiling(x$y), ceiling(x$z))
}

#' @export
floor.Coord1D <- function(x) {
	as_coord1d(floor(x$x))
}

#' @export
floor.Coord2D <- function(x) {
	as_coord2d(floor(x$x), floor(x$y))
}

#' @export
floor.Coord3D <- function(x) {
	as_coord3d(floor(x$x), floor(x$y), floor(x$z))
}

#' @export
round.Coord1D <- function(x, digits = 0, ...) {
	chkDots(...)
	as_coord1d(round(x$x, digits))
}

#' @export
round.Coord2D <- function(x, digits = 0, ...) {
	chkDots(...)
	as_coord2d(round(x$x, digits), round(x$y, digits))
}

#' @export
round.Coord3D <- function(x, digits = 0, ...) {
	chkDots(...)
	as_coord3d(round(x$x, digits), round(x$y, digits), round(x$z, digits))
}

#' @export
signif.Coord1D <- function(x, digits = 6) {
	as_coord1d(signif(x$x, digits))
}

#' @export
signif.Coord2D <- function(x, digits = 6) {
	as_coord2d(signif(x$x, digits), signif(x$y, digits))
}

#' @export
signif.Coord3D <- function(x, digits = 6) {
	as_coord3d(signif(x$x, digits), signif(x$y, digits), signif(x$z, digits))
}

#' @export
trunc.Coord1D <- function(x, ...) {
	chkDots(...)
	as_coord1d(trunc(x$x))
}

#' @export
trunc.Coord2D <- function(x, ...) {
	chkDots(...)
	as_coord2d(trunc(x$x), trunc(x$y))
}

#' @export
trunc.Coord3D <- function(x, ...) {
	chkDots(...)
	as_coord3d(trunc(x$x), trunc(x$y), trunc(x$z))
}

# Group "Complex"

#' @export
Re.Coord2D <- function(z) {
	z$x
}

#' @export
Im.Coord2D <- function(z) {
	z$y
}

#' @export
Mod.Coord2D <- function(z) {
	sqrt(rowSums(z$xyw[, 1:2, drop = FALSE]^2))
}

#' @export
Conj.Coord2D <- function(z) {
	complex(real = z$x, imaginary = -z$y)
}

#' @export
Arg.Coord2D <- function(z) {
	atan2(z$y, z$x)
}

#' Compute 3D vector cross product
#'
#' `cross_product3d()` computes the cross product of two [Coord3D] class vectors.
#' @param x A [Coord3D] class vector.
#' @param y A [Coord3D] class vector.
#' @return A [Coord3D] class vector
#' @examples
#' x <- as_coord3d(2, 3, 4)
#' y <- as_coord3d(5, 6, 7)
#' cross_product3d(x, y)
#' if (getRversion() >= "4.4.0") {
#'   crossprod(x, y)
#' }
#' @seealso [dot_product3d()] for the dot product of two [Coord3D] vectors.
#' @export
cross_product3d <- function(x, y) {
	stopifnot(is_coord3d(x), is_coord3d(y))
	n <- max(length(x), length(y))
	x <- rep_len(x, n)
	y <- rep_len(y, n)
	m <- matrix(1, nrow = n, ncol = 4, dimnames = list(NULL, c("x", "y", "z", "w")))
	m[, 1] <- x$y * y$z - x$z * y$y
	m[, 2] <- x$z * y$x - x$x * y$z
	m[, 3] <- x$x * y$y - x$y * y$x
	Coord3D$new(m)
}

#' @rawNamespace if (getRversion() >= "4.4.0") {
#'   S3method("crossprod",Coord3D)
#' }
crossprod.Coord3D <- function(x, y, ...) {
	chkDots(...)
	cross_product3d(x, y)
}

# nolint start
# #' Scalar projections
# #'
# #' `as.double.Coord2D()` and `as.double.Coord3D()` computes the scalar projections of [Coord2D]
# #' and [Coord3D] vectors.
# #' The scalar projection of a vector `x` onto a vector `y` is also
# #' known as the component of `x` in the direction of `y`.
# #'
# #' @param x [Coord2D] or [Coord3D] object
# #' @param y [Coord2D] or [Coord3D] object of length one to project onto.
# #' @inheritParams project2d
# #' @param ... If `y` is not a coordinate vector passed to either [as_coord2d()] or [as_coord3d()].
# #' @examples
# #' p2 <- as_coord2d(rnorm(5), rnorm(5))
# #' as.numeric(p2)
# #' p3 <- as_coord2d(rnorm(5), rnorm(5), rnorm(5))
# #' as.numeric(p3)
# #' @name scalar_projection
# #' @return A double vector
# #' @export
# as.double.Coord2D <- function(x, y = as_coord2d(1, 0), ...) {
#     if (!is_coord2d(y))
#         y <- as_coord2d(y, ...)
#     stopifnot(length(y) == 1)
#     (x * y) / abs(y)
# }
#
# #' @rdname scalar_projection
# #' @export
# as.double.Coord3D <- function(x, y = as_coord3d(1, 0, 0), ...) {
#     if (!is_coord3d(y))
#         y <- as_coord3d(y, ...)
#     stopifnot(length(y) == 1)
#     (x * y) / abs(y)
# }
# nolint end

test_that("coord1d()", {
	x <- c(2, 5, 7)
	p1 <- as_coord1d(x = x)
	expect_true(is_coord1d(p1))
	expect_equal(p1$x, x)
	expect_length(p1, 3)

	expect_equal(p1[2:3], as_coord1d(x = x[2:3]))

	p2 <- as_coord1d("origin")
	expect_equal(rep_len(p2, 4), as_coord1d(rep(0, 4)))
	expect_true(is_equivalent(p2, "origin"))
})

test_that("coord2d()", {
	x <- c(2, 5, 7)
	y <- c(3, 4, 6)
	p1 <- as_coord2d(x = x, y = y)
	expect_true(is_coord2d(p1))
	expect_equal(p1$x, x)
	expect_equal(p1$y, y)
	expect_length(p1, 3)

	expect_equal(p1[2:3], as_coord2d(x = x[2:3], y = y[2:3]))

	p2 <- as_coord2d("origin")
	expect_equal(rep_len(p2, 4), as_coord2d(rep(0, 4), rep(0, 4)))
	expect_true(is_equivalent(p2, "origin"))
})

test_that("coord3d()", {
	x <- c(2, 5, 7)
	y <- c(3, 4, 6)
	z <- c(2, 7, 3)
	p1 <- as_coord3d(x = x, y = y, z = z)

	expect_true(is_coord3d(p1))
	expect_equal(p1$x, x)
	expect_equal(p1$y, y)
	expect_equal(p1$z, z)
	expect_length(p1, 3)

	expect_equal(p1[2:3], as_coord3d(x = x[2:3], y = y[2:3], z = z[2:3]))

	p3 <- as_coord3d("origin")
	expect_equal(rep_len(p3, 4), as_coord3d(rep(0, 4), rep(0, 4), rep(0, 4)))
	expect_true(is_equivalent(p3, "origin"))
})

test_that("convex_hull2d()", {
	p <- as_coord2d(x = rnorm(25), y = rnorm(25))
	hull <- convex_hull2d(p)
	expect_true(is_polygon2d(hull))
	expect_true(hull$is_convex)
	# Same coordinates as manual chull calculation
	expected <- p[rev(grDevices::chull(as.list(p)))]
	expect_equal(hull$x, expected$x)
	expect_equal(hull$y, expected$y)
})

test_that("mean()", {
	x <- c(2, 5, 7)
	y <- c(3, 4, 6)
	z <- c(2, 7, 3)
	p1 <- as_coord1d(x = x)
	expect_equal(mean(p1), as_coord1d(mean(x)))
	expect_equal(mean(p1), sum(p1, na.rm = TRUE) / length(p1))

	p2 <- as_coord2d(x = x, y = y)
	expect_equal(mean(p2), as_coord2d(mean(x), mean(y)))
	expect_equal(mean(p2), sum(p2, na.rm = TRUE) / length(p2))

	p3 <- as_coord3d(x = x, y = y, z = z)
	expect_equal(mean(p3), as_coord3d(mean(x), mean(y), mean(z)))
	expect_equal(mean(p3), sum(p3, na.rm = TRUE) / length(p3))
})

test_that("as_coord1d()", {
	expect_warning(as_coord1d("foobar"))

	l <- list(x = 1:10)
	p <- as_coord1d(l)
	expect_equal(p$x, l$x)

	m <- as.matrix(1:10)
	p <- as_coord1d(m)
	expect_equal(p$x, 1:10)
	expect_true(is.data.frame(as.data.frame(p)))
	expect_true(is.list(as.list(p)))

	expect_equal(as_coord1d(p), p)
	expect_equal(as_coord1d(as.data.frame(p)), p)

	# Projections form Coord2D
	x <- c(2, 5, 7)
	y <- c(3, 4, 6)
	p1 <- as_coord2d(x = x, y = y)
	expect_equal(as_coord1d(p1, line = "x-axis")$x, x)
	expect_equal(as_coord1d(p1, scale = 0.5)$x, x + 0.5 * y)
	expect_equal(as_coord1d(p1, permutation = "yx")$x, y)
	expect_equal(as_coord1d(p1, permutation = "yx", scale = 0.5)$x, y + 0.5 * x)

	line <- as_line2d(a = 0, b = -1, c = 2)
	expect_equal(as_coord1d(p1, line = line)$x, x)
	expect_equal(as_coord1d(p1, line = line, scale = 0.5)$x, x + 0.5 * (y - 2))

	line <- as_line2d(a = 1, b = 0, c = -2)
	expect_equal(as_coord1d(p1, line = line)$x, y)

	x6 <- c(-5, 5, 5)
	y6 <- c(5, -5, 5)
	l6 <- as_line2d(a = 1, b = -1, c = 0)
	p6 <- as_coord2d(x = x6, y = y6)
	expect_equal(as_coord1d(p6, line = l6)$x, c(0, 0, sqrt(50)))

	l7 <- as_line2d(a = 1, b = -1, c = 2)
	p7 <- as_coord2d(x = x6, y = y6)
	expect_equal(as_coord1d(p7, line = l7)$x, c(0, 0, sqrt(50)))
})

test_that("as_coord2d()", {
	v1 <- as_coord2d(angle(90, "degrees"), radius = 2)
	expect_equal(v1$x, 0)
	expect_equal(v1$y, 2)

	x <- c(2, 5, 7)
	y <- c(3, 4, 6)
	p1 <- as_coord2d(x = x, y = y)
	expect_equal(p1 == p1, rep_len(TRUE, 3))
	expect_equal(p1 != p1, rep_len(FALSE, 3))

	expect_equal(p1, as_coord2d(as.data.frame(p1)))
	expect_true(is.matrix(as.matrix(p1)))
	expect_true(is.data.frame(as.data.frame(p1)))
	expect_true(is.list(as.list(p1)))

	expect_equal(p1, as_coord2d(p1))

	l <- list(x = x, y = y)
	expect_equal(p1, as_coord2d(l))
	m <- as.matrix(as.data.frame(l))
	expect_equal(p1, as_coord2d(m))

	expect_equal(as_coord2d("origin"), as_coord2d(0, 0))
	expect_equal(as_coord2d("x-axis"), as_coord2d(1, 0))
	expect_equal(as_coord2d("y-axis"), as_coord2d(0, 1))
	expect_warning(as_coord2d("foobar"))

	# Projections from Coord3D
	x <- c(2, 5, 7)
	y <- c(3, 4, 6)
	z <- c(4, 9, 3)
	p <- as_coord3d(x, y, z)
	p1 <- as_coord2d(p)
	expect_equal(p1$x, x)
	expect_equal(p1$y, y)

	p2 <- as_coord2d(p, "xzy")
	expect_equal(p2$x, x)
	expect_equal(p2$y, z)

	p3 <- as_coord2d(p, "xzy", plane = "xy-plane", scale = 0.5, alpha = 60, unit = "degrees")
	scale <- 0.5
	alpha <- angle(60, "degrees")
	expect_equal(p3$x, x + scale * cos(alpha) * y)
	expect_equal(p3$y, z + scale * sin(alpha) * y)

	# projection to plane parallel to xy-plane
	plane4 <- as_plane3d(a = 0, b = 0, c = 1, d = -3)
	p4 <- as_coord2d(p, "xzy", plane = plane4, scale = 0.5, alpha = 60, unit = "degrees")
	expect_equal(p4$x, x + scale * cos(alpha) * (y - 3))
	expect_equal(p4$y, z + scale * sin(alpha) * (y - 3))

	p5 <- as_coord2d(p, "xzy", plane = plane4)
	expect_equal(p5$x, x)
	expect_equal(p5$y, z)

	p6 <- as_coord2d(p, plane = plane4)
	expect_equal(p6$x, x)
	expect_equal(p6$y, y)

	# roll rotates the in-plane coordinate frame around the plane normal
	roll <- angle(90, "degrees")
	p_roll <- as_coord2d(p, roll = roll)
	expect_equal(p_roll$x, -y)
	expect_equal(p_roll$y, x)

	# oblique projection onto non-xy-plane (xz-plane, normal = (0,-1,0))
	# aligned in-plane coords are (-z, x); depth component is -y
	p7 <- as_coord2d(p, plane = "xz-plane", scale = 0.5, alpha = 60, unit = "degrees")
	expect_equal(p7$x, -z - scale * cos(alpha) * y)
	expect_equal(p7$y, x - scale * sin(alpha) * y)

	# orthographic projections to axis-aligned planes through origin
	p_xy1 <- as_coord2d(p, plane = "xy-plane")
	p_xy2 <- as_coord2d(p, permutation = "xyz")
	expect_equal(p_xy1$x, x)
	expect_equal(p_xy1$y, y)
	expect_equal(p_xy1, p_xy2)

	p_yx1 <- as_coord2d(p, plane = "yx-plane")$rotate(degrees(-90))
	p_yx2 <- as_coord2d(p, permutation = "yxz")
	expect_equal(p_yx1$x, y)
	expect_equal(p_yx1$y, x)
	expect_equal(p_yx1, p_yx2)

	p_xz1 <- as_coord2d(p, plane = "xz-plane")$rotate(degrees(-90))
	p_xz2 <- as_coord2d(p, permutation = "xzy")
	expect_equal(p_xz1$x, x)
	expect_equal(p_xz1$y, z)
	expect_equal(p_xz1, p_xz2)

	p_zx1 <- as_coord2d(p, plane = "zx-plane")$rotate(degrees(180))
	p_zx2 <- as_coord2d(p, permutation = "zxy")
	expect_equal(p_zx1$x, z)
	expect_equal(p_zx1$y, x)
	expect_equal(p_zx1, p_zx2)

	p_yz1 <- as_coord2d(p, plane = "yz-plane")$rotate(degrees(-90))
	p_yz2 <- as_coord2d(p, permutation = "yzx")
	expect_equal(p_yz1$x, y)
	expect_equal(p_yz1$y, z)
	expect_equal(p_yz1, p_yz2)

	p_zy1 <- as_coord2d(p, plane = "zy-plane")$rotate(degrees(180))
	p_zy2 <- as_coord2d(p, permutation = "zyx")
	expect_equal(p_zy1$x, z)
	expect_equal(p_zy1$y, y)
	expect_equal(p_zy1, p_zy2)
})

test_that("as_coord3d()", {
	x <- c(2, 5, 7)
	y <- c(3, 4, 6)
	z <- c(2, 7, 3)
	p1 <- as_coord3d(x = x, y = y, z = z)

	expect_equal(p1, as_coord3d(as.data.frame(p1)))
	expect_equal(p1 == p1, rep_len(TRUE, 3))
	expect_equal(p1 != p1, rep_len(FALSE, 3))
	expect_true(is.matrix(as.matrix(p1)))
	expect_true(is.data.frame(as.data.frame(p1)))
	expect_true(is.list(as.list(p1)))

	expect_equal(p1, as_coord3d(p1))

	l <- list(x = x, y = y, z = z)
	expect_equal(p1, as_coord3d(l))
	m <- as.matrix(as.data.frame(l))
	expect_equal(p1, as_coord3d(m))
	l <- list(x = x, y = y)
	expect_equal(as_coord3d(x, y, 0), as_coord3d(l))
	expect_equal(as_coord3d(x, y, z), as_coord3d(l, z = z))

	expect_equal(as_coord3d("origin"), as_coord3d(0, 0, 0))
	expect_equal(as_coord3d("x-axis"), as_coord3d(1, 0, 0))
	expect_equal(normal3d("yz-plane"), as_coord3d(1, 0, 0))
	expect_equal(normal3d("zy-plane"), as_coord3d(-1, 0, 0))
	expect_equal(as_coord3d("y-axis"), as_coord3d(0, 1, 0))
	expect_equal(normal3d("xz-plane"), as_coord3d(0, -1, 0))
	expect_equal(normal3d("zx-plane"), as_coord3d(0, 1, 0))
	expect_equal(as_coord3d("z-axis"), as_coord3d(0, 0, 1))
	expect_equal(normal3d("xy-plane"), as_coord3d(0, 0, 1))
	expect_equal(normal3d("yx-plane"), as_coord3d(0, 0, -1))
	expect_warning(as_coord3d("foobar"))

	# spherical coordinates
	expect_equal(
		as_coord3d(
			angle(1 / 3, "pi-radians"),
			radius = 8,
			inclination = 1 / 6,
			unit = "pi-radians"
		),
		as_coord3d(2, 2 * sqrt(3), 4 * sqrt(3))
	)
	# cylindrical coordinates
	expect_equal(as_coord3d(arctangent(-3), radius = sqrt(10), z = "5"), as_coord3d(1, -3, 5))
})

test_that("complex", {
	c1 <- complex(real = 1:4, imaginary = 1:4)
	p1 <- as_coord2d(c1)
	expect_equal(c1, as.complex(p1))
	expect_equal(Re(c1), Re(p1))
	expect_equal(Im(c1), Im(p1))
	expect_equal(Mod(c1), Mod(p1))
	expect_equal(Arg(c1), Arg(p1))
	expect_equal(Conj(c1), Conj(p1))
	expect_equal(abs(c1), abs(p1))
	expect_equal(abs(p1), sqrt(p1 * p1))

	p2 <- as_coord3d(p1, z = 0)
	expect_equal(abs(p1), abs(p2))
	expect_equal(abs(p2), sqrt(p2 * p2))
})

test_that("Ops", {
	x <- c(2, 5, 7)
	y <- c(3, 4, 6)
	z <- c(2, 7, 3)
	p1 <- as_coord1d(x = x)
	p2 <- as_coord2d(x = x, y = y)
	p3 <- as_coord3d(x = x, y = y, z = z)

	expect_true(all(p1 == p1))
	expect_true(all(p2 == p2))
	expect_true(all(p3 == p3))

	expect_false(any(p1 != p1))
	expect_false(any(p2 != p2))
	expect_false(any(p3 != p3))

	expect_equal(p1, +p1)
	expect_equal(p2, +p2)
	expect_equal(p3, +p3)

	expect_equal(p1 + as_coord1d(1), as_coord1d(x + 1))
	expect_equal(as_coord1d(1) + p1, as_coord1d(x + 1))
	expect_equal(p2 + as_coord2d(1, 1), as_coord2d(x + 1, y + 1))
	expect_equal(as_coord2d(1, 1) + p2, as_coord2d(x + 1, y + 1))
	expect_equal(p3 + as_coord3d(1, 1, 1), as_coord3d(x + 1, y + 1, z + 1))
	expect_equal(as_coord3d(1, 1, 1) + p3, as_coord3d(x + 1, y + 1, z + 1))

	expect_equal(p1$clone()$scale(-1), -p1)
	expect_equal(p2$clone()$scale(-1), -p2)
	expect_equal(p3$clone()$scale(-1), -p3)

	expect_equal(p1 - as_coord1d(1), as_coord1d(x - 1))
	expect_error(as_coord1d(1) - p1)
	expect_equal(p2 - as_coord2d(1, 1), as_coord2d(x - 1, y - 1))
	expect_error(as_coord2d(1, 1) - p2)
	expect_equal(p3 - as_coord3d(1, 1, 1), as_coord3d(x - 1, y - 1, z - 1))
	expect_error(as_coord3d(1, 1, 1) - p3)

	expect_equal(p1 * p1, x * x)

	expect_equal(p1, 1 * p1)
	expect_equal(p2, 1 * p2)
	expect_equal(p3, 1 * p3)

	expect_equal(p1, p1 * 1)
	expect_equal(p2, p2 * 1)
	expect_equal(p3, p3 * 1)

	expect_error(p1 * "foobar")
	expect_error(p2 * "foobar")
	expect_error(p3 * "foobar")

	expect_error(p1 / "foobar")
	expect_error(p2 / "foobar")
	expect_error(p3 / "foobar")
})

test_that("rounding S3 generics", {
	expect_equal(as_coord1d(1), floor(as_coord1d(1.2)))
	expect_equal(as_coord2d(1, 1), floor(as_coord2d(1.2, 1.8)))
	expect_equal(as_coord3d(1, 1, 3), floor(as_coord3d(1.2, 1.8, 3.2)))

	expect_equal(as_coord1d(2), ceiling(as_coord1d(1.2)))
	expect_equal(as_coord2d(2, 2), ceiling(as_coord2d(1.2, 1.8)))
	expect_equal(as_coord3d(2, 2, 4), ceiling(as_coord3d(1.2, 1.8, 3.2)))

	expect_equal(as_coord1d(1), round(as_coord1d(1.2)))
	expect_equal(as_coord2d(1, 2), round(as_coord2d(1.2, 1.8)))
	expect_equal(as_coord3d(1, 2, 3), round(as_coord3d(1.2, 1.8, 3.2)))

	expect_equal(as_coord1d(1.2), signif(as_coord1d(1.2)))
	expect_equal(as_coord2d(1.2, 1.8), signif(as_coord2d(1.2, 1.8)))
	expect_equal(as_coord3d(1.2, 1.8, 3.2), signif(as_coord3d(1.2, 1.8, 3.2)))

	expect_equal(as_coord1d(1), trunc(as_coord1d(1.2)))
	expect_equal(as_coord2d(1, 1), trunc(as_coord2d(1.2, 1.8)))
	expect_equal(as_coord3d(1, 1, 3), trunc(as_coord3d(1.2, 1.8, 3.2)))
})

test_that("`c.Coord1D()`, `c.Coord2D()`, and `c.Coord3D()`", {
	p1 <- as_coord1d(1:4)
	p2 <- as_coord1d(5:8)
	expect_equal(c(p1, p2), as_coord1d(1:8))

	p1 <- as_coord2d(1:4, 1:4)
	p2 <- as_coord2d(5:8, 5:8)
	expect_equal(c(p1, p2), as_coord2d(1:8, 1:8))

	p1 <- as_coord3d(1:4, 1:4, 1:4)
	p2 <- as_coord3d(5:8, 5:8, 5:8)
	expect_equal(c(p1, p2), as_coord3d(1:8, 1:8, 1:8))
})

test_that("print.Coord1D()", {
	x <- c(2, 5, 7)
	p1 <- as_coord1d(x = x)
	expect_snapshot(print(p1))
	expect_snapshot(print(as_coord1d(numeric(0))))
})

test_that("print.Coord2D()", {
	x <- c(2, 5, 7)
	y <- c(3, 4, 6)
	p1 <- as_coord2d(x = x, y = y)
	expect_snapshot(print(p1))
	expect_snapshot(print(as_coord2d(numeric(0))))
})

test_that("print.Coord3D()", {
	x <- c(2, 5, 7)
	y <- c(3, 4, 6)
	z <- c(2, 7, 3)
	p1 <- as_coord3d(x = x, y = y)
	expect_snapshot(print(p1))
	expect_snapshot(print(as_coord3d(numeric(0))))
})

# nolint start
# test_that("scalar projections", {
#     x <- c(2, 5, 7)
#     y <- c(3, 4, 6)
#     z <- c(4, 9, 3)
#     p1 <- as_coord2d(x = x, y = y)
#     expect_equal(as.double(p1), x)
#
#     p3 <- as_coord3d(x = x, y = y, z = z)
#     expect_equal(as.double(p3), x)
#
#     y <- as_coord2d(1, -1)
#     n <- normal2d(as_coord2d(0.5, 0.5))
#     expect_equal(n * p1,
#                  as.double(p1, y))
#
#     p2 <- as_coord2d(2, 1)
#     n <- as_coord2d(0, -1)
#     expect_equal(n * p2,
#                  as.double(p2, n))
# })
# nolint end

test_that("dot_product", {
	p1 <- as_coord1d(2)
	p2 <- as_coord1d(3)
	expect_equal(dot_product1d(p1, p2), 6)
	expect_equal(p1 * p2, 6)

	p1 <- as_coord2d(1, 2)
	p2 <- as_coord2d(3, 4)
	expect_equal(dot_product2d(p1, p2), 11)
	expect_equal(p1 * p2, 11)

	p1 <- as_coord3d(1, 2, 3)
	p2 <- as_coord3d(4, 5, 6)
	expect_equal(dot_product3d(p1, p2), 32)
	expect_equal(p1 * p2, 32)

	skip_if_not(getRversion() >= "4.3.0")
	expect_equal(as_coord1d(2) %*% as_coord1d(3), 6)
	expect_equal(as_coord2d(1, 2) %*% as_coord2d(3, 4), 11)
	expect_equal(as_coord3d(1, 2, 3) %*% as_coord3d(4, 5, 6), 32)
})

test_that("cross_product3d", {
	x <- as_coord3d(2, 3, 4)
	y <- as_coord3d(5, 6, 7)
	expect_equal(cross_product3d(x, y), as_coord3d(-3, 6, -3))

	skip_if_not(getRversion() >= "4.4.0")
	expect_equal(crossprod(x, y), as_coord3d(-3, 6, -3))
})

test_that("normal2d", {
	expect_equal(normal2d(as_coord2d(x = 5, y = 3), normalize = FALSE), as_coord2d(3, -5))
	expect_equal(normal2d(as_line2d(angle(0, "degrees"))), as_coord2d(0, -1))
})

test_that("normal3d", {
	expect_equal(normal3d("xy-plane"), as_coord3d(0, 0, 1))
	expect_equal(normal3d(as_coord3d("x-axis"), cross = "y-axis"), as_coord3d(0, 0, 1))
	expect_equal(normal3d(as_coord3d(2, 0, 0), cross = as_coord3d(0, 2, 0)), as_coord3d(0, 0, 1))
	expect_warning(normal3d("foobar"))
})

test_that("is functions", {
	p <- as_coord1d(x = c(2, NaN, Inf, NA_real_))
	expect_equal(is.na(p), c(FALSE, TRUE, FALSE, TRUE))
	expect_equal(is.nan(p), c(FALSE, TRUE, FALSE, FALSE))
	expect_equal(is.infinite(p), c(FALSE, FALSE, TRUE, FALSE))
	expect_equal(is.finite(p), c(TRUE, FALSE, FALSE, FALSE))
	p <- as_coord2d(x = c(2, NaN, Inf, NA_real_), y = 0)
	expect_equal(is.na(p), c(FALSE, TRUE, FALSE, TRUE))
	expect_equal(is.nan(p), c(FALSE, TRUE, FALSE, FALSE))
	expect_equal(is.infinite(p), c(FALSE, FALSE, TRUE, FALSE))
	expect_equal(is.finite(p), c(TRUE, FALSE, FALSE, FALSE))
	p <- as_coord3d(x = c(2, NaN, Inf, NA_real_), y = 0, z = 0)
	expect_equal(is.na(p), c(FALSE, TRUE, FALSE, TRUE))
	expect_equal(is.nan(p), c(FALSE, TRUE, FALSE, FALSE))
	expect_equal(is.infinite(p), c(FALSE, FALSE, TRUE, FALSE))
	expect_equal(is.finite(p), c(TRUE, FALSE, FALSE, FALSE))
})

test_that("range()", {
	x <- c(2, 5, 7)
	y <- c(3, 4, 6)
	z <- c(2, 7, 3)
	expect_equal(range(as_coord1d(x), na.rm = TRUE), as_coord1d(c(2, 7)))
	expect_equal(range(as_coord2d(x, y), na.rm = TRUE), as_coord2d(c(2, 7), c(3, 6)))
	expect_equal(range(as_coord3d(x, y, z), na.rm = TRUE), as_coord3d(c(2, 7), c(3, 6), c(2, 7)))
})

test_that("sort.Coord2D() orders farthest first by default", {
	p <- as_coord2d(x = c(1, 3, 2), y = c(3, 1, 2))
	p_sorted <- sort(p, alpha = degrees(45))
	depths <- painter_depth(p, scale = 1, alpha = degrees(45))
	expect_r6_class(p_sorted, "Coord2D")
	expect_equal(length(p_sorted), length(p))
	expect_equal(painter_depth(p_sorted, scale = 1, alpha = degrees(45)), sort(depths))
})

test_that("sort.Coord3D() orthographic sorts by z by default", {
	p <- as_coord3d(x = 1:3, y = 1:3, z = c(3, 1, 2))
	p_sorted <- sort(p)
	expect_r6_class(p_sorted, "Coord3D")
	expect_equal(p_sorted$z, c(1, 2, 3))
})

#' Test if 1D affine transformation matrix
#'
#' `is_transform1d()` tests if object is a [transform1d()] affine transformation matrix
#'
#' @param x An object
#' @return A logical value
#' @examples
#' m <- transform1d(diag(2L))
#' is_transform1d(m)
#' is_transform1d(diag(2L))
#' @export
is_transform1d <- function(x) {
	inherits(x, "transform1d")
}

#' Test if 2D affine transformation matrix
#'
#' `is_transform2d()` tests if object is a [transform2d()] affine transformation matrix
#'
#' @param x An object
#' @return A logical value
#' @examples
#' m <- transform2d(diag(3L))
#' is_transform2d(m)
#' is_transform2d(diag(3L))
#' @export
is_transform2d <- function(x) {
	inherits(x, "transform2d")
}

#' Test if 3D affine transformation matrix
#'
#' `is_transform3d()` tests if object is a [transform3d()] affine transformation matrix
#'
#' @param x An object
#' @return A logical value
#' @examples
#' m <- transform3d(diag(4L))
#' is_transform3d(m)
#' is_transform3d(diag(4L))
#' @export
is_transform3d <- function(x) {
	inherits(x, "transform3d")
}

#' Cast to 1D affine transformation matrix
#'
#' `as_transform1d()` casts to a [transform1d()] affine transformation matrix
#'
#' @param x An object that can be cast to a
#' @param ... Further arguments passed to or from other methods
#' @return A [transform1d()] object
#' @examples
#' m <- diag(2L)
#' as_transform1d(m)
#' @export
as_transform1d <- function(x, ...) {
	UseMethod("as_transform1d")
}

#' Cast to 2D affine transformation matrix
#'
#' `as_transform2d()` casts to a [transform2d()] affine transformation matrix
#'
#' @param x An object that can be cast to a
#' @param ... Further arguments passed to or from other methods
#' @return A [transform2d()] object
#' @examples
#' m <- diag(3L)
#' as_transform2d(m)
#' @export
as_transform2d <- function(x, ...) {
	UseMethod("as_transform2d")
}

#' Cast to 3D affine transformation matrix
#'
#' `as_transform3d()` casts to a [transform3d()] affine transformation matrix
#'
#' @param x An object that can be cast to a
#' @param ... Further arguments passed to or from other methods
#' @return A [transform3d()] object
#' @examples
#' m <- diag(4L)
#' as_transform3d(m)
#' @export
as_transform3d <- function(x, ...) {
	UseMethod("as_transform3d")
}

#' @rdname as_transform1d
#' @export
as_transform1d.transform1d <- function(x, ...) {
	chkDots(...)
	x
}

#' @rdname as_transform2d
#' @export
as_transform2d.transform2d <- function(x, ...) {
	chkDots(...)
	x
}

#' @rdname as_transform3d
#' @export
as_transform3d.transform3d <- function(x, ...) {
	chkDots(...)
	x
}

#' @rdname as_transform1d
#' @export
as_transform1d.default <- function(x, ...) {
	transform1d(as.matrix(x, ...))
}

#' @rdname as_transform2d
#' @export
as_transform2d.default <- function(x, ...) {
	transform2d(as.matrix(x, ...))
}

#' @rdname as_transform3d
#' @export
as_transform3d.default <- function(x, ...) {
	transform3d(as.matrix(x, ...))
}

#' @export
t.at_matrix <- function(x) {
	class(x) <- NULL
	NextMethod()
}

is_at_matrix <- function(x) inherits(x, "at_matrix")

#' @export
Ops.at_matrix <- function(e1, e2) {
	if (is_at_matrix(e1)) {
		class(e1) <- NULL
	}
	if (!missing(e2) && is_at_matrix(e2)) {
		class(e2) <- NULL
	}
	NextMethod()
}

#' @export
Complex.at_matrix <- function(z) {
	class(z) <- NULL
	NextMethod()
}

#' @export
Math.at_matrix <- function(x, ...) {
	class(x) <- NULL
	NextMethod()
}

#' @export
solve.at_matrix <- function(a, b, ...) {
	m <- NextMethod()
	if (missing(b) && !is_at_matrix(m)) {
		class(m) <- class(a)
	}
	m
}

#' @rawNamespace if (getRversion() >= "4.3.0") {
#'   S3method("%*%",at_matrix)
#' }
`%*%.at_matrix` <- function(x, y) {
	m <- NextMethod()
	if (is_at_matrix(x) && is_at_matrix(y) && !is_at_matrix(m)) {
		class(m) <- class(x)
	}
	m
}

#' @export
as.matrix.at_matrix <- function(x, ...) {
	chkDots(...)
	class(x) <- NULL
	x
}

#' Test whether two objects are congruent
#'
#' `is_congruent()` is a S3 generic that tests whether two different objects are \dQuote{congruent}.
#' The `is_congruent()` method for [angle()] classes tests whether two angles are congruent.
#'
#' @param x,y Two objects to test whether they are \dQuote{"congruent"}.
#' @param ... Further arguments passed to or from other methods.
#' @return A logical vector
#'
#' @examples
#'   # Use `is_congruent()` to check if two angles are "congruent"
#'   a1 <- angle(180, "degrees")
#'   a2 <- angle(pi, "radians")
#'   a3 <- angle(-180, "degrees") # Only congruent modulus full turns
#'   a1 == a2
#'   isTRUE(all.equal(a1, a2))
#'   is_congruent(a1, a2)
#'   is_congruent(a1, a2, mod_turns = FALSE)
#'   a1 == a3
#'   isTRUE(all.equal(a1, a3))
#'   is_congruent(a1, a3)
#'   is_congruent(a1, a3, mod_turns = FALSE)
#'
#' @export
is_congruent <- function(x, y, ...) {
	UseMethod("is_congruent")
}

#' @rdname is_congruent
#' @export
is_congruent.numeric <- function(x, y, ..., tolerance = sqrt(.Machine$double.eps)) {
	chkDots(...)
	abs(x - y) < tolerance
}

#' @rdname is_congruent
#' @param mod_turns If `TRUE` angles that are congruent modulo full turns will be considered \dQuote{congruent}.
#' @param tolerance Angles (coerced to half-turns) or numerics with differences smaller
#'                  than `tolerance` will be considered \dQuote{congruent}.
#' @export
is_congruent.angle <- function(x, y, ..., mod_turns = TRUE, tolerance = sqrt(.Machine$double.eps)) {
	chkDots(...)
	if (!is_angle(y)) {
		y <- as_angle(y, unit = angular_unit(x))
	}
	x <- as.numeric(x, "pi-radians")
	y <- as.numeric(y, "pi-radians")
	if (mod_turns) {
		n <- max(length(x), length(y))
		x <- rep_len(x, n) %% 2
		y <- rep_len(y, n) %% 2
		# Half-turn angles near two and zero (within tolerance) are "equivalent"
		id1 <- which(x < tolerance & y + tolerance > 2 + x)
		y[id1] <- y[id1] - 2
		id2 <- which(y < tolerance & x + tolerance > 2 + y)
		x[id2] <- x[id2] - 2
	}
	abs(x - y) < tolerance
}

#' Test whether an object has a Coord1D class
#'
#' `is_coord1d()` tests whether an object has a "Coord1D" class
#'
#' @param x An object
#' @return A logical value
#'
#' @examples
#' p <- as_coord1d(x = sample.int(10, 3))
#' is_coord1d(p)
#' @export
is_coord1d <- function(x) {
	inherits(x, "Coord1D")
}

#' Test whether an object has a Coord2D class
#'
#' `is_coord2d()` tests whether an object has a "Coord2D" class
#'
#' @param x An object
#' @return A logical value
#'
#' @examples
#' p <- as_coord2d(x = sample.int(10, 3), y = sample.int(10, 3))
#' is_coord2d(p)
#' @export
is_coord2d <- function(x) {
	inherits(x, "Coord2D")
}

#' Test whether an object has a Coord3D class
#'
#' `is_coord3d()` tests whether an object has a "Coord3D" class
#'
#' @param x An object
#' @return A logical value
#'
#' @examples
#' p <- as_coord3d(x = sample.int(10, 3),
#'                 y = sample.int(10, 3),
#'                 z = sample.int(10, 3))
#' is_coord3d(p)
#' @export
is_coord3d <- function(x) {
	inherits(x, "Coord3D")
}

#' Test whether an object has a Point1D class
#'
#' `is_point1d()` tests whether an object has a "Point1D" class
#'
#' @param x An object
#' @return A logical value
#'
#' @examples
#' p <- as_point1d(a = 1, b = 5)
#' is_point1d(p)
#' @export
is_point1d <- function(x) inherits(x, "Point1D")

#' Test whether an object has a Line2D class
#'
#' `is_line2d()` tests whether an object has a "Line2D" class
#'
#' @param x An object
#' @return A logical value
#'
#' @examples
#' l <- as_line2d(a = 1, b = 2, c = 3)
#' is_line2d(l)
#' @export
is_line2d <- function(x) inherits(x, "Line2D")

#' Test whether an object has a Plane3D class
#'
#' `is_plane3d()` tests whether an object has a "Plane3D" class
#'
#' @param x An object
#' @return A logical value
#'
#' @examples
#' p <- as_plane3d(a = 1, b = 2, c = 3, 4)
#' is_plane3d(p)
#' @export
is_plane3d <- function(x) inherits(x, "Plane3D")

#' Test whether an object has a Polygon2D class
#'
#' `is_polygon2d()` tests whether an object has a "Polygon2D" class
#'
#' @param x An object
#' @return A logical value
#'
#' @examples
#' p <- as_polygon2d(as_coord2d(x = c(0, 1, 1, 0), y = c(0, 0, 1, 1)))
#' is_polygon2d(p)
#' @export
is_polygon2d <- function(x) inherits(x, "Polygon2D")

#' Test whether an object has a Segment2D class
#'
#' `is_segment2d()` tests whether an object has a "Segment2D" class
#'
#' @param x An object
#' @return A logical value
#'
#' @examples
#' p1 <- as_coord2d(x = c(0, 1), y = c(0, 0))
#' p2 <- as_coord2d(x = c(1, 1), y = c(0, 1))
#' s <- as_segment2d(p1, p2 = p2)
#' is_segment2d(s)
#' @export
is_segment2d <- function(x) inherits(x, "Segment2D")

#' Test whether an object has an Ellipse2D class
#'
#' `is_ellipse2d()` tests whether an object has an "Ellipse2D" class
#'
#' @param x An object
#' @return A logical value
#'
#' @examples
#' c1 <- as_ellipse2d(as_coord2d(0.5, 0.5), r = 0.5)
#' is_ellipse2d(c1)
#' is_ellipse2d(c1) && all(c1$is_circle)
#' e1 <- as_ellipse2d(as_coord2d(0, 0), rx = 2, ry = 1)
#' is_ellipse2d(e1)
#' is_ellipse2d(e1) && all(e1$is_circle)
#' @export
is_ellipse2d <- function(x) inherits(x, "Ellipse2D")

#' Painter's depth, order, and sort methods
#'
#' `painter_depth()` computes depth values for use with the painter's algorithm.
#' For parallel (orthographic or oblique) projections, depth is a dot product
#' of the point with the projection direction vector derived from the plane and
#' oblique parameters.
#' `painter_order()` wraps [painter_depth()] and [order()] to return the integer
#' indices that sort objects farthest-first (the draw order for the painter's
#' algorithm).
#' `sort.Coord2D()` and `sort.Coord3D()` wrap
#' `painter_order()` to return the sorted object directly.
#' `sort.Coord2D()` also handles [Segment2D] objects via inheritance.
#'
#' @param x Object to compute painter's depth/order for, or to sort.
#' @param ... Passed to [as_plane3d()] or [as_angle()] as needed.
#' @param scale Oblique projection foreshortening scale factor.
#'   The 2D methods default to `1` since `0` yields zero depth for all
#'   `Coord2D` points (z is always 0), making depth-based ordering impossible.
#'   The `Coord3D` method defaults to `0` (orthographic projection, depth = z).
#'   A value of `0.5` is used by a \dQuote{cabinet projection}
#'   while a value of `1.0` is used by a \dQuote{cavalier projection}.
#'   All positive values produce the same ordering.
#' @param alpha Oblique projection angle (the angle the off-axis direction is
#'   projected going off at).
#'   An [angle()] object or one coercible to one with `as_angle(alpha, ...)`.
#'   Popular angles are 45 degrees, 60 degrees, and `arctangent(2)` degrees.
#' @param decreasing If `TRUE` sort/order closest first instead of farthest
#'   first.
#' @examples
#' # Coord2D: oblique projection with scale = 1, alpha = 45 degrees
#' p <- as_coord2d(x = c(1, 2, 3), y = c(3, 1, 2))
#' painter_depth(p)
#' painter_order(p)
#' sort(p)
#'
#' # Coord3D: orthographic projection onto xy-plane (depth = z)
#' p3 <- as_coord3d(x = 1:3, y = 1:3, z = c(3, 1, 2))
#' painter_depth(p3)
#' painter_order(p3)
#' sort(p3)
#'
#' # Segment2D: depth/order at midpoints
#' p1 <- as_coord2d(x = c(0, 2), y = c(0, 0))
#' p2 <- as_coord2d(x = c(2, 2), y = c(0, 2))
#' s <- as_segment2d(p1, p2 = p2)
#' painter_depth(s)
#' painter_order(s)
#' sort(s)
#'
#' # Polygon2D: depth/order of each edge
#' vertices <- as_coord2d(x = c(0, 0.5, 1, 0.5), y = c(0.5, 1, 0.5, 0))
#' poly <- as_polygon2d(vertices)
#' painter_depth(poly)
#' painter_order(poly)
#' @rdname painter_depth
#' @export
painter_depth <- function(x, ...) {
	UseMethod("painter_depth")
}

#' @rdname painter_depth
#' @export
painter_depth.Coord2D <- function(x, ..., scale = 1, alpha = angle(45, "degrees")) {
	if (!is_angle(alpha)) {
		alpha <- as_angle(alpha, ...)
	}
	-scale * cos(alpha) * x$x - scale * sin(alpha) * x$y
}

#' @rdname painter_depth
#' @param permutation Either "xyz" (no permutation), "xzy" (permute y and z axes),
#'                    "yxz" (permute x and y axes), "yzx" (x becomes z, y becomes x, z becomes y),
#'                    "zxy" (x becomes y, y becomes z, z becomes x), "zyx" (permute x and z axes).
#'                    This permutation is applied before the projection.
#' @param plane A [Plane3D] class object representing the plane
#'         projected onto, or an object coercible to one using
#'         `as_plane3d(plane, ...)` such as "xy-plane", "xz-plane", or "yz-plane".
#' @param roll Rotation of the in-plane coordinate frame around the plane normal
#'             after the azimuth/inclination alignment.
#'             An [angle()] object or one coercible to one with `as_angle(roll, ...)`.
#'             Defaults to `angle(0)` (no roll), which preserves the azimuth/inclination convention.
#' @export
painter_depth.Coord3D <- function(
	x,
	permutation = c("xyz", "xzy", "yxz", "yzx", "zyx", "zxy"),
	...,
	plane = as_plane3d("xy-plane"),
	scale = 0,
	alpha = angle(45, "degrees"),
	roll = angle(0, "degrees")
) {
	if (!is_plane3d(plane)) {
		plane <- as_plane3d(plane, ...)
	}
	if (!is_angle(alpha)) {
		alpha <- as_angle(alpha, ...)
	}
	if (!is_angle(roll)) {
		roll <- as_angle(roll, ...)
	}
	stopifnot(length(plane) == 1, length(alpha) == 1, length(roll) == 1)
	permutation <- match.arg(permutation)

	denom <- plane$a^2 + plane$b^2 + plane$c^2
	closest <- as_coord3d(
		-plane$a * plane$d / denom,
		-plane$b * plane$d / denom,
		-plane$c * plane$d / denom
	)
	azimuth <- as_angle(plane, type = "azimuth")
	inclination <- as_angle(plane, type = "inclination")
	z_axis <- Coord3D$new(matrix(
		c(0, 0, 1, 1),
		nrow = 1,
		dimnames = list(NULL, c("x", "y", "z", "w"))
	))
	y_axis <- Coord3D$new(matrix(
		c(0, 1, 0, 1),
		nrow = 1,
		dimnames = list(NULL, c("x", "y", "z", "w"))
	))
	p <- x$clone()$permute(permutation)$translate(-closest)$rotate(z_axis, -azimuth)$rotate(
		y_axis,
		-inclination
	)$rotate(z_axis, roll)
	p$z - scale * cos(alpha) * p$x - scale * sin(alpha) * p$y
}

#' @rdname painter_depth
#' @export
painter_depth.Segment2D <- function(x, ..., scale = 1, alpha = angle(45, "degrees")) {
	painter_depth(x$mid_point, ..., scale = scale, alpha = alpha)
}

#' @rdname painter_depth
#' @export
painter_depth.Polygon2D <- function(x, ..., scale = 1, alpha = angle(45, "degrees")) {
	painter_depth(x$edges, ..., scale = scale, alpha = alpha)
}

#' @rdname painter_depth
#' @return `painter_order()` returns an integer vector of indices.
#' @export
painter_order <- function(x, ...) {
	UseMethod("painter_order")
}

#' @rdname painter_depth
#' @export
painter_order.Coord2D <- function(
	x,
	...,
	decreasing = FALSE,
	scale = 1,
	alpha = angle(45, "degrees")
) {
	order(painter_depth(x, ..., scale = scale, alpha = alpha), decreasing = decreasing)
}

#' @rdname painter_depth
#' @export
painter_order.Coord3D <- function(
	x,
	permutation = c("xyz", "xzy", "yxz", "yzx", "zyx", "zxy"),
	...,
	decreasing = FALSE,
	plane = as_plane3d("xy-plane"),
	scale = 0,
	alpha = angle(45, "degrees"),
	roll = angle(0, "degrees")
) {
	order(
		painter_depth(
			x,
			permutation = permutation,
			...,
			plane = plane,
			scale = scale,
			alpha = alpha,
			roll = roll
		),
		decreasing = decreasing
	)
}

#' @rdname painter_depth
#' @export
painter_order.Polygon2D <- function(
	x,
	...,
	decreasing = FALSE,
	scale = 1,
	alpha = angle(45, "degrees")
) {
	order(painter_depth(x, ..., scale = scale, alpha = alpha), decreasing = decreasing)
}

#' @rdname painter_depth
#' @return `sort.Coord2D()` and `sort.Coord3D()` return a sorted object of the
#'   same class as `x`.
#' @export
sort.Coord2D <- function(x, decreasing = FALSE, ..., scale = 1, alpha = angle(45, "degrees")) {
	x[painter_order(x, ..., decreasing = decreasing, scale = scale, alpha = alpha)]
}

#' @rdname painter_depth
#' @export
sort.Coord3D <- function(
	x,
	decreasing = FALSE,
	...,
	permutation = c("xyz", "xzy", "yxz", "yzx", "zyx", "zxy"),
	plane = as_plane3d("xy-plane"),
	scale = 0,
	alpha = angle(45, "degrees"),
	roll = angle(0, "degrees")
) {
	x[painter_order(
		x,
		...,
		decreasing = decreasing,
		permutation = permutation,
		plane = plane,
		scale = scale,
		alpha = alpha,
		roll = roll
	)]
}

test_that("painter_depth.Coord2D() scale = 0 gives zero depth", {
	p <- as_coord2d(x = 1:3, y = 1:3)
	expect_equal(painter_depth(p, scale = 0), rep(0, 3))
})

test_that("painter_depth.Coord2D() oblique matches formula", {
	p <- as_coord2d(x = c(1, 2), y = c(1, 1))
	s <- 0.5
	a <- degrees(45)
	expected <- -s * cos(a) * p$x - s * sin(a) * p$y
	expect_equal(painter_depth(p, scale = s, alpha = a), expected)
})

test_that("painter_depth.Coord3D() orthographic onto xy-plane gives z", {
	p <- as_coord3d(x = 1:3, y = 1:3, z = c(3, 1, 2))
	expect_equal(painter_depth(p), p$z)
})

test_that("painter_depth.Coord3D() oblique onto xy-plane matches formula", {
	p <- as_coord3d(x = c(1, 2), y = c(1, 1), z = c(0, 1))
	s <- 0.5
	a <- degrees(45)
	expected <- -s * cos(a) * p$x - s * sin(a) * p$y + p$z
	expect_equal(painter_depth(p, scale = s, alpha = a), expected)
})

test_that("painter_depth.Coord3D() respects non-default plane", {
	p <- as_coord3d(x = c(1, 2), y = c(3, 4), z = c(5, 6))
	d <- painter_depth(p, plane = "xz-plane")
	expect_equal(d, -p$y)
})

test_that("painter_depth.Coord3D() roll rotates in-plane frame, leaving depth unchanged", {
	p <- as_coord3d(x = c(1, 2), y = c(1, 1), z = c(2, 3))
	d0 <- painter_depth(p)
	d_roll <- painter_depth(p, roll = degrees(90))
	expect_equal(d0, p$z)
	expect_equal(d_roll, p$z)
})

test_that("painter_depth.Segment2D() uses midpoints", {
	p1 <- as_coord2d(x = c(0, 2), y = c(0, 0))
	p2 <- as_coord2d(x = c(2, 2), y = c(0, 2))
	s <- as_segment2d(p1, p2 = p2)
	expect_equal(painter_depth(s), painter_depth(s$mid_point))
})

test_that("painter_depth.Polygon2D() gives one depth per edge", {
	vertices <- as_coord2d(x = c(0, 0.5, 1, 0.5), y = c(0.5, 1, 0.5, 0))
	poly <- as_polygon2d(vertices)
	d <- painter_depth(poly, scale = 0.5, alpha = degrees(45))
	expect_length(d, 4L)
	expect_equal(d, painter_depth(poly$edges, scale = 0.5, alpha = degrees(45)))
})

test_that("painter_order() returns order() of painter_depth()", {
	p <- as_coord2d(x = c(1, 3, 2), y = c(3, 1, 2))
	expect_equal(painter_order(p), order(painter_depth(p)))
	expect_equal(painter_order(p, decreasing = TRUE), order(painter_depth(p), decreasing = TRUE))
})

test_that("painter_order.Coord3D() returns order() of painter_depth()", {
	p <- as_coord3d(x = 1:3, y = 1:3, z = c(3, 1, 2))
	expect_equal(painter_order(p), order(painter_depth(p)))
})

test_that("painter_order() dispatches painter_depth.Segment2D() via Coord2D inheritance", {
	p1 <- as_coord2d(x = c(0, 2), y = c(0, 0))
	p2 <- as_coord2d(x = c(2, 2), y = c(0, 2))
	s <- as_segment2d(p1, p2 = p2)
	expect_equal(painter_order(s), order(painter_depth(s)))
})

test_that("painter_order.Polygon2D() returns order() of painter_depth()", {
	vertices <- as_coord2d(x = c(0, 0.5, 1, 0.5), y = c(0.5, 1, 0.5, 0))
	poly <- as_polygon2d(vertices)
	expect_equal(painter_order(poly), order(painter_depth(poly)))
})

#' The intersection of two objects.
#'
#' `intersection()` is an S3 method that returns the intersection of two objects.
#'
#' `affiner::intersection()` has the same S3 signature as
#' `euclid::intersection()` (so it shouldn't matter if one masks the other).
#'
#' @param x,y The two objects to compute intersection for.
#' @param ... Passed to other methods (or ignored).
#' @return A list of the object intersections (or `NULL` if no intersection).
#'   For [Line2D]-[Ellipse2D] intersections each list element is a [Coord2D]
#'   of length 2 (two crossing points), length 1 (tangent), or `NULL`
#'   (no intersection).
#' @examples
#' line1 <- as_line2d("x-axis")
#' line2 <- as_line2d("y-axis")
#' line3 <- as_line2d(a = 0, b = 1, c = 2) # y + 2 = 0
#' intersection(line1, line1)
#' intersection(line1, line2)
#' intersection(line1, line3)
#'
#' # Unit circle vs x-axis: two points at (1, 0) and (-1, 0)
#' circ <- as_ellipse2d(as_coord2d("origin"), r = 1)
#' intersection(circ, line1)
#'
#' # Tangent: circle vs line y = 1 (touches top of circle)
#' line_top <- as_line2d(a = 0, b = 1, c = -1) # y - 1 = 0
#' intersection(circ, line_top)
#'
#' # No intersection: circle vs y = 2
#' line_above <- as_line2d(a = 0, b = 1, c = -2) # y - 2 = 0
#' intersection(circ, line_above)
#' @export
intersection <- function(x, y, ...) {
	UseMethod("intersection")
}

#' @rdname intersection
#' @export
intersection.Point1D <- function(x, y, ..., tolerance = sqrt(.Machine$double.eps)) {
	if (!is_point1d(y)) {
		y <- as_point1d(y)
	}
	stopifnot(!any(is_degenerate(x)), !any(is_degenerate(y)))
	n <- max(length(x), length(y))
	x <- rep(standardize(x), length.out = n)
	y <- rep(standardize(y), length.out = n)
	l <- vector("list", n)
	for (i in seq_len(n)) {
		xi <- x[i]
		yi <- y[i]
		if (is_equivalent(xi, yi, tolerance = tolerance)) {
			l[[i]] <- xi
		} # else already NULL
	}
	l
}

#' @rdname intersection
#' @param tolerance Numerics with differences smaller
#'                  than `tolerance` will be considered \dQuote{equivalent}.
#' @export
intersection.Line2D <- function(x, y, ..., tolerance = sqrt(.Machine$double.eps)) {
	if (is_ellipse2d(y)) {
		return(intersection_Ellipse2D_Line2D(y, x, ..., tolerance = tolerance))
	}
	if (is_coord2d(y)) {
		return(intersection_Coord2D_Line2D(y, x, ..., tolerance = tolerance))
	}
	if (!is_line2d(y)) {
		y <- as_line2d(y)
	}
	stopifnot(!any(is_degenerate(x)), !any(is_degenerate(y)))
	n <- max(length(x), length(y))
	x <- rep(standardize(x), length.out = n)
	y <- rep(standardize(y), length.out = n)
	l <- vector("list", n)
	for (i in seq_len(n)) {
		xi <- x[i]
		yi <- y[i]
		if (is_equivalent(xi, yi, tolerance = tolerance)) {
			l[[i]] <- xi
		} else if (!is_parallel(xi, yi, tolerance = tolerance)) {
			m <- matrix(c(xi$a, xi$b, yi$a, yi$b), nrow = 2L, ncol = 2L, byrow = TRUE)
			v <- matrix(c(-xi$c, -yi$c))
			s <- solve(m, v)
			l[[i]] <- as_coord2d(x = s[1L], y = s[2L])
		} # else already NULL
	}
	l
}

intersection_Coord2D_Line2D <- function(x, y, ..., tolerance = sqrt(.Machine$double.eps)) {
	stopifnot(!any(is_degenerate(x)), !any(is_degenerate(y)))
	n <- max(length(x), length(y))
	x <- rep(x, length.out = n)
	y <- rep(y, length.out = n)
	l <- vector("list", n)
	for (i in seq_len(n)) {
		xi <- x[i] # Coord2D
		yi <- y[i] # Line2D
		if (is_equivalent(yi$a * xi$x + yi$b * xi$y, -yi$c, tolerance = tolerance)) {
			l[[i]] <- xi
		} # else already NULL
	}
	l
}

#' @rdname intersection
#' @export
intersection.Plane3D <- function(x, y, ..., tolerance = sqrt(.Machine$double.eps)) {
	if (is_coord3d(y)) {
		return(intersection_Coord3D_Plane3D(y, x, ..., tolerance = tolerance))
	}
	if (!is_plane3d(y)) {
		y <- as_plane3d(y)
	}
	stopifnot(!any(is_degenerate(x)), !any(is_degenerate(y)))
	n <- max(length(x), length(y))
	x <- rep(standardize(x), length.out = n)
	y <- rep(standardize(y), length.out = n)
	l <- vector("list", n)
	for (i in seq_len(n)) {
		xi <- x[i]
		yi <- y[i]
		if (is_equivalent(xi, yi, tolerance = tolerance)) {
			l[[i]] <- xi
		} else if (!is_parallel(xi, yi, tolerance = tolerance)) {
			stop("We have not implemented a `Line3D` class.")
		} # else already NULL
	}
	l
}

intersection_Coord3D_Plane3D <- function(x, y, ..., tolerance = sqrt(.Machine$double.eps)) {
	stopifnot(!any(is_degenerate(x)), !any(is_degenerate(y)))
	n <- max(length(x), length(y))
	x <- rep(x, length.out = n)
	y <- rep(y, length.out = n)
	l <- vector("list", n)
	for (i in seq_len(n)) {
		xi <- x[i] # Coord3D
		yi <- y[i] # Plane3D
		if (is_equivalent(yi$a * xi$x + yi$b * xi$y + yi$c * xi$z, -yi$d, tolerance = tolerance)) {
			l[[i]] <- xi
		} # else already NULL
	}
	l
}

# Vectorised wrapper: each (e[i], l[j]) pair (recycled to max length)
intersection_Ellipse2D_Line2D <- function(e, l, ..., tolerance = sqrt(.Machine$double.eps)) {
	ne <- length(e$rx)
	nl <- length(l)
	n <- max(ne, nl)
	ei <- rep_len(seq_len(ne), n)
	li <- rep_len(seq_len(nl), n)
	result <- vector("list", n)
	for (k in seq_len(n)) {
		r <- ellipse_line_intersection_single(e[ei[k]], l[li[k]], tolerance)
		if (!is.null(r)) result[[k]] <- r
	}
	result
}

# Single Ellipse2D (length 1) vs single Line2D intersection.
# Line: a*x + b*y + c = 0
# Ellipse parametric (affiner convention, y up, CCW angle theta):
#   x(t) = cx + rx*cos(t)*cos(theta) - ry*sin(t)*sin(theta)
#   y(t) = cy + rx*cos(t)*sin(theta) + ry*sin(t)*cos(theta)
# Substituting into the line equation gives A*cos(t) + B*sin(t) + C = 0
# where A = rx*(a*cos(theta) + b*sin(theta)),
#       B = ry*(b*cos(theta) - a*sin(theta)),
#       C = a*cx + b*cy + c.
# Rewriting as R*cos(t - phi) = -C and solving for t.
ellipse_line_intersection_single <- function(e, l, tolerance = sqrt(.Machine$double.eps)) {
	cx <- e$x
	cy <- e$y
	rx <- e$rx
	ry <- e$ry
	theta <- e$theta
	ct <- cos(theta)
	st <- sin(theta)
	a <- l$a
	b <- l$b
	cc <- l$c
	A <- rx * (a * ct + b * st)
	B <- ry * (b * ct - a * st)
	C <- a * cx + b * cy + cc
	R <- sqrt(A^2 + B^2)
	if (R < tolerance) {
		return(NULL) # degenerate ellipse
	}
	ratio <- -C / R
	if (ratio < -(1 + tolerance) || ratio > (1 + tolerance)) {
		return(NULL) # no intersection
	}
	ratio <- max(-1, min(1, ratio))
	phi <- atan2(B, A)
	da <- acos(ratio)
	t1 <- phi + da
	t2 <- phi - da
	x1 <- cx + rx * cos(t1) * ct - ry * sin(t1) * st
	y1 <- cy + rx * cos(t1) * st + ry * sin(t1) * ct
	x2 <- cx + rx * cos(t2) * ct - ry * sin(t2) * st
	y2 <- cy + rx * cos(t2) * st + ry * sin(t2) * ct
	if (abs(x1 - x2) < tolerance && abs(y1 - y2) < tolerance) {
		as_coord2d(x = x1, y = y1) # tangent: one point
	} else {
		# │ condition │        consequence        │
		# │ b > 0     │ x1 < x2                   │
		# │ b < 0     │ x1 > x2                   │
		# │ b = 0     │ x1 = x2 (vertical line)   │
		# │ a < 0     │ y1 < y2                   │
		# │ a > 0     │ y1 > y2                   │
		# │ a = 0     │ y1 = y2 (horizontal line) │
		as_coord2d(x = c(x1, x2), y = c(y1, y2))
	}
}

#' @rdname intersection
#' @export
intersection.Coord1D <- function(x, y, ..., tolerance = sqrt(.Machine$double.eps)) {
	if (!is_coord1d(y)) {
		y <- as_coord1d(y)
	}
	stopifnot(!any(is_degenerate(x)), !any(is_degenerate(y)))
	n <- max(length(x), length(y))
	x <- rep(x, length.out = n)
	y <- rep(y, length.out = n)
	l <- vector("list", n)
	for (i in seq_len(n)) {
		xi <- x[i]
		yi <- y[i]
		if (is_equivalent(xi, yi, tolerance = tolerance)) {
			l[[i]] <- xi
		} # else already NULL
	}
	l
}

#' @rdname intersection
#' @export
intersection.Coord2D <- function(x, y, ..., tolerance = sqrt(.Machine$double.eps)) {
	if (is_line2d(y)) {
		return(intersection_Coord2D_Line2D(x, y, ..., tolerance = tolerance))
	}
	if (!is_coord2d(y)) {
		y <- as_coord2d(y)
	}
	stopifnot(!any(is_degenerate(x)), !any(is_degenerate(y)))
	n <- max(length(x), length(y))
	x <- rep(x, length.out = n)
	y <- rep(y, length.out = n)
	l <- vector("list", n)
	for (i in seq_len(n)) {
		xi <- x[i]
		yi <- y[i]
		if (is_equivalent(xi, yi, tolerance = tolerance)) {
			l[[i]] <- xi
		} # else already NULL
	}
	l
}

#' @rdname intersection
#' @export
intersection.Ellipse2D <- function(x, y, ..., tolerance = sqrt(.Machine$double.eps)) {
	if (is_line2d(y)) {
		return(intersection_Ellipse2D_Line2D(x, y, ..., tolerance = tolerance))
	}
	NextMethod()
}

#' @rdname intersection
#' @export
intersection.Coord3D <- function(x, y, ..., tolerance = sqrt(.Machine$double.eps)) {
	if (is_plane3d(y)) {
		return(intersection_Coord3D_Plane3D(x, y, ..., tolerance = tolerance))
	}
	if (!is_coord3d(y)) {
		y <- as_coord3d(y)
	}
	stopifnot(!any(is_degenerate(x)), !any(is_degenerate(y)))
	n <- max(length(x), length(y))
	x <- rep(x, length.out = n)
	y <- rep(y, length.out = n)
	l <- vector("list", n)
	for (i in seq_len(n)) {
		xi <- x[i]
		yi <- y[i]
		if (is_equivalent(xi, yi, tolerance = tolerance)) {
			l[[i]] <- xi
		} # else already NULL
	}
	l
}

#' Isometric cube grob
#'
#' `isometricCube()` is a grid grob function to render
#' isometric cube faces by automatically wrapping around `affineGrob()`.
#'
#' Any `ggplot2` objects are coerced to grobs by [ggplot2::ggplotGrob()].  Depending on what you'd like
#' to do you may want to instead manually convert a ggplot2 object `gg` to a grob with `gtable::gtable_filter(ggplot2::ggplotGrob(gg), "panel")`.
#'
#' `r affine_transformation_support`
#' @param top A grid grob object to use as the top side of the cube.  ggplot2 objects will be coerced by [ggplot2::ggplotGrob()].
#' @param right A grid grob object to use as the right side of the cube.  ggplot2 objects will be coerced by [ggplot2::ggplotGrob()].
#' @param left A grid grob object to use as the left side of the cube.  ggplot2 objects will be coerced by [ggplot2::ggplotGrob()].
#' @param gp_border A [grid::gpar()] object for the [polygonGrob()] used to draw borders around
#'                  the cube faces.
#' @param name A character identifier (for grid).
#' @param gp A [grid::gpar()] object.
#' @param vp A [grid::viewport()] object (or `NULL`).
#' @return A [grid::gTree()] (grob) object of class "isocube".
#'         As a side effect `grid.isocube()` draws to the active graphics device.
#' @examples
#' # Only works if active graphics device supports affine transformations
#' # such as `png(type="cairo")` on R 4.2+
#' can_run_grid_example <- require("grid", quietly = TRUE) &&
#'   getRversion() >= "4.2.0" &&
#'   isTRUE(dev.capabilities()$transformations)
#' if (can_run_grid_example) {
#'   grid.newpage()
#'   gp_text <- gpar(fontsize = 72)
#'   grid.isocube(top = textGrob("top", gp = gp_text),
#'                right = textGrob("right", gp = gp_text),
#'                left = textGrob("left", gp = gp_text))
#' }
#' if (can_run_grid_example) {
#'   colors <- c("#D55E00", "#009E73", "#56B4E9")
#'   spacings <- c(0.25, 0.2, 0.25)
#'   texts <- c("pkgname", "left\nface", "right\nface")
#'   rots <- c(45, 0, 0)
#'   fontsizes <- c(52, 80, 80)
#'   sides <- c("top", "left", "right")
#'   types <- gridpattern::names_polygon_tiling[c(5, 7, 9)]
#'   l_grobs <- list()
#'   grid.newpage()
#'   for (i in 1:3) {
#'       if (requireNamespace("gridpattern", quietly = TRUE)) {
#'           bg <- gridpattern::grid.pattern_polygon_tiling(
#'                      colour = "grey80",
#'                      fill = c(colors[i], "white"),
#'                      type = types[i],
#'                      spacing = spacings[i],
#'                      draw = FALSE)
#'       } else {
#'           bg <- rectGrob(gp = gpar(col = NA, fill = colors[i]))
#'       }
#'       text <- textGrob(texts[i], rot = rots[i],
#'                        gp = gpar(fontsize = fontsizes[i]))
#'       l_grobs[[sides[i]]] <- grobTree(bg, text)
#'   }
#'   grid.newpage()
#'   grid.isocube(top = l_grobs$top,
#'                right = l_grobs$right,
#'                left = l_grobs$left)
#' }
#' \donttest{# May take more than 5 seconds on CRAN machines
#' can_run_artsy_example <- can_run_grid_example &&
#'   require("aRtsy", quietly = TRUE) &&
#'   require("ggplot2", quietly = TRUE) &&
#'   requireNamespace("gtable", quietly = TRUE)
#' if (can_run_artsy_example) {
#'   gg <- canvas_planet(colorPalette("lava"), threshold = 3) +
#'     scale_x_continuous(expand = c(0, 0)) +
#'     scale_y_continuous(expand = c(0, 0))
#'   grob <- ggplotGrob(gg)
#'   grob <- gtable::gtable_filter(grob, "panel") # grab just the panel
#'   grid.newpage()
#'   grid.isocube(top = grob, left = grob, right = grob,
#'                gp_border = grid::gpar(col = "darkorange", lwd = 12))
#' }
#' }
#' @export
isocubeGrob <- function(
	top,
	right,
	left,
	gp_border = grid::gpar(col = "black", lwd = 12),
	name = NULL,
	gp = grid::gpar(),
	vp = NULL
) {
	stopifnot(getRversion() >= "4.2.0")
	if (inherits(top, "ggplot")) {
		top <- ggplot2::ggplotGrob(top)
	}
	if (inherits(right, "ggplot")) {
		right <- ggplot2::ggplotGrob(right)
	}
	if (inherits(left, "ggplot")) {
		left <- ggplot2::ggplotGrob(left)
	}

	xy <- as_coord2d(angle(seq(90, 360 + 90, by = 60), "degrees"), radius = c(rep(0.488, 6), 0))
	xy$translate(x = 0.5, y = 0.5)
	l_xy <- list()
	l_xy$top <- xy[c(1, 2, 7, 6)]
	l_xy$right <- xy[c(7, 4, 5, 6)]
	l_xy$left <- xy[c(2, 3, 4, 7)]

	vp_define <- grid::viewport(width = grid::unit(1, "snpc"), height = grid::unit(1, "snpc"))

	grid::gTree(
		top = top,
		right = right,
		left = left,
		gp_border = gp_border,
		l_xy = l_xy,
		vp_define = vp_define,
		name = name,
		gp = gp,
		vp = vp,
		cl = "isocube"
	)
}

#' @importFrom grid makeContent
#' @export
makeContent.isocube <- function(x) {
	gl <- grid::gList()
	sides <- c("top", "right", "left")
	for (i in 1:3) {
		side <- sides[[i]]
		xy_side <- x$l_xy[[side]]
		settings <- affine_settings(xy_side, unit = "snpc")
		grob <- x[[side]]
		gl[[i]] <- affineGrob(
			grob,
			vp_define = x$vp_define,
			transform = settings$transform,
			vp_use = settings$vp
		)
	}

	x$gp_border$fill <- "transparent"
	for (i in 1:3) {
		side <- sides[[i]]
		xy_side <- x$l_xy[[side]]
		gl[[i + 3L]] <- grid::polygonGrob(
			xy_side$x,
			xy_side$y,
			gp = x$gp_border,
			default.units = "snpc"
		)
	}

	grid::setChildren(x, gl)
}

#' @rdname isocubeGrob
#' @param ... Passed to `isocubeGrob()`
#' @export
grid.isocube <- function(...) {
	ic_grob <- isocubeGrob(...)
	grid::grid.draw(ic_grob)
	invisible(ic_grob)
}

test_that("isocubeGrob() coerces ggplot objects", {
	skip_if_not(getRversion() >= "4.2.0")
	skip_if_not_installed("ggplot2", "4.0.2")
	gg <- ggplot2::ggplot()
	grob <- isocubeGrob(top = gg, right = gg, left = gg)
	expect_true(grid::is.grob(grob$top))
	expect_true(grid::is.grob(grob$right))
	expect_true(grid::is.grob(grob$left))
})

test_that("isocubeGrob() works", {
	skip_if_not(getRversion() >= "4.3.0")
	skip_if_not(isTRUE(all(capabilities(c("cairo", "png")))))
	f <- tempfile(fileext = ".png")
	png(f, type = "cairo")
	gp_text <- grid::gpar(fontsize = 72)
	grid.isocube(
		top = grid::textGrob("top", gp = gp_text),
		right = grid::textGrob("right", gp = gp_text),
		left = grid::textGrob("left", gp = gp_text)
	)
	dev.off()
	expect_true(file.size(f) > 0)
	unlink(f)
})

#' 2D ellipse R6 Class
#'
#' `Ellipse2D` is an [R6::R6Class()] object representing one or more
#' two-dimensional ellipses.  It inherits from [Coord2D] so its center(s)
#' can be used with `$x` / `$y` / `$xyw` etc.  The semi-axes `rx`/`ry`
#' and orientation angle `theta` are updated automatically by each
#' transformation method.
#'
#' When `rx == ry` (within floating-point tolerance) the ellipse is a
#' circle, which can be tested with the `$is_circle` active binding.
#'
#' @examples
#' # An ellipse
#' e <- as_ellipse2d(as_coord2d(0, 0), rx = 2, ry = 1, theta = degrees(45))
#' print(e)
#' e$is_circle
#'
#' # A circle (special case)
#' c1 <- as_ellipse2d(as_coord2d(0.5, 0.5), rx = 0.5)
#' print(c1)
#' c1$is_circle
#' @export
Ellipse2D <- R6Class(
	"Ellipse2D",
	inherit = Coord2D,
	public = list(
		#' @param xyw A matrix with three columns for homogeneous center coordinates
		#'   (`"x"`, `"y"`, `"w"`).
		#' @param rx Numeric vector of x-axis semi-radii.
		#' @param ry Numeric vector of y-axis semi-radii.
		#' @param theta An [angle()] vector (or numeric, interpreted using
		#'   the default angular unit) of rotation angles.
		initialize = function(xyw, rx, ry, theta) {
			super$initialize(xyw)
			private$rx_ <- as.numeric(rx)
			private$ry_ <- as.numeric(ry)
			if (!is_angle(theta)) {
				theta <- as_angle(theta)
			}
			private$theta_ <- theta
		},
		#' @param n Number of ellipses to print.  If `NULL` print all.
		#' @param ... Passed to [format.default()].
		print = function(n = NULL, ...) {
			nlen <- nrow(private$mat_xyw)
			unit <- angular_unit(self$theta)
			cat("<Ellipse2D[", nlen, "] (theta in ", unit, ")>\n", sep = "")
			m <- cbind(
				x = self$x,
				y = self$y,
				rx = self$rx,
				ry = self$ry,
				theta = as.numeric(self$theta)
			)
			if (!is.null(n) && n < nlen) {
				m <- m[seq_len(n), , drop = FALSE]
			}
			print_mat(m, ...)
			invisible(self)
		},
		#' @param mat `r r2i_transform2d_mat`
		transform = function(mat = transform2d()) {
			if (!is_transform2d(mat)) {
				mat <- as_transform2d(mat)
			}
			# Flush any pending lazy center transform first so centers and
			# shape parameters stay in sync.
			private$apply_any_delayed_transformations()
			# Apply center transform eagerly (bypassing lazy queue).
			private$mat_xyw <- private$mat_xyw %*% mat
			colnames(private$mat_xyw) <- c("x", "y", "w")
			# Apply the linear 2x2 part of mat to each ellipse's shape.
			M <- mat[1:2, 1:2]
			n <- length(self$rx)
			new_rx <- numeric(n)
			new_ry <- numeric(n)
			new_theta <- numeric(n)
			for (i in seq_len(n)) {
				# shape matrix S = R(theta) %*% diag(rx, ry) maps unit circle to ellipse
				# M %*% S is new shape matrix that maps unit circle to ellipse
				sv <- svd(M %*% ellipse_shape_matrix(self$rx[i], self$ry[i], self$theta[i]))
				# 2x2 matrix factors as U %*% diag(d) %*% t(V)
				# U rotates the result into its final orientation in world coordinates
				# diag(d) stretches unit circle into an axis-alined ellipse with semi-axes d[1] >= d[2]
				# t(V) rotates/reflects but doesn't change ellipse shape
				new_rx[i] <- sv$d[1L]
				new_ry[i] <- sv$d[2L]
				# angle of the first column of U which is direction of the major axis
				new_theta[i] <- atan2(sv$u[2L, 1L], sv$u[1L, 1L])
			}
			private$rx_ <- new_rx
			private$ry_ <- new_ry
			private$theta_ <- radians(new_theta)
			invisible(self)
		}
	),
	active = list(
		#' @field rx Numeric vector of x-axis semi-radii (in the ellipse local frame).
		rx = function() private$rx_,
		#' @field ry Numeric vector of y-axis semi-radii (in the ellipse local frame).
		ry = function() private$ry_,
		#' @field theta An [angle()] vector of rotation angles of the ellipse
		#'   x-axis relative to the global x-axis.
		theta = function() private$theta_,
		#' @field is_circle Logical vector; `TRUE` for each ellipse where
		#'   `rx == ry` within floating-point tolerance.
		is_circle = function() {
			tol <- .Machine$double.eps^0.5
			r_max <- pmax(abs(self$rx), abs(self$ry))
			abs(self$rx - self$ry) <= tol * pmax(r_max, 1)
		}
	),
	private = list(
		rx_ = NULL,
		ry_ = NULL,
		theta_ = NULL
	)
)

# Internal: build the 2x2 shape matrix S = R(theta) %*% diag(rx, ry).
# Maps the unit circle to the (uncentered) ellipse.
# theta may be an angle() object or plain numeric radians.
ellipse_shape_matrix <- function(rx, ry, theta) {
	ct <- cos(theta)
	st <- sin(theta)
	matrix(c(ct * rx, st * rx, -st * ry, ct * ry), 2L, 2L)
}

#' Cast to Ellipse2D object
#'
#' `as_ellipse2d()` casts to an [Ellipse2D] object.
#'
#' @param x Object to cast.  Either a [Coord2D] object of center(s) or a
#'   numeric vector of x-coordinates of center(s).
#' @param ... Ignored; only included for S3 method consistency.
#' @return An [Ellipse2D] object.
#' @examples
#' # Ellipse from Coord2D center
#' e <- as_ellipse2d(as_coord2d(0, 0), rx = 2, ry = 1, theta = degrees(30))
#' plot(e)
#'
#' # Multiple ellipses
#' e2 <- as_ellipse2d(x = c(0, 1), y = c(0, 1), rx = c(1, 2), ry = c(0.5, 1))
#' plot(e2)
#'
#' # Multiple circles from x, y, r vectors
#' c2 <- as_ellipse2d(x = c(0, 1), y = c(0, 1), rx = c(0.3, 0.4))
#' plot(c2)
#' @export
as_ellipse2d <- function(x, ...) {
	UseMethod("as_ellipse2d")
}

#' @rdname as_ellipse2d
#' @param r Numeric vector of (circular) radii (default `0.5`).
#' @param rx Numeric vector of x-axis semi-radii (default `r`).
#' @param ry Numeric vector of y-axis semi-radii (default `r`).
#' @param theta An [angle()] vector (or numeric interpreted using the default
#'   angular unit) of rotation angles (default `angle(0)`).
#' @export
as_ellipse2d.Coord2D <- function(x, ..., r = 0.5, rx = r, ry = r, theta = angle(0)) {
	chkDots(...)
	if (!is_angle(theta)) {
		theta <- as_angle(theta)
	}
	n <- max(length(x), length(rx), length(ry), length(theta))
	xyw <- x$xyw
	if (nrow(xyw) < n) {
		xyw <- xyw[rep_len(seq_len(nrow(xyw)), n), , drop = FALSE]
	}
	Ellipse2D$new(
		xyw,
		rx = rep_len(as.numeric(rx), n),
		ry = rep_len(as.numeric(ry), n),
		theta = rep(theta, length.out = n)
	)
}

#' @rdname as_ellipse2d
#' @param y Numeric vector of y-coordinates of center(s) (used when `x` is
#'   numeric).
#' @export
as_ellipse2d.numeric <- function(x, y = 0, ..., r = 0.5, rx = r, ry = rx, theta = angle(0)) {
	chkDots(...)
	as_ellipse2d(as_coord2d(x = x, y = y), rx = rx, ry = ry, theta = theta)
}

#' Plot coordinates, points, lines, polygons, ellipses, and segments
#'
#' [plot()] plots [Coord1D], [Coord2D], [Polygon2D], [Ellipse2D], and
#' [Segment2D] class objects.
#' [points()] draws [Coord1D] and [Coord2D] class objects to an existing plot.
#' [lines()] draws [Coord2D], [Ellipse2D], [Point1D], [Line2D], and [Segment2D]
#' class objects to an existing plot.
#' If the suggested [ggplot2][ggplot2::ggplot2] and [rgl][rgl::rgl] packages
#' are available we also register [ggplot2::autolayer()] methods for [Coord1D],
#' [Coord2D], [Ellipse2D], [Line2D], [Point1D], [Polygon2D], and [Segment2D]
#' class objects and [rgl::plot3d()] methods for [Coord3D] and [Plane3D] class
#' objects.
#'
#' @param x A supported object to plot.
#' @param ... Passed to the underlying plot method.
#' @return Used for its side effect of drawing to the graphics device.
#' @examples
#' c1 <- as_coord2d(x = 0, y = 1:10)
#' l <- as_line2d(a = 1, b = -1, c = 0) # y = x
#' c2 <- c1$clone()$reflect(l)
#' plot(c1, xlim = c(-1, 11), ylim = c(-1, 11),
#'      main = "2D reflection across a line")
#' lines(l)
#' points(c2, col = "red")
#'
#' c1 <- as_coord2d(x = 1:10, y = 1:10)
#' l <- as_line2d(a = -1, b = 0, c = 0) # x = 0
#' c2 <- c1$clone()$project(l)
#' if (require("ggplot2", quietly = TRUE,
#'             include.only = c("ggplot", "autolayer", "labs"))) {
#'   ggplot() +
#'       autolayer(c1) +
#'       autolayer(l) +
#'       autolayer(c2, color = "red") +
#'       labs(title = "2D projection onto a line")
#' }
#'
#' c1 <- as_coord1d(x = seq.int(-4, -1))
#' pt <- as_point1d(a = 1, b = 0) # x = 0
#' c2 <- c1$clone()$reflect(pt)
#' plot(c1, xlim = c(-5, 5), main = "1D reflection across a point")
#' lines(pt)
#' points(c2, col = "red")
#'
#' # 3D reflection across a plane
#' c1 <- as_coord3d(x = 1:10, y = 1:10, z = 1:10)
#' pl <- as_plane3d(a = 0, b = 0, c = -1, d = 2) # z = 2
#' c2 <- c1$clone()$reflect(pl)
#' if (require("rgl", quietly = TRUE, include.only = "plot3d")) {
#'   plot3d(c1, col = "blue", size = 8)
#'   plot3d(pl, color = "grey", alpha = 0.6)
#'   plot3d(c2, add = TRUE, col = "red", size = 8)
#' }
#' @name graphics
#' @export
plot.Coord1D <- function(x, ...) {
	plot(data.frame(x = x$x, y = 0), ..., ylab = "", yaxt = "n")
}

#' @rdname graphics
#' @importFrom graphics points
#' @export
points.Coord1D <- function(x, ...) {
	points(data.frame(x = x$x, y = 0), ...)
}

#' @rdname graphics
#' @importFrom graphics lines
#' @export
lines.Point1D <- function(x, ...) {
	graphics::abline(v = -x$b / x$a, ...)
}

#' @rdname graphics
#' @param asp the y/x aspect ratio.
#' @export
plot.Coord2D <- function(x, ..., asp = 1) {
	plot(as.data.frame(x)[, 1:2], ..., asp = asp)
}

#' @rdname graphics
#' @importFrom graphics points
#' @export
points.Coord2D <- function(x, ...) {
	points(as.data.frame(x)[, 1:2], ...)
}

#' @rdname graphics
#' @importFrom graphics lines
#' @export
lines.Coord2D <- function(x, ...) {
	graphics::lines(x$x, x$y, ...)
	invisible(x)
}

#' @rdname graphics
#' @importFrom graphics lines
#' @export
lines.Polygon2D <- function(x, ...) {
	graphics::lines(c(x$x, x$x[1L]), c(x$y, x$y[1L]), ...)
	invisible(x)
}

#' @rdname graphics
#' @param n Number of vertices used to approximate each ellipse (default `60L`).
#' @importFrom graphics lines
#' @export
lines.Ellipse2D <- function(x, n = 60L, ...) {
	for (i in seq_len(length(x))) {
		p <- as_polygon2d(x[i], n = n)
		graphics::lines(c(p$x, p$x[1L]), c(p$y, p$y[1L]), ...)
	}
	invisible(x)
}

#' @rdname graphics
#' @importFrom graphics polygon
#' @param col,border,density,angle,fillOddEven Passed to [graphics::polygon()].
#' @export
plot.Polygon2D <- function(
	x,
	...,
	col = NA,
	border = NULL,
	density = NULL,
	angle = 45,
	fillOddEven = FALSE,
	asp = 1
) {
	plot(data.frame(x = x$x, y = x$y), type = "n", ..., asp = asp)
	graphics::polygon(
		x$x,
		x$y,
		col = col,
		border = border,
		density = density,
		angle = angle,
		fillOddEven = fillOddEven
	)
	invisible(x)
}

#' @rdname graphics
#' @importFrom graphics polygon
#' @export
plot.Ellipse2D <- function(
	x,
	n = 60L,
	...,
	col = NA,
	border = NULL,
	density = NULL,
	angle = 45,
	fillOddEven = FALSE,
	asp = 1
) {
	polys <- lapply(seq_len(length(x)), function(i) as_polygon2d(x[i], n = n))
	all_x <- unlist(lapply(polys, `[[`, "x"))
	all_y <- unlist(lapply(polys, `[[`, "y"))
	plot(data.frame(x = all_x, y = all_y), type = "n", ..., asp = asp)
	for (p in polys) {
		graphics::polygon(
			p$x,
			p$y,
			col = col,
			border = border,
			density = density,
			angle = angle,
			fillOddEven = fillOddEven
		)
	}
	invisible(x)
}

#' @rdname graphics
#' @importFrom graphics points
#' @export
lines.Line2D <- function(x, ...) {
	index_v <- which(x$b == 0)
	l <- list()
	if (length(index_v)) {
		graphics::abline(v = -x$c / x$a, ...)
		x <- x[-index_v]
	}
	if (length(x)) {
		slope <- -x$a / x$b
		intercept <- -x$c / x$b
		# a,b only allow single values
		for (i in seq_len(length(x))) {
			graphics::abline(a = intercept[i], b = slope[i], ...)
		}
	}
}

#' @rdname graphics
#' @importFrom graphics segments
#' @export
plot.Segment2D <- function(x, ..., asp = 1) {
	xlim <- range(c(x$p1$x, x$p2$x))
	ylim <- range(c(x$p1$y, x$p2$y))
	graphics::plot.default(NULL, xlim = xlim, ylim = ylim, asp = asp)
	graphics::segments(x$p1$x, x$p1$y, x$p2$x, x$p2$y, ...)
	invisible(x)
}

#' @rdname graphics
#' @importFrom graphics segments
#' @export
lines.Segment2D <- function(x, ...) {
	graphics::segments(x$p1$x, x$p1$y, x$p2$x, x$p2$y, ...)
	invisible(x)
}

#' @exportS3Method rgl::plot3d
plot3d.Coord3D <- function(x, ...) {
	rgl::plot3d(as.data.frame(x), ...)
}

#' @exportS3Method rgl::plot3d
plot3d.Plane3D <- function(x, ...) {
	rgl::planes3d(a = x$a, b = x$b, c = x$c, d = x$d, ...)
}

#' Whether two objects intersect
#'
#' `has_intersection()` is an S3 method that returns whether two objects intersect.
#'
#' `affiner::has_intersection()` has the same S3 signature and default method as
#' `euclid::has_intersection()` (so it shouldn't matter if one masks the other).
#' @examples
#' line1 <- as_line2d("x-axis")
#' line2 <- as_line2d("y-axis")
#' line3 <- as_line2d(a = 0, b = 1, c = 2) # y + 2 = 0
#' has_intersection(line1, line1)
#' has_intersection(line1, line2)
#' has_intersection(line1, line3)
#' @param x,y The two objects to check if they intersect.
#' @param ... Passed to other methods (or ignored).
#' @return A logical vector.
#' @export
has_intersection <- function(x, y, ...) {
	UseMethod("has_intersection")
}

#' @rdname has_intersection
#' @param tolerance Numerics with differences smaller
#'                  than `tolerance` will be considered \dQuote{equivalent}.
#' @export
has_intersection.default <- function(x, y, ...) {
	!vapply(intersection(x, y, ...), is.null, logical(1L))
}

#' @rdname has_intersection
#' @export
has_intersection.Point1D <- function(x, y, ..., tolerance = sqrt(.Machine$double.eps)) {
	if (!is_point1d(y)) {
		y <- as_point1d(y)
	}
	stopifnot(!any(is_degenerate(x)), !any(is_degenerate(y)))
	n <- max(length(x), length(y))
	x <- rep(standardize(x), length.out = n)
	y <- rep(standardize(y), length.out = n)
	is_equivalent(x, y, tolerance = tolerance)
}

#' @rdname has_intersection
#' @param tolerance Numerics with differences smaller
#'                  than `tolerance` will be considered \dQuote{equivalent}.
#' @export
has_intersection.Line2D <- function(x, y, ..., tolerance = sqrt(.Machine$double.eps)) {
	if (is_coord2d(y)) {
		return(has_intersection.default(x, y, ..., tolerance = tolerance))
	}
	if (!is_line2d(y)) {
		y <- as_line2d(y)
	}
	stopifnot(!any(is_degenerate(x)), !any(is_degenerate(y)))
	n <- max(length(x), length(y))
	x <- rep(standardize(x), length.out = n)
	y <- rep(standardize(y), length.out = n)
	is_equivalent(x, y, tolerance = tolerance) |
		!is_parallel(x, y, tolerance = tolerance)
}

#' @rdname has_intersection
#' @export
has_intersection.Plane3D <- function(x, y, ..., tolerance = sqrt(.Machine$double.eps)) {
	if (is_coord3d(y)) {
		return(has_intersection.default(x, y, ..., tolerance = tolerance))
	}
	if (!is_plane3d(y)) {
		y <- as_plane3d(y)
	}
	stopifnot(!any(is_degenerate(x)), !any(is_degenerate(y)))
	n <- max(length(x), length(y))
	x <- rep(standardize(x), length.out = n)
	y <- rep(standardize(y), length.out = n)
	is_equivalent(x, y, tolerance = tolerance) |
		!is_parallel(x, y, tolerance = tolerance)
}

# Rotation matrix to Axis-angle representation
# https://en.wikipedia.org/wiki/Axis-angle_representation

#' Convert from 3D rotation matrix to axis-angle representation.
#'
#' `rotate3d_to_AA()` converts from (post-multiplied) rotation matrix
#' to an axis-angle representation of 3D rotations.
#'
#' @examples
#'  # axis-angle representation of 90 degree rotation about the x-axis
#'  rotate3d_to_AA(rotate3d("x-axis", 90, unit = "degrees"))
#'
#'  # find Axis-Angle representation of first rotating about x-axis 180 degrees
#'  # and then rotating about z-axis 45 degrees
#'  R <- rotate3d("x-axis", 180, unit = "degrees") %*%
#'         rotate3d("z-axis", 45, unit = "degrees")
#'  AA <- rotate3d_to_AA(R)
#'
#'  # Can use `rotate3d()` to convert back to rotation matrix representation
#'  all.equal(R, do.call(rotate3d, AA))
#'
#' @seealso \url{https://en.wikipedia.org/wiki/Axis-angle_representation} for more details
#'   about the Axis-angle representation of 3D rotations.
#'   [rotate3d()] can be used to convert from an axis-angle representation to a rotation matrix.
#' @param mat 3D rotation matrix (post-multiplied).
#'            If you have a pre-multiplied rotation matrix
#'            simply transpose it with [t()] to get a post-multiplied rotation matrix.
#' @inheritParams angle
#' @export
rotate3d_to_AA <- function(mat = diag(4), unit = getOption("affiner_angular_unit", "degrees")) {
	unit <- standardize_angular_unit(unit)
	if (!all(dim(mat) == c(3, 3)) && !is_transform3d(mat)) {
		mat <- transform3d(mat)
	}
	R <- mat[1:3, 1:3]

	theta <- arccosine(0.5 * (trace(R) - 1), unit = unit)
	if (all(is_congruent(R, diag(3)))) {
		# no rotation
		theta <- angle(0, unit)
		e <- c(0, 0, 1)
	} else if (is_congruent(theta, angle(180, "degrees"))) {
		# 180 degree rotation
		theta <- angle(from_piradians(1, unit), unit)
		B <- 0.5 * (R + diag(3))
		e <- sqrt(diag(B))
		sB <- sign(B)
		if (all(is_congruent(sB, ppn))) {
			e[3] <- -e[3]
			theta <- angle(from_piradians(-1, unit), unit)
		} else if (all(is_congruent(sB, pnp))) {
			e[2] <- -e[2]
		} else if (all(is_congruent(sB, npp))) {
			e[1] <- -e[1]
		}
	} else {
		e <- numeric(3)
		e[1] <- R[3, 2] - R[2, 3]
		e[2] <- R[1, 3] - R[3, 1]
		e[3] <- R[2, 1] - R[1, 2]
		e <- e / (2 * sin(-theta))
	}
	if (e[3] < 0) {
		# Force z-axis element positive
		e <- -e
		theta <- -theta
	}
	list(axis = as_coord3d(e[1], e[2], e[3]), theta = theta)
}

# Sign matrices for "B" matrix used to tell signs for axis unit vector (up to sign ambiguity) when angle = 180 degrees
# signs for ppp and nnn don't need to be changed
# ppn and npp
ppn <- matrix(c(1, 1, -1, 1, 1, -1, -1, -1, 1), ncol = 3, byrow = TRUE)
# pnp and npn
pnp <- matrix(c(1, -1, 1, -1, 1, -1, 1, -1, 1), ncol = 3, byrow = TRUE)
# pnn and npp
npp <- matrix(c(1, -1, -1, -1, 1, 1, -1, 1, 1), ncol = 3, byrow = TRUE)

# trace of a (square) matrix
trace <- function(m) sum(diag(m))

#' Regular n-gon polygon
#'
#' `regular_ngon_polygon2d()` creates a regular n-gon [Polygon2D] object
#' centered at `(x, y)` with circumradius `radius`.
#'
#' @param n Number of vertices.
#' @param x X-coordinate of the center (default `0`).
#' @param y Y-coordinate of the center (default `0`).
#' @param radius Circumradius: distance from center to each vertex
#'   (default `0.5`).
#' @param theta Angle of the first vertex (default `degrees(90)`).
#'   Will be coerced by [degrees()].
#' @return A [Polygon2D] object with `$is_convex == TRUE`.
#' @seealso [isotoxal_2ngon_polygon2d()] for isotoxal star polygons.
#'   [rectangle_polygon2d()] for rectangles.
#' @examples
#' # A regular hexagon
#' p <- regular_ngon_polygon2d(6)
#' p$is_convex
#' plot(p)
#' @export
regular_ngon_polygon2d <- function(n, x = 0, y = 0, radius = 0.5, theta = degrees(90)) {
	t <- degrees(theta) + degrees(seq(0, by = 360 / n, length.out = n))
	as_polygon2d(as_coord2d(x + radius * cos(t), y + radius * sin(t)), convex = TRUE)
}

# ---
# repo: trevorld/affiner
# file: standalone-affine-settings.r
# last-updated: 2023-04-25
# license: https://unlicense.org
# imports: grid
# ---
#
# nocov start
#
# You may copy this source into your own R package
# by either using `usethis::use_standalone("trevorld/affiner", "standalone-affine-settings.r")`
# or simply copying this file into your `R` directory and adding `grid` to the `Imports` of
# your `DESCRIPTION` file.

#' Compute `grid` affine transformation feature viewports and transformation functions
#'
#' `affine_settings` computes `grid` group affine transformation feature viewport and transformation
#' function settings given the (x,y) coordinates of the corners of the
#' affine transformed "viewport" one wishes to draw in.
#'
#' @param xy An R object with named elements `x` and `y` representing the (x,y) coordinates
#'           of the affine transformed "viewport" one wishes to draw in.
#'           The (x,y) coordinates of the "viewport" should be in
#'           "upper left", "lower left", "lower right", and "upper right" order
#'           (this ordering should be from the perspective of **before**
#'           the "affine transformation" of the "viewport").
#' @param unit Which [grid::unit()] to assume the `xy` "x" and "y" coordinates are expressed in.
#' @return A named list with the following group affine transformation feature viewport
#'         and functions settings:\describe{
#'   \item{transform}{An affine transformation function to pass to [affineGrob()] or [useGrob()].
#'                    If `getRversion()` is less than `"4.2.0"` will instead return `NULL`.}
#'   \item{vp}{A [grid::viewport()] object to pass to [affineGrob()] or [useGrob()].}
#'   \item{sx}{x-axis sx factor}
#'   \item{flipX}{whether the affine transformed "viewport" is "flipped" horizontally}
#'   \item{x}{x-coordinate for viewport}
#'   \item{y}{y-coordinate for viewport}
#'   \item{width}{Width of viewport}
#'   \item{height}{Height of viewport}
#'   \item{default.units}{Default [grid::unit()] for viewport}
#'   \item{angle}{angle for viewport}
#' }
#' @seealso Intended for use with [affineGrob()] or perhaps [grid::useGrob()].
#'          See <https://www.stat.auckland.ac.nz/~paul/Reports/GraphicsEngine/groups/groups.html>
#'          for more information about the group affine transformation feature.
#' @keywords internal
#' @noRd
affine_settings <- function(
	xy = data.frame(x = c(0, 0, 1, 1), y = c(1, 0, 0, 1)),
	unit = getOption("affiner_grid_unit", "inches")
) {
	df <- data.frame(x = xy$x, y = xy$y)
	stopifnot(
		nrow(df) == 4,
		unit %in%
			c(
				"snpc",
				"inches",
				"inch",
				"in",
				"cm",
				"centimeter",
				"centimetre",
				"mm",
				"points",
				"bigpts"
			)
	)
	df <- at_label_face_coords(df)
	x <- at_x(df)
	y <- at_y(df)
	flipX <- at_is_flipped(df)
	origin <- at_origin(df, flipX)
	df <- at_translate_to_origin(df, origin)
	angle <- at_get_angle(df, flipX)
	df <- at_rotate(df, angle)
	width <- at_width(df)
	height <- at_height(df)
	sx <- at_shear_sx(df, flipX, height, width)

	vp <- grid::viewport(
		x = x,
		y = y,
		width = width,
		height = height,
		default.units = unit,
		angle = angle
	)
	if (getRversion() >= "4.2.0") {
		transform <- function(group, ...) {
			grid::viewportTransform(
				group,
				...,
				shear = grid::groupShear(sx = sx),
				flip = grid::groupFlip(flipX = flipX)
			)
		}
	} else {
		transform <- NULL
	}

	list(
		transform = transform,
		vp = vp,
		sx = sx,
		flipX = flipX,
		x = x,
		y = y,
		width = width,
		height = height,
		default.units = unit,
		angle = angle
	)
}

# given (x,y) coordinates of four points in (affine transformed) rectangle
# label points before and after transformation by position in rectangle
at_label_face_coords <- function(df) {
	df$before <- c("upper_left", "lower_left", "lower_right", "upper_right")
	df$after <- ""
	i_left <- order(df$x, -df$y) # leftmost points with tie-brakes by y

	# "lower-left" will be the lowest of the two left-most vertices
	if (df[i_left[1], "y"] < df[i_left[2], "y"]) {
		i_ll <- i_left[1]
	} else {
		i_ll <- i_left[2]
	}
	df[i_ll, "after"] <- "lower_left"
	df[shift_idx(i_ll, 2), "after"] <- "upper_right"
	# translate lower-left to origin and rotate so lower-left and upper-right on x-axis
	# then upper-left will be above x-axis and lower-right will be below x-axis
	df0 <- at_translate_to_origin(df, df[i_ll, ])
	df0 <- at_rotate(df0, df0[shift_idx(i_ll, 2), "theta"])
	if (df0[shift_idx(i_ll, 1), "y"] > 0) {
		df[shift_idx(i_ll, 1), "after"] <- "upper_left"
		df[shift_idx(i_ll, 3), "after"] <- "lower_right"
	} else {
		df[shift_idx(i_ll, 1), "after"] <- "lower_right"
		df[shift_idx(i_ll, 3), "after"] <- "upper_left"
	}
	df
}

shift_idx <- function(x, i) {
	x2 <- (x + i) %% 4
	if (x2 == 0) {
		4
	} else {
		x2
	}
}

at_is_flipped <- function(df) {
	i_ul <- which(df$after == "upper_left")
	i_next <- ifelse(i_ul < 4L, i_ul + 1L, 1L)
	df[i_next, "after"] != "lower_left"
}

# translate "lower left" (after flipping but before rotation) to (0,0)
# add polar r and theta coordinates
at_translate_to_origin <- function(df, origin) {
	df$x <- df$x - origin$x
	df$y <- df$y - origin$y
	df$theta <- to_t(df$x, df$y)
	df$r <- to_r(df$x, df$y)
	df
}

at_shear_sx <- function(df, flipped, height, width) {
	if (flipped) {
		i_ll <- which(df$before == "lower_right")
		i_ul <- which(df$before == "upper_right")
	} else {
		i_ll <- which(df$before == "lower_left")
		i_ul <- which(df$before == "upper_left")
	}
	(df[i_ul, "x"] - df[i_ll, "x"]) / height
}

at_origin <- function(df, flipped = FALSE) {
	if (flipped) {
		i_ll <- which(df$before == "lower_right")
	} else {
		i_ll <- which(df$before == "lower_left")
	}

	df[i_ll, ]
}

at_x <- function(df) {
	mean(df$x)
}

at_y <- function(df) {
	mean(df$y)
}

at_get_angle <- function(df, flipped = FALSE) {
	if (flipped) {
		i_lr <- which(df$before == "lower_left")
	} else {
		i_lr <- which(df$before == "lower_right")
	}
	df[i_lr, "theta"]
}

at_rotate <- function(df, angle) {
	df$theta <- df$theta - angle
	df$x <- to_x(df$theta, df$r)
	df$y <- to_y(df$theta, df$r)
	df
}

at_width <- function(df) {
	x_ll <- df[which(df$before == "lower_left"), "x"]
	x_lr <- df[which(df$before == "lower_right"), "x"]
	abs(x_lr - x_ll)
}

at_height <- function(df) {
	y_ll <- df[which(df$before == "lower_left"), "y"]
	y_ul <- df[which(df$before == "upper_left"), "y"]
	abs(y_ul - y_ll)
}

# Convert between Cartesian and polar coordinates
to_x <- function(t, r) {
	r * cos(pi * t / 180)
}
to_y <- function(t, r) {
	r * sin(pi * t / 180)
}
to_r <- function(x, y) {
	sqrt(x^2 + y^2)
}
to_t <- function(x, y) {
	180 * atan2(y, x) / pi
}

# nocov end

#' 2D line segment R6 Class
#'
#' `Segment2D` is an [R6::R6Class()] object representing a vector of two-dimensional line segments.
#' It inherits from [Coord2D] using the first endpoint `p1` as the coordinate data, so all transformation methods (e.g. `$rotate()`, `$translate()`) and arithmetic operators work directly on the segment vector.
#' The full edge vector to the second endpoint is stored privately and kept consistent under every transformation.
#'
#' @examples
#' p1 <- as_coord2d(x = c(0, 1), y = c(0, 0))
#' p2 <- as_coord2d(x = c(1, 1), y = c(0, 1))
#' s <- as_segment2d(p1, p2 = p2)
#' print(s)
#' s$p1
#' s$p2
#' s$mid_point
#' length(s)
#' s[1]
#' @export
Segment2D <- R6Class(
	"Segment2D",
	inherit = Coord2D,
	public = list(
		#' @param xyw A matrix with three columns for homogeneous `p1` coordinates.
		#'   Column names should be `"x"`, `"y"`, and `"w"`.
		#' @param vec A two-column matrix of edge vectors `p2 - p1`.
		#'   Column names should be `"x"` and `"y"`.
		initialize = function(xyw, vec) {
			super$initialize(xyw)
			private$vec <- vec
		},
		#' @param n Number of segments to print.  If `NULL` print all.
		#' @param ... Passed to [format.default()].
		print = function(n = NULL, ...) {
			ns <- nrow(private$mat_xyw)
			if (is.null(n) || n > ns) {
				n <- ns
			}
			cat("<Segment2D[", ns, "]>\n", sep = "")
			if (n > 0) {
				xyw <- self$xyw
				m <- cbind(
					"p1$x" = xyw[, 1L],
					"p1$y" = xyw[, 2L],
					"p2$x" = xyw[, 1L] + private$vec[, 1L],
					"p2$y" = xyw[, 2L] + private$vec[, 2L]
				)
				print_mat(m[1:n, , drop = FALSE], ...)
			}
			invisible(self)
		},
		#' @param mat `r r2i_transform2d_mat`
		transform = function(mat = transform2d()) {
			if (!is_transform2d(mat)) {
				mat <- as_transform2d(mat)
			}
			private$mid_cache <- NULL
			private$vec <- private$vec %*% mat[1:2, 1:2]
			super$transform(mat)
		}
	),
	private = list(
		vec = NULL,
		mid_cache = NULL
	),
	active = list(
		#' @field p1 A [Coord2D] object of the first endpoints.
		p1 = function() {
			Coord2D$new(self$xyw)
		},
		#' @field p2 A [Coord2D] object of the second endpoints.
		p2 = function() {
			xyw <- self$xyw
			Coord2D$new(cbind(
				x = xyw[, 1L] + private$vec[, 1L],
				y = xyw[, 2L] + private$vec[, 2L],
				w = xyw[, 3L]
			))
		},
		#' @field mid_point A [Coord2D] object of the segment midpoints
		#'   (computed on demand and cached until the next transformation).
		mid_point = function() {
			if (!is.null(private$mid_cache)) {
				return(private$mid_cache)
			}
			xyw <- self$xyw
			m <- Coord2D$new(cbind(
				x = xyw[, 1L] + private$vec[, 1L] / 2,
				y = xyw[, 2L] + private$vec[, 2L] / 2,
				w = xyw[, 3L]
			))
			private$mid_cache <- m
			m
		}
	)
)

#' Cast to Segment2D object
#'
#' `as_segment2d()` creates a [Segment2D] object.
#'
#' @param p1 A [Coord2D] object of first endpoints.
#' @param ... Ignored.
#' @param p2 A [Coord2D] object of second endpoints.
#'   If missing, `vec` must be supplied.
#' @param vec A [Coord2D] object of edge vectors (`p2 - p1`).
#'   Used only when `p2` is missing.
#' @return A [Segment2D] object.
#' @examples
#' p1 <- as_coord2d(x = c(0, 1), y = c(0, 0))
#' p2 <- as_coord2d(x = c(1, 1), y = c(0, 1))
#' s <- as_segment2d(p1, p2 = p2)
#' s$p1
#' s$p2
#' s$mid_point
#'
#' # From a polygon's edges
#' poly <- as_polygon2d(as_coord2d(x = c(0, 1, 1, 0), y = c(0, 0, 1, 1)))
#' as_segment2d(poly)
#' @export
as_segment2d <- function(...) {
	UseMethod("as_segment2d")
}

#' @rdname as_segment2d
#' @export
as_segment2d.Coord2D <- function(p1, ..., p2, vec) {
	chkDots(...)
	xyw1 <- p1$xyw
	if (missing(p2)) {
		if (!is_coord2d(vec)) {
			vec <- as_coord2d(vec)
		}
		stopifnot(length(p1) == length(vec))
		v <- cbind(x = vec$x, y = vec$y)
	} else {
		if (!is_coord2d(p2)) {
			p2 <- as_coord2d(p2)
		}
		stopifnot(length(p1) == length(p2))
		xyw2 <- p2$xyw
		v <- cbind(
			x = xyw2[, 1L] - xyw1[, 1L],
			y = xyw2[, 2L] - xyw1[, 2L]
		)
	}
	Segment2D$new(xyw1, v)
}

#' @rdname as_segment2d
#' @param p A [Polygon2D] object whose edges should be returned.
#' @export
as_segment2d.Polygon2D <- function(p, ...) {
	chkDots(...)
	p$edges
}

#' 2D normal vectors
#'
#' `normal2d()` is an S3 generic that computes a 2D normal vector.
#'
#' @examples
#'   p <- as_coord2d(x = 2, y = 3)
#'   normal2d(p)
#'   normal2d(p, normalize = FALSE)
#' @return A [Coord2D] (normal) vector
#' @rdname normal2d
#' @param x Object to compute a 2D normal vector for
#'          such as a [Line2D] object.
#' @param normalize If `TRUE` coerce to a normalize vector
#' @param ... Passed to or from other methods.
#' @export
normal2d <- function(x, ...) {
	UseMethod("normal2d")
}

#' @rdname normal2d
#' @export
normal2d.Coord2D <- function(x, ..., normalize = TRUE) {
	chkDots(...)
	# Two possible normals:
	# 1: (x, y) => (y, -x) / abs((x, y))
	# 2: (x, y) => (-y, x) / abs((x, y))
	# we arbitrarily went with option 1
	n <- x$clone()$scale(-1, 1)$permute("yx")
	if (normalize) {
		n$scale(1 / abs(x))
	}
	n
}

#' 3D normal vectors
#'
#' `normal3d()` is an S3 generic that computes a 3D normal vector.
#'
#' @examples
#' normal3d("xy-plane")
#' normal3d(as_coord3d(2, 0, 0), cross = as_coord3d(0, 2, 0))
#' @return A [Coord3D] (normal) vector
#' @export
normal3d <- function(x, ...) {
	UseMethod("normal3d")
}

#' @param cross A [Coord3D] vector.
#'              We'll compute the normal of `x` and `cross` by taking their cross product.
#' @param normalize If `TRUE` normalize to a unit vector
#' @param ... Passed to other methods such as [as_coord3d()].
#' @rdname normal3d
#' @export
normal3d.Coord3D <- function(x, cross, ..., normalize = TRUE) {
	if (!is_coord3d(cross)) {
		cross <- as_coord3d(cross, ...)
	}
	n <- cross_product3d(x, cross)
	if (normalize) {
		n$scale(1 / abs(n))
	}
	n
}

#' @param x Object to compute a 3D normal vector for
#'          such as a [Plane3D] object
#' @rdname normal3d
#' @export
normal3d.character <- function(x, ..., normalize = TRUE) {
	chkDots(...)
	xc <- vapply(x, normal3d_character_x, double(1), USE.NAMES = FALSE)
	yc <- vapply(x, normal3d_character_y, double(1), USE.NAMES = FALSE)
	zc <- vapply(x, normal3d_character_z, double(1), USE.NAMES = FALSE)
	p <- as_coord3d(xc, yc, zc)
	if (any(is.na(p) & !is.na(x))) {
		warning("NAs introduced by coercion")
	}
	p$scale(1 / abs(p))
	p
}

normal3d_character_x <- function(x) {
	switch(
		x,
		"yz-plane" = 1,
		"zy-plane" = -1,
		"xz-plane" = 0,
		"zx-plane" = 0,
		"xy-plane" = 0,
		"yx-plane" = 0,
		NA_real_
	)
}

normal3d_character_y <- function(x) {
	switch(
		x,
		"yz-plane" = 0,
		"zy-plane" = 0,
		"xz-plane" = -1,
		"zx-plane" = 1,
		"xy-plane" = 0,
		"yx-plane" = 0,
		NA_real_
	)
}

normal3d_character_z <- function(x) {
	switch(
		x,
		"yz-plane" = 0,
		"zy-plane" = 0,
		"xz-plane" = 0,
		"zx-plane" = 0,
		"xy-plane" = 1,
		"yx-plane" = -1,
		NA_real_
	)
}

#' @rdname normal2d
#' @export
normal2d.Line2D <- function(x, ..., normalize = TRUE) {
	chkDots(...)
	n <- as_coord2d(x$a, x$b)
	n$scale(1 / abs(n))
	n
}

#' @rdname normal3d
#' @export
normal3d.Plane3D <- function(x, ..., normalize = TRUE) {
	chkDots(...)
	n <- as_coord3d(x$a, x$b, x$c)
	n$scale(1 / abs(n))
	n
}

#' Compute Euclidean norm
#'
#' `abs()` computes the Euclidean norm for [Coord2D] class objects and [Coord3D] class objects.
#' @param x A [Coord2D] class object or [Coord2D] class object.
#' @examples
#'   z <- complex(real = 1:4, imaginary = 1:4)
#'   p <- as_coord2d(z)
#'   abs(p) # Euclidean norm
#'   # Less efficient ways to calculate same Euclidean norms
#'   sqrt(p * p) # `*` dot product
#'   distance2d(p, as_coord2d(0, 0, 0))
#'
#'   # In {base} R `abs()` calculates Euclidean norm of complex numbers
#'   all.equal(abs(p), abs(z))
#'   all.equal(Mod(p), Mod(z))
#'
#'   p3 <- as_coord3d(x = 1:4, y = 1:4, z = 1:4)
#'   abs(p3)
#' @return A numeric vector
#' @rdname abs.Coord
#' @export
abs.Coord1D <- function(x) {
	abs(x$x)
}


#' @rdname abs.Coord
#' @export
abs.Coord2D <- function(x) {
	sqrt(rowSums(x$xyw[, 1:2, drop = FALSE]^2))
}

#' @rdname abs.Coord
#' @export
abs.Coord3D <- function(x) {
	sqrt(rowSums(x$xyzw[, 1:3, drop = FALSE]^2))
}

#' 1D Euclidean distances
#'
#' `distance1d()` computes 1D Euclidean distances.
#'
#' @param x Either a [Coord1D] or [Point1D] class object
#' @param y Either a [Coord1D] or [Point1D] class object
#' @examples
#' p <- as_coord1d(x = 1:4)
#' distance1d(p, as_coord1d(0))
#' @export
distance1d <- function(x, y) {
	if (is_coord1d(x)) {
		distance1d_coord1d(x, y)
	} else if (is_point1d(x)) {
		distance1d_point1d(x, y)
	} else {
		stop(paste("Don't know how to handle", dQuote(class(x)), "class object"))
	}
}

distance1d_coord1d <- function(x, y) {
	if (is_coord1d(y)) {
		distance1d_coord1d_coord1d(x, y)
	} else if (is_point1d(y)) {
		distance1d_coord1d_point1d(x, y)
	} else {
		stop(paste("Don't know how to handle", dQuote(class(y)), "class object"))
	}
}

distance1d_point1d <- function(x, y) {
	if (is_coord1d(y)) {
		distance1d_coord1d_point1d(y, x)
	} else {
		stop(paste("Don't know how to handle", dQuote(class(y)), "class object"))
	}
}

distance1d_coord1d_coord1d <- function(x, y) {
	n <- max(length(x), length(y))
	x <- rep_len(x, n)
	y <- rep_len(y, n)
	abs(x$x - y$x)
}

distance1d_coord1d_point1d <- function(x, y) {
	distance1d_coord1d_coord1d(x, as_coord1d(y))
}

#' 2D Euclidean distances
#'
#' `distance2d()` computes 2D Euclidean distances.
#'
#' @param x Either a [Coord2D] or [Line2D] class object
#' @param y Either a [Coord2D] or [Line2D] class object
#' @examples
#' p <- as_coord2d(x = 1:4, y = 1:4)
#' distance2d(p, as_coord2d(0, 0))
#' @export
distance2d <- function(x, y) {
	if (is_coord2d(x)) {
		distance2d_coord2d(x, y)
	} else if (is_line2d(x)) {
		distance2d_line2d(x, y)
	} else {
		stop(paste("Don't know how to handle", dQuote(class(x)), "class object"))
	}
}

distance2d_coord2d <- function(x, y) {
	if (is_coord2d(y)) {
		distance2d_coord2d_coord2d(x, y)
	} else if (is_line2d(y)) {
		distance2d_coord2d_line2d(x, y)
	} else {
		stop(paste("Don't know how to handle", dQuote(class(y)), "class object"))
	}
}

distance2d_line2d <- function(x, y) {
	if (is_coord2d(y)) {
		distance2d_coord2d_line2d(y, x)
	} else {
		stop(paste("Don't know how to handle", dQuote(class(y)), "class object"))
	}
}

distance2d_coord2d_coord2d <- function(x, y) {
	n <- max(length(x), length(y))
	x <- rep_len(x, n)
	y <- rep_len(y, n)
	sqrt(rowSums((x$xyw[, 1:2, drop = FALSE] - y$xyw[, 1:2, drop = FALSE])^2))
}

distance2d_coord2d_line2d <- function(x, y) {
	abs(y$a * x$x + y$b * x$y + y$c) / sqrt(y$a^2 + y$b^2)
}

#' 3D Euclidean distances
#'
#' `distance3d()` computes 3D Euclidean distances.
#'
#' @param x Either a [Coord3D] or [Plane3D] class object
#' @param y Either a [Coord3D] or [Plane3D] class object
#' @examples
#' p <- as_coord3d(x = 1:4, y = 1:4, z = 1:4)
#' distance3d(p, as_coord3d("origin"))
#' @export
distance3d <- function(x, y) {
	if (is_coord3d(x)) {
		distance3d_coord3d(x, y)
	} else if (is_plane3d(x)) {
		distance3d_plane3d(x, y)
	} else {
		stop(paste("Don't know how to handle", dQuote(class(x)), "class object"))
	}
}

distance3d_coord3d <- function(x, y) {
	if (is_coord3d(y)) {
		distance3d_coord3d_coord3d(x, y)
	} else if (is_plane3d(y)) {
		distance3d_coord3d_plane3d(x, y)
	} else {
		stop(paste("Don't know how to handle", dQuote(class(y)), "class object"))
	}
}

distance3d_plane3d <- function(x, y) {
	if (is_coord3d(y)) {
		distance3d_coord3d_plane3d(y, x)
	} else {
		stop(paste("Don't know how to handle", dQuote(class(y)), "class object"))
	}
}

distance3d_coord3d_coord3d <- function(x, y) {
	n <- max(length(x), length(y))
	x <- rep_len(x, n)
	y <- rep_len(y, n)
	sqrt(rowSums((x$xyzw[, 1:3, drop = FALSE] - y$xyzw[, 1:3, drop = FALSE])^2))
}

distance3d_coord3d_plane3d <- function(x, y) {
	n <- max(length(x), length(y))
	x <- rep_len(x, n)
	y <- rep_len(y, n)
	abs(y$a * x$x + y$b * x$y + y$c * x$z + y$d) / sqrt(y$a^2 + y$b^2 + y$c^2)
}

#' 1D points R6 Class
#'
#' `Point1D` is an [R6::R6Class()] object representing one-dimensional points.
#'
#' @examples
#' p1 <- as_point1d(a = 1, b = 5)
#' @field a Numeric vector that parameterizes the point via the equation `a * x + b = 0`.
#' @field b Numeric vector that parameterizes the point via the equation `a * x + b = 0`.
#' @export
Point1D <- R6Class(
	"Point1D",
	public = list(
		#' @param a Numeric vector that parameterizes the line via the equation `a * x + b = 0`.
		#' @param b Numeric vector that parameterizes the line via the equation `a * x + b = 0`.
		initialize = function(a, b) {
			stopifnot(length(a) == length(b), !any(is_congruent(a, 0), na.rm = TRUE))
			self$a <- a
			self$b <- b
		},
		#' @param n Number of lines to print.  If `NULL` print all of them.
		#' @param ... Passed to [format.default()].
		print = function(n = NULL, ...) {
			if (is.null(n) || n > length(self$a)) {
				n <- length(self$a)
			}
			cat("<Point1D[", length(self$a), "]>\n", sep = "")
			if (n > 0) {
				print_mat(cbind(a = self$a, b = self$b), ...)
			}
			invisible(self)
		},
		a = NULL,
		b = NULL
	)
)

#' 2D lines R6 Class
#'
#' `Line2D` is an [R6::R6Class()] object representing two-dimensional lines.
#'
#' @examples
#' p1 <- as_coord2d(x = 5, y = 10)
#' p2 <- as_coord2d(x = 7, y = 12)
#' theta <- degrees(45)
#' as_line2d(theta, p1)
#' as_line2d(p1, p2)
#' @field a Numeric vector that parameterizes the line via the equation `a * x + b * y + c = 0`.
#' @field b Numeric vector that parameterizes the line via the equation `a * x + b * y + c = 0`.
#' @field c Numeric vector that parameterizes the line via the equation `a * x + b * y + c = 0`.
#' @export
Line2D <- R6Class(
	"Line2D",
	public = list(
		#' @param a Numeric vector that parameterizes the line via the equation `a * x + b * y + c = 0`.
		#' @param b Numeric vector that parameterizes the line via the equation `a * x + b * y + c = 0`.
		#' @param c Numeric vector that parameterizes the line via the equation `a * x + b * y + c = 0`.
		initialize = function(a, b, c) {
			stopifnot(length(a) == length(b) && length(b) == length(c))
			self$a <- a
			self$b <- b
			self$c <- c
		},
		#' @param n Number of lines to print.  If `NULL` print all of them.
		#' @param ... Passed to [format.default()].
		print = function(n = NULL, ...) {
			if (is.null(n) || n > length(self$a)) {
				n <- length(self$a)
			}
			cat("<Line2D[", length(self$a), "]>\n", sep = "")
			if (n > 0) {
				print_mat(cbind(a = self$a, b = self$b, c = self$c), ...)
			}
			invisible(self)
		},
		a = NULL,
		b = NULL,
		c = NULL
	)
)

#' 3D planes R6 Class
#'
#' `Plane3D` is an [R6::R6Class()] object representing three-dimensional planes.
#'
#' @field a Numeric vector that parameterizes the plane via the equation `a * x + b * y + c * z + d = 0`.
#' @field b Numeric vector that parameterizes the plane via the equation `a * x + b * y + c * z + d = 0`.
#' @field c Numeric vector that parameterizes the plane via the equation `a * x + b * y + c * z + d = 0`.
#' @field d Numeric vector that parameterizes the plane via the equation `a * x + b * y + c * z + d = 0`.
#' @export
Plane3D <- R6Class(
	"Plane3D",
	public = list(
		#' @param a Numeric vector that parameterizes the plane via the equation `a * x + b * y + c * z + d = 0`.
		#' @param b Numeric vector that parameterizes the plane via the equation `a * x + b * y + c * z + d = 0`.
		#' @param c Numeric vector that parameterizes the plane via the equation `a * x + b * y + c * z + d = 0`.
		#' @param d Numeric vector that parameterizes the plane via the equation `a * x + b * y + c * z + d = 0`.
		initialize = function(a, b, c, d) {
			stopifnot(length(a) == length(b) && length(b) == length(c) && length(c) == length(d))
			self$a <- a
			self$b <- b
			self$c <- c
			self$d <- d
		},
		#' @param n Number of lines to print.  If `NULL` print all of them.
		#' @param ... Passed to [format.default()].
		print = function(n = NULL, ...) {
			if (is.null(n) || n > length(self$a)) {
				n <- length(self$a)
			}
			cat("<Plane3D[", length(self$a), "]>\n", sep = "")
			if (n > 0) {
				print(cbind(a = self$a, b = self$b, c = self$c, d = self$d), ...)
			}
			invisible(self)
		},
		a = NULL,
		b = NULL,
		c = NULL,
		d = NULL
	)
)

#' Cast to Point1D object
#'
#' `as_point1d()` casts to a [Point1D] object.
#'
#' @examples
#' p1 <- as_point1d(a = 1, b = 0)
#' @param ... Passed to other function such as `as_coord2d()`.
#' @export
as_point1d <- function(...) {
	UseMethod("as_point1d")
}

#' Cast to Line2D object
#'
#' `as_line2d()` casts to a [Line2D] object.
#'
#' @examples
#' p1 <- as_coord2d(x = 5, y = 10)
#' p2 <- as_coord2d(x = 7, y = 12)
#' theta <- degrees(45)
#' as_line2d(theta, p1)
#' as_line2d(p1, p2)
#' @param ... Passed to other function such as `as_coord2d()`.
#' @export
as_line2d <- function(...) {
	UseMethod("as_line2d")
}

#' Cast to Plane3D object
#'
#' `as_plane3d()` casts to a [Plane3D] object.
#'
#' @param ... Passed to other function such as `as_coord2d()`.
#' @export
as_plane3d <- function(...) {
	UseMethod("as_plane3d")
}

#' @rdname as_point1d
#' @param a,b Numeric vectors that parameterize the point via the equation `a * x + b = 0`.
#'            Note this means that `x = -b / a`.
#' @export
as_point1d.numeric <- function(a, b, ...) {
	n <- max(length(a), length(b))
	a <- rep_len(a, n)
	b <- rep_len(b, n)
	Point1D$new(a, b)
}

#' @rdname as_line2d
#' @param a,b,c Numeric vectors that parameterize the line via the equation `a * x + b * y + c = 0`.
#'              Note if `y = m * x + b` then `m * x + 1 * y + -b = 0`.
#' @export
as_line2d.numeric <- function(a, b, c, ...) {
	n <- max(length(a), length(b), length(c))
	a <- rep_len(a, n)
	b <- rep_len(b, n)
	c <- rep_len(c, n)
	Line2D$new(a, b, c)
}

#' @rdname as_plane3d
#' @param a,b,c,d Numeric vectors that parameterize the plane via the equation `a * x + b * y + c * z + d = 0`.
#' @export
as_plane3d.numeric <- function(a, b, c, d, ...) {
	n <- max(length(a), length(b), length(c), length(d))
	a <- rep_len(a, n)
	b <- rep_len(b, n)
	c <- rep_len(c, n)
	d <- rep_len(d, n)
	Plane3D$new(a, b, c, d)
}

#' @rdname as_line2d
#' @param theta Angle of the line represented by an [angle()] vector.
#' @param p1 Point on the line represented by a [Coord2D] class object.
#' @export
as_line2d.angle <- function(theta, p1 = as_coord2d("origin"), ...) {
	if (!is_coord2d(p1)) {
		p1 <- as_coord2d(p1, ...)
	}
	# a * x + b * y + c = 0
	# sin(theta) * x + -cos(theta) * y + c = 0
	a <- sin(theta)
	b <- -cos(theta)
	c <- -a * p1$x + -b * p1$y
	Line2D$new(a, b, c)
}

#' @rdname as_point1d
#' @param x A (character) vector to be cast to a [Point1D] object
#' @export
as_point1d.character <- function(x, ...) {
	a <- as_point1d_character_a(x)
	p <- as_point1d(a, 0)
	if (any(is.na(p) & !is.na(x))) {
		warning("NAs introduced by coercion")
	}
	p
}

as_point1d_character_a <- function(x) {
	switch(x, "origin" = 1, "zero" = 1, NA_real_)
}

#' @rdname as_line2d
#' @param x A (character) vector to be cast to a [Line2D] object
#' @export
as_line2d.character <- function(x, ...) {
	a <- as_line2d_character_a(x)
	b <- as_line2d_character_b(x)
	l <- as_line2d(a, b, 0)
	if (any(is.na(l) & !is.na(x))) {
		warning("NAs introduced by coercion")
	}
	l
}

as_line2d_character_a <- function(x) {
	switch(x, "x-axis" = 0, "y-axis" = 1, NA_real_)
}

as_line2d_character_b <- function(x) {
	switch(x, "x-axis" = -1, "y-axis" = 0, NA_real_)
}

#' @rdname as_plane3d
#' @param x A (character) vector to be cast to a [Plane3D] object
#' @export
as_plane3d.character <- function(x, ...) {
	a <- normal3d_character_x(x)
	b <- normal3d_character_y(x)
	c <- normal3d_character_z(x)
	p <- as_plane3d(a, b, c, 0)
	if (any(is.na(p) & !is.na(x))) {
		warning("NAs introduced by coercion")
	}
	p
}

#' @rdname as_point1d
#' @param normal [Coord1D] class object.
#' @export
as_point1d.Coord1D <- function(normal, ...) {
	as_point1d(a = 1, b = -normal$x)
}

#' @rdname as_line2d
#' @param normal Normal vector to the line represented by a [Coord2D] class object. `p2` should be missing.
#' @param p2 Another point on the line represented by a [Coord2D] class object.
#' @export
as_line2d.Coord2D <- function(normal, p1 = as_coord3d("origin"), p2, ...) {
	if (!is_coord2d(p1)) {
		p1 <- as_coord2d(p1, ...)
	}
	if (!missing(normal)) {
		stopifnot(missing(p2))
		n <- max(length(p1), length(normal))
		p1 <- rep_len(p1, n)
		normal <- rep_len(normal, n)
	} else {
		if (!is_coord2d(p2)) {
			p2 <- as_coord2d(p2, ...)
		}
		stopifnot(all(p1 != p2))
		normal <- normal2d(p2 - p1)
	}
	Line2D$new(normal$x, normal$y, -(normal * p1))
}

#' @rdname as_plane3d
#' @param p1 Point on the plane represented by a [Coord3D] class object.
#' @param p2,p3 Points on the plane represented by [Coord3D] class objects.  `normal` should be missing.
#' @param normal Normal vector to the plane represented by a [Coord3D] class object. `p2` and `p3` should be missing.
#' @export
as_plane3d.Coord3D <- function(normal, p1 = as_coord3d("origin"), p2, p3, ...) {
	if (!is_coord3d(p1)) {
		p1 <- as_coord3d(p1, ...)
	}
	if (!missing(normal)) {
		stopifnot(missing(p2), missing(p3))
		n <- max(length(p1), length(normal))
		p1 <- rep_len(p1, n)
		normal <- rep_len(normal, n)
	} else {
		if (!is_coord3d(p2)) {
			p2 <- as_coord3d(p2, ...)
		}
		if (!is_coord3d(p3)) {
			p3 <- as_coord3d(p3, ...)
		}
		v1 <- p2 - p1
		v2 <- p3 - p1
		normal <- normal3d(v1, v2)
	}
	Plane3D$new(normal$x, normal$y, normal$z, -(normal * p1))
}

#' @rdname as_point1d
#' @param point A [Point1D] object
#' @export
as_point1d.Point1D <- function(point, ...) {
	point
}

#' @rdname as_line2d
#' @param line A [Line2D] object
#' @export
as_line2d.Line2D <- function(line, ...) {
	line
}

#' @rdname as_line2d
#' @param point A [Point1D] object
#' @export
as_line2d.Point1D <- function(point, b = 0, ...) {
	as_line2d(point$a, b, point$b)
}

#' @rdname as_plane3d
#' @param plane A [Plane3D] object
#' @export
as_plane3d.Plane3D <- function(plane, ...) {
	plane
}

#' @rdname as_plane3d
#' @param point A [Point1D] object
#' @export
as_plane3d.Point1D <- function(point, b = 0, c = 0, ...) {
	as_plane3d(point$a, b, c, point$b)
}

#' @rdname as_plane3d
#' @param line A [Line2D] object
#' @export
as_plane3d.Line2D <- function(line, c = 0, ...) {
	as_plane3d(line$a, line$b, c, line$c)
}

test_that("as_line2d()", {
	skip_if_not_installed("withr")
	withr::local_options(affiner_options(default = TRUE))

	x <- c(2, 5, 7)
	y <- c(3, 4, 6)
	p1 <- as_coord2d(x = x, y = y)
	p0 <- as_coord2d("origin")

	l1 <- as_line2d(p1 = p0, p2 = as.data.frame(p1))
	l2 <- as_line2d(as_angle(p1), "origin")

	expect_equal(normal2d(as_line2d(1, 1, 0)), as_coord2d(0.5 * sqrt(2), 0.5 * sqrt(2)))

	expect_equal(l1, l2)
	expect_equal(l1, l1[1:3])
	expect_equal(l1, as_line2d(l1))
	expect_snapshot(print(l1))

	expect_equal(length(rep_len(l1, 3)), 3L)
	expect_equal(length(rep(l1, 2)), 6L)
	expect_equal(length(c(l1, l2)), 6L)
	expect_equal(is.na(l1), rep(FALSE, 3L))
	expect_equal(is.nan(l1), rep(FALSE, 3L))
	expect_equal(is.infinite(l1), rep(FALSE, 3L))
	expect_equal(is.finite(l1), rep(TRUE, 3L))
	expect_true(is.data.frame(as.data.frame(l1)))
	expect_true(is.list(as.list(l1)))

	lxa <- as_line2d("x-axis")
	expect_equal(lxa, as_line2d(0, -1, 0))

	lya <- as_line2d("y-axis")
	expect_equal(lya, as_line2d(1, 0, 0))

	expect_true(is_equivalent(lxa, "x-axis"))
	expect_false(is_equivalent(lxa, "y-axis"))

	expect_warning(as_line2d("boo"))

	expect_true(is_equivalent(as_angle(lxa), degrees(0)))
	expect_true(is_congruent(as_angle(lya), degrees(90)))

	expect_true(is_congruent(
		2 * as_angle(as_line2d(0, -1, 1), "degrees"),
		2 * as_angle(0, "degrees")
	))

	expect_true(is_congruent(
		2 * as_angle(as_line2d(1, 0, 1), "degrees"),
		2 * as_angle(90, "degrees")
	))

	expect_true(is_congruent(
		2 * as_angle(as_line2d(0, 1, 1), "degrees"),
		2 * as_angle(180, "degrees")
	))

	expect_true(is_congruent(
		2 * as_angle(as_line2d(-1, 0, 1), "degrees"),
		2 * as_angle(270, "degrees")
	))

	expect_true(is_congruent(
		2 * as_angle(as_line2d(-1, 1, 1), "degrees"),
		2 * as_angle(45, "degrees")
	))

	expect_true(is_congruent(
		2 * as_angle(as_line2d(1, 1, 1), "degrees"),
		2 * as_angle(-45, "degrees")
	))

	pt <- as_point1d(a = 1, b = -1)
	expect_equal(as_line2d(pt, b = 3), as_line2d(a = 1, b = 3, c = -1))

	# Conversions to and from normals
	expect_equal(as_line2d(as_coord2d(1, 0)), as_line2d(a = 1, b = 0, c = 0))

	expect_equal(normal2d(as_line2d(a = 1, b = 1, c = 0)), as_coord2d(sqrt(2) / 2, sqrt(2) / 2))
})

test_that("`is_parallel()`", {
	line1 <- as_line2d("x-axis")
	line2 <- as_line2d("y-axis")
	line3 <- as_line2d(a = 0, b = 1, c = 2) # y + 2 = 0
	expect_true(is_parallel(line1, line1))
	expect_false(is_parallel(line1, line2))
	expect_false(is_parallel(line1, "y-axis"))
	expect_true(is_parallel(line1, line3))
	expect_false(is_equivalent(line1, line3))

	plane1 <- as_plane3d("xy-plane")
	plane2 <- as_plane3d("xz-plane")
	plane3 <- as_plane3d(a = 0, b = 0, c = 1, d = 2) # z + 2 = 0
	expect_true(is_parallel(plane1, plane1))
	expect_false(is_parallel(plane1, plane2))
	expect_false(is_parallel(plane1, "xz-plane"))
	expect_true(is_parallel(plane1, plane3))
	expect_false(is_equivalent(plane1, plane3))
})

test_that("as_plane3d()", {
	p1 <- as_plane3d(1, 2, 3, 4)
	expect_true(is_plane3d(p1))

	expect_equal(as_plane3d("xy-plane"), as_plane3d(0, 0, 1, 0))
	expect_true(is_equivalent(as_plane3d(0, 0, 1, 0), "xy-plane"))

	expect_equal(p1, as_plane3d(p1))
	expect_equal(p1, p1[1L])
	expect_warning(as_plane3d("boo"))

	expect_equal(as_plane3d("xy-plane"), as_plane3d(normal3d("xy-plane"), "origin"))

	expect_equal(as_plane3d(as_line2d(1, 0, -2)), as_plane3d(1, 0, 0, -2))

	expect_false(is.na(p1))
	expect_false(is.nan(p1))
	expect_false(is.infinite(p1))
	expect_true(is.finite(p1))
	expect_length(p1, 1L)
	expect_length(rep(p1, 3), 3L)
	expect_length(c(p1, p1), 2L)
	expect_true(is.data.frame(as.data.frame(p1)))
	expect_true(is.list(as.list(p1)))

	expect_snapshot(print(p1))

	expect_equal(
		as_plane3d(p1 = as_coord3d("origin"), p2 = "x-axis", p3 = "y-axis"),
		as_plane3d("xy-plane")
	)

	pt <- as_point1d(a = 1, b = -1)
	expect_equal(as_plane3d(pt, b = 3, c = 4), as_plane3d(a = 1, b = 3, c = 4, d = -1))
})

---
title: "Geometry in affiner"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Geometry in affiner}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

### Table of Contents

* [Angle objects](#angles)
* [Trigonometry](#trigonometry)
* [2D Coordinates](#2d)
* [3D Coordinates](#3d)
* [Orthographic/Axonometric and Oblique Projections](#projection)

## <a name="angles">Angle objects</a>

In `{affiner}` angles are represented by the `angle()` class:

* Supports the following [angular units](https://en.wikipedia.org/wiki/Angle#Units) (note we ignore any punctuation and space characters as well as any trailing s's e.g. "half turns" will be treated as equivalent to "halfturn"):

  + "deg" or "degree"
  + "half-revolution", "half-turn", or "pi-radian"
  + "gon", "grad", "grade", or "gradian"
  + "rad" or "radian"
  + "rev", "revolution", "tr", or "turn"

* `degrees()`, `gradians()`, `pi_radians()`, `radians()`, `turns()` are convenience wrappers around `as_angle.()` that specifies the angular unit.

* One can use the `affiner_angular_unit` global option to set the default angular unit used by this package from "degrees" to "gradians", (multiples of) "pi-radians", "radians", or "turns".

* Use `is_congruent()` to check if two angles are congruent modulo full turns.

```{r angles}
library("affiner")
as_angle(90, "degrees") + turns(1)
is_congruent(degrees(180), radians(pi))
as.numeric(turns(1/3), "radians")
```

## <a name="trigonometry">Trigonometry</a>

`{affiner}` provides several `angle()` class aware trigonometric functions:

* `sine()`, `cosine()`, `tangent()`, `secant()`, `cosecant()`, `cotangent()`,
* `arcsine()`, `arccosine()`, `arctangent()`, `arcsecant()`, `arccosecant()`, and `arccotangent()`.
    `arcsine()` and `arccosine()` also feature a `tolerance` value so that values that exceed the `1` / `-1` cutoffs by a small tolerance are rounded to those values.
* One can also use the base S3 methods `sin()`, `cos()`, and `tan()` on `angle()` objects.

```{r trig}
library("affiner")
sin(2 * pi)
sine(degrees(360))
arctangent(x = 0, y = 1)
```

## <a name="2d">2D Coordinates</a>

In `{affiner}` 2D Coordinates are represented by a `Coord2D` R6 class:

* Create `Coord2D` objects with `as_coord2d()`
* `Coord2D` R6 objects supports several affine transformation methods that can be chained:

  + `permute()`
  + `project()`
  + `reflect()`
  + `rotate()`
  + `scale()`
  + `shear()`
  + `translate()`
  + `transform()`
  + R6 method chained affine transformation matrices are auto-multiplied
    so you don't need to manually multiply them for efficiency reasons.
  + `{affiner}` affine transformations are post-multiplied so affine transformations
    can be applied in an intuitive order.
  +  `abs()` computes Euclidean norm and `distance2d()` computes Euclidean distances
  + `convex_hull2d()` computes the convex hull.
  * `dot_product2d()` computes dot (inner) products.  You may also use the `*` operator and if R >= 4.3.0 you may also use the `%*%` operator to compute dot (inner) products.
  * `range()` computes the axis-aligned bounding box ranges.

```{r 2d}
# Cartesian coordinates
library("affiner")
p <- as_coord2d(x = 1:10, y = 1:10)
print(p)

p2 <- p$
    clone()$
    scale(x = 0.5)$
    rotate(degrees(90))$
    reflect(as_line2d("y-axis"))$
    translate(as_coord2d(x = 0.5, y = 0.5))$
    print()

# Polar coordinates
theta <- degrees(seq(0, 300, by = 60))
radius <- 1
p <- as_coord2d(theta, radius = radius)
is_congruent(as_angle(p), theta) |> all()
is_congruent(abs(p), radius) |> all()
```

## <a name="3d">3D Coordinates</a>

In `{affiner}` 3D Coordinates are represented by a `Coord3D` R6 class:

* Create `Coord3D` objects with `as_coord3d()`
* `Coord3D` R6 objects supports several affine transformation methods that can be chained:

  + `permute()`
  + `project()`
  + `reflect()`
  + `rotate()`
  + `scale()`
  + `shear()`
  + `translate()`
  + `transform()`
  + R6 method chained affine transformation matrices are auto-multiplied
    so you don't need to manually pre-multiply them for efficiency reasons.
  + `{affiner}` affine transformations are post-multiplied so affine transformations
    can be applied in an intuitive order.
  +  `abs()` computes Euclidean norm and `distance3d()` computes Euclidean distances
  * `range()` computes the axis-aligned bounding box ranges.
  * `cross_product3d()` computes cross products.  If R >= 4.4.0 you may also use `crossprod()`.
  * `dot_product3d()` computes dot (inner) products.  You may also use the `*` operator and if R >= 4.3.0 you may also use the `%*%` operator to compute dot (inner) products.

```{r 3d}
# Cartesian coordinates
library("affiner")
p <- as_coord3d(x = 1:10, y = 1:10, z = 1:10)
print(p)

p2 <- p$
    clone()$
    scale(z = 0.5)$
    rotate(axis = as_coord3d("z-axis"), theta = degrees(90))$
    reflect(as_plane3d("yz-plane"))$
    shear(xy_shear = 0.5)$
    translate(as_coord3d(x = 0.5, y = 0.5, z = 0.5))$
    print()

# Spherical coordinates
inclination <- as_angle(p, type = "inclination")
azimuth <- as_angle(p, type = "azimuth")
radius <- abs(p)
ps <- as_coord3d(azimuth, radius = radius, inclination = inclination)
all.equal(p, ps)

# Cylindrical coordinates
radius <- as_coord2d(p, plane = "xy-plane") |> abs()
pc <- as_coord3d(azimuth, radius = radius, z = p$z)
all.equal(p, pc)
```

## <a name="projection">Orthographic/Axonometric and Oblique Projections</a>

`{affiner}` can project `Coord3D` objects to `Coord2D` objects using orthographic/axonometric and oblique projections:

* For a multiview/primary orthographic projection onto the xy-plane use `as_coord2d(x)`
* For a multiview/primary orthographic projection onto the xz-plane use `as_coord2d(x, permutation = "xzy")`
* For a "cabinet" oblique projection onto the xy-plane use `as_coord2d(x, scale = 0.5)`
* For a "cabinet" oblique projection onto the xz-plane use `as_coord2d(x, permutation = "xzy", scale = 0.5)`
* For other oblique projections manipulate the `scale` parameter (usually from 0.5 to 1.0) and the `alpha` angle parameter
  (usually from 30° to 45°).
* For one "isometric" axonometric projection one can use

  ```
  x$
    clone()$
    translate(-mean(x)$
    rotate("z-axis", degrees(45))$
    rotate("x-axis", degrees(-90 + 35.264)) |>
    as_coord2d()
  ```

* Other axonometric projections can be achieved with the right 3D rotations
* See [`vignette("affiner", package = "affiner")`](affiner.html) for some visual examples
* Recall one can use "scale" affine transformation to flip signs of x/y/z axes and "permute" affine transformation to switch order of x/y/z coordinates

---
title: "Introduction to affiner"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Introduction to affiner}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

### Table of Contents

* [Overview](#overview)
* [Examples](#examples)

  + [Isometric-cube hex logo](#logo)
  + [Render an "illustrated" d6 dice using oblique and isometric projections](#dice)

## <a name="overview">Overview</a>

`{affiner}` is an extraction and improvement of the low-level geometric and R 4.2 affine transformation feature functionality used in [{piecepackr}](https://github.com/piecepackr/piecepackr/) to render board game pieces in `{grid}` using a 3D oblique projection.

<img src="https://github.com/piecepackr/piecepackr/raw/master/man/figures/README-polyhedral-1.png" alt="Polyhedral dice">

The current goals are to:

1. Make it easier for `{piecepackr}` users to use this low-level geometric functionality without exporting it in `{piecepackr}` (which already has a large API) in case they want to do things like implement custom polyhedral dice.
2. Make it easier for other R developers to support fancier 3D renderings in their packages using `{grid}` (and perhaps `{ggplot2}`).
3. Refactor and improve this code base.

Some particular intended strengths compared to other R geometry packages:

1. Focus on reducing pain points in rendering "illustrated" (game) pieces in a 3D parallel projections in `{grid}` (e.g. oblique projections and isometric projections).
2. Helpers for using the R 4.2 affine transformation feature to render "illustrated" 3D faces in `{grid}`.
   The `affine_settings()` function which reverse engineers `useGrob()`'s `vp` and `transformation` arguments
   is even available as a "standalone" file that can be copied
   over into other R packages under the permissive [Unlicense](https://unlicense.org/).
3. Functions to convert between [axis-angle representation](https://en.wikipedia.org/wiki/Axis%E2%80%93angle_representation) and [rotation matrix](https://en.wikipedia.org/wiki/Rotation_matrix).
4. Allows users to use whichever angular unit is most convenient for them including degrees, radians, turns, half-turns aka (multiples-of-)pi-radians, and gradians.
5. Light dependencies: only non-base R package dependency is [{R6}](https://github.com/r-lib/R6/) which is a pure R package with no other dependencies.

## <a name="examples">Examples</a>

```{r hidden, echo = FALSE}
knitr::opts_chunk$set(fig.cap = "", dev = "ragg_png")
```

### <a name="logo">Isometric-cube hex logo</a>

`isocubeGrob()` / `grid.isocube()` provides a convenience wrapper around `affineGrob()` for the isometric cube case:

```{r hex-logo2, fig.width = 4.5, fig.height = 4.5, fig.alt = "Isometric-cube hex logo", eval = getRversion() >= "4.2.0" && requireNamespace("ggplot2", quietly=TRUE) && requireNamespace("aRtsy", quietly=TRUE) && requireNamespace("gtable", quietly=TRUE), message = FALSE}
library("aRtsy")
library("ggplot2")

gg <- canvas_planet(colorPalette("lava"), threshold = 3) +
  scale_x_continuous(expand=c(0, 0)) +
  scale_y_continuous(expand=c(0, 0))
grob <- ggplotGrob(gg)
grob <- gtable::gtable_filter(grob, "panel") # grab just the panel
affiner::grid.isocube(top = grob, left = grob, right = grob,
  gp_border = grid::gpar(col = "darkorange", lwd = 12))
```

* Here's an example using `affine_settings()` and `grid.affine()` directly to make an isometric cube logo.
* The three visible faces of an isometrically projected cube are parallelograms.  Additionally each projected vertex will match one of the six vertices of a regular hexagon or its center.
* The `as_coord2d()` method for `angle()` objects lets you compute the regular polygon vertices and center using [polar coordinates](https://en.wikipedia.org/wiki/Polar_coordinate_system)
* Use `affine_settings()` and `affineGrob()`  or `grid.affine()` to render arbitrary "illustrated" grobs within each of these parallelograms.

```{r hex-logo, fig.width = 4.5, fig.height = 4.5, fig.alt = "Isometric-cube hex logo", eval = getRversion() >= "4.2.0"}
library("affiner")
library("grid")

xy <- as_coord2d(angle(seq(90, 360 + 90, by = 60), "degrees"),
                 radius = c(rep(0.488, 6), 0))
xy$translate(x = 0.5, y = 0.5)
l_xy <- list()
l_xy$top <- xy[c(1, 2, 7, 6)]
l_xy$right <- xy[c(7, 4, 5, 6)]
l_xy$left <- xy[c(2, 3, 4, 7)]

gp_border <- gpar(fill = NA, col = "black", lwd = 12)
vp_define <- viewport(width = unit(3, "inches"), height = unit(3, "inches"))

colors <- c("#D55E00", "#56B4E9", "#009E73")
spacings <- c(0.25, 0.25, 0.2)
texts <- c("pkgname", "right\nface", "left\nface")
rots <- c(45, 0, 0)
fontsizes <- c(52, 80, 80)
sides <- c("top", "right", "left")
types <- gridpattern::names_polygon_tiling[c(5, 9, 7)]
l_grobs <- list()
grid.newpage()
for (i in 1:3) {
    side <- sides[i]
    xy_side <- l_xy[[side]]
    if (requireNamespace("gridpattern", quietly = TRUE)) {
        bg <- gridpattern::grid.pattern_polygon_tiling(
                   colour = "grey80",
                   fill = c(colors[i], "white"),
                   type = types[i],
                   spacing = spacings[i],
                   draw = FALSE)
    } else {
        bg <- rectGrob(gp = gpar(col = NA, fill = colors[i]))
    }
    text <- textGrob(texts[i], rot = rots[i],
                     gp = gpar(fontsize = fontsizes[i]))
    settings <- affine_settings(xy_side, unit = "snpc")
    grob <- l_grobs[[side]] <- grobTree(bg, text)
    grid.affine(grob,
                vp_define = vp_define,
                transform = settings$transform,
                vp_use = settings$vp)
    grid.polygon(xy_side$x, xy_side$y, gp = gp_border)
}
```

### <a name="dice">Render an "illustrated" d6 dice using oblique and isometric projections</a>

Our high-level strategy for rendering 3D objects is as follows:

* Figure out the "physical" 3D coordinates of the cube face vertices in "inches".  These cube faces will correspond to target "3D viewports" we'll want to render the illustrated cube face grobs into.

  + For my piecepack diagrams I make with [{piecepackr}](https://github.com/piecepackr/piecepackr/) I figuratively think as my graphics device as a piece of paper (bottom left corner is the origin)  and calculate the 3D coordinates in inches as if my board game pieces were physically sitting on top of the graphics device (since piecepack tiles are 2 inches by 2 inches this is usually a straightforward calculation).
  
* Project these 3D coordinates onto a "physical" xy-plane (corresponding to our graphics device) in "inches" using a parallel projection (in this example we'll do a couple oblique projections and an isometric projection).  Note since all parallel projections are affine transformations we know the projected vertices of a square "3D viewport" will project to the 2D coordinates of a "parallelogram viewport".
* (If they don't already do so) translate these parallelograms so they lie within the graphics device view (i.e. the "parallelogram viewport" vertices are all in the upper right quadrant of the xy-plane).

  + If you use an oblique projection to the xy-plane with an `alpha` angle between 0 and 90 degrees then any flat faces lying directly on the xy-plane will stay where they are and any flat faces on a parallel higher plane will only be shifted up/right.  So in this case one usually doesn't need to such a translation assuming all your "objects" were placed in the upper quadrant of the xy-plane to begin with.

* Use `affine_settings()` and `grid.affine()` / `affineGrob()` to render the illustrated cube face "grobs" within these affine transformed "parallelogram viewports".  The order these are drawn is important but in this example we manually sorted them ahead of time in an order that worked for our target projections.

```{r die-faces, fig.width = 4.0, fig.height = 3.0, fig.alt = "The six die faces", eval = getRversion() >= "4.2.0"}
library("affiner")
library("grid")
xyz_face <- as_coord3d(x = c(0, 0, 1, 1) - 0.5, y = c(1, 0, 0, 1) - 0.5, z = 0.5)
l_faces <- list() # order faces for our target projections
l_faces$bottom <- xyz_face$clone()$
                    rotate("z-axis", angle(180, "degrees"))$
                    rotate("y-axis", angle(180, "degrees"))
l_faces$north <- xyz_face$clone()$
                    rotate("z-axis", angle(90, "degrees"))$
                    rotate("x-axis", angle(-90, "degrees"))
l_faces$east <- xyz_face$clone()$
                    rotate("z-axis", angle(90, "degrees"))$
                    rotate("y-axis", angle(90, "degrees"))
l_faces$west <- xyz_face$clone()$
                    rotate("y-axis", angle(-90, "degrees"))
l_faces$south <- xyz_face$clone()$
                    rotate("z-axis", angle(180, "degrees"))$
                    rotate("x-axis", angle(90, "degrees"))
l_faces$top <- xyz_face$clone()$
                    rotate("z-axis", angle(-90, "degrees"))

colors <- c("#D55E00", "#009E73", "#56B4E9", "#E69F00", "#CC79A7", "#0072B2")
spacings <- c(0.25, 0.2, 0.25, 0.25, 0.25, 0.25)
die_face_grob <- function(digit) {
    if (requireNamespace("gridpattern", quietly = TRUE)) {
        bg <- gridpattern::grid.pattern_polygon_tiling(
                   colour = "grey80",
                   fill = c(colors[digit], "white"),
                   type = gridpattern::names_polygon_tiling[digit],
                   spacing = spacings[digit],
                   draw = FALSE)
    } else {
        bg <- rectGrob(gp = gpar(col = NA, fill = colors[digit]))
    }
    digit <- textGrob(digit, gp = gpar(fontsize = 72))
    grobTree(bg, digit)
}
l_face_grobs <- lapply(1:6, function(i) die_face_grob(i))
grid.newpage()
for (i in 1:6) {
    vp <- viewport(x = unit((i - 1) %% 3 + 1, "inches"),
                   y = unit(3 - ((i - 1) %/% 3 + 1), "inches"),
                   width = unit(1, "inches"), height = unit(1, "inches"))
    pushViewport(vp)
    grid.draw(l_face_grobs[[i]])
    popViewport()
    grid.text("The six die faces", y = 0.9, 
              gp = gpar(fontsize = 18, face = "bold"))
}
```
```{r projected-die, fig.width = 3.0, fig.height = 3.0, fig.alt = "Parallel projection of a die", eval = getRversion() >= "4.2.0"}
# re-order face grobs for our target projections
# bottom = 6, north = 4, east = 5, west = 2, south = 3, top = 1
l_face_grobs <- l_face_grobs[c(6, 4, 5, 2, 3, 1)]
draw_die <- function(l_xy, l_face_grobs) {
    min_x <- min(vapply(l_xy, function(x) min(x$x), numeric(1)))
    min_y <- min(vapply(l_xy, function(x) min(x$y), numeric(1)))
    l_xy <- lapply(l_xy, function(xy) {
        xy$translate(x = -min_x + 0.5, y = -min_y + 0.5)
    })
    grid.newpage()
    vp_define <- viewport(width = unit(1, "inches"), height = unit(1, "inches"))
    gp_border <- gpar(col = "black", lwd = 4, fill = NA)
    for (i in 1:6) {
        xy <- l_xy[[i]]
        settings <- affine_settings(xy, unit = "inches")
        grid.affine(l_face_grobs[[i]],
                    vp_define = vp_define,
                    transform = settings$transform,
                    vp_use = settings$vp)
        grid.polygon(xy$x, xy$y, default.units = "inches", gp = gp_border)
    }
}
# oblique projection of dice onto xy-plane
l_xy_oblique1 <- lapply(l_faces, function(xyz) {
    xyz$clone() |>
        as_coord2d(scale = 0.5)
})
draw_die(l_xy_oblique1, l_face_grobs)
grid.text("Oblique projection\n(onto xy-plane)", y = 0.9,
          gp = gpar(fontsize = 18, face = "bold"))

# oblique projection of dice on xz-plane
l_xy_oblique2 <- lapply(l_faces, function(xyz) {
    xyz$clone()$
        permute("xzy") |>
        as_coord2d(scale = 0.5, alpha = angle(135, "degrees"))
})
draw_die(l_xy_oblique2, l_face_grobs)
grid.text("Oblique projection\n(onto xz-plane)", y = 0.9,
          gp = gpar(fontsize = 18, face = "bold"))

# isometric projection
l_xy_isometric <- lapply(l_faces, function(xyz) {
    xyz$clone()$
        rotate("z-axis", angle(45, "degrees"))$
        rotate("x-axis", angle(-(90 - 35.264), "degrees")) |>
        as_coord2d()
})

draw_die(l_xy_isometric, l_face_grobs)
grid.text("Isometric projection", y = 0.9,
          gp = gpar(fontsize = 18, face = "bold"))
```

# print.Coord1D()

    Code
      print(p1)
    Output
      <Coord1D[3]>
           x w
      [1,] 2 1
      [2,] 5 1
      [3,] 7 1

---

    Code
      print(as_coord1d(numeric(0)))
    Output
      <Coord1D[0]>

# print.Coord2D()

    Code
      print(p1)
    Output
      <Coord2D[3]>
           x y w
      [1,] 2 3 1
      [2,] 5 4 1
      [3,] 7 6 1

---

    Code
      print(as_coord2d(numeric(0)))
    Output
      <Coord2D[0]>

# print.Coord3D()

    Code
      print(p1)
    Output
      <Coord3D[3]>
           x y z w
      [1,] 2 3 0 1
      [2,] 5 4 0 1
      [3,] 7 6 0 1

---

    Code
      print(as_coord3d(numeric(0)))
    Output
      <Coord3D[0]>

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/z-affine-settings-docs.R
\name{affine_settings}
\alias{affine_settings}
\title{Compute \code{grid} affine transformation feature viewports and transformation functions}
\usage{
affine_settings(
  xy = data.frame(x = c(0, 0, 1, 1), y = c(1, 0, 0, 1)),
  unit = getOption("affiner_grid_unit", "inches")
)
}
\arguments{
\item{xy}{An R object with named elements \code{x} and \code{y} representing the (x,y) coordinates
of the affine transformed "viewport" one wishes to draw in.
The (x,y) coordinates of the "viewport" should be in
"upper left", "lower left", "lower right", and "upper right" order
(this ordering should be from the perspective of \strong{before}
the "affine transformation" of the "viewport").}

\item{unit}{Which \code{\link[grid:unit]{grid::unit()}} to assume the \code{xy} "x" and "y" coordinates are expressed in.}
}
\value{
A named list with the following group affine transformation feature viewport
and functions settings:\describe{
\item{transform}{An affine transformation function to pass to \code{\link[=affineGrob]{affineGrob()}} or \code{\link[=useGrob]{useGrob()}}.
If \code{getRversion()} is less than \code{"4.2.0"} will instead be \code{NULL}.}
\item{vp}{A \code{\link[grid:viewport]{grid::viewport()}} object to pass to \code{\link[=affineGrob]{affineGrob()}} or \code{\link[=useGrob]{useGrob()}}.}
\item{sx}{x-axis sx factor}
\item{flipX}{whether the affine transformed "viewport" is "flipped" horizontally}
\item{x}{x-coordinate for viewport}
\item{y}{y-coordinate for viewport}
\item{width}{Width of viewport}
\item{height}{Height of viewport}
\item{default.units}{Default \code{\link[grid:unit]{grid::unit()}} for viewport}
\item{angle}{angle for viewport}
}
}
\description{
\code{affine_settings()} computes \code{grid} group affine transformation feature viewport and transformation
function settings given the (x,y) coordinates of the corners of the
affine transformed "viewport" one wishes to draw in.
}
\section{Usage in other packages}{


To avoid taking a dependency on \code{affiner} you may copy the source of \code{affine_settings()}
into your own package under the permissive Unlicense.  Either use
\code{usethis::use_standalone("trevorld/affiner", "standalone-affine-settings.r")} or
copy the file \code{standalone-affine-settings.r} into your \code{R} directory and add \code{grid}
to the \code{Imports} of your \code{DESCRIPTION} file.
}

\examples{
if (require("grid")) {
  grob <- grobTree(rectGrob(gp = gpar(fill = "blue", col = NA)),
                   circleGrob(gp=gpar(fill="yellow", col = NA)),
                   textGrob("RSTATS", gp=gpar(fontsize=32)))
  grid.newpage()
  pushViewport(viewport(width=unit(4, "in"), height=unit(2, "in")))
  grid.draw(grob)
  popViewport()
}
if (require("grid") &&
    getRversion() >= "4.2.0" &&
    isTRUE(dev.capabilities()$transformations)) {
  # Only works if active graphics device supports affine transformations
  # such as `png(type="cairo")` on R 4.2+
  vp_define <- viewport(width=unit(2, "in"), height=unit(2, "in"))
  settings <- affine_settings(xy = list(x = c(1/3, 0/3, 2/3, 3/3),
                                        y = c(2/3, 1/3, 1/3, 2/3)),
                              unit = "snpc")
  affine <- affineGrob(grob,
                       vp_define=vp_define,
                       transform = settings$transform,
                       vp_use = settings$vp)
  grid.newpage()
  grid.draw(affine)
}
if (require("grid") &&
    getRversion() >= "4.2.0" &&
    isTRUE(dev.capabilities()$transformations)) {
  # Only works if active graphics device supports affine transformations
  # such as `png(type="cairo")` on R 4.2+
  settings <- affine_settings(xy = list(x = c(3/3, 2/3, 0/3, 1/3),
                                        y = c(2/3, 1/3, 1/3, 2/3)),
                              unit = "snpc")
  affine <- affineGrob(grob,
                       vp_define=vp_define,
                       transform = settings$transform,
                       vp_use = settings$vp)
  grid.newpage()
  grid.draw(affine)
}
}
\seealso{
Intended for use with \code{\link[=affineGrob]{affineGrob()}} and \code{\link[grid:useGrob]{grid::useGrob()}}.
See \url{https://www.stat.auckland.ac.nz/~paul/Reports/GraphicsEngine/groups/groups.html}
for more information about the group affine transformation feature.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/affine-methods.R
\name{is_transform2d}
\alias{is_transform2d}
\title{Test if 2D affine transformation matrix}
\usage{
is_transform2d(x)
}
\arguments{
\item{x}{An object}
}
\value{
A logical value
}
\description{
\code{is_transform2d()} tests if object is a \code{\link[=transform2d]{transform2d()}} affine transformation matrix
}
\examples{
m <- transform2d(diag(3L))
is_transform2d(m)
is_transform2d(diag(3L))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/normal.R
\name{normal3d}
\alias{normal3d}
\alias{normal3d.Coord3D}
\alias{normal3d.character}
\alias{normal3d.Plane3D}
\title{3D normal vectors}
\usage{
normal3d(x, ...)

\method{normal3d}{Coord3D}(x, cross, ..., normalize = TRUE)

\method{normal3d}{character}(x, ..., normalize = TRUE)

\method{normal3d}{Plane3D}(x, ..., normalize = TRUE)
}
\arguments{
\item{x}{Object to compute a 3D normal vector for
such as a \link{Plane3D} object}

\item{...}{Passed to other methods such as \code{\link[=as_coord3d]{as_coord3d()}}.}

\item{cross}{A \link{Coord3D} vector.
We'll compute the normal of \code{x} and \code{cross} by taking their cross product.}

\item{normalize}{If \code{TRUE} normalize to a unit vector}
}
\value{
A \link{Coord3D} (normal) vector
}
\description{
\code{normal3d()} is an S3 generic that computes a 3D normal vector.
}
\examples{
normal3d("xy-plane")
normal3d(as_coord3d(2, 0, 0), cross = as_coord3d(0, 2, 0))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/polygon.R
\name{as_polygon2d}
\alias{as_polygon2d}
\alias{as_polygon2d.Coord2D}
\alias{as_polygon2d.Polygon2D}
\alias{as_polygon2d.numeric}
\alias{as_polygon2d.Ellipse2D}
\title{Cast to Polygon2D object}
\usage{
as_polygon2d(x, ...)

\method{as_polygon2d}{Coord2D}(x, convex = NA, ...)

\method{as_polygon2d}{Polygon2D}(x, convex = NA, ...)

\method{as_polygon2d}{numeric}(x, y = 0, convex = NA, ...)

\method{as_polygon2d}{Ellipse2D}(x, n = 60L, type = c("inner", "outer"), ...)
}
\arguments{
\item{x}{Object to cast.  Either a \link{Coord2D} object of vertices or a
numeric vector of x-coordinates.}

\item{...}{Ignored; only included for S3 method consistency.}

\item{convex}{\code{NA} to auto-detect (default), \code{TRUE} to assert convex,
\code{FALSE} to mark as concave.}

\item{y}{Numeric vector of y-coordinates (used when \code{x} is numeric).}

\item{n}{Number of vertices in the approximating polygon (default \code{60L}).}

\item{type}{\code{"inner"} (default) for an inscribed polygon whose vertices lie
on the ellipse, or \code{"outer"} for a circumscribed polygon whose edges are
tangent to the ellipse.  Both polygons are convex.}
}
\value{
A \link{Polygon2D} object.
}
\description{
\code{as_polygon2d()} casts to a \link{Polygon2D} object.
}
\examples{
vertices <- as_coord2d(x = c(0, 1, 1, 0), y = c(0, 0, 1, 1))
p <- as_polygon2d(vertices)
p$is_convex
print(p)
plot(p)
}
\seealso{
\code{\link[=isotoxal_2ngon_polygon2d]{isotoxal_2ngon_polygon2d()}} and \code{\link[=regular_ngon_polygon2d]{regular_ngon_polygon2d()}} to
directly construct common polygon shapes.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distance.R
\name{distance2d}
\alias{distance2d}
\title{2D Euclidean distances}
\usage{
distance2d(x, y)
}
\arguments{
\item{x}{Either a \link{Coord2D} or \link{Line2D} class object}

\item{y}{Either a \link{Coord2D} or \link{Line2D} class object}
}
\description{
\code{distance2d()} computes 2D Euclidean distances.
}
\examples{
p <- as_coord2d(x = 1:4, y = 1:4)
distance2d(p, as_coord2d(0, 0))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/is_methods.R
\name{is_polygon2d}
\alias{is_polygon2d}
\title{Test whether an object has a Polygon2D class}
\usage{
is_polygon2d(x)
}
\arguments{
\item{x}{An object}
}
\value{
A logical value
}
\description{
\code{is_polygon2d()} tests whether an object has a "Polygon2D" class
}
\examples{
p <- as_polygon2d(as_coord2d(x = c(0, 1, 1, 0), y = c(0, 0, 1, 1)))
is_polygon2d(p)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coord-methods.R
\name{bounding_ranges}
\alias{bounding_ranges}
\alias{range.Coord1D}
\alias{range.Coord2D}
\alias{range.Coord3D}
\title{Compute axis-aligned ranges}
\usage{
\method{range}{Coord1D}(..., na.rm = FALSE)

\method{range}{Coord2D}(..., na.rm = FALSE)

\method{range}{Coord3D}(..., na.rm = FALSE)
}
\arguments{
\item{...}{\link{Coord1D}, \link{Coord2D}, or \link{Coord3D} object(s)}

\item{na.rm}{logical, indicating if \code{NA}'s should be omitted}
}
\value{
Either a \link{Coord1D}, \link{Coord2D}, or \link{Coord3D} object of length two.
The first element will have the minimum x/y(/z) coordinates
and the second element will have the maximum x/y(/z) coordinates
of the axis-aligned ranges.
}
\description{
\code{range()} computes axis-aligned ranges for
\link{Coord1D}, \link{Coord2D}, and \link{Coord3D} class objects.
}
\examples{
range(as_coord2d(rnorm(5), rnorm(5)))
range(as_coord3d(rnorm(5), rnorm(5), rnorm(5)))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coord-methods.R
\name{convex_hull2d}
\alias{convex_hull2d}
\alias{convex_hull2d.Coord2D}
\alias{convex_hull2d.Polygon2D}
\title{Compute 2D convex hulls}
\usage{
convex_hull2d(x, ...)

\method{convex_hull2d}{Coord2D}(x, ...)

\method{convex_hull2d}{Polygon2D}(x, ...)
}
\arguments{
\item{x}{An object to compute the convex hull of,
such as a \link{Coord2D} or \link{Polygon2D} object.}

\item{...}{Further arguments passed to or from other methods.}
}
\value{
A \link{Polygon2D} object representing the convex hull.
}
\description{
\code{convex_hull2d()} is a S3 generic for computing the convex hull of an
object.  It is implemented for \link{Coord2D} and \link{Polygon2D} objects using
\code{\link[grDevices:chull]{grDevices::chull()}} and returns a \link{Polygon2D} whose vertices are in
counter-clockwise order.
}
\examples{
pts <- as_coord2d(x = rnorm(20), y = rnorm(20))
hull <- convex_hull2d(pts)
is_polygon2d(hull)
hull$is_convex
plot(hull)
points(pts)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distance.R
\name{abs.Coord1D}
\alias{abs.Coord1D}
\alias{abs.Coord2D}
\alias{abs.Coord3D}
\title{Compute Euclidean norm}
\usage{
\method{abs}{Coord1D}(x)

\method{abs}{Coord2D}(x)

\method{abs}{Coord3D}(x)
}
\arguments{
\item{x}{A \link{Coord2D} class object or \link{Coord2D} class object.}
}
\value{
A numeric vector
}
\description{
\code{abs()} computes the Euclidean norm for \link{Coord2D} class objects and \link{Coord3D} class objects.
}
\examples{
  z <- complex(real = 1:4, imaginary = 1:4)
  p <- as_coord2d(z)
  abs(p) # Euclidean norm
  # Less efficient ways to calculate same Euclidean norms
  sqrt(p * p) # `*` dot product
  distance2d(p, as_coord2d(0, 0, 0))

  # In {base} R `abs()` calculates Euclidean norm of complex numbers
  all.equal(abs(p), abs(z))
  all.equal(Mod(p), Mod(z))

  p3 <- as_coord3d(x = 1:4, y = 1:4, z = 1:4)
  abs(p3)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coord-methods.R
\name{cross_product3d}
\alias{cross_product3d}
\title{Compute 3D vector cross product}
\usage{
cross_product3d(x, y)
}
\arguments{
\item{x}{A \link{Coord3D} class vector.}

\item{y}{A \link{Coord3D} class vector.}
}
\value{
A \link{Coord3D} class vector
}
\description{
\code{cross_product3d()} computes the cross product of two \link{Coord3D} class vectors.
}
\examples{
x <- as_coord3d(2, 3, 4)
y <- as_coord3d(5, 6, 7)
cross_product3d(x, y)
if (getRversion() >= "4.4.0") {
  crossprod(x, y)
}
}
\seealso{
\code{\link[=dot_product3d]{dot_product3d()}} for the dot product of two \link{Coord3D} vectors.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as_coord.R
\name{as_coord1d}
\alias{as_coord1d}
\alias{as_coord1d.character}
\alias{as_coord1d.Coord2D}
\alias{as_coord1d.data.frame}
\alias{as_coord1d.list}
\alias{as_coord1d.matrix}
\alias{as_coord1d.numeric}
\alias{as_coord1d.Coord1D}
\alias{as_coord1d.Point1D}
\title{Cast to coord1d object}
\usage{
as_coord1d(x, ...)

\method{as_coord1d}{character}(x, ...)

\method{as_coord1d}{Coord2D}(
  x,
  permutation = c("xy", "yx"),
  ...,
  line = as_line2d("x-axis"),
  scale = 0
)

\method{as_coord1d}{data.frame}(x, ...)

\method{as_coord1d}{list}(x, ...)

\method{as_coord1d}{matrix}(x, ...)

\method{as_coord1d}{numeric}(x, ...)

\method{as_coord1d}{Coord1D}(x, ...)

\method{as_coord1d}{Point1D}(x, ...)
}
\arguments{
\item{x}{An object that can be cast to a \link{Coord1D} class object
such as a numeric vector of x-coordinates.}

\item{...}{Further arguments passed to or from other methods}

\item{permutation}{Either "xy" (no permutation) or "yx" (permute x and y axes)}

\item{line}{A \link{Line2D} object of length one representing the line
you with to reflect across or project to or an object coercible to one by \code{as_line2d(line, ...)}
such as "x-axis" or "y-axis".}

\item{scale}{Oblique projection scale factor.
A degenerate \code{0} value indicates an orthogonal projection.}
}
\value{
A \link{Coord1D} class object
}
\description{
\code{as_coord1d()} casts to a \link{Coord1D} class object
}
\examples{
as_coord1d(x = rnorm(10))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ellipse.R
\name{as_ellipse2d}
\alias{as_ellipse2d}
\alias{as_ellipse2d.Coord2D}
\alias{as_ellipse2d.numeric}
\title{Cast to Ellipse2D object}
\usage{
as_ellipse2d(x, ...)

\method{as_ellipse2d}{Coord2D}(x, ..., r = 0.5, rx = r, ry = r, theta = angle(0))

\method{as_ellipse2d}{numeric}(x, y = 0, ..., r = 0.5, rx = r, ry = rx, theta = angle(0))
}
\arguments{
\item{x}{Object to cast.  Either a \link{Coord2D} object of center(s) or a
numeric vector of x-coordinates of center(s).}

\item{...}{Ignored; only included for S3 method consistency.}

\item{r}{Numeric vector of (circular) radii (default \code{0.5}).}

\item{rx}{Numeric vector of x-axis semi-radii (default \code{r}).}

\item{ry}{Numeric vector of y-axis semi-radii (default \code{r}).}

\item{theta}{An \code{\link[=angle]{angle()}} vector (or numeric interpreted using the default
angular unit) of rotation angles (default \code{angle(0)}).}

\item{y}{Numeric vector of y-coordinates of center(s) (used when \code{x} is
numeric).}
}
\value{
An \link{Ellipse2D} object.
}
\description{
\code{as_ellipse2d()} casts to an \link{Ellipse2D} object.
}
\examples{
# Ellipse from Coord2D center
e <- as_ellipse2d(as_coord2d(0, 0), rx = 2, ry = 1, theta = degrees(30))
plot(e)

# Multiple ellipses
e2 <- as_ellipse2d(x = c(0, 1), y = c(0, 1), rx = c(1, 2), ry = c(0.5, 1))
plot(e2)

# Multiple circles from x, y, r vectors
c2 <- as_ellipse2d(x = c(0, 1), y = c(0, 1), rx = c(0.3, 0.4))
plot(c2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/is_parallel.R
\name{is_parallel}
\alias{is_parallel}
\alias{is_parallel.Line2D}
\alias{is_parallel.Plane3D}
\title{Whether two objects are parallel}
\usage{
is_parallel(x, y, ...)

\method{is_parallel}{Line2D}(x, y, ..., tolerance = sqrt(.Machine$double.eps))

\method{is_parallel}{Plane3D}(x, y, ..., tolerance = sqrt(.Machine$double.eps))
}
\arguments{
\item{x, y}{The two objects to compute if they are parallel.}

\item{...}{Passed to other methods (or ignored).}

\item{tolerance}{Numerics with differences smaller
than \code{tolerance} will be considered \dQuote{equivalent}.}
}
\value{
A logical vector.
}
\description{
\code{is_parallel()} is a S3 method that tests whether two objects are parallel.
}
\examples{
line1 <- as_line2d("x-axis")
line2 <- as_line2d("y-axis")
line3 <- as_line2d(a = 0, b = 1, c = 2) # y + 2 = 0
is_parallel(line1, line1)
is_parallel(line1, line2)
is_parallel(line1, line3)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/regular_ngon_polygon2d.R
\name{regular_ngon_polygon2d}
\alias{regular_ngon_polygon2d}
\title{Regular n-gon polygon}
\usage{
regular_ngon_polygon2d(n, x = 0, y = 0, radius = 0.5, theta = degrees(90))
}
\arguments{
\item{n}{Number of vertices.}

\item{x}{X-coordinate of the center (default \code{0}).}

\item{y}{Y-coordinate of the center (default \code{0}).}

\item{radius}{Circumradius: distance from center to each vertex
(default \code{0.5}).}

\item{theta}{Angle of the first vertex (default \code{degrees(90)}).
Will be coerced by \code{\link[=degrees]{degrees()}}.}
}
\value{
A \link{Polygon2D} object with \verb{$is_convex == TRUE}.
}
\description{
\code{regular_ngon_polygon2d()} creates a regular n-gon \link{Polygon2D} object
centered at \verb{(x, y)} with circumradius \code{radius}.
}
\examples{
# A regular hexagon
p <- regular_ngon_polygon2d(6)
p$is_convex
plot(p)
}
\seealso{
\code{\link[=isotoxal_2ngon_polygon2d]{isotoxal_2ngon_polygon2d()}} for isotoxal star polygons.
\code{\link[=rectangle_polygon2d]{rectangle_polygon2d()}} for rectangles.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coord.R
\name{Coord1D}
\alias{Coord1D}
\title{1D coordinate vector R6 Class}
\description{
\code{Coord1D} is an \code{\link[R6:R6Class]{R6::R6Class()}} object representing one-dimensional points
represented by Cartesian Coordinates.
}
\examples{
p <- as_coord1d(x = rnorm(100, 2))
print(p, n = 10L)
pc <- mean(p) # Centroid
# method chained affine transformation matrices are auto-pre-multiplied
p$
  translate(-pc)$
  reflect("origin")$
  print(n = 10L)
}
\section{Active bindings}{
  \if{html}{\out{<div class="r6-active-bindings">}}
  \describe{
    \item{\code{xw}}{A two-column matrix representing the homogeneous coordinates.
The first column is the "x" coordinates
and the second column is all ones.}

    \item{\code{x}}{A numeric vector of x-coordinates.}
  }
  \if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-Coord1D-initialize}{\code{Coord1D$new()}}
    \item \href{#method-Coord1D-print}{\code{Coord1D$print()}}
    \item \href{#method-Coord1D-project}{\code{Coord1D$project()}}
    \item \href{#method-Coord1D-reflect}{\code{Coord1D$reflect()}}
    \item \href{#method-Coord1D-scale}{\code{Coord1D$scale()}}
    \item \href{#method-Coord1D-translate}{\code{Coord1D$translate()}}
    \item \href{#method-Coord1D-transform}{\code{Coord1D$transform()}}
    \item \href{#method-Coord1D-clone}{\code{Coord1D$clone()}}
  }
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Coord1D-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-Coord1D-initialize}{}}}
\subsection{\code{Coord1D$new()}}{
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Coord1D$new(xw)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{xw}}{A matrix with three columns representing (homogeneous) coordinates.
The first column represents x coordinates and
the last column is all ones.
Column names should be "x" and "w".}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Coord1D-print"></a>}}
\if{latex}{\out{\hypertarget{method-Coord1D-print}{}}}
\subsection{\code{Coord1D$print()}}{
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Coord1D$print(n = NULL, ...)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{n}}{Number of coordinates to print. If \code{NULL} print all of them.}
      \item{\code{...}}{Passed to \code{\link[=format.default]{format.default()}}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Coord1D-project"></a>}}
\if{latex}{\out{\hypertarget{method-Coord1D-project}{}}}
\subsection{\code{Coord1D$project()}}{
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Coord1D$project(point = as_point1d("origin"), ...)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{point}}{A \link{Point1D} object of length one representing the point
you with to reflect across or project to or an object coercible to one by \code{as_point1d(point, ...)}
such as "origin".}
      \item{\code{...}}{Passed to \code{\link[=project1d]{project1d()}}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Coord1D-reflect"></a>}}
\if{latex}{\out{\hypertarget{method-Coord1D-reflect}{}}}
\subsection{\code{Coord1D$reflect()}}{
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Coord1D$reflect(point = as_point1d("origin"), ...)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{point}}{A \link{Point1D} object of length one representing the point
you with to reflect across or project to or an object coercible to one by \code{as_point1d(point, ...)}
such as "origin".}
      \item{\code{...}}{Passed to \code{\link[=reflect1d]{reflect1d()}}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Coord1D-scale"></a>}}
\if{latex}{\out{\hypertarget{method-Coord1D-scale}{}}}
\subsection{\code{Coord1D$scale()}}{
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Coord1D$scale(x_scale = 1)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{x_scale}}{Scaling factor to apply to x coordinates}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Coord1D-translate"></a>}}
\if{latex}{\out{\hypertarget{method-Coord1D-translate}{}}}
\subsection{\code{Coord1D$translate()}}{
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Coord1D$translate(x = as_coord1d(0), ...)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{x}}{A \link{Coord1D} object of length one or an object coercible to one by \code{as_coord1d(x, ...)}.}
      \item{\code{...}}{Passed to \code{as_coord1d(x, ...)} if \code{x} is not a \link{Coord1D} object}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Coord1D-transform"></a>}}
\if{latex}{\out{\hypertarget{method-Coord1D-transform}{}}}
\subsection{\code{Coord1D$transform()}}{
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Coord1D$transform(mat = transform1d())}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{mat}}{A 2x2 matrix representing a post-multiplied affine transformation matrix.
The last \strong{column} must be equal to \code{c(0, 1)}.
If the last \strong{row} is \code{c(0, 1)} you may need to transpose it
to convert it from a pre-multiplied affine transformation matrix to a post-multiplied one.
If a 1x1 matrix we'll quietly add a final column/row equal to \code{c(0, 1)}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Coord1D-clone"></a>}}
\if{latex}{\out{\hypertarget{method-Coord1D-clone}{}}}
\subsection{\code{Coord1D$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Coord1D$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/line_plane.R
\name{as_line2d}
\alias{as_line2d}
\alias{as_line2d.numeric}
\alias{as_line2d.angle}
\alias{as_line2d.character}
\alias{as_line2d.Coord2D}
\alias{as_line2d.Line2D}
\alias{as_line2d.Point1D}
\title{Cast to Line2D object}
\usage{
as_line2d(...)

\method{as_line2d}{numeric}(a, b, c, ...)

\method{as_line2d}{angle}(theta, p1 = as_coord2d("origin"), ...)

\method{as_line2d}{character}(x, ...)

\method{as_line2d}{Coord2D}(normal, p1 = as_coord3d("origin"), p2, ...)

\method{as_line2d}{Line2D}(line, ...)

\method{as_line2d}{Point1D}(point, b = 0, ...)
}
\arguments{
\item{...}{Passed to other function such as \code{as_coord2d()}.}

\item{a, b, c}{Numeric vectors that parameterize the line via the equation \code{a * x + b * y + c = 0}.
Note if \code{y = m * x + b} then \code{m * x + 1 * y + -b = 0}.}

\item{theta}{Angle of the line represented by an \code{\link[=angle]{angle()}} vector.}

\item{p1}{Point on the line represented by a \link{Coord2D} class object.}

\item{x}{A (character) vector to be cast to a \link{Line2D} object}

\item{normal}{Normal vector to the line represented by a \link{Coord2D} class object. \code{p2} should be missing.}

\item{p2}{Another point on the line represented by a \link{Coord2D} class object.}

\item{line}{A \link{Line2D} object}

\item{point}{A \link{Point1D} object}
}
\description{
\code{as_line2d()} casts to a \link{Line2D} object.
}
\examples{
p1 <- as_coord2d(x = 5, y = 10)
p2 <- as_coord2d(x = 7, y = 12)
theta <- degrees(45)
as_line2d(theta, p1)
as_line2d(p1, p2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coord-methods.R
\name{dot_product}
\alias{dot_product}
\alias{dot_product1d}
\alias{dot_product2d}
\alias{dot_product3d}
\title{Compute dot (inner) products}
\usage{
dot_product1d(x, y)

dot_product2d(x, y)

dot_product3d(x, y)
}
\arguments{
\item{x}{A \link{Coord1D}, \link{Coord2D}, or \link{Coord3D} object.}

\item{y}{A \link{Coord1D}, \link{Coord2D}, or \link{Coord3D} object.}
}
\value{
A numeric vector of dot products.
}
\description{
\code{dot_product1d()}, \code{dot_product2d()}, and \code{dot_product3d()} compute the
dot (inner) product of two coordinate vectors.
You may also use the \code{*} operator and if R >= 4.3.0 you may also use the \code{\%*\%} operator to compute the dot (inner) product.
}
\examples{
p1 <- as_coord2d(x = c(1, 2), y = c(3, 4))
p2 <- as_coord2d(x = c(5, 6), y = c(7, 8))
dot_product2d(p1, p2)
p1 * p2 # equivalent
if (getRversion() >= "4.3.0") {
  p1 \%*\% p2
}
}
\seealso{
\code{\link[=cross_product3d]{cross_product3d()}} for the cross product of two \link{Coord3D} vectors.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/angle.R
\name{as_angle}
\alias{as_angle}
\alias{as_angle.angle}
\alias{as_angle.character}
\alias{as_angle.complex}
\alias{as_angle.Coord2D}
\alias{as_angle.Coord3D}
\alias{as_angle.Line2D}
\alias{as_angle.Plane3D}
\alias{as_angle.numeric}
\title{Cast to angle vector}
\usage{
as_angle(x, unit = getOption("affiner_angular_unit", "degrees"), ...)

\method{as_angle}{angle}(x, unit = getOption("affiner_angular_unit", "degrees"), ...)

\method{as_angle}{character}(x, unit = getOption("affiner_angular_unit", "degrees"), ...)

\method{as_angle}{complex}(x, unit = getOption("affiner_angular_unit", "degrees"), ...)

\method{as_angle}{Coord2D}(x, unit = getOption("affiner_angular_unit", "degrees"), ...)

\method{as_angle}{Coord3D}(
  x,
  unit = getOption("affiner_angular_unit", "degrees"),
  type = c("azimuth", "inclination"),
  ...
)

\method{as_angle}{Line2D}(x, unit = getOption("affiner_angular_unit", "degrees"), ...)

\method{as_angle}{Plane3D}(
  x,
  unit = getOption("affiner_angular_unit", "degrees"),
  type = c("azimuth", "inclination"),
  ...
)

\method{as_angle}{numeric}(x, unit = getOption("affiner_angular_unit", "degrees"), ...)
}
\arguments{
\item{x}{An R object to convert to a \code{\link[=angle]{angle()}} vector}

\item{unit}{A string of the desired angular unit.  Supports the following strings
(note we ignore any punctuation and space characters as well as any trailing \code{s}'s
e.g. "half turns" will be treated as equivalent to "halfturn"):
\itemize{
\item "deg" or "degree"
\item "half-revolution", "half-turn", or "pi-radian"
\item "gon", "grad", "grade", or "gradian"
\item "rad" or "radian"
\item "rev", "revolution", "tr", or "turn"
}}

\item{...}{Further arguments passed to or from other methods}

\item{type}{Use "azimuth" to calculate the azimuthal angle and "inclination" to calculate the inclination angle aka polar angle.}
}
\value{
An \code{\link[=angle]{angle()}} vector
}
\description{
\code{as_angle()} casts to an \code{\link[=angle]{angle()}} vector
}
\examples{
as_angle(angle(pi, "radians"), "pi-radians")
as_angle(complex(real = 0, imaginary = 1), "degrees")
as_angle(as_coord2d(x = 0, y = 1), "turns")
as_angle(200, "gradians")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/is_methods.R
\name{is_segment2d}
\alias{is_segment2d}
\title{Test whether an object has a Segment2D class}
\usage{
is_segment2d(x)
}
\arguments{
\item{x}{An object}
}
\value{
A logical value
}
\description{
\code{is_segment2d()} tests whether an object has a "Segment2D" class
}
\examples{
p1 <- as_coord2d(x = c(0, 1), y = c(0, 0))
p2 <- as_coord2d(x = c(1, 1), y = c(0, 1))
s <- as_segment2d(p1, p2 = p2)
is_segment2d(s)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/isocubeGrob.R
\name{isocubeGrob}
\alias{isocubeGrob}
\alias{grid.isocube}
\title{Isometric cube grob}
\usage{
isocubeGrob(
  top,
  right,
  left,
  gp_border = grid::gpar(col = "black", lwd = 12),
  name = NULL,
  gp = grid::gpar(),
  vp = NULL
)

grid.isocube(...)
}
\arguments{
\item{top}{A grid grob object to use as the top side of the cube.  ggplot2 objects will be coerced by \code{\link[ggplot2:ggplotGrob]{ggplot2::ggplotGrob()}}.}

\item{right}{A grid grob object to use as the right side of the cube.  ggplot2 objects will be coerced by \code{\link[ggplot2:ggplotGrob]{ggplot2::ggplotGrob()}}.}

\item{left}{A grid grob object to use as the left side of the cube.  ggplot2 objects will be coerced by \code{\link[ggplot2:ggplotGrob]{ggplot2::ggplotGrob()}}.}

\item{gp_border}{A \code{\link[grid:gpar]{grid::gpar()}} object for the \code{\link[=polygonGrob]{polygonGrob()}} used to draw borders around
the cube faces.}

\item{name}{A character identifier (for grid).}

\item{gp}{A \code{\link[grid:gpar]{grid::gpar()}} object.}

\item{vp}{A \code{\link[grid:viewport]{grid::viewport()}} object (or \code{NULL}).}

\item{...}{Passed to \code{isocubeGrob()}}
}
\value{
A \code{\link[grid:gTree]{grid::gTree()}} (grob) object of class "isocube".
As a side effect \code{grid.isocube()} draws to the active graphics device.
}
\description{
\code{isometricCube()} is a grid grob function to render
isometric cube faces by automatically wrapping around \code{affineGrob()}.
}
\details{
Any \code{ggplot2} objects are coerced to grobs by \code{\link[ggplot2:ggplotGrob]{ggplot2::ggplotGrob()}}.  Depending on what you'd like
to do you may want to instead manually convert a ggplot2 object \code{gg} to a grob with \code{gtable::gtable_filter(ggplot2::ggplotGrob(gg), "panel")}.

Not all graphics devices provided by \code{grDevices} or other R packages support the \href{https://www.stat.auckland.ac.nz/~paul/Reports/GraphicsEngine/groups/groups.html}{affine transformation feature introduced in R 4.2}.
If \code{isTRUE(getRversion() >= '4.2.0')} then the active graphics device should support this feature if \code{isTRUE(grDevices::dev.capabilities()$transformations)}.
In particular the following graphics devices should support the affine transformation feature:
\itemize{
\item R's \code{\link[grDevices:pdf]{grDevices::pdf()}} device
\item R's 'cairo' devices e.g. \code{\link[grDevices:cairo_pdf]{grDevices::cairo_pdf()}}, \code{grDevices::png(type = 'cairo')}, \code{\link[grDevices:svg]{grDevices::svg()}}, \code{grDevices::x11(type = 'cairo')}, etc. If \code{isTRUE(capabilities('cairo'))} then R was compiled with support for the 'cairo' devices .
\item R's 'quartz' devices (since R 4.3.0) e.g. \code{\link[grDevices:quartz]{grDevices::quartz()}}, \code{grDevices::png(type = 'quartz')}, etc. If \code{isTRUE(capabilities('aqua'))} then R was compiled with support for the 'quartz' devices (generally only \code{TRUE} on macOS systems).
\item \code{ragg}'s devices (since v1.3.0) e.g. \code{\link[ragg:agg_png]{ragg::agg_png()}}, \code{\link[ragg:agg_capture]{ragg::agg_capture()}}, etc.
}
}
\examples{
# Only works if active graphics device supports affine transformations
# such as `png(type="cairo")` on R 4.2+
can_run_grid_example <- require("grid", quietly = TRUE) &&
  getRversion() >= "4.2.0" &&
  isTRUE(dev.capabilities()$transformations)
if (can_run_grid_example) {
  grid.newpage()
  gp_text <- gpar(fontsize = 72)
  grid.isocube(top = textGrob("top", gp = gp_text),
               right = textGrob("right", gp = gp_text),
               left = textGrob("left", gp = gp_text))
}
if (can_run_grid_example) {
  colors <- c("#D55E00", "#009E73", "#56B4E9")
  spacings <- c(0.25, 0.2, 0.25)
  texts <- c("pkgname", "left\nface", "right\nface")
  rots <- c(45, 0, 0)
  fontsizes <- c(52, 80, 80)
  sides <- c("top", "left", "right")
  types <- gridpattern::names_polygon_tiling[c(5, 7, 9)]
  l_grobs <- list()
  grid.newpage()
  for (i in 1:3) {
      if (requireNamespace("gridpattern", quietly = TRUE)) {
          bg <- gridpattern::grid.pattern_polygon_tiling(
                     colour = "grey80",
                     fill = c(colors[i], "white"),
                     type = types[i],
                     spacing = spacings[i],
                     draw = FALSE)
      } else {
          bg <- rectGrob(gp = gpar(col = NA, fill = colors[i]))
      }
      text <- textGrob(texts[i], rot = rots[i],
                       gp = gpar(fontsize = fontsizes[i]))
      l_grobs[[sides[i]]] <- grobTree(bg, text)
  }
  grid.newpage()
  grid.isocube(top = l_grobs$top,
               right = l_grobs$right,
               left = l_grobs$left)
}
\donttest{# May take more than 5 seconds on CRAN machines
can_run_artsy_example <- can_run_grid_example &&
  require("aRtsy", quietly = TRUE) &&
  require("ggplot2", quietly = TRUE) &&
  requireNamespace("gtable", quietly = TRUE)
if (can_run_artsy_example) {
  gg <- canvas_planet(colorPalette("lava"), threshold = 3) +
    scale_x_continuous(expand = c(0, 0)) +
    scale_y_continuous(expand = c(0, 0))
  grob <- ggplotGrob(gg)
  grob <- gtable::gtable_filter(grob, "panel") # grab just the panel
  grid.newpage()
  grid.isocube(top = grob, left = grob, right = grob,
               gp_border = grid::gpar(col = "darkorange", lwd = 12))
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/angle.R
\name{angle}
\alias{angle}
\alias{degrees}
\alias{gradians}
\alias{pi_radians}
\alias{radians}
\alias{turns}
\title{Angle vectors}
\usage{
angle(x = numeric(), unit = getOption("affiner_angular_unit", "degrees"))

degrees(x)

gradians(x)

pi_radians(x)

radians(x)

turns(x)
}
\arguments{
\item{x}{An angle vector or an object to convert to it (such as a numeric vector)}

\item{unit}{A string of the desired angular unit.  Supports the following strings
(note we ignore any punctuation and space characters as well as any trailing \code{s}'s
e.g. "half turns" will be treated as equivalent to "halfturn"):
\itemize{
\item "deg" or "degree"
\item "half-revolution", "half-turn", or "pi-radian"
\item "gon", "grad", "grade", or "gradian"
\item "rad" or "radian"
\item "rev", "revolution", "tr", or "turn"
}}
}
\value{
A numeric vector of class "angle".
Its "unit" attribute is a standardized string of the specified angular unit.
}
\description{
\code{angle()} creates angle vectors with user specified angular unit.
around \code{\link[=as_angle]{as_angle()}} for those angular units.
}
\examples{
  # Different representations of the "same" angle
  angle(180, "degrees")
  angle(pi, "radians")
  angle(0.5, "turns")
  angle(200, "gradians")
  pi_radians(1)

  a1 <- angle(180, "degrees")
  angular_unit(a1)
  is_angle(a1)
  as.numeric(a1, "radians")
  cos(a1)

  a2 <- as_angle(a1, "radians")
  angular_unit(a2)
  is_congruent(a1, a2)
}
\seealso{
\code{\link[=as_angle]{as_angle()}}, \code{\link[=angular_unit]{angular_unit()}}, and \link{angle-methods}.
\url{https://en.wikipedia.org/wiki/Angle#Units} for more information about angular units.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/is_equivalent.R
\name{is_equivalent}
\alias{is_equivalent}
\alias{is_equivalent.angle}
\alias{is_equivalent.numeric}
\alias{is_equivalent.Coord1D}
\alias{is_equivalent.Coord2D}
\alias{is_equivalent.Coord3D}
\alias{is_equivalent.Point1D}
\alias{is_equivalent.Line2D}
\alias{is_equivalent.Plane3D}
\title{Test whether two objects are equivalent}
\usage{
is_equivalent(x, y, ...)

\method{is_equivalent}{angle}(
  x,
  y,
  ...,
  mod_turns = TRUE,
  tolerance = sqrt(.Machine$double.eps)
)

\method{is_equivalent}{numeric}(x, y, ..., tolerance = sqrt(.Machine$double.eps))

\method{is_equivalent}{Coord1D}(x, y, ..., tolerance = sqrt(.Machine$double.eps))

\method{is_equivalent}{Coord2D}(x, y, ..., tolerance = sqrt(.Machine$double.eps))

\method{is_equivalent}{Coord3D}(x, y, ..., tolerance = sqrt(.Machine$double.eps))

\method{is_equivalent}{Point1D}(x, y, ..., tolerance = sqrt(.Machine$double.eps))

\method{is_equivalent}{Line2D}(x, y, ..., tolerance = sqrt(.Machine$double.eps))

\method{is_equivalent}{Plane3D}(x, y, ..., tolerance = sqrt(.Machine$double.eps))
}
\arguments{
\item{x, y}{Two objects to test whether they are \dQuote{"equivalent"}.}

\item{...}{Further arguments passed to or from other methods.}

\item{mod_turns}{If \code{TRUE} angles that are congruent modulo full turns will be considered \dQuote{congruent}.}

\item{tolerance}{Numerics with differences smaller
than \code{tolerance} will be considered \dQuote{equivalent}.}
}
\value{
A logical vector
}
\description{
\code{is_equivalent()} is a S3 generic that tests whether two different objects are \dQuote{equivalent}.
The \code{is_equivalent()} method for \code{\link[=angle]{angle()}} classes tests whether two angles are congruent.
The \code{is_equivalent()} method for \link{Point1D}, \link{Line2D}, \link{Plane3D} classes tests whether they are the same point/line/plane after standardization.
}
\examples{
line1 <- as_line2d(a = 1, b = 2, c = 3) # 1 * x + 2 * y + 3 = 0
line2 <- as_line2d(a = 2, b = 4, c = 6) # 2 * x + 4 * y + 6 = 0
is_equivalent(line1, line2)
}
\seealso{
\code{\link[=is_congruent]{is_congruent()}}, \code{\link[=all.equal]{all.equal()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/is_methods.R
\name{is_point1d}
\alias{is_point1d}
\title{Test whether an object has a Point1D class}
\usage{
is_point1d(x)
}
\arguments{
\item{x}{An object}
}
\value{
A logical value
}
\description{
\code{is_point1d()} tests whether an object has a "Point1D" class
}
\examples{
p <- as_point1d(a = 1, b = 5)
is_point1d(p)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/segment.R
\name{as_segment2d}
\alias{as_segment2d}
\alias{as_segment2d.Coord2D}
\alias{as_segment2d.Polygon2D}
\title{Cast to Segment2D object}
\usage{
as_segment2d(...)

\method{as_segment2d}{Coord2D}(p1, ..., p2, vec)

\method{as_segment2d}{Polygon2D}(p, ...)
}
\arguments{
\item{...}{Ignored.}

\item{p1}{A \link{Coord2D} object of first endpoints.}

\item{p2}{A \link{Coord2D} object of second endpoints.
If missing, \code{vec} must be supplied.}

\item{vec}{A \link{Coord2D} object of edge vectors (\code{p2 - p1}).
Used only when \code{p2} is missing.}

\item{p}{A \link{Polygon2D} object whose edges should be returned.}
}
\value{
A \link{Segment2D} object.
}
\description{
\code{as_segment2d()} creates a \link{Segment2D} object.
}
\examples{
p1 <- as_coord2d(x = c(0, 1), y = c(0, 0))
p2 <- as_coord2d(x = c(1, 1), y = c(0, 1))
s <- as_segment2d(p1, p2 = p2)
s$p1
s$p2
s$mid_point

# From a polygon's edges
poly <- as_polygon2d(as_coord2d(x = c(0, 1, 1, 0), y = c(0, 0, 1, 1)))
as_segment2d(poly)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/line_plane.R
\name{as_plane3d}
\alias{as_plane3d}
\alias{as_plane3d.numeric}
\alias{as_plane3d.character}
\alias{as_plane3d.Coord3D}
\alias{as_plane3d.Plane3D}
\alias{as_plane3d.Point1D}
\alias{as_plane3d.Line2D}
\title{Cast to Plane3D object}
\usage{
as_plane3d(...)

\method{as_plane3d}{numeric}(a, b, c, d, ...)

\method{as_plane3d}{character}(x, ...)

\method{as_plane3d}{Coord3D}(normal, p1 = as_coord3d("origin"), p2, p3, ...)

\method{as_plane3d}{Plane3D}(plane, ...)

\method{as_plane3d}{Point1D}(point, b = 0, c = 0, ...)

\method{as_plane3d}{Line2D}(line, c = 0, ...)
}
\arguments{
\item{...}{Passed to other function such as \code{as_coord2d()}.}

\item{a, b, c, d}{Numeric vectors that parameterize the plane via the equation \code{a * x + b * y + c * z + d = 0}.}

\item{x}{A (character) vector to be cast to a \link{Plane3D} object}

\item{normal}{Normal vector to the plane represented by a \link{Coord3D} class object. \code{p2} and \code{p3} should be missing.}

\item{p1}{Point on the plane represented by a \link{Coord3D} class object.}

\item{p2, p3}{Points on the plane represented by \link{Coord3D} class objects.  \code{normal} should be missing.}

\item{plane}{A \link{Plane3D} object}

\item{point}{A \link{Point1D} object}

\item{line}{A \link{Line2D} object}
}
\description{
\code{as_plane3d()} casts to a \link{Plane3D} object.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/is_methods.R
\name{is_line2d}
\alias{is_line2d}
\title{Test whether an object has a Line2D class}
\usage{
is_line2d(x)
}
\arguments{
\item{x}{An object}
}
\value{
A logical value
}
\description{
\code{is_line2d()} tests whether an object has a "Line2D" class
}
\examples{
l <- as_line2d(a = 1, b = 2, c = 3)
is_line2d(l)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/trigonometry.R
\name{inverse-trigonometric-functions}
\alias{inverse-trigonometric-functions}
\alias{arcsine}
\alias{arccosine}
\alias{arctangent}
\alias{arcsecant}
\alias{arccosecant}
\alias{arccotangent}
\title{Angle vector aware inverse trigonometric functions}
\usage{
arcsine(
  x,
  unit = getOption("affiner_angular_unit", "degrees"),
  tolerance = sqrt(.Machine$double.eps)
)

arccosine(
  x,
  unit = getOption("affiner_angular_unit", "degrees"),
  tolerance = sqrt(.Machine$double.eps)
)

arctangent(x, unit = getOption("affiner_angular_unit", "degrees"), y = NULL)

arcsecant(x, unit = getOption("affiner_angular_unit", "degrees"))

arccosecant(x, unit = getOption("affiner_angular_unit", "degrees"))

arccotangent(x, unit = getOption("affiner_angular_unit", "degrees"))
}
\arguments{
\item{x}{A numeric vector}

\item{unit}{A string of the desired angular unit.  Supports the following strings
(note we ignore any punctuation and space characters as well as any trailing \code{s}'s
e.g. "half turns" will be treated as equivalent to "halfturn"):
\itemize{
\item "deg" or "degree"
\item "half-revolution", "half-turn", or "pi-radian"
\item "gon", "grad", "grade", or "gradian"
\item "rad" or "radian"
\item "rev", "revolution", "tr", or "turn"
}}

\item{tolerance}{If \code{x} greater than 1 (or less than -1) but is within a \code{tolerance}
of 1 (or -1) then it will be treated as 1 (or -1)}

\item{y}{A numeric vector or \code{NULL}.
If \code{NULL} (default) we compute the 1-argument arctangent
else we compute the 2-argument arctangent.
For positive coordinates \verb{(x, y)} then \code{arctangent(x = y/x) == arctangent(x = x, y = y)}.}
}
\value{
An \code{\link[=angle]{angle()}} vector
}
\description{
\code{arcsine()}, \code{arccosine()}, \code{arctangent()},
\code{arcsecant()}, \code{arccosecant()}, and \code{arccotangent()} are
inverse trigonometric functions that return \code{\link[=angle]{angle()}} vectors
with a user chosen angular unit.
}
\examples{
arccosine(-1, "degrees")
arcsine(0, "turns")
arctangent(0, "gradians")
arccosecant(-1, "degrees")
arcsecant(1, "degrees")
arccotangent(1, "half-turns")

# `base::atan2(y, x)` computes the angle of the vector from origin to (x, y)
as_angle(as_coord2d(x = 1, y = 1), "degrees")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/affineGrob.R
\name{affineGrob}
\alias{affineGrob}
\alias{grid.affine}
\title{Affine transformation grob}
\usage{
affineGrob(
  grob,
  vp_define = NULL,
  transform = NULL,
  vp_use = NULL,
  name = NULL,
  gp = grid::gpar(),
  vp = NULL
)

grid.affine(...)
}
\arguments{
\item{grob}{A grid grob to perform affine transformations on.  Passed to \code{\link[grid:defineGrob]{grid::defineGrob()}} as its \code{src} argument.}

\item{vp_define}{\code{\link[grid:viewport]{grid::viewport()}} to define grid group in.  Passed to \code{\link[grid:defineGrob]{grid::defineGrob()}} as its \code{vp} argument.
This will cumulative with the current viewport and the \code{vp} argument (if any),
if this cumulative viewport falls outside the graphics device drawing area this
grob may be clipped on certain graphics devices.}

\item{transform}{An affine transformation function.
If \code{NULL} default to \code{\link[grid:viewportTransform]{grid::viewportTransform()}}.
Passed to \code{\link[grid:useGrob]{grid::useGrob()}} as its \code{transform} argument.}

\item{vp_use}{\code{\link[grid:viewport]{grid::viewport()}} passed to \code{\link[grid:useGrob]{grid::useGrob()}} as its \code{vp} argument.}

\item{name}{A character identifier (for grid).}

\item{gp}{A \code{\link[grid:gpar]{grid::gpar()}} object.}

\item{vp}{A \code{\link[grid:viewport]{grid::viewport()}} object (or \code{NULL}).}

\item{...}{Passed to \code{affineGrob()}}
}
\value{
A \code{\link[grid:gTree]{grid::gTree()}} (grob) object of class "affine".
As a side effect \code{grid.affine()} draws to the active graphics device.
}
\description{
\code{affineGrob()} is a grid grob function to facilitate
using the group affine transformation features introduced in R 4.2.
}
\details{
Not all graphics devices provided by \code{grDevices} or other R packages support the \href{https://www.stat.auckland.ac.nz/~paul/Reports/GraphicsEngine/groups/groups.html}{affine transformation feature introduced in R 4.2}.
If \code{isTRUE(getRversion() >= '4.2.0')} then the active graphics device should support this feature if \code{isTRUE(grDevices::dev.capabilities()$transformations)}.
In particular the following graphics devices should support the affine transformation feature:
\itemize{
\item R's \code{\link[grDevices:pdf]{grDevices::pdf()}} device
\item R's 'cairo' devices e.g. \code{\link[grDevices:cairo_pdf]{grDevices::cairo_pdf()}}, \code{grDevices::png(type = 'cairo')}, \code{\link[grDevices:svg]{grDevices::svg()}}, \code{grDevices::x11(type = 'cairo')}, etc. If \code{isTRUE(capabilities('cairo'))} then R was compiled with support for the 'cairo' devices .
\item R's 'quartz' devices (since R 4.3.0) e.g. \code{\link[grDevices:quartz]{grDevices::quartz()}}, \code{grDevices::png(type = 'quartz')}, etc. If \code{isTRUE(capabilities('aqua'))} then R was compiled with support for the 'quartz' devices (generally only \code{TRUE} on macOS systems).
\item \code{ragg}'s devices (since v1.3.0) e.g. \code{\link[ragg:agg_png]{ragg::agg_png()}}, \code{\link[ragg:agg_capture]{ragg::agg_capture()}}, etc.
}
}
\examples{
if (require("grid")) {
  grob <- grobTree(rectGrob(gp = gpar(fill = "blue", col = NA)),
                   circleGrob(gp=gpar(fill="yellow", col = NA)),
                   textGrob("RSTATS", gp=gpar(fontsize=32)))
  grid.newpage()
  pushViewport(viewport(width=unit(4, "in"), height=unit(2, "in")))
  grid.draw(grob)
  popViewport()
}

if (require("grid") &&
    getRversion() >= "4.2.0" &&
    isTRUE(dev.capabilities()$transformations)) {
  # Only works if active graphics device supports affine transformations
  # such as `png(type="cairo")` on R 4.2+
  vp_define <- viewport(width=unit(2, "in"), height=unit(2, "in"))
  affine <- affineGrob(grob, vp_define=vp_define)
  grid.newpage()
  pushViewport(viewport(width=unit(4, "in"), height=unit(2, "in")))
  grid.draw(affine)
  popViewport()
}
if (require("grid") &&
    getRversion() >= "4.2.0" &&
    isTRUE(dev.capabilities()$transformations)) {
  # Only works if active graphics device supports affine transformations
  # such as `png(type="cairo")` on R 4.2+
  settings <- affine_settings(xy = list(x = c(3/3, 2/3, 0/3, 1/3),
                                        y = c(2/3, 1/3, 1/3, 2/3)),
                              unit = "snpc")
  affine <- affineGrob(grob,
                       vp_define = vp_define,
                       transform = settings$transform,
                       vp_use = settings$vp)
  grid.newpage()
  grid.draw(affine)
}
}
\seealso{
See \code{\link[=affine_settings]{affine_settings()}} for computing good \code{transform} and \code{vp_use} settings.
See \url{https://www.stat.auckland.ac.nz/~paul/Reports/GraphicsEngine/groups/groups.html}
for more information about the group affine transformation feature.
See \code{\link[=isocubeGrob]{isocubeGrob()}} which wraps this function to render isometric cubes.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/affine.R
\name{transform1d}
\alias{transform1d}
\alias{project1d}
\alias{reflect1d}
\alias{scale1d}
\alias{translate1d}
\title{1D affine transformation matrices}
\usage{
transform1d(mat = diag(2L))

project1d(point = as_point1d("origin"), ...)

reflect1d(point = as_point1d("origin"), ...)

scale1d(x_scale = 1)

translate1d(x = as_coord1d(0), ...)
}
\arguments{
\item{mat}{A 2x2 matrix representing a post-multiplied affine transformation matrix.
The last \strong{column} must be equal to \code{c(0, 1)}.
If the last \strong{row} is \code{c(0, 1)} you may need to transpose it
to convert it from a pre-multiplied affine transformation matrix to a post-multiplied one.
If a 1x1 matrix we'll quietly add a final column/row equal to \code{c(0, 1)}.}

\item{point}{A \link{Point1D} object of length one representing the point
you with to reflect across or project to or an object coercible to one by \code{as_point1d(point, ...)}
such as "origin".}

\item{...}{Passed to \code{\link[=as_coord1d]{as_coord1d()}}.}

\item{x_scale}{Scaling factor to apply to x coordinates}

\item{x}{A \link{Coord1D} object of length one or an object coercible to one by \code{as_coord1d(x, ...)}.}
}
\value{
A 2x2 post-multiplied affine transformation matrix with classes "transform1d" and "at_matrix"
}
\description{
\code{transform1d()}, \code{reflect1d()}, \code{scale1d()},
and \code{translate1d()} create 1D affine transformation matrix objects.
}
\details{
\describe{
\item{\code{transform1d()}}{User supplied (post-multiplied) affine transformation matrix}.
\item{\code{reflect1d()}}{Reflections across a point.}
\item{\code{scale1d()}}{Scale the x-coordinates by multiplicative scale factors.}
\item{\code{translate1d()}}{Translate the coordinates by a \link{Coord1D} class object parameter.}
}

\code{transform1d()} 1D affine transformation matrix objects are meant to be
post-multiplied and therefore should \strong{not} be multiplied in reverse order.
Note the \link{Coord1D} class object methods auto-pre-multiply affine transformations
when "method chaining" so pre-multiplying affine transformation matrices
to do a single cumulative transformation instead of a method chain of multiple transformations
will not improve performance as much as it does in other R packages.

To convert a pre-multiplied 1D affine transformation matrix to a post-multiplied one
simply compute its transpose using \code{\link[=t]{t()}}.  To get an inverse transformation matrix
from an existing transformation matrix that does the opposite transformations
simply compute its inverse using \code{\link[=solve]{solve()}}.
}
\examples{
p <- as_coord1d(x = sample(1:10, 3))

# {affiner} affine transformation matrices are post-multiplied
# and therefore should **not** go in reverse order
mat <- transform1d(diag(2)) \%*\%
         scale1d(2) \%*\%
         translate1d(x = -1)
p1 <- p$
  clone()$
  transform(mat)

# The equivalent result applying affine transformations via method chaining
p2 <- p$
  clone()$
  transform(diag(2))$
  scale(2)$
  translate(x = -1)

all.equal(p1, p2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/line_plane.R
\name{Plane3D}
\alias{Plane3D}
\title{3D planes R6 Class}
\description{
\code{Plane3D} is an \code{\link[R6:R6Class]{R6::R6Class()}} object representing three-dimensional planes.
}
\section{Public fields}{
  \if{html}{\out{<div class="r6-fields">}}
  \describe{
    \item{\code{a}}{Numeric vector that parameterizes the plane via the equation \code{a * x + b * y + c * z + d = 0}.}

    \item{\code{b}}{Numeric vector that parameterizes the plane via the equation \code{a * x + b * y + c * z + d = 0}.}

    \item{\code{c}}{Numeric vector that parameterizes the plane via the equation \code{a * x + b * y + c * z + d = 0}.}

    \item{\code{d}}{Numeric vector that parameterizes the plane via the equation \code{a * x + b * y + c * z + d = 0}.}
  }
  \if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-Plane3D-initialize}{\code{Plane3D$new()}}
    \item \href{#method-Plane3D-print}{\code{Plane3D$print()}}
    \item \href{#method-Plane3D-clone}{\code{Plane3D$clone()}}
  }
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Plane3D-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-Plane3D-initialize}{}}}
\subsection{\code{Plane3D$new()}}{
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Plane3D$new(a, b, c, d)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{a}}{Numeric vector that parameterizes the plane via the equation \code{a * x + b * y + c * z + d = 0}.}
      \item{\code{b}}{Numeric vector that parameterizes the plane via the equation \code{a * x + b * y + c * z + d = 0}.}
      \item{\code{c}}{Numeric vector that parameterizes the plane via the equation \code{a * x + b * y + c * z + d = 0}.}
      \item{\code{d}}{Numeric vector that parameterizes the plane via the equation \code{a * x + b * y + c * z + d = 0}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Plane3D-print"></a>}}
\if{latex}{\out{\hypertarget{method-Plane3D-print}{}}}
\subsection{\code{Plane3D$print()}}{
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Plane3D$print(n = NULL, ...)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{n}}{Number of lines to print.  If \code{NULL} print all of them.}
      \item{\code{...}}{Passed to \code{\link[=format.default]{format.default()}}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Plane3D-clone"></a>}}
\if{latex}{\out{\hypertarget{method-Plane3D-clone}{}}}
\subsection{\code{Plane3D$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Plane3D$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/painter_depth.R
\name{painter_depth}
\alias{painter_depth}
\alias{painter_depth.Coord2D}
\alias{painter_depth.Coord3D}
\alias{painter_depth.Segment2D}
\alias{painter_depth.Polygon2D}
\alias{painter_order}
\alias{painter_order.Coord2D}
\alias{painter_order.Coord3D}
\alias{painter_order.Polygon2D}
\alias{sort.Coord2D}
\alias{sort.Coord3D}
\title{Painter's depth, order, and sort methods}
\usage{
painter_depth(x, ...)

\method{painter_depth}{Coord2D}(x, ..., scale = 1, alpha = angle(45, "degrees"))

\method{painter_depth}{Coord3D}(
  x,
  permutation = c("xyz", "xzy", "yxz", "yzx", "zyx", "zxy"),
  ...,
  plane = as_plane3d("xy-plane"),
  scale = 0,
  alpha = angle(45, "degrees"),
  roll = angle(0, "degrees")
)

\method{painter_depth}{Segment2D}(x, ..., scale = 1, alpha = angle(45, "degrees"))

\method{painter_depth}{Polygon2D}(x, ..., scale = 1, alpha = angle(45, "degrees"))

painter_order(x, ...)

\method{painter_order}{Coord2D}(
  x,
  ...,
  decreasing = FALSE,
  scale = 1,
  alpha = angle(45, "degrees")
)

\method{painter_order}{Coord3D}(
  x,
  permutation = c("xyz", "xzy", "yxz", "yzx", "zyx", "zxy"),
  ...,
  decreasing = FALSE,
  plane = as_plane3d("xy-plane"),
  scale = 0,
  alpha = angle(45, "degrees"),
  roll = angle(0, "degrees")
)

\method{painter_order}{Polygon2D}(
  x,
  ...,
  decreasing = FALSE,
  scale = 1,
  alpha = angle(45, "degrees")
)

\method{sort}{Coord2D}(x, decreasing = FALSE, ..., scale = 1, alpha = angle(45, "degrees"))

\method{sort}{Coord3D}(
  x,
  decreasing = FALSE,
  ...,
  permutation = c("xyz", "xzy", "yxz", "yzx", "zyx", "zxy"),
  plane = as_plane3d("xy-plane"),
  scale = 0,
  alpha = angle(45, "degrees"),
  roll = angle(0, "degrees")
)
}
\arguments{
\item{x}{Object to compute painter's depth/order for, or to sort.}

\item{...}{Passed to \code{\link[=as_plane3d]{as_plane3d()}} or \code{\link[=as_angle]{as_angle()}} as needed.}

\item{scale}{Oblique projection foreshortening scale factor.
The 2D methods default to \code{1} since \code{0} yields zero depth for all
\code{Coord2D} points (z is always 0), making depth-based ordering impossible.
The \code{Coord3D} method defaults to \code{0} (orthographic projection, depth = z).
A value of \code{0.5} is used by a \dQuote{cabinet projection}
while a value of \code{1.0} is used by a \dQuote{cavalier projection}.
All positive values produce the same ordering.}

\item{alpha}{Oblique projection angle (the angle the off-axis direction is
projected going off at).
An \code{\link[=angle]{angle()}} object or one coercible to one with \code{as_angle(alpha, ...)}.
Popular angles are 45 degrees, 60 degrees, and \code{arctangent(2)} degrees.}

\item{permutation}{Either "xyz" (no permutation), "xzy" (permute y and z axes),
"yxz" (permute x and y axes), "yzx" (x becomes z, y becomes x, z becomes y),
"zxy" (x becomes y, y becomes z, z becomes x), "zyx" (permute x and z axes).
This permutation is applied before the projection.}

\item{plane}{A \link{Plane3D} class object representing the plane
projected onto, or an object coercible to one using
\code{as_plane3d(plane, ...)} such as "xy-plane", "xz-plane", or "yz-plane".}

\item{roll}{Rotation of the in-plane coordinate frame around the plane normal
after the azimuth/inclination alignment.
An \code{\link[=angle]{angle()}} object or one coercible to one with \code{as_angle(roll, ...)}.
Defaults to \code{angle(0)} (no roll), which preserves the azimuth/inclination convention.}

\item{decreasing}{If \code{TRUE} sort/order closest first instead of farthest
first.}
}
\value{
\code{painter_order()} returns an integer vector of indices.

\code{sort.Coord2D()} and \code{sort.Coord3D()} return a sorted object of the
same class as \code{x}.
}
\description{
\code{painter_depth()} computes depth values for use with the painter's algorithm.
For parallel (orthographic or oblique) projections, depth is a dot product
of the point with the projection direction vector derived from the plane and
oblique parameters.
\code{painter_order()} wraps \code{\link[=painter_depth]{painter_depth()}} and \code{\link[=order]{order()}} to return the integer
indices that sort objects farthest-first (the draw order for the painter's
algorithm).
\code{sort.Coord2D()} and \code{sort.Coord3D()} wrap
\code{painter_order()} to return the sorted object directly.
\code{sort.Coord2D()} also handles \link{Segment2D} objects via inheritance.
}
\examples{
# Coord2D: oblique projection with scale = 1, alpha = 45 degrees
p <- as_coord2d(x = c(1, 2, 3), y = c(3, 1, 2))
painter_depth(p)
painter_order(p)
sort(p)

# Coord3D: orthographic projection onto xy-plane (depth = z)
p3 <- as_coord3d(x = 1:3, y = 1:3, z = c(3, 1, 2))
painter_depth(p3)
painter_order(p3)
sort(p3)

# Segment2D: depth/order at midpoints
p1 <- as_coord2d(x = c(0, 2), y = c(0, 0))
p2 <- as_coord2d(x = c(2, 2), y = c(0, 2))
s <- as_segment2d(p1, p2 = p2)
painter_depth(s)
painter_order(s)
sort(s)

# Polygon2D: depth/order of each edge
vertices <- as_coord2d(x = c(0, 0.5, 1, 0.5), y = c(0.5, 1, 0.5, 0))
poly <- as_polygon2d(vertices)
painter_depth(poly)
painter_order(poly)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distance.R
\name{distance3d}
\alias{distance3d}
\title{3D Euclidean distances}
\usage{
distance3d(x, y)
}
\arguments{
\item{x}{Either a \link{Coord3D} or \link{Plane3D} class object}

\item{y}{Either a \link{Coord3D} or \link{Plane3D} class object}
}
\description{
\code{distance3d()} computes 3D Euclidean distances.
}
\examples{
p <- as_coord3d(x = 1:4, y = 1:4, z = 1:4)
distance3d(p, as_coord3d("origin"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distance.R
\name{distance1d}
\alias{distance1d}
\title{1D Euclidean distances}
\usage{
distance1d(x, y)
}
\arguments{
\item{x}{Either a \link{Coord1D} or \link{Point1D} class object}

\item{y}{Either a \link{Coord1D} or \link{Point1D} class object}
}
\description{
\code{distance1d()} computes 1D Euclidean distances.
}
\examples{
p <- as_coord1d(x = 1:4)
distance1d(p, as_coord1d(0))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/normal.R
\name{normal2d}
\alias{normal2d}
\alias{normal2d.Coord2D}
\alias{normal2d.Line2D}
\title{2D normal vectors}
\usage{
normal2d(x, ...)

\method{normal2d}{Coord2D}(x, ..., normalize = TRUE)

\method{normal2d}{Line2D}(x, ..., normalize = TRUE)
}
\arguments{
\item{x}{Object to compute a 2D normal vector for
such as a \link{Line2D} object.}

\item{...}{Passed to or from other methods.}

\item{normalize}{If \code{TRUE} coerce to a normalize vector}
}
\value{
A \link{Coord2D} (normal) vector
}
\description{
\code{normal2d()} is an S3 generic that computes a 2D normal vector.
}
\examples{
  p <- as_coord2d(x = 2, y = 3)
  normal2d(p)
  normal2d(p, normalize = FALSE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rectangle_polygon2d.R
\name{rectangle_polygon2d}
\alias{rectangle_polygon2d}
\title{Rectangle polygon}
\usage{
rectangle_polygon2d(width = 1, height = 1, x = 0, y = 0, theta = degrees(0))
}
\arguments{
\item{width}{Width of the rectangle (default \code{1}).}

\item{height}{Height of the rectangle (default \code{1}).}

\item{x}{X-coordinate of the center (default \code{0}).}

\item{y}{Y-coordinate of the center (default \code{0}).}

\item{theta}{Rotation angle (default \code{degrees(0)}).
Will be coerced by \code{\link[=degrees]{degrees()}}.}
}
\value{
A \link{Polygon2D} object with \verb{$is_convex == TRUE}.
}
\description{
\code{rectangle_polygon2d()} creates a rectangle \link{Polygon2D} object
centered at \verb{(x, y)}.
}
\examples{
# A unit square
p <- rectangle_polygon2d()
p$is_convex
plot(p)

# A rotated rectangle
p2 <- rectangle_polygon2d(width = 2, height = 1, theta = degrees(45))
plot(p2)
}
\seealso{
\code{\link[=isotoxal_2ngon_polygon2d]{isotoxal_2ngon_polygon2d()}} for isotoxal star polygons.
\code{\link[=regular_ngon_polygon2d]{regular_ngon_polygon2d()}} for regular convex polygons.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/affine.R
\name{transform2d}
\alias{transform2d}
\alias{permute2d}
\alias{project2d}
\alias{reflect2d}
\alias{rotate2d}
\alias{scale2d}
\alias{shear2d}
\alias{translate2d}
\title{2D affine transformation matrices}
\usage{
transform2d(mat = diag(3L))

permute2d(permutation = c("xy", "yx"))

project2d(line = as_line2d("x-axis"), ..., scale = 0)

reflect2d(line = as_line2d("x-axis"), ...)

rotate2d(theta = angle(0), ...)

scale2d(x_scale = 1, y_scale = x_scale)

shear2d(xy_shear = 0, yx_shear = 0)

translate2d(x = as_coord2d(0, 0), ...)
}
\arguments{
\item{mat}{A 3x3 matrix representing a post-multiplied affine transformation matrix.
The last \strong{column} must be equal to \code{c(0, 0, 1)}.
If the last \strong{row} is \code{c(0, 0, 1)} you may need to transpose it
to convert it from a pre-multiplied affine transformation matrix to a post-multiplied one.
If a 2x2 matrix (such as a 2x2 post-multiplied 2D rotation matrix)
we'll quietly add a final column/row equal to \code{c(0, 0, 1)}.}

\item{permutation}{Either "xy" (no permutation) or "yx" (permute x and y axes)}

\item{line}{A \link{Line2D} object of length one representing the line
you with to reflect across or project to or an object coercible to one by \code{as_line2d(line, ...)}
such as "x-axis" or "y-axis".}

\item{...}{Passed to \code{\link[=as_angle]{as_angle()}} or \code{\link[=as_coord2d]{as_coord2d()}}.}

\item{scale}{Oblique projection scale factor.
A degenerate \code{0} value indicates an orthogonal projection.}

\item{theta}{An \code{\link[=angle]{angle()}} object of length one or an object coercible to one by \code{as_angle(theta, ...)}.}

\item{x_scale}{Scaling factor to apply to x coordinates}

\item{y_scale}{Scaling factor to apply to y coordinates}

\item{xy_shear}{Horizontal shear factor: \code{x = x + xy_shear * y}}

\item{yx_shear}{Vertical shear factor: \code{y = yx_shear * x + y}}

\item{x}{A \link{Coord2D} object of length one or an object coercible to one by \code{as_coord2d(x, ...)}.}
}
\value{
A 3x3 post-multiplied affine transformation matrix with classes "transform2d" and "at_matrix"
}
\description{
\code{transform2d()}, \code{project2d()}, \code{reflect2d()}, \code{rotate2d()}, \code{scale2d()}, \code{shear2d()},
and \code{translate2d()} create 2D affine transformation matrix objects.
}
\details{
\describe{
\item{\code{transform2d()}}{User supplied (post-multiplied) affine transformation matrix}.
\item{\code{project2d()}}{Oblique vector projections onto a line parameterized by
an oblique projection scale factor.
A (degenerate) scale factor of zero results in an orthogonal projection.}
\item{\code{reflect2d()}}{Reflections across a line.
To "flip" across both the x-axis and the y-axis use \code{scale2d(-1)}.}
\item{\code{rotate2d()}}{Rotations around the origin parameterized by an \code{\link[=angle]{angle()}}.}
\item{\code{scale2d()}}{Scale the x-coordinates and/or the y-coordinates by multiplicative scale factors.}
\item{\code{shear2d()}}{Shear the x-coordinates and/or the y-coordinates using shear factors.}
\item{\code{translate2d()}}{Translate the coordinates by a \link{Coord2D} class object parameter.}
}

\code{transform2d()} 2D affine transformation matrix objects are meant to be
post-multiplied and therefore should \strong{not} be multiplied in reverse order.
Note the \link{Coord2D} class object methods auto-pre-multiply affine transformations
when "method chaining" so pre-multiplying affine transformation matrices
to do a single cumulative transformation instead of a method chain of multiple transformations
will not improve performance as much as it does in other R packages.

To convert a pre-multiplied 2D affine transformation matrix to a post-multiplied one
simply compute its transpose using \code{\link[=t]{t()}}.  To get an inverse transformation matrix
from an existing transformation matrix that does the opposite transformations
simply compute its inverse using \code{\link[=solve]{solve()}}.
}
\examples{
p <- as_coord2d(x = sample(1:10, 3), y = sample(1:10, 3))

# {affiner} affine transformation matrices are post-multiplied
# and therefore should **not** go in reverse order
mat <- transform2d(diag(3)) \%*\%
         reflect2d(as_coord2d(-1, 1)) \%*\%
         rotate2d(90, "degrees") \%*\%
         scale2d(1, 2) \%*\%
         shear2d(0.5, 0.5) \%*\%
         translate2d(x = -1, y = -1)
p1 <- p$
  clone()$
  transform(mat)

# The equivalent result applying affine transformations via method chaining
p2 <- p$
  clone()$
  transform(diag(3L))$
  reflect(as_coord2d(-1, 1))$
  rotate(90, "degrees")$
  scale(1, 2)$
  shear(0.5, 0.5)$
  translate(x = -1, y = -1)

all.equal(p1, p2)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/segment.R
\name{Segment2D}
\alias{Segment2D}
\title{2D line segment R6 Class}
\description{
\code{Segment2D} is an \code{\link[R6:R6Class]{R6::R6Class()}} object representing a vector of two-dimensional line segments.
It inherits from \link{Coord2D} using the first endpoint \code{p1} as the coordinate data, so all transformation methods (e.g. \verb{$rotate()}, \verb{$translate()}) and arithmetic operators work directly on the segment vector.
The full edge vector to the second endpoint is stored privately and kept consistent under every transformation.
}
\examples{
p1 <- as_coord2d(x = c(0, 1), y = c(0, 0))
p2 <- as_coord2d(x = c(1, 1), y = c(0, 1))
s <- as_segment2d(p1, p2 = p2)
print(s)
s$p1
s$p2
s$mid_point
length(s)
s[1]
}
\section{Super class}{
\code{\link[affiner:Coord2D]{Coord2D}} -> \code{Segment2D}
}
\section{Active bindings}{
  \if{html}{\out{<div class="r6-active-bindings">}}
  \describe{
    \item{\code{p1}}{A \link{Coord2D} object of the first endpoints.}

    \item{\code{p2}}{A \link{Coord2D} object of the second endpoints.}

    \item{\code{mid_point}}{A \link{Coord2D} object of the segment midpoints
(computed on demand and cached until the next transformation).}
  }
  \if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-Segment2D-initialize}{\code{Segment2D$new()}}
    \item \href{#method-Segment2D-print}{\code{Segment2D$print()}}
    \item \href{#method-Segment2D-transform}{\code{Segment2D$transform()}}
    \item \href{#method-Segment2D-clone}{\code{Segment2D$clone()}}
  }
}
\if{html}{\out{<details><summary>Inherited methods</summary>
<ul>
  <li><span class="pkg-link" data-pkg="affiner" data-topic="Coord2D" data-id="permute"><a href='../../affiner/html/Coord2D.html#method-Coord2D-permute'><code>Coord2D$permute()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="affiner" data-topic="Coord2D" data-id="project"><a href='../../affiner/html/Coord2D.html#method-Coord2D-project'><code>Coord2D$project()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="affiner" data-topic="Coord2D" data-id="reflect"><a href='../../affiner/html/Coord2D.html#method-Coord2D-reflect'><code>Coord2D$reflect()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="affiner" data-topic="Coord2D" data-id="rotate"><a href='../../affiner/html/Coord2D.html#method-Coord2D-rotate'><code>Coord2D$rotate()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="affiner" data-topic="Coord2D" data-id="scale"><a href='../../affiner/html/Coord2D.html#method-Coord2D-scale'><code>Coord2D$scale()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="affiner" data-topic="Coord2D" data-id="shear"><a href='../../affiner/html/Coord2D.html#method-Coord2D-shear'><code>Coord2D$shear()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="affiner" data-topic="Coord2D" data-id="translate"><a href='../../affiner/html/Coord2D.html#method-Coord2D-translate'><code>Coord2D$translate()</code></a></span></li>
</ul>
</details>}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Segment2D-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-Segment2D-initialize}{}}}
\subsection{\code{Segment2D$new()}}{
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Segment2D$new(xyw, vec)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{xyw}}{A matrix with three columns for homogeneous \code{p1} coordinates.
Column names should be \code{"x"}, \code{"y"}, and \code{"w"}.}
      \item{\code{vec}}{A two-column matrix of edge vectors \code{p2 - p1}.
Column names should be \code{"x"} and \code{"y"}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Segment2D-print"></a>}}
\if{latex}{\out{\hypertarget{method-Segment2D-print}{}}}
\subsection{\code{Segment2D$print()}}{
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Segment2D$print(n = NULL, ...)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{n}}{Number of segments to print.  If \code{NULL} print all.}
      \item{\code{...}}{Passed to \code{\link[=format.default]{format.default()}}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Segment2D-transform"></a>}}
\if{latex}{\out{\hypertarget{method-Segment2D-transform}{}}}
\subsection{\code{Segment2D$transform()}}{
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Segment2D$transform(mat = transform2d())}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{mat}}{A 3x3 matrix representing a post-multiplied affine transformation matrix.
The last \strong{column} must be equal to \code{c(0, 0, 1)}.
If the last \strong{row} is \code{c(0, 0, 1)} you may need to transpose it
to convert it from a pre-multiplied affine transformation matrix to a post-multiplied one.
If a 2x2 matrix (such as a 2x2 post-multiplied 2D rotation matrix)
we'll quietly add a final column/row equal to \code{c(0, 0, 1)}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Segment2D-clone"></a>}}
\if{latex}{\out{\hypertarget{method-Segment2D-clone}{}}}
\subsection{\code{Segment2D$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Segment2D$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/angle.R
\name{is_angle}
\alias{is_angle}
\title{Test whether an object is an angle vector}
\usage{
is_angle(x)
}
\arguments{
\item{x}{An object}
}
\value{
A logical value
}
\description{
\code{is_angle()} tests whether an object is an angle vector
}
\examples{
a <- angle(180, "degrees")
is_angle(a)
is_angle(pi)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/angle.R
\name{angle-methods}
\alias{angle-methods}
\alias{as.double.angle}
\alias{as.complex.angle}
\alias{format.angle}
\alias{print.angle}
\alias{abs.angle}
\title{Implemented base methods for angle vectors}
\usage{
\method{as.double}{angle}(x, unit = angular_unit(x), ...)

\method{as.complex}{angle}(x, modulus = 1, ...)

\method{format}{angle}(x, unit = angular_unit(x), ..., use_unicode = is_utf8_output())

\method{print}{angle}(x, unit = angular_unit(x), ..., use_unicode = is_utf8_output())

\method{abs}{angle}(x)
}
\arguments{
\item{x}{\code{\link[=angle]{angle()}} vector}

\item{unit}{A string of the desired angular unit.  Supports the following strings
(note we ignore any punctuation and space characters as well as any trailing \code{s}'s
e.g. "half turns" will be treated as equivalent to "halfturn"):
\itemize{
\item "deg" or "degree"
\item "half-revolution", "half-turn", or "pi-radian"
\item "gon", "grad", "grade", or "gradian"
\item "rad" or "radian"
\item "rev", "revolution", "tr", or "turn"
}}

\item{...}{Passed to \code{\link[=print.default]{print.default()}}}

\item{modulus}{Numeric vector representing the complex numbers' modulus}

\item{use_unicode}{If \code{TRUE} use Unicode symbols as appropriate.}
}
\value{
Typical values as usually returned by these base generics.
}
\description{
We implemented methods for several base generics for the \code{\link[=angle]{angle()}} vectors.
}
\details{
\itemize{
\item Mathematical \link{Ops} (in particular \code{+} and \code{-})
for two angle vectors will (if necessary)
set the second vector's \code{\link[=angular_unit]{angular_unit()}} to match the first.
\item \code{\link[=as.numeric]{as.numeric()}} takes a \code{unit} argument which can be used to convert angles into other angular units
e.g. \code{angle(x, "degrees") |> as.numeric("radians")} to cast a numeric vector \code{x} from degrees to radians.
\item \code{\link[=abs]{abs()}} will calculate the angle modulo full turns.
\item Use \code{\link[=is_congruent]{is_congruent()}} to test if two angles are congruent instead of \code{==} or \code{all.equal()}.
\item Not all implemented methods are documented here and since \code{\link[=angle]{angle()}} is a
\code{\link[=numeric]{numeric()}} class many other S3 generics
besides the explicitly implemented ones should also work with it.
}
}
\examples{
  # Two "congruent" angles
  a1 <- angle(180, "degrees")
  a2 <- angle(pi, "radians")

  print(a1)
  print(a1, unit = "radians")
  print(a1, unit = "pi-radians")

  cos(a1)
  sin(a1)
  tan(a1)

  # mathematical operations will coerce second `angle()` object to
  # same `angular_unit()` as the first one
  a1 + a2
  a1 - a2

  as.numeric(a1)
  as.numeric(a1, "radians")
  as.numeric(a1, "turns")

  # Use `is_congruent()` to check if two angles are "congruent"
  a1 == a2
  isTRUE(all.equal(a1, a2))
  is_congruent(a1, a2)
  is_congruent(a1, a2, mod_turns = FALSE)
  a3 <- angle(-180, "degrees") # Only congruent modulus full turns
  a1 == a3
  isTRUE(all.equal(a1, a2))
  is_congruent(a1, a3)
  is_congruent(a1, a3, mod_turns = FALSE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coord-methods.R
\name{centroid}
\alias{centroid}
\alias{mean.Coord1D}
\alias{mean.Coord2D}
\alias{mean.Coord3D}
\title{Compute centroids of coordinates}
\usage{
\method{mean}{Coord1D}(x, ...)

\method{mean}{Coord2D}(x, ...)

\method{mean}{Coord3D}(x, ...)
}
\arguments{
\item{x}{A \link{Coord1D}, \link{Coord2D}, or \link{Coord3D} object}

\item{...}{Passed to \code{\link[base:mean]{base::mean()}}}
}
\value{
A \link{Coord1D}, \link{Coord2D}, or \link{Coord3D} class object of length one
}
\description{
\code{mean()}computes centroids for \link{Coord1D}, \link{Coord2D}, and \link{Coord3D} class objects
}
\examples{
p <- as_coord2d(x = 1:4, y = 1:4)
print(mean(p))
print(sum(p) / length(p)) # less efficient alternative

p <- as_coord3d(x = 1:4, y = 1:4, z = 1:4)
print(mean(p))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/R_to_AA.R
\name{rotate3d_to_AA}
\alias{rotate3d_to_AA}
\title{Convert from 3D rotation matrix to axis-angle representation.}
\usage{
rotate3d_to_AA(
  mat = diag(4),
  unit = getOption("affiner_angular_unit", "degrees")
)
}
\arguments{
\item{mat}{3D rotation matrix (post-multiplied).
If you have a pre-multiplied rotation matrix
simply transpose it with \code{\link[=t]{t()}} to get a post-multiplied rotation matrix.}

\item{unit}{A string of the desired angular unit.  Supports the following strings
(note we ignore any punctuation and space characters as well as any trailing \code{s}'s
e.g. "half turns" will be treated as equivalent to "halfturn"):
\itemize{
\item "deg" or "degree"
\item "half-revolution", "half-turn", or "pi-radian"
\item "gon", "grad", "grade", or "gradian"
\item "rad" or "radian"
\item "rev", "revolution", "tr", or "turn"
}}
}
\description{
\code{rotate3d_to_AA()} converts from (post-multiplied) rotation matrix
to an axis-angle representation of 3D rotations.
}
\examples{
 # axis-angle representation of 90 degree rotation about the x-axis
 rotate3d_to_AA(rotate3d("x-axis", 90, unit = "degrees"))

 # find Axis-Angle representation of first rotating about x-axis 180 degrees
 # and then rotating about z-axis 45 degrees
 R <- rotate3d("x-axis", 180, unit = "degrees") \%*\%
        rotate3d("z-axis", 45, unit = "degrees")
 AA <- rotate3d_to_AA(R)

 # Can use `rotate3d()` to convert back to rotation matrix representation
 all.equal(R, do.call(rotate3d, AA))

}
\seealso{
\url{https://en.wikipedia.org/wiki/Axis-angle_representation} for more details
about the Axis-angle representation of 3D rotations.
\code{\link[=rotate3d]{rotate3d()}} can be used to convert from an axis-angle representation to a rotation matrix.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ellipse.R
\name{Ellipse2D}
\alias{Ellipse2D}
\title{2D ellipse R6 Class}
\description{
\code{Ellipse2D} is an \code{\link[R6:R6Class]{R6::R6Class()}} object representing one or more
two-dimensional ellipses.  It inherits from \link{Coord2D} so its center(s)
can be used with \verb{$x} / \verb{$y} / \verb{$xyw} etc.  The semi-axes \code{rx}/\code{ry}
and orientation angle \code{theta} are updated automatically by each
transformation method.
}
\details{
When \code{rx == ry} (within floating-point tolerance) the ellipse is a
circle, which can be tested with the \verb{$is_circle} active binding.
}
\examples{
# An ellipse
e <- as_ellipse2d(as_coord2d(0, 0), rx = 2, ry = 1, theta = degrees(45))
print(e)
e$is_circle

# A circle (special case)
c1 <- as_ellipse2d(as_coord2d(0.5, 0.5), rx = 0.5)
print(c1)
c1$is_circle
}
\section{Super class}{
\code{\link[affiner:Coord2D]{Coord2D}} -> \code{Ellipse2D}
}
\section{Active bindings}{
  \if{html}{\out{<div class="r6-active-bindings">}}
  \describe{
    \item{\code{rx}}{Numeric vector of x-axis semi-radii (in the ellipse local frame).}

    \item{\code{ry}}{Numeric vector of y-axis semi-radii (in the ellipse local frame).}

    \item{\code{theta}}{An \code{\link[=angle]{angle()}} vector of rotation angles of the ellipse
x-axis relative to the global x-axis.}

    \item{\code{is_circle}}{Logical vector; \code{TRUE} for each ellipse where
\code{rx == ry} within floating-point tolerance.}
  }
  \if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-Ellipse2D-initialize}{\code{Ellipse2D$new()}}
    \item \href{#method-Ellipse2D-print}{\code{Ellipse2D$print()}}
    \item \href{#method-Ellipse2D-transform}{\code{Ellipse2D$transform()}}
    \item \href{#method-Ellipse2D-clone}{\code{Ellipse2D$clone()}}
  }
}
\if{html}{\out{<details><summary>Inherited methods</summary>
<ul>
  <li><span class="pkg-link" data-pkg="affiner" data-topic="Coord2D" data-id="permute"><a href='../../affiner/html/Coord2D.html#method-Coord2D-permute'><code>Coord2D$permute()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="affiner" data-topic="Coord2D" data-id="project"><a href='../../affiner/html/Coord2D.html#method-Coord2D-project'><code>Coord2D$project()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="affiner" data-topic="Coord2D" data-id="reflect"><a href='../../affiner/html/Coord2D.html#method-Coord2D-reflect'><code>Coord2D$reflect()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="affiner" data-topic="Coord2D" data-id="rotate"><a href='../../affiner/html/Coord2D.html#method-Coord2D-rotate'><code>Coord2D$rotate()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="affiner" data-topic="Coord2D" data-id="scale"><a href='../../affiner/html/Coord2D.html#method-Coord2D-scale'><code>Coord2D$scale()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="affiner" data-topic="Coord2D" data-id="shear"><a href='../../affiner/html/Coord2D.html#method-Coord2D-shear'><code>Coord2D$shear()</code></a></span></li>
  <li><span class="pkg-link" data-pkg="affiner" data-topic="Coord2D" data-id="translate"><a href='../../affiner/html/Coord2D.html#method-Coord2D-translate'><code>Coord2D$translate()</code></a></span></li>
</ul>
</details>}}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Ellipse2D-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-Ellipse2D-initialize}{}}}
\subsection{\code{Ellipse2D$new()}}{
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Ellipse2D$new(xyw, rx, ry, theta)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{xyw}}{A matrix with three columns for homogeneous center coordinates
(\code{"x"}, \code{"y"}, \code{"w"}).}
      \item{\code{rx}}{Numeric vector of x-axis semi-radii.}
      \item{\code{ry}}{Numeric vector of y-axis semi-radii.}
      \item{\code{theta}}{An \code{\link[=angle]{angle()}} vector (or numeric, interpreted using
the default angular unit) of rotation angles.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Ellipse2D-print"></a>}}
\if{latex}{\out{\hypertarget{method-Ellipse2D-print}{}}}
\subsection{\code{Ellipse2D$print()}}{
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Ellipse2D$print(n = NULL, ...)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{n}}{Number of ellipses to print.  If \code{NULL} print all.}
      \item{\code{...}}{Passed to \code{\link[=format.default]{format.default()}}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Ellipse2D-transform"></a>}}
\if{latex}{\out{\hypertarget{method-Ellipse2D-transform}{}}}
\subsection{\code{Ellipse2D$transform()}}{
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Ellipse2D$transform(mat = transform2d())}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{mat}}{A 3x3 matrix representing a post-multiplied affine transformation matrix.
The last \strong{column} must be equal to \code{c(0, 0, 1)}.
If the last \strong{row} is \code{c(0, 0, 1)} you may need to transpose it
to convert it from a pre-multiplied affine transformation matrix to a post-multiplied one.
If a 2x2 matrix (such as a 2x2 post-multiplied 2D rotation matrix)
we'll quietly add a final column/row equal to \code{c(0, 0, 1)}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Ellipse2D-clone"></a>}}
\if{latex}{\out{\hypertarget{method-Ellipse2D-clone}{}}}
\subsection{\code{Ellipse2D$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Ellipse2D$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/intersection.R
\name{intersection}
\alias{intersection}
\alias{intersection.Point1D}
\alias{intersection.Line2D}
\alias{intersection.Plane3D}
\alias{intersection.Coord1D}
\alias{intersection.Coord2D}
\alias{intersection.Ellipse2D}
\alias{intersection.Coord3D}
\title{The intersection of two objects.}
\usage{
intersection(x, y, ...)

\method{intersection}{Point1D}(x, y, ..., tolerance = sqrt(.Machine$double.eps))

\method{intersection}{Line2D}(x, y, ..., tolerance = sqrt(.Machine$double.eps))

\method{intersection}{Plane3D}(x, y, ..., tolerance = sqrt(.Machine$double.eps))

\method{intersection}{Coord1D}(x, y, ..., tolerance = sqrt(.Machine$double.eps))

\method{intersection}{Coord2D}(x, y, ..., tolerance = sqrt(.Machine$double.eps))

\method{intersection}{Ellipse2D}(x, y, ..., tolerance = sqrt(.Machine$double.eps))

\method{intersection}{Coord3D}(x, y, ..., tolerance = sqrt(.Machine$double.eps))
}
\arguments{
\item{x, y}{The two objects to compute intersection for.}

\item{...}{Passed to other methods (or ignored).}

\item{tolerance}{Numerics with differences smaller
than \code{tolerance} will be considered \dQuote{equivalent}.}
}
\value{
A list of the object intersections (or \code{NULL} if no intersection).
For \link{Line2D}-\link{Ellipse2D} intersections each list element is a \link{Coord2D}
of length 2 (two crossing points), length 1 (tangent), or \code{NULL}
(no intersection).
}
\description{
\code{intersection()} is an S3 method that returns the intersection of two objects.
}
\details{
\code{affiner::intersection()} has the same S3 signature as
\code{euclid::intersection()} (so it shouldn't matter if one masks the other).
}
\examples{
line1 <- as_line2d("x-axis")
line2 <- as_line2d("y-axis")
line3 <- as_line2d(a = 0, b = 1, c = 2) # y + 2 = 0
intersection(line1, line1)
intersection(line1, line2)
intersection(line1, line3)

# Unit circle vs x-axis: two points at (1, 0) and (-1, 0)
circ <- as_ellipse2d(as_coord2d("origin"), r = 1)
intersection(circ, line1)

# Tangent: circle vs line y = 1 (touches top of circle)
line_top <- as_line2d(a = 0, b = 1, c = -1) # y - 1 = 0
intersection(circ, line_top)

# No intersection: circle vs y = 2
line_above <- as_line2d(a = 0, b = 1, c = -2) # y - 2 = 0
intersection(circ, line_above)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as_coord.R
\name{as_coord2d}
\alias{as_coord2d}
\alias{as_coord2d.angle}
\alias{as_coord2d.character}
\alias{as_coord2d.complex}
\alias{as_coord2d.Coord3D}
\alias{as_coord2d.data.frame}
\alias{as_coord2d.list}
\alias{as_coord2d.matrix}
\alias{as_coord2d.numeric}
\alias{as_coord2d.Coord2D}
\title{Cast to coord2d object}
\usage{
as_coord2d(x, ...)

\method{as_coord2d}{angle}(x, radius = 1, ...)

\method{as_coord2d}{character}(x, ...)

\method{as_coord2d}{complex}(x, ...)

\method{as_coord2d}{Coord3D}(
  x,
  permutation = c("xyz", "xzy", "yxz", "yzx", "zyx", "zxy"),
  ...,
  plane = as_plane3d("xy-plane"),
  scale = 0,
  alpha = angle(45, "degrees"),
  roll = angle(0, "degrees")
)

\method{as_coord2d}{data.frame}(x, ...)

\method{as_coord2d}{list}(x, ...)

\method{as_coord2d}{matrix}(x, ...)

\method{as_coord2d}{numeric}(x, y = rep_len(0, length(x)), ...)

\method{as_coord2d}{Coord2D}(x, ...)
}
\arguments{
\item{x}{An object that can be cast to a \link{Coord2D} class object
such as a matrix or data frame of coordinates.}

\item{...}{Further arguments passed to or from other methods}

\item{radius}{A numeric vector of radial distances.}

\item{permutation}{Either "xyz" (no permutation), "xzy" (permute y and z axes),
"yxz" (permute x and y axes), "yzx" (x becomes z, y becomes x, z becomes y),
"zxy" (x becomes y, y becomes z, z becomes x), "zyx" (permute x and z axes).
This permutation is applied before the (oblique) projection.}

\item{plane}{A \link{Plane3D} class object representing the plane
you wish to project to or an object coercible to one using \code{as_plane3d(plane, ...)}
such as "xy-plane", "xz-plane", or "yz-plane".}

\item{scale}{Oblique projection foreshortening scale factor.
A (degenerate) \code{0} value indicates an orthographic projection.
A value of \code{0.5} is used by a \dQuote{cabinet projection}
while a value of \code{1.0} is used by a \dQuote{cavalier projection}.}

\item{alpha}{Oblique projection angle (the angle the third axis is projected going off at).
An \code{\link[=angle]{angle()}} object or one coercible to one with \code{as_angle(alpha, ...)}.
Popular angles are 45 degrees, 60 degrees, and \code{arctangent(2)} degrees.}

\item{roll}{Rotation of the in-plane coordinate frame around the plane normal
after the azimuth/inclination alignment.
An \code{\link[=angle]{angle()}} object or one coercible to one with \code{as_angle(roll, ...)}.
Defaults to \code{angle(0)} (no roll), which preserves the azimuth/inclination convention.}

\item{y}{Numeric vector of y-coordinates to be used.}
}
\value{
A \link{Coord2D} class object
}
\description{
\code{as_coord2d()} casts to a \link{Coord2D} class object
}
\examples{
df <- data.frame(x = sample.int(10, 3),
                 y = sample.int(10, 3))
as_coord2d(df)
as_coord2d(complex(real = 3, imaginary = 2))
as_coord2d(angle(90, "degrees"), radius = 2)
as_coord2d(as_coord3d(1, 2, 2), alpha = degrees(90), scale = 0.5)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/has_intersection.R
\name{has_intersection}
\alias{has_intersection}
\alias{has_intersection.default}
\alias{has_intersection.Point1D}
\alias{has_intersection.Line2D}
\alias{has_intersection.Plane3D}
\title{Whether two objects intersect}
\usage{
has_intersection(x, y, ...)

\method{has_intersection}{default}(x, y, ...)

\method{has_intersection}{Point1D}(x, y, ..., tolerance = sqrt(.Machine$double.eps))

\method{has_intersection}{Line2D}(x, y, ..., tolerance = sqrt(.Machine$double.eps))

\method{has_intersection}{Plane3D}(x, y, ..., tolerance = sqrt(.Machine$double.eps))
}
\arguments{
\item{x, y}{The two objects to check if they intersect.}

\item{...}{Passed to other methods (or ignored).}

\item{tolerance}{Numerics with differences smaller
than \code{tolerance} will be considered \dQuote{equivalent}.}
}
\value{
A logical vector.
}
\description{
\code{has_intersection()} is an S3 method that returns whether two objects intersect.
}
\details{
\code{affiner::has_intersection()} has the same S3 signature and default method as
\code{euclid::has_intersection()} (so it shouldn't matter if one masks the other).
}
\examples{
line1 <- as_line2d("x-axis")
line2 <- as_line2d("y-axis")
line3 <- as_line2d(a = 0, b = 1, c = 2) # y + 2 = 0
has_intersection(line1, line1)
has_intersection(line1, line2)
has_intersection(line1, line3)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/is_congruent.R
\name{is_congruent}
\alias{is_congruent}
\alias{is_congruent.numeric}
\alias{is_congruent.angle}
\title{Test whether two objects are congruent}
\usage{
is_congruent(x, y, ...)

\method{is_congruent}{numeric}(x, y, ..., tolerance = sqrt(.Machine$double.eps))

\method{is_congruent}{angle}(
  x,
  y,
  ...,
  mod_turns = TRUE,
  tolerance = sqrt(.Machine$double.eps)
)
}
\arguments{
\item{x, y}{Two objects to test whether they are \dQuote{"congruent"}.}

\item{...}{Further arguments passed to or from other methods.}

\item{tolerance}{Angles (coerced to half-turns) or numerics with differences smaller
than \code{tolerance} will be considered \dQuote{congruent}.}

\item{mod_turns}{If \code{TRUE} angles that are congruent modulo full turns will be considered \dQuote{congruent}.}
}
\value{
A logical vector
}
\description{
\code{is_congruent()} is a S3 generic that tests whether two different objects are \dQuote{congruent}.
The \code{is_congruent()} method for \code{\link[=angle]{angle()}} classes tests whether two angles are congruent.
}
\examples{
  # Use `is_congruent()` to check if two angles are "congruent"
  a1 <- angle(180, "degrees")
  a2 <- angle(pi, "radians")
  a3 <- angle(-180, "degrees") # Only congruent modulus full turns
  a1 == a2
  isTRUE(all.equal(a1, a2))
  is_congruent(a1, a2)
  is_congruent(a1, a2, mod_turns = FALSE)
  a1 == a3
  isTRUE(all.equal(a1, a3))
  is_congruent(a1, a3)
  is_congruent(a1, a3, mod_turns = FALSE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/is_methods.R
\name{is_coord1d}
\alias{is_coord1d}
\title{Test whether an object has a Coord1D class}
\usage{
is_coord1d(x)
}
\arguments{
\item{x}{An object}
}
\value{
A logical value
}
\description{
\code{is_coord1d()} tests whether an object has a "Coord1D" class
}
\examples{
p <- as_coord1d(x = sample.int(10, 3))
is_coord1d(p)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/trigonometry.R
\name{trigonometric-functions}
\alias{trigonometric-functions}
\alias{sine}
\alias{cosine}
\alias{tangent}
\alias{secant}
\alias{cosecant}
\alias{cotangent}
\title{Angle vector aware trigonometric functions}
\usage{
sine(x, unit = getOption("affiner_angular_unit", "degrees"))

cosine(x, unit = getOption("affiner_angular_unit", "degrees"))

tangent(x, unit = getOption("affiner_angular_unit", "degrees"))

secant(x, unit = getOption("affiner_angular_unit", "degrees"))

cosecant(x, unit = getOption("affiner_angular_unit", "degrees"))

cotangent(x, unit = getOption("affiner_angular_unit", "degrees"))
}
\arguments{
\item{x}{An angle vector or an object to convert to it (such as a numeric vector)}

\item{unit}{A string of the desired angular unit.  Supports the following strings
(note we ignore any punctuation and space characters as well as any trailing \code{s}'s
e.g. "half turns" will be treated as equivalent to "halfturn"):
\itemize{
\item "deg" or "degree"
\item "half-revolution", "half-turn", or "pi-radian"
\item "gon", "grad", "grade", or "gradian"
\item "rad" or "radian"
\item "rev", "revolution", "tr", or "turn"
}}
}
\value{
A numeric vector
}
\description{
\code{sine()}, \code{cosine()}, \code{tangent()}, \code{secant()}, \code{cosecant()}, and \code{cotangent()} are
\code{\link[=angle]{angle()}} aware trigonometric functions that allow for a user chosen angular unit.
}
\examples{
sine(pi, "radians")
cosine(180, "degrees")
tangent(0.5, "turns")

a <- angle(0.5, "turns")
secant(a)
cosecant(a)
cotangent(a)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/line_plane.R
\name{as_point1d}
\alias{as_point1d}
\alias{as_point1d.numeric}
\alias{as_point1d.character}
\alias{as_point1d.Coord1D}
\alias{as_point1d.Point1D}
\title{Cast to Point1D object}
\usage{
as_point1d(...)

\method{as_point1d}{numeric}(a, b, ...)

\method{as_point1d}{character}(x, ...)

\method{as_point1d}{Coord1D}(normal, ...)

\method{as_point1d}{Point1D}(point, ...)
}
\arguments{
\item{...}{Passed to other function such as \code{as_coord2d()}.}

\item{a, b}{Numeric vectors that parameterize the point via the equation \code{a * x + b = 0}.
Note this means that \code{x = -b / a}.}

\item{x}{A (character) vector to be cast to a \link{Point1D} object}

\item{normal}{\link{Coord1D} class object.}

\item{point}{A \link{Point1D} object}
}
\description{
\code{as_point1d()} casts to a \link{Point1D} object.
}
\examples{
p1 <- as_point1d(a = 1, b = 0)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/angle.R
\name{angular_unit}
\alias{angular_unit}
\alias{angular_unit<-}
\title{Get/set angular unit of angle vectors}
\usage{
angular_unit(x)

angular_unit(x) <- value
}
\arguments{
\item{x}{An \code{\link[=angle]{angle()}} vector}

\item{value}{A string of the desired angular unit.  See \code{\link[=angle]{angle()}} for supported strings.}
}
\value{
\code{angular_unit()} returns a string of \code{x}'s angular unit.
}
\description{
\code{angular_unit()} gets/sets the angular unit of \code{\link[=angle]{angle()}} vectors.
}
\examples{
a <- angle(seq(0, 360, by = 90), "degrees")
angular_unit(a)
print(a)
angular_unit(a) <- "turns"
angular_unit(a)
print(a)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/is_methods.R
\name{is_ellipse2d}
\alias{is_ellipse2d}
\title{Test whether an object has an Ellipse2D class}
\usage{
is_ellipse2d(x)
}
\arguments{
\item{x}{An object}
}
\value{
A logical value
}
\description{
\code{is_ellipse2d()} tests whether an object has an "Ellipse2D" class
}
\examples{
c1 <- as_ellipse2d(as_coord2d(0.5, 0.5), r = 0.5)
is_ellipse2d(c1)
is_ellipse2d(c1) && all(c1$is_circle)
e1 <- as_ellipse2d(as_coord2d(0, 0), rx = 2, ry = 1)
is_ellipse2d(e1)
is_ellipse2d(e1) && all(e1$is_circle)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/affiner_options.R
\name{affiner_options}
\alias{affiner_options}
\title{Get affiner options}
\usage{
affiner_options(..., default = FALSE)
}
\arguments{
\item{...}{\code{affiner} package options using \code{name = value}.
The return list will use any of these instead of the current/default values.}

\item{default}{If \code{TRUE} return the default values instead of current values.}
}
\value{
A list of option values.
Note this function \strong{does not} set option values itself but
this list can be passed to \code{\link[=options]{options()}}, \code{\link[withr:local_options]{withr::local_options()}}, or \code{\link[withr:with_options]{withr::with_options()}}.
}
\description{
\code{affiner_options()} returns the \code{affiner} package's global options.
}
\examples{
  affiner_options()

  affiner_options(default = TRUE)

  affiner_options(affiner_angular_unit = "pi-radians")
}
\seealso{
\link{affiner} for a high-level description of relevant global options.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as_coord.R
\name{as_coord3d}
\alias{as_coord3d}
\alias{as_coord3d.angle}
\alias{as_coord3d.character}
\alias{as_coord3d.data.frame}
\alias{as_coord3d.list}
\alias{as_coord3d.matrix}
\alias{as_coord3d.numeric}
\alias{as_coord3d.Coord3D}
\alias{as_coord3d.Coord2D}
\title{Cast to coord3d object}
\usage{
as_coord3d(x, ...)

\method{as_coord3d}{angle}(x, radius = 1, inclination = NULL, z = NULL, ...)

\method{as_coord3d}{character}(x, ...)

\method{as_coord3d}{data.frame}(x, ..., z = NULL)

\method{as_coord3d}{list}(x, ..., z = NULL)

\method{as_coord3d}{matrix}(x, ...)

\method{as_coord3d}{numeric}(x, y = rep_len(0, length(x)), z = rep_len(0, length(x)), ...)

\method{as_coord3d}{Coord3D}(x, ...)

\method{as_coord3d}{Coord2D}(x, z = rep_len(0, length(x)), ...)
}
\arguments{
\item{x}{An object that can be cast to a \link{Coord3D} class object
such as a matrix or data frame of coordinates.}

\item{...}{Further arguments passed to or from other methods}

\item{radius}{A numeric vector.  If \code{inclination} is not \code{NULL} represents spherical distances
of spherical coordinates and if \code{z} is not \code{NULL} represents
radial distances of cylindrical coordinates.}

\item{inclination}{Spherical coordinates inclination angle aka polar angle.
\code{x} represents the azimuth aka azimuthal angle.}

\item{z}{Numeric vector of z-coordinates to be used}

\item{y}{Numeric vector of y-coordinates to be used
if \code{hasName(x, "z")} is \code{FALSE}.}
}
\value{
A \link{Coord3D} class object
}
\description{
\code{as_coord3d()} casts to a \link{Coord3D} class object
}
\examples{
as_coord3d(x = 1, y = 2, z = 3)
df <- data.frame(x = sample.int(10, 3),
                 y = sample.int(10, 3),
                 z = sample.int(10, 3))
as_coord3d(df)
# Cylindrical coordinates
as_coord3d(degrees(90), z = 1, radius = 1)
# Spherical coordinates
as_coord3d(degrees(90), inclination = degrees(90), radius = 1)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/affiner-package.R
\docType{package}
\name{affiner-package}
\alias{affiner}
\alias{affiner-package}
\title{affiner: A Finer Way to Render 3D Illustrated Objects in 'grid' Using Affine Transformations}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

Dilate, permute, project, reflect, rotate, shear, and translate 2D and 3D points. Supports parallel projections including oblique projections such as the cabinet projection as well as axonometric projections such as the isometric projection. Use 'grid's "affine transformation" feature to render illustrated flat surfaces.
}
\section{Package options}{

The following affiner function arguments may be set globally via \code{\link[base:options]{base::options()}}:
\describe{
\item{affiner_angular_unit}{The default for the \code{unit} argument used by \code{\link[=angle]{angle()}} and \code{\link[=as_angle]{as_angle()}}.
The default for this option is "degrees".}
\item{affiner_grid_unit}{The default for the \code{unit} argument used by \code{\link[=affine_settings]{affine_settings()}}.
The default for this option is "inches".}
}
The following \code{cli} options may also be of interest:
\describe{
\item{cli.unicode}{Whether UTF-8 character support should be assumed.
Along with \code{\link[=l10n_info]{l10n_info()}} used to determine the default of the
\code{use_unicode} argument of \code{\link[=format.angle]{format.angle()}} and \code{\link[=print.angle]{print.angle()}}.}
}
}

\seealso{
Useful links:
\itemize{
  \item \url{https://trevorldavis.com/R/affiner/}
  \item Report bugs at \url{https://github.com/trevorld/affiner/issues}
}

}
\author{
\strong{Maintainer}: Trevor L. Davis \email{trevor.l.davis@gmail.com} (\href{https://orcid.org/0000-0001-6341-4639}{ORCID})

Authors:
\itemize{
  \item Trevor L. Davis \email{trevor.l.davis@gmail.com} (\href{https://orcid.org/0000-0001-6341-4639}{ORCID})
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coord.R
\name{Coord3D}
\alias{Coord3D}
\title{3D coordinate vector R6 Class}
\description{
\code{Coord3D} is an \code{\link[R6:R6Class]{R6::R6Class()}} object representing three-dimensional points
represented by Cartesian Coordinates.
}
\examples{
p <- as_coord3d(x = rnorm(100, 2), y = rnorm(100, 2), z = rnorm(100, 2))
print(p, n = 10)
pc <- mean(p) # Centroid
# method chained affine transformation matrices are auto-pre-multiplied
p$
  translate(-pc)$
  reflect("xy-plane")$
  rotate("z-axis", degrees(90))$
  print(n = 10)
}
\section{Active bindings}{
  \if{html}{\out{<div class="r6-active-bindings">}}
  \describe{
    \item{\code{xyzw}}{A four-column matrix representing the homogeneous coordinates.
The first three columns are "x", "y", and "z" coordinates
and the fourth column is all ones.}

    \item{\code{x}}{A numeric vector of x-coordinates.}

    \item{\code{y}}{A numeric vector of y-coordinates.}

    \item{\code{z}}{A numeric vector of z-coordinates.}
  }
  \if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-Coord3D-initialize}{\code{Coord3D$new()}}
    \item \href{#method-Coord3D-permute}{\code{Coord3D$permute()}}
    \item \href{#method-Coord3D-print}{\code{Coord3D$print()}}
    \item \href{#method-Coord3D-project}{\code{Coord3D$project()}}
    \item \href{#method-Coord3D-reflect}{\code{Coord3D$reflect()}}
    \item \href{#method-Coord3D-rotate}{\code{Coord3D$rotate()}}
    \item \href{#method-Coord3D-scale}{\code{Coord3D$scale()}}
    \item \href{#method-Coord3D-shear}{\code{Coord3D$shear()}}
    \item \href{#method-Coord3D-translate}{\code{Coord3D$translate()}}
    \item \href{#method-Coord3D-transform}{\code{Coord3D$transform()}}
    \item \href{#method-Coord3D-clone}{\code{Coord3D$clone()}}
  }
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Coord3D-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-Coord3D-initialize}{}}}
\subsection{\code{Coord3D$new()}}{
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Coord3D$new(xyzw)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{xyzw}}{A matrix with four columns representing (homogeneous) coordinates.
The first three columns represent x, y, and z coordinates and
the last column is all ones.
Column names should be "x", "y", "z", and "w".}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Coord3D-permute"></a>}}
\if{latex}{\out{\hypertarget{method-Coord3D-permute}{}}}
\subsection{\code{Coord3D$permute()}}{
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Coord3D$permute(permutation = c("xyz", "xzy", "yxz", "yzx", "zyx", "zxy"))}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{permutation}}{Either "xyz" (no permutation), "xzy" (permute y and z axes),
"yxz" (permute x and y axes), "yzx" (x becomes z, y becomes x, z becomes y),
"zxy" (x becomes y, y becomes z, z becomes x), "zyx" (permute x and z axes)}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Coord3D-print"></a>}}
\if{latex}{\out{\hypertarget{method-Coord3D-print}{}}}
\subsection{\code{Coord3D$print()}}{
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Coord3D$print(n = NULL, ...)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{n}}{Number of coordinates to print.  If \code{NULL} print all of them.}
      \item{\code{...}}{Passed to \code{\link[=format.default]{format.default()}}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Coord3D-project"></a>}}
\if{latex}{\out{\hypertarget{method-Coord3D-project}{}}}
\subsection{\code{Coord3D$project()}}{
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Coord3D$project(
  plane = as_plane3d("xy-plane"),
  ...,
  scale = 0,
  alpha = angle(45, "degrees")
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{plane}}{A \link{Plane3D} object of length one representing the plane
you wish to reflect across or project to or an object coercible to one using \code{as_plane3d(plane, ...)}
such as "xy-plane", "xz-plane", or "yz-plane".}
      \item{\code{...}}{Passed to \code{\link[=project3d]{project3d()}}.}
      \item{\code{scale}}{Oblique projection foreshortening scale factor.
A (degenerate) \code{0} value indicates an orthographic projection.
A value of \code{0.5} is used by a \dQuote{cabinet projection}
while a value of \code{1.0} is used by a \dQuote{cavalier projection}.}
      \item{\code{alpha}}{Oblique projection angle (the angle the third axis is projected going off at).
An \code{\link[=angle]{angle()}} object or one coercible to one with \code{as_angle(alpha, ...)}.
Popular angles are 45 degrees, 60 degrees, and \code{arctangent(2)} degrees.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Coord3D-reflect"></a>}}
\if{latex}{\out{\hypertarget{method-Coord3D-reflect}{}}}
\subsection{\code{Coord3D$reflect()}}{
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Coord3D$reflect(plane = as_plane3d("xy-plane"), ...)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{plane}}{A \link{Plane3D} object of length one representing the plane
you wish to reflect across or project to or an object coercible to one using \code{as_plane3d(plane, ...)}
such as "xy-plane", "xz-plane", or "yz-plane".}
      \item{\code{...}}{Passed to \code{\link[=reflect3d]{reflect3d()}}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Coord3D-rotate"></a>}}
\if{latex}{\out{\hypertarget{method-Coord3D-rotate}{}}}
\subsection{\code{Coord3D$rotate()}}{
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Coord3D$rotate(axis = as_coord3d("z-axis"), theta = angle(0), ...)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{axis}}{A \link{Coord3D} class object or one that can coerced to one by \code{as_coord3d(axis, ...)}.
The \code{axis} represents the axis to be rotated around.}
      \item{\code{theta}}{An \code{\link[=angle]{angle()}} object of length one or an object coercible to one by \code{as_angle(theta, ...)}.}
      \item{\code{...}}{Passed to \code{\link[=rotate3d]{rotate3d()}}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Coord3D-scale"></a>}}
\if{latex}{\out{\hypertarget{method-Coord3D-scale}{}}}
\subsection{\code{Coord3D$scale()}}{
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Coord3D$scale(x_scale = 1, y_scale = x_scale, z_scale = x_scale)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{x_scale}}{Scaling factor to apply to x coordinates}
      \item{\code{y_scale}}{Scaling factor to apply to y coordinates}
      \item{\code{z_scale}}{Scaling factor to apply to z coordinates}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Coord3D-shear"></a>}}
\if{latex}{\out{\hypertarget{method-Coord3D-shear}{}}}
\subsection{\code{Coord3D$shear()}}{
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Coord3D$shear(
  xy_shear = 0,
  xz_shear = 0,
  yx_shear = 0,
  yz_shear = 0,
  zx_shear = 0,
  zy_shear = 0
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{xy_shear}}{Shear factor: \code{x = x + xy_shear * y + xz_shear * z}}
      \item{\code{xz_shear}}{Shear factor: \code{x = x + xy_shear * y + xz_shear * z}}
      \item{\code{yx_shear}}{Shear factor: \code{y = yx_shear * x + y + yz_shear * z}}
      \item{\code{yz_shear}}{Shear factor: \code{y = yx_shear * x + y + yz_shear * z}}
      \item{\code{zx_shear}}{Shear factor: \code{z = zx_shear * x + zy_shear * y + z}}
      \item{\code{zy_shear}}{Shear factor: \code{z = zx_shear * x + zy_shear * y + z}}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Coord3D-translate"></a>}}
\if{latex}{\out{\hypertarget{method-Coord3D-translate}{}}}
\subsection{\code{Coord3D$translate()}}{
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Coord3D$translate(x = as_coord3d(0, 0, 0), ...)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{x}}{A \link{Coord3D} object of length one or an object coercible to one by \code{as_coord3d(x, ...)}.}
      \item{\code{...}}{Passed to \code{as_coord3d(x, ...)} if \code{x} is not a \link{Coord3D} object}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Coord3D-transform"></a>}}
\if{latex}{\out{\hypertarget{method-Coord3D-transform}{}}}
\subsection{\code{Coord3D$transform()}}{
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Coord3D$transform(mat = transform3d())}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{mat}}{A 4x4 matrix representing a post-multiplied affine transformation matrix.
The last \strong{column} must be equal to \code{c(0, 0, 0, 1)}.
If the last \strong{row} is \code{c(0, 0, 0, 1)} you may need to transpose it
to convert it from a pre-multiplied affine transformation matrix to a post-multiplied one.
If a 3x3 matrix (such as a 3x3 post-multiplied 3D rotation matrix)
we'll quietly add a final column/row equal to \code{c(0, 0, 0, 1)}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Coord3D-clone"></a>}}
\if{latex}{\out{\hypertarget{method-Coord3D-clone}{}}}
\subsection{\code{Coord3D$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Coord3D$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/affine.R
\name{transform3d}
\alias{transform3d}
\alias{permute3d}
\alias{project3d}
\alias{reflect3d}
\alias{rotate3d}
\alias{scale3d}
\alias{shear3d}
\alias{translate3d}
\title{3D affine transformation matrices}
\usage{
transform3d(mat = diag(4L))

permute3d(permutation = c("xyz", "xzy", "yxz", "yzx", "zyx", "zxy"))

project3d(
  plane = as_plane3d("xy-plane"),
  ...,
  scale = 0,
  alpha = angle(45, "degrees")
)

reflect3d(plane = as_plane3d("xy-plane"), ...)

rotate3d(axis = as_coord3d("z-axis"), theta = angle(0), ...)

scale3d(x_scale = 1, y_scale = x_scale, z_scale = x_scale)

shear3d(
  xy_shear = 0,
  xz_shear = 0,
  yx_shear = 0,
  yz_shear = 0,
  zx_shear = 0,
  zy_shear = 0
)

translate3d(x = as_coord3d(0, 0, 0), ...)
}
\arguments{
\item{mat}{A 4x4 matrix representing a post-multiplied affine transformation matrix.
The last \strong{column} must be equal to \code{c(0, 0, 0, 1)}.
If the last \strong{row} is \code{c(0, 0, 0, 1)} you may need to transpose it
to convert it from a pre-multiplied affine transformation matrix to a post-multiplied one.
If a 3x3 matrix (such as a 3x3 post-multiplied 3D rotation matrix)
we'll quietly add a final column/row equal to \code{c(0, 0, 0, 1)}.}

\item{permutation}{Either "xyz" (no permutation), "xzy" (permute y and z axes),
"yxz" (permute x and y axes), "yzx" (x becomes z, y becomes x, z becomes y),
"zxy" (x becomes y, y becomes z, z becomes x), "zyx" (permute x and z axes)}

\item{plane}{A \link{Plane3D} object of length one representing the plane
you wish to reflect across or project to or an object coercible to one using \code{as_plane3d(plane, ...)}
such as "xy-plane", "xz-plane", or "yz-plane".}

\item{...}{Passed to \code{\link[=as_angle]{as_angle()}} or \code{\link[=as_coord3d]{as_coord3d()}}.}

\item{scale}{Oblique projection foreshortening scale factor.
A (degenerate) \code{0} value indicates an orthographic projection.
A value of \code{0.5} is used by a \dQuote{cabinet projection}
while a value of \code{1.0} is used by a \dQuote{cavalier projection}.}

\item{alpha}{Oblique projection angle (the angle the third axis is projected going off at).
An \code{\link[=angle]{angle()}} object or one coercible to one with \code{as_angle(alpha, ...)}.
Popular angles are 45 degrees, 60 degrees, and \code{arctangent(2)} degrees.}

\item{axis}{A \link{Coord3D} class object or one that can coerced to one by \code{as_coord3d(axis, ...)}.
The \code{axis} represents the axis to be rotated around.}

\item{theta}{An \code{\link[=angle]{angle()}} object of length one or an object coercible to one by \code{as_angle(theta, ...)}.}

\item{x_scale}{Scaling factor to apply to x coordinates}

\item{y_scale}{Scaling factor to apply to y coordinates}

\item{z_scale}{Scaling factor to apply to z coordinates}

\item{xy_shear}{Shear factor: \code{x = x + xy_shear * y + xz_shear * z}}

\item{xz_shear}{Shear factor: \code{x = x + xy_shear * y + xz_shear * z}}

\item{yx_shear}{Shear factor: \code{y = yx_shear * x + y + yz_shear * z}}

\item{yz_shear}{Shear factor: \code{y = yx_shear * x + y + yz_shear * z}}

\item{zx_shear}{Shear factor: \code{z = zx_shear * x + zy_shear * y + z}}

\item{zy_shear}{Shear factor: \code{z = zx_shear * x + zy_shear * y + z}}

\item{x}{A \link{Coord3D} object of length one or an object coercible to one by \code{as_coord3d(x, ...)}.}
}
\value{
A 4x4 post-multiplied affine transformation matrix with classes "transform3d" and "at_matrix"
}
\description{
\code{transform3d()}, \code{project3d()}, \code{reflect3d()}, \code{rotate3d()}, \code{scale3d()}, \code{shear3d()},
and \code{translate3d()} create 3D affine transformation matrix objects.
}
\details{
\describe{
\item{\code{transform3d()}}{User supplied (post-multiplied) affine transformation matrix}.
\item{\code{scale3d()}}{Scale the x-coordinates and/or the y-coordinates and/or the z-coordinates
by multiplicative scale factors.}
\item{\code{shear3d()}}{Shear the x-coordinates and/or the y-coordinates
and/or the z-coordinates using shear factors.}
\item{\code{translate3d()}}{Translate the coordinates by a \link{Coord3D} class object parameter.}
}

\code{transform3d()} 3D affine transformation matrix objects are meant to be
post-multiplied and therefore should \strong{not} be multiplied in reverse order.
Note the \link{Coord3D} class object methods auto-pre-multiply affine transformations
when "method chaining" so pre-multiplying affine transformation matrices
to do a single cumulative transformation instead of a method chain of multiple transformations
will not improve performance as much as it does in other R packages.

To convert a pre-multiplied 3D affine transformation matrix to a post-multiplied one
simply compute its transpose using \code{\link[=t]{t()}}.  To get an inverse transformation matrix
from an existing transformation matrix that does the opposite transformations
simply compute its inverse using \code{\link[=solve]{solve()}}.
}
\examples{
p <- as_coord3d(x = sample(1:10, 3), y = sample(1:10, 3), z = sample(1:10, 3))

# {affiner} affine transformation matrices are post-multiplied
# and therefore should **not** go in reverse order
mat <- transform3d(diag(4L)) \%*\%
         rotate3d("z-axis", degrees(90)) \%*\%
         scale3d(1, 2, 1) \%*\%
         translate3d(x = -1, y = -1, z = -1)
p1 <- p$
  clone()$
  transform(mat)

# The equivalent result applying affine transformations via method chaining
p2 <- p$
  clone()$
  transform(diag(4L))$
  rotate("z-axis", degrees(90))$
  scale(1, 2, 1)$
  translate(x = -1, y = -1, z = -1)

all.equal(p1, p2)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/isotoxal_2ngon_inner_radius.R
\name{isotoxal_2ngon_inner_radius}
\alias{isotoxal_2ngon_inner_radius}
\alias{star_inner_radius}
\title{Isotoxal \verb{2n}-gon inner radius}
\usage{
isotoxal_2ngon_inner_radius(
  n,
  outer_radius = 1,
  ...,
  alpha = NULL,
  beta_ext = NULL,
  d = NULL
)

star_inner_radius(
  n,
  outer_radius = 1,
  ...,
  alpha = NULL,
  beta_ext = NULL,
  d = NULL
)
}
\arguments{
\item{n}{The number of outer vertices.}

\item{outer_radius}{The outer radius of the isotoxal \verb{2n}-gon.}

\item{...}{Ignored.}

\item{alpha}{The interior angle of an outer vertex.  Will be coerced by \code{\link[=degrees]{degrees()}}.}

\item{beta_ext}{The exterior angle of an inner vertex.  Will be coerced by \code{\link[=degrees]{degrees()}}.}

\item{d}{The density aka winding number of the regular star polygon (outline) in which case this star is represented by \verb{|n/d|}.}
}
\value{
A numeric vector
}
\description{
\code{isotoxal_2ngon_inner_radius()} computes the inner radius
of an isotoxal \verb{2n}-gon polygon.
\code{star_inner_radius()} is an alias.
}
\details{
Isotoxal \verb{2n}-gon polygons are polygons with:
\itemize{
\item \verb{2n} vertices alternating between \code{n} "outer" vertices evenly spaced on one circle and \code{n} "inner" vertices evenly spaced on smaller circle with the same center.
\item Each edge of the polygon is of the same length.
\item The outer vertices all have the same interior angle \code{alpha} and the
inner vertices all have the same interior angle \code{beta}.
\item They are a generalization of (the outlines of) concave simple "star" polygons that also includes convex polygons with an even number of vertices.
}
}
\examples{
# |8/3| star has outer vertex internal angle 45 degrees
# and inner vertex external angle 90 degrees
isotoxal_2ngon_inner_radius(8, d = 3)
isotoxal_2ngon_inner_radius(8, alpha = 45)
isotoxal_2ngon_inner_radius(8, beta_ext = 90)
}
\seealso{
\code{\link[=isotoxal_2ngon_polygon2d]{isotoxal_2ngon_polygon2d()}} to construct an isotoxal \verb{2n}-gon
\link{Polygon2D} object.
\url{https://en.wikipedia.org/wiki/Isotoxal_figure#Isotoxal_polygons} and
\url{https://en.wikipedia.org/wiki/Star_polygon#Isotoxal_star_simple_polygons}
for more information on isotoxal polygons.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/isotoxal_2ngon_polygon2d.R
\name{isotoxal_2ngon_polygon2d}
\alias{isotoxal_2ngon_polygon2d}
\alias{star_polygon2d}
\title{Isotoxal \verb{2n}-gon and star polygon}
\usage{
isotoxal_2ngon_polygon2d(
  n,
  x = 0,
  y = 0,
  radius = 0.5,
  radial_scale = isotoxal_2ngon_inner_radius(n, alpha = alpha, beta_ext = beta_ext, d =
    d),
  theta = degrees(90),
  ...,
  alpha = NULL,
  beta_ext = NULL,
  d = NULL
)

star_polygon2d(
  n,
  x = 0,
  y = 0,
  radius = 0.5,
  radial_scale = isotoxal_2ngon_inner_radius(n, alpha = alpha, beta_ext = beta_ext, d =
    d),
  theta = degrees(90),
  ...,
  alpha = NULL,
  beta_ext = NULL,
  d = NULL
)
}
\arguments{
\item{n}{Number of outer vertices.}

\item{x}{X-coordinate of the center (default \code{0}).}

\item{y}{Y-coordinate of the center (default \code{0}).}

\item{radius}{Outer circumradius (default \code{0.5}).}

\item{radial_scale}{Inner radius as a fraction of \code{radius}.  Defaults to
\code{isotoxal_2ngon_inner_radius(n, alpha = alpha, beta_ext = beta_ext, d = d)}.
Exactly one of \code{radial_scale}, \code{alpha}, \code{beta_ext}, or \code{d} must be
supplied.}

\item{theta}{Angle of the first outer vertex (default \code{degrees(90)}).
Will be coerced by \code{\link[=degrees]{degrees()}}.}

\item{...}{Ignored.}

\item{alpha}{Interior angle of an outer vertex.  Will be coerced by
\code{\link[=degrees]{degrees()}}.}

\item{beta_ext}{Exterior angle of an inner vertex.  Will be coerced by
\code{\link[=degrees]{degrees()}}.}

\item{d}{Density (winding number) of the star polygon \verb{|n/d|}.}
}
\value{
A \link{Polygon2D} object.
}
\description{
\code{isotoxal_2ngon_polygon2d()} creates an isotoxal \verb{2n}-gon \link{Polygon2D}
object centered at \verb{(x, y)}.
\code{star_polygon2d()} is an alias.
}
\details{
Isotoxal \verb{2n}-gon polygons have \verb{2n} vertices alternating between \code{n}
outer vertices on a circle of radius \code{radius} and \code{n} inner vertices on
a concentric circle of radius \code{radial_scale * radius}.
}
\examples{
# |5/2| star (the verda stelo)
p <- isotoxal_2ngon_polygon2d(5, d = 2)
p$is_convex
plot(p, col = "#008000", border = NA)

# `star_polygon2d()` is an alias
p2 <- star_polygon2d(5, d = 2)
all.equal(p, p2)
}
\seealso{
\code{\link[=isotoxal_2ngon_inner_radius]{isotoxal_2ngon_inner_radius()}} to compute the radial scale.
\code{\link[=rectangle_polygon2d]{rectangle_polygon2d()}} for rectangles.
\code{\link[=regular_ngon_polygon2d]{regular_ngon_polygon2d()}} for regular convex polygons.
\url{https://en.wikipedia.org/wiki/Isotoxal_figure#Isotoxal_polygons} and
\url{https://en.wikipedia.org/wiki/Star_polygon#Isotoxal_star_simple_polygons}
for more information on isotoxal polygons.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/affine-methods.R
\name{as_transform1d}
\alias{as_transform1d}
\alias{as_transform1d.transform1d}
\alias{as_transform1d.default}
\title{Cast to 1D affine transformation matrix}
\usage{
as_transform1d(x, ...)

\method{as_transform1d}{transform1d}(x, ...)

\method{as_transform1d}{default}(x, ...)
}
\arguments{
\item{x}{An object that can be cast to a}

\item{...}{Further arguments passed to or from other methods}
}
\value{
A \code{\link[=transform1d]{transform1d()}} object
}
\description{
\code{as_transform1d()} casts to a \code{\link[=transform1d]{transform1d()}} affine transformation matrix
}
\examples{
m <- diag(2L)
as_transform1d(m)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/has_overlap2d.R
\name{has_overlap2d}
\alias{has_overlap2d}
\alias{has_overlap2d.Polygon2D}
\alias{has_overlap2d.Ellipse2D}
\title{Whether two 2D shapes overlap}
\usage{
has_overlap2d(x, y, ...)

\method{has_overlap2d}{Polygon2D}(x, y, ..., n = 64L, tol = sqrt(.Machine$double.eps))

\method{has_overlap2d}{Ellipse2D}(x, y, ..., n = 64L, tol = sqrt(.Machine$double.eps))
}
\arguments{
\item{x, y}{The two shapes to check.  Supported classes are \link{Polygon2D}
and \link{Ellipse2D}.}

\item{...}{Ignored; only included for S3 method consistency.}

\item{n}{Number of vertices used to approximate non-circular \link{Ellipse2D}
objects (default \code{64L}).  Larger values give tighter bounds.}

\item{tol}{Numeric tolerance for overlap comparisons (default \code{0}).
A positive value requires projections to overlap by more than \code{tol}
before overlap is declared, which avoids false positives caused by
floating-point error accumulating in rotated coordinates.
\code{sqrt(.Machine$double.eps)} is a reasonable choice when shapes are
constructed via trigonometric transformations.}
}
\value{
A logical vector (or \code{NA}) of length equal to the longer of \code{x}
and \code{y} (for \link{Ellipse2D} objects; \link{Polygon2D} objects are always scalar).
}
\description{
\code{has_overlap2d()} is an S3 generic that returns whether two 2D shapes
have a non-zero-area overlap (i.e. their interiors intersect).
}
\details{
For two convex shapes the function uses the
\href{https://en.wikipedia.org/wiki/Hyperplane_separation_theorem}{Separating Axis Theorem (SAT)}.
For concave \link{Polygon2D} objects the function first checks whether the
axis-aligned bounding box and convex hull overlap; if neither rules out
overlap it returns \code{NA} with a warning because exact detection is not yet
supported.
For non-circular \link{Ellipse2D} objects the function approximates the ellipse
with inner and outer polygons: if the outer polygon does not overlap the
result is \code{FALSE}; if the inner polygon overlaps the result is \code{TRUE};
otherwise \code{NA} is returned with a warning.
}
\examples{
# Two overlapping squares
sq1 <- as_polygon2d(as_coord2d(
  x = c(0, 1, 1, 0),
  y = c(0, 0, 1, 1)
))
sq2 <- as_polygon2d(as_coord2d(
  x = c(0.5, 1.5, 1.5, 0.5),
  y = c(0.5, 0.5, 1.5, 1.5)
))
sq3 <- as_polygon2d(as_coord2d(
  x = c(2, 3, 3, 2),
  y = c(2, 2, 3, 3)
))
has_overlap2d(sq1, sq2)
has_overlap2d(sq1, sq3)

# Circle vs polygon
circ <- as_ellipse2d(as_coord2d(0.5, 0.5), rx = 0.4)
has_overlap2d(sq1, circ)
has_overlap2d(sq3, circ)

# Two circles
c1 <- as_ellipse2d(as_coord2d(0, 0), rx = 1)
c2 <- as_ellipse2d(as_coord2d(1.5, 0), rx = 1)
c3 <- as_ellipse2d(as_coord2d(3, 0), rx = 1)
has_overlap2d(c1, c2)
has_overlap2d(c1, c3)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coord.R
\name{Coord2D}
\alias{Coord2D}
\title{2D coordinate vector R6 Class}
\description{
\code{Coord2D} is an \code{\link[R6:R6Class]{R6::R6Class()}} object representing two-dimensional points
represented by Cartesian Coordinates.
}
\examples{
p <- as_coord2d(x = rnorm(100, 2), y = rnorm(100, 2))
print(p, n = 10)
pc <- mean(p) # Centroid
# method chained affine transformation matrices are auto-pre-multiplied
p$
  translate(-pc)$
  shear(x = 1, y = 0)$
  reflect("x-axis")$
  rotate(90, "degrees")$
  print(n = 10)
}
\section{Active bindings}{
  \if{html}{\out{<div class="r6-active-bindings">}}
  \describe{
    \item{\code{xyw}}{A three-column matrix representing the homogeneous coordinates.
The first two columns are "x" and "y" coordinates
and the third column is all ones.}

    \item{\code{x}}{A numeric vector of x-coordinates.}

    \item{\code{y}}{A numeric vector of y-coordinates.}
  }
  \if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-Coord2D-initialize}{\code{Coord2D$new()}}
    \item \href{#method-Coord2D-permute}{\code{Coord2D$permute()}}
    \item \href{#method-Coord2D-print}{\code{Coord2D$print()}}
    \item \href{#method-Coord2D-project}{\code{Coord2D$project()}}
    \item \href{#method-Coord2D-reflect}{\code{Coord2D$reflect()}}
    \item \href{#method-Coord2D-rotate}{\code{Coord2D$rotate()}}
    \item \href{#method-Coord2D-scale}{\code{Coord2D$scale()}}
    \item \href{#method-Coord2D-shear}{\code{Coord2D$shear()}}
    \item \href{#method-Coord2D-translate}{\code{Coord2D$translate()}}
    \item \href{#method-Coord2D-transform}{\code{Coord2D$transform()}}
    \item \href{#method-Coord2D-clone}{\code{Coord2D$clone()}}
  }
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Coord2D-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-Coord2D-initialize}{}}}
\subsection{\code{Coord2D$new()}}{
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Coord2D$new(xyw)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{xyw}}{A matrix with three columns representing (homogeneous) coordinates.
The first two columns represent x and y coordinates and
the last column is all ones.
Column names should be "x", "y", and "w".}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Coord2D-permute"></a>}}
\if{latex}{\out{\hypertarget{method-Coord2D-permute}{}}}
\subsection{\code{Coord2D$permute()}}{
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Coord2D$permute(permutation = c("xy", "yx"))}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{permutation}}{Either "xy" (no permutation) or "yx" (permute x and y axes)}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Coord2D-print"></a>}}
\if{latex}{\out{\hypertarget{method-Coord2D-print}{}}}
\subsection{\code{Coord2D$print()}}{
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Coord2D$print(n = NULL, ...)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{n}}{Number of coordinates to print.  If \code{NULL} print all of them.}
      \item{\code{...}}{Passed to \code{\link[=format.default]{format.default()}}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Coord2D-project"></a>}}
\if{latex}{\out{\hypertarget{method-Coord2D-project}{}}}
\subsection{\code{Coord2D$project()}}{
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Coord2D$project(line = as_line2d("x-axis"), ..., scale = 0)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{line}}{A \link{Line2D} object of length one representing the line
you with to reflect across or project to or an object coercible to one by \code{as_line2d(line, ...)}
such as "x-axis" or "y-axis".}
      \item{\code{...}}{Passed to \code{\link[=project2d]{project2d()}}}
      \item{\code{scale}}{Oblique projection scale factor.
A degenerate \code{0} value indicates an orthogonal projection.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Coord2D-reflect"></a>}}
\if{latex}{\out{\hypertarget{method-Coord2D-reflect}{}}}
\subsection{\code{Coord2D$reflect()}}{
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Coord2D$reflect(line = as_line2d("x-axis"), ...)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{line}}{A \link{Line2D} object of length one representing the line
you with to reflect across or project to or an object coercible to one by \code{as_line2d(line, ...)}
such as "x-axis" or "y-axis".}
      \item{\code{...}}{Passed to \code{\link[=reflect2d]{reflect2d()}}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Coord2D-rotate"></a>}}
\if{latex}{\out{\hypertarget{method-Coord2D-rotate}{}}}
\subsection{\code{Coord2D$rotate()}}{
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Coord2D$rotate(theta = angle(0), ...)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{theta}}{An \code{\link[=angle]{angle()}} object of length one or an object coercible to one by \code{as_angle(theta, ...)}.}
      \item{\code{...}}{Passed to \code{\link[=as_angle]{as_angle()}}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Coord2D-scale"></a>}}
\if{latex}{\out{\hypertarget{method-Coord2D-scale}{}}}
\subsection{\code{Coord2D$scale()}}{
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Coord2D$scale(x_scale = 1, y_scale = x_scale)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{x_scale}}{Scaling factor to apply to x coordinates}
      \item{\code{y_scale}}{Scaling factor to apply to y coordinates}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Coord2D-shear"></a>}}
\if{latex}{\out{\hypertarget{method-Coord2D-shear}{}}}
\subsection{\code{Coord2D$shear()}}{
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Coord2D$shear(xy_shear = 0, yx_shear = 0)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{xy_shear}}{Horizontal shear factor: \code{x = x + xy_shear * y}}
      \item{\code{yx_shear}}{Vertical shear factor: \code{y = yx_shear * x + y}}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Coord2D-translate"></a>}}
\if{latex}{\out{\hypertarget{method-Coord2D-translate}{}}}
\subsection{\code{Coord2D$translate()}}{
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Coord2D$translate(x = as_coord2d(0, 0), ...)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{x}}{A \link{Coord2D} object of length one or an object coercible to one by \code{as_coord2d(x, ...)}.}
      \item{\code{...}}{Passed to \code{as_coord2d(x, ...)} if \code{x} is not a \link{Coord2D} object}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Coord2D-transform"></a>}}
\if{latex}{\out{\hypertarget{method-Coord2D-transform}{}}}
\subsection{\code{Coord2D$transform()}}{
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Coord2D$transform(mat = transform2d())}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{mat}}{A 3x3 matrix representing a post-multiplied affine transformation matrix.
The last \strong{column} must be equal to \code{c(0, 0, 1)}.
If the last \strong{row} is \code{c(0, 0, 1)} you may need to transpose it
to convert it from a pre-multiplied affine transformation matrix to a post-multiplied one.
If a 2x2 matrix (such as a 2x2 post-multiplied 2D rotation matrix)
we'll quietly add a final column/row equal to \code{c(0, 0, 1)}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Coord2D-clone"></a>}}
\if{latex}{\out{\hypertarget{method-Coord2D-clone}{}}}
\subsection{\code{Coord2D$clone()}}{
  The objects of this class are cloneable with this method.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{Coord2D$clone(deep = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{deep}}{Whether to make a deep clone.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{graphics}
\alias{graphics}
\alias{plot.Coord1D}
\alias{points.Coord1D}
\alias{lines.Point1D}
\alias{plot.Coord2D}
\alias{points.Coord2D}
\alias{lines.Coord2D}
\alias{lines.Polygon2D}
\alias{lines.Ellipse2D}
\alias{plot.Polygon2D}
\alias{plot.Ellipse2D}
\alias{lines.Line2D}
\alias{plot.Segment2D}
\alias{lines.Segment2D}
\title{Plot coordinates, points, lines, polygons, ellipses, and segments}
\usage{
\method{plot}{Coord1D}(x, ...)

\method{points}{Coord1D}(x, ...)

\method{lines}{Point1D}(x, ...)

\method{plot}{Coord2D}(x, ..., asp = 1)

\method{points}{Coord2D}(x, ...)

\method{lines}{Coord2D}(x, ...)

\method{lines}{Polygon2D}(x, ...)

\method{lines}{Ellipse2D}(x, n = 60L, ...)

\method{plot}{Polygon2D}(
  x,
  ...,
  col = NA,
  border = NULL,
  density = NULL,
  angle = 45,
  fillOddEven = FALSE,
  asp = 1
)

\method{plot}{Ellipse2D}(
  x,
  n = 60L,
  ...,
  col = NA,
  border = NULL,
  density = NULL,
  angle = 45,
  fillOddEven = FALSE,
  asp = 1
)

\method{lines}{Line2D}(x, ...)

\method{plot}{Segment2D}(x, ..., asp = 1)

\method{lines}{Segment2D}(x, ...)
}
\arguments{
\item{x}{A supported object to plot.}

\item{...}{Passed to the underlying plot method.}

\item{asp}{the y/x aspect ratio.}

\item{n}{Number of vertices used to approximate each ellipse (default \code{60L}).}

\item{col, border, density, angle, fillOddEven}{Passed to \code{\link[graphics:polygon]{graphics::polygon()}}.}
}
\value{
Used for its side effect of drawing to the graphics device.
}
\description{
\code{\link[=plot]{plot()}} plots \link{Coord1D}, \link{Coord2D}, \link{Polygon2D}, \link{Ellipse2D}, and
\link{Segment2D} class objects.
\code{\link[=points]{points()}} draws \link{Coord1D} and \link{Coord2D} class objects to an existing plot.
\code{\link[=lines]{lines()}} draws \link{Coord2D}, \link{Ellipse2D}, \link{Point1D}, \link{Line2D}, and \link{Segment2D}
class objects to an existing plot.
If the suggested \link[ggplot2:ggplot2]{ggplot2} and \link[rgl:rgl]{rgl} packages
are available we also register \code{\link[ggplot2:autolayer]{ggplot2::autolayer()}} methods for \link{Coord1D},
\link{Coord2D}, \link{Ellipse2D}, \link{Line2D}, \link{Point1D}, \link{Polygon2D}, and \link{Segment2D}
class objects and \code{\link[rgl:plot3d]{rgl::plot3d()}} methods for \link{Coord3D} and \link{Plane3D} class
objects.
}
\examples{
c1 <- as_coord2d(x = 0, y = 1:10)
l <- as_line2d(a = 1, b = -1, c = 0) # y = x
c2 <- c1$clone()$reflect(l)
plot(c1, xlim = c(-1, 11), ylim = c(-1, 11),
     main = "2D reflection across a line")
lines(l)
points(c2, col = "red")

c1 <- as_coord2d(x = 1:10, y = 1:10)
l <- as_line2d(a = -1, b = 0, c = 0) # x = 0
c2 <- c1$clone()$project(l)
if (require("ggplot2", quietly = TRUE,
            include.only = c("ggplot", "autolayer", "labs"))) {
  ggplot() +
      autolayer(c1) +
      autolayer(l) +
      autolayer(c2, color = "red") +
      labs(title = "2D projection onto a line")
}

c1 <- as_coord1d(x = seq.int(-4, -1))
pt <- as_point1d(a = 1, b = 0) # x = 0
c2 <- c1$clone()$reflect(pt)
plot(c1, xlim = c(-5, 5), main = "1D reflection across a point")
lines(pt)
points(c2, col = "red")

# 3D reflection across a plane
c1 <- as_coord3d(x = 1:10, y = 1:10, z = 1:10)
pl <- as_plane3d(a = 0, b = 0, c = -1, d = 2) # z = 2
c2 <- c1$clone()$reflect(pl)
if (require("rgl", quietly = TRUE, include.only = "plot3d")) {
  plot3d(c1, col = "blue", size = 8)
  plot3d(pl, color = "grey", alpha = 0.6)
  plot3d(c2, add = TRUE, col = "red", size = 8)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/affine-methods.R
\name{as_transform3d}
\alias{as_transform3d}
\alias{as_transform3d.transform3d}
\alias{as_transform3d.default}
\title{Cast to 3D affine transformation matrix}
\usage{
as_transform3d(x, ...)

\method{as_transform3d}{transform3d}(x, ...)

\method{as_transform3d}{default}(x, ...)
}
\arguments{
\item{x}{An object that can be cast to a}

\item{...}{Further arguments passed to or from other methods}
}
\value{
A \code{\link[=transform3d]{transform3d()}} object
}
\description{
\code{as_transform3d()} casts to a \code{\link[=transform3d]{transform3d()}} affine transformation matrix
}
\examples{
m <- diag(4L)
as_transform3d(m)
}
